/* fast_lock.h */
/* share/epicsH @(#)fast_lock.h	1.8     9/24/91 */

/*
 *      Author:          Jeff Hill
 *      Date:            4-11-89
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01 joh 041189	initial release into middle age
 * .02 joh 071989	added fast lock test
 * .03 joh 090391	VRTX kernel ifdef for V5 vxWorks
 * .04 joh 091091	Now uses V5 vxWorks binary semaphore
 * .05 joh 092491	Dont use V5 semaphore on V3 EPICS yet
 */

/*
 * 
 *
 *	Macros and data structures for a much faster (than vrtx semaphore) 
 *	mutual exclusion lock/unlock.
 *	(semaphore in this case is only used to wait for unlock)
 *
 *      On a 68020 the following times include lock and unlock
 *
 *	FASTLOCK-FASTUNLOCK     8 microsecs
 *	semTake-semGive(Macro) 80 microsecs
 *	semTake-semGive	       92 microsecs
 *
 */

#ifndef INCLfast_lockh
#define INCLfast_lockh

#ifndef INCLsemLibh
#include <semLib.h>
#endif

#ifndef INCLtaskLibh
#include <taskLib.h>
#endif

#include <vxLib.h>
/*
 *        Macro equivalent of vxWorks glue for better performance
 */
#ifdef VRTX_KERNEL
#	define semGive(SEMID)\
	{    if ((SEMID)->count == 0)vrtxPost (&((SEMID)->count), 1);    }

#	define semTake(SEMID)\
	{int dummy;    vrtxPend (&((SEMID)->count), 0, &dummy);    }
#endif

typedef struct{
#ifdef V5_vxWorks
  	SEM_ID		ppend;		/* wait for lock sem */
#else
  	SEMAPHORE	pend;		/* wait for lock sem */
#endif
  	unsigned short	count;		/* cnt of tasks waiting for lock */
  	unsigned short	lock;		/* test and set lock bit */
}FAST_LOCK;





/*
 *	extra paren avoids order of ops problems
 *	(returns what semBCreate returns on v5 vxWorks)
 */
#ifdef V5_vxWorks
#	define FASTLOCKINIT(PFAST_LOCK)\
	(\
	((FAST_LOCK *)(PFAST_LOCK))->count =0, \
	((FAST_LOCK *)(PFAST_LOCK))->lock =0, \
	((FAST_LOCK *)(PFAST_LOCK))->ppend = \
		semBCreate(SEM_Q_PRIORITY, SEM_EMPTY) \
	)
#else
#	define FASTLOCKINIT(PFAST_LOCK)\
	{\
	((FAST_LOCK *)(PFAST_LOCK))->count =0;\
	((FAST_LOCK *)(PFAST_LOCK))->lock =0;\
	semInit( &( ((FAST_LOCK *)(PFAST_LOCK))->pend ) );\
	}
#endif

/* 
 * new requirement with v5 vxWorks
 */
#ifdef V5_vxWorks
#	define FASTLOCKFREE(PFAST_LOCK)\
  		semDelete( ((FAST_LOCK *)(PFAST_LOCK))->ppend )
#else
#	define FASTLOCKFREE(PFAST_LOCK)	(0)
#endif

#if CPU==I80486
#	define FASTLOCK(PFAST_LOCK)\
	{\
	taskLock();\
	((FAST_LOCK *)(PFAST_LOCK))->count++;\
	taskUnlock();\
	while(!vxTas( &( ((FAST_LOCK *)(PFAST_LOCK))->lock ) ))\
  		semTake(((FAST_LOCK *)(PFAST_LOCK))->ppend, WAIT_FOREVER);\
	taskLock();\
	(  ((FAST_LOCK *)(PFAST_LOCK))->count)--;\
	taskUnlock();\
	}
#elif defined(V5_vxWorks)
#	define FASTLOCK(PFAST_LOCK)\
	{\
	((FAST_LOCK *)(PFAST_LOCK))->count++;\
	while(!vxTas( &( ((FAST_LOCK *)(PFAST_LOCK))->lock ) ))\
  		semTake(((FAST_LOCK *)(PFAST_LOCK))->ppend, WAIT_FOREVER);\
	(  ((FAST_LOCK *)(PFAST_LOCK))->count)--;\
	}
#else
#	define FASTLOCK(PFAST_LOCK)\
	{\
	((FAST_LOCK *)(PFAST_LOCK))->count++;\
	while(!vxTas( &( ((FAST_LOCK *)(PFAST_LOCK))->lock ) ))\
   		semTake(&(((FAST_LOCK *)(PFAST_LOCK))->pend));\
	(  ((FAST_LOCK *)(PFAST_LOCK))->count)--;\
	}
#endif

#ifdef V5_vxWorks
#	define FASTUNLOCK(PFAST_LOCK)\
	{\
	((FAST_LOCK *)(PFAST_LOCK))->lock = FALSE;\
	if( ((FAST_LOCK *)(PFAST_LOCK))->count )\
  		semGive(((FAST_LOCK *)(PFAST_LOCK))->ppend);\
	};
#else
#	define FASTUNLOCK(PFAST_LOCK)\
	{\
	((FAST_LOCK *)(PFAST_LOCK))->lock = FALSE;\
	if( ((FAST_LOCK *)(PFAST_LOCK))->count )\
		semGive( &( ((FAST_LOCK *)(PFAST_LOCK))->pend ) );\
	};
#endif

#define FASTLOCKTEST(PFAST_LOCK)\
( ((FAST_LOCK *)(PFAST_LOCK))->lock )


/*
 *	Nothing after this endif
 */
#endif
