/*	/share/epicsH  @(#)seq.h	1.8     1/29/92
 *
 *	DESCRIPTION: Definitions for the run-time sequencer.
 *
 *      Author:         Andy Kozubal
 *      Date:           
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  07-03-91        ajk	Changed SSCB semaphore id names.
 * .02	07-05-91	ajk	Added function prototypes.
 * .03	12-16-91	ajk	Setting up for VxWorks version 5.
 */

#include	"cadef.h"
#include	"alarm.h"

/* Macros to handle set & clear event bits */
#define NBITS           32		/* # bits in bit mask word */
#define	NWRDS		32		/* # words in bit fields */
typedef long            bitMask;

#define bitSet(word, bitnum)	(word[bitnum/NBITS] |= (1<<(bitnum%NBITS)))

#define bitClear(word, bitnum)	(word[bitnum/NBITS] &= ~(1<<(bitnum%NBITS)))

#define bitTest(word, bitnum)	((word[bitnum/NBITS] & (1<<(bitnum%NBITS))) != 0)

/* Symbol table entry */
typedef	struct {
	char	*Name;		/* Symbol name */
	int	Type;		/* Type of symbol, V_... */
	int	Size;		/* Size of indivual item (int, long, ...) */
	int	NumElements;	/* Number of elements in array */
	int	Offset;		/* Offset from beginning of structure */
	int	ChanIndex;	/* -1 of not associated with a channel */
} SymTabItem;

/* Variable types */
#define	V_NONE		0		/* not defined */
#define	V_CHAR		1		/* char */
#define	V_SHORT		2		/* short */
#define	V_INT		3		/* int (but converted to long) */
#define	V_LONG		4		/* long */
#define	V_FLOAT		5		/* float */
#define	V_DOUBLE	6		/* double */
#define	V_STRING	7		/* strings (array of char) */
#define	V_EVFLAG	8		/* event flag */
#define	V_FUNC		9		/* function (not a variable) */
#define	V_UCHAR		11		/* unsigned char */
#define	V_USHORT	12		/* unsigned short */
#define	V_UINT		13		/* unsigned int (converted to unsigned long) */
#define	V_ULONG		14		/* unsigned long */

/* Structure to hold information about database channels */
struct	db_channel
{
	short		index;		/* index in array of db channels */
	char		*db_uname;	/* unexpanded channel name */
	char		*db_name;	/* channel name after macro expansion */
	char		*var;		/* ptr to variable (or an offset) */
	chid		chanid;		/* ptr to channel id (from ca_search()) */
	BOOL		connected;	/* TRUE only if channel is connected */
	BOOL		get_complete;	/* TRUE if previous pvGet completed */
	short		status;		/* last db access status code */
	short		severity;	/* last db access severity code */
	short		size;		/* size (in bytes) of single variable element */
	short		get_type;	/* db get type (e.g. DBR_STS_INT) */
	short		put_type;	/* db put type (e.g. DBR_INT) */
	int		count;		/* element count for db requests */
	BOOL		mon_flag;	/* TRUE if channel is to be monitored */
	BOOL		monitored;	/* TRUE if channel IS monitored */
	float		delta;		/* monitor dead-band */
	float		timeout;	/* monitor timeout */
	evid		eventid;	/* event id (supplied by CA) */
	SIDNT		getSemId;	/* semaphore for syncronous pvGet() */
	struct state_program *sprog;	/* state program that owns this structure */
};
typedef	struct db_channel CHAN;

/* Structure to hold information about a state */
struct	state_info_block
{
	FPTR		action_func;	/* ptr to action routine for this state */
	FPTR		event_func;	/* ptr to event routine for this state */
	FPTR		delay_func;	/* ptr to delay routine for this state */
	bitMask		eventMask[NWRDS];/* event mask for this state */
	char		*name;		/* state name (for debugging) */
};
typedef	struct	state_info_block STATE;

#define	MAX_NDELAY	20		/* max # delays allowed in each SS */
/* Structure to hold information about a State Set */
struct	state_set_control_block
{
	int		task_id;	/* task id */
	int		task_priority;	/* task priority */
	SIDNT		syncSemId;	/* semaphore for event sync */
	SIDNT		getSemId;	/* semaphore for synchronous pvGet() */
	int		num_states;	/* number of states */
	STATE		*states;	/* ptr to array of state blocks */
	short		current_state;	/* current state index */
	short		next_state;	/* next state index */
	short		prev_state;	/* previous state index */
	short		error_state;	/* error state index (-1 if none defined) */
	short		trans_num;	/* highest priority trans. # that triggered */
	BOOL		action_complete;/* TRUE if action complete
					   FALSE if waiting for an event */
	bitMask		events[NWRDS];	/* outstanding events */
	bitMask		*pMask;		/* current event mask */
	int		ndelay;		/* number of delays activated */
	ULONG		timeout[MAX_NDELAY]; /* time-out for delay() */
	ULONG		time;		/* time that a state was entered */
	char		*name;		/* state set name (for debugging) */
};
typedef	struct	state_set_control_block SSCB;

/* Macro table */
typedef	struct	macro {
	char	*name;
	char	*value;
} MACRO;

/* All information about a state program.
	The address of this structure is passed to the run-time sequencer:
 */
struct	state_program
{
	int		magic;		/* magic number so we don't s... on system */
	int		task_id;	/* task id (main task) */
	char		*dyn_ptr;	/* ptr to allocated area at run time */
	BOOL		task_is_deleted;/* TRUE if main task has been deleted */
	int		task_priority;	/* task priority */
	SIDNT		caSemId;	/* semiphore for locking CA events */
	CHAN		*channels;	/* table of channels */
	int		nchan;		/* number of db channels */
	int		conn_count;	/* number of channels connected */
	SSCB		*sscb;		/* array of state set control blocks */
	int		nss;		/* number of state sets */
	STATE		*states;	/* ptr to dyn array of state blocks */
	int		nstates;	/* number of states */
	char		*user_area;	/* ptr to user variable area */
	int		user_size;	/* # bytes in user variable area */
	SymTabItem	*sym_table;	/* ptr to symbol table */
	int		num_symbols;	/* number of symbols */
	MACRO		*mac_ptr;	/* ptr to macro table */
	char		*scr_ptr;	/* ptr to start of scratch area */
	int		scr_nleft;	/* # bytes left in scratch area */
	char		*name;		/* program name (for debugging) */
	char		*params;	/* program paramters */
	BOOL		async_flag;	/* async. gets flag */
	BOOL		debug_flag;	/* debug flag */
	BOOL		reent_flag;	/* reentrant flag */
	BOOL		conn_flag;	/* connection wait flag */
	FPTR		exit_func;	/* exit function */
	SIDNT		logSemId;	/* logfile locking semaphore */
	int		logFd;		/* logfile file descr. */
	int		seq_token;	/* sequencer redundancy token */
	long		prog_time;	/* sequence program starting tick */
};
typedef	struct	state_program SPROG;

#define	MAX_MACROS	20

char	*seqAlloc(SPROG *pSP, int nChar);
char	*seqMacValGet(char *macName, int macNameLth, MACRO *macTbl);

/* #define	MAGIC 910726			/ * magic number for SPROG */
#define MAGIC 940421			/* Magic number of new version */

/* Macros for special function calls */
#define	seq_params		sprog, ss_ptr

#define	pvPut(db_index)		seq_pvPut(seq_params, &sprog->channels[db_index])

#define	pvGet(db_index)		seq_pvGet(seq_params, &sprog->channels[db_index])

#define	pvMonitor(db_index)	seq_pvMonitor(seq_params, &sprog->channels[db_index])

#define	pvStopMonitor(db_index)	seq_pvStopMonitor(seq_params, &sprog->channels[db_index])

#define	pvStatus(db_index)	(sprog->channels[db_index].status)

#define	pvSeverity(db_index)	(sprog->channels[db_index].severity)

#define	pvConnected(db_index)	(sprog->channels[db_index].connected)

#define	pvGetComplete(db_index)	(sprog->channels[db_index].get_complete)

#define	pvChannelCount()	(sprog->nchan)

#define	pvConnectCount()	(sprog->conn_count)

#define	pvFlush()		seq_pvFlush(seq_params)

#define	pvCount(db_index)	(sprog->channels[db_index].count)

#define	pvIndex(db_index)	(db_index)

#define	efSet(ef_name)		seq_efSet(seq_params, ef_name)

#define	efTest(ef_name)		seq_efTest(seq_params, ef_name)

#define	test_delay(delay_id)	seq_test_delay(seq_params, delay_id)

#define	start_delay(delay_id, delay)\
	seq_start_delay(seq_params, delay_id, (float)delay)

#define	macValueGet(mac_name)\
	seqMacValGet(mac_name, strlen(mac_name), sprog->mac_ptr)

#define	flagGet(flag)		seq_flagGet(sprog, flag)


#ifdef vxWorks
/* Task parameters */
#define SPAWN_STACK_SIZE	10000
#define SPAWN_OPTIONS		VX_DEALLOC_STACK | VX_FP_TASK | VX_STDIO
#define SPAWN_PRIORITY		165

/* Pre-VxWorks 5.0 defs */
#ifndef	V5_vxWorks
#define	semTake(semId, timeout)	seq_semTake(semId, timeout)
/* #define	SEM_INVERSION_SAFE	0
#define	SEM_DELETE_SAFE		0 */
#define	SEM_FULL		2
#define	SEM_EMPTY		1
/* #define	SEM_Q_PRIORITY		0
#define	SEM_Q_FIFO		0 */
#define	WAIT_FOREVER		(-1)
#define	NO_WAIT			0
#endif	V5_vxWorks

int	seq_connect(SPROG *);
VOID	seq_event_handler(struct event_handler_args);
VOID	seq_conn_handler(struct connection_handler_args);
VOID	seq_efSet(SPROG *, int, int);
int	seq_efTest(SPROG *, SSCB *, int);
int	seq_pvGet(SPROG *, SSCB *, CHAN *);
VOID	seq_callback_handler(struct event_handler_args);
VOID	seq_pvFlush();
int	seq_pvPut(SPROG *, SSCB *, CHAN *);
int	seq_pvMonitor(SPROG *, SSCB *, CHAN *);
int	seq_pvStopMonitor(SPROG *, SSCB *, CHAN *);
int	seq(SPROG *, char *, int, char *);
VOID	seq_log (SPROG *pSP, char *fmt, int arg1, int arg2, int arg3, int arg4,
		 int arg5, int arg6, int arg7, int arg8);
char	*seqAlloc(SPROG *, int);
STS	seqLog();
STATUS	seqAddProg (SPROG *pSP);
STATUS	seqDelProg (SPROG *pSP);
int	sequencer (SPROG *pSP, int stack_size, char *ptask_name, char *resName);
VOID	ss_entry(SPROG *, SSCB *);
VOID	seq_start_delay (SPROG *pSP, SSCB *pSS, int delay_id, float delay);
BOOL	seq_test_delay (SPROG *pSP, SSCB *pSS, int delay_id);
int	sprog_delete (int tid);
SPROG	*seqQryFind (int);
void	seqMacTblInit (MACRO *pMac);
void	seqMacEval (char *pInStr, char *pOutStr, int maxChar, MACRO *macTbl);
char	*seqMacValGet (char *macName, int macNameLth, MACRO *macTbl);
int	seqMacParse (char *pMacStr, SPROG *pSP);
int	seq_sendState (SPROG *pSP);
VOID	ss_entry (SPROG *pSP, SSCB *pSS);

#ifndef	V5_vxWorks
SIDNT	semMCreate(int);
SIDNT	semBCreate(int);
#endif	V5_vxWorks

#endif vxWorks
