/****
*
*	File name :		mv90.c
*	Author :		Steve Johnson
*	Creation date :	December 22, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		mv90_init
*		mv90_publish
*		mv90_accept
*		mv90_write
*		mv90_read
*		mv90_poll
*		mv90_shutdown
*		mv90_exit
*
*	Siemens mv90 Protocol implementation
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssqueues.h>
#include <ssnames.h>
#include <sslevel1.h>
#include "mv90.h"

typedef enum {
	CLASS_PUBLISHED = 1,
	CLASS_ACCEPTED	= 2,
	CLASS_CONNECTED = 3,
} MV90_CLASS;

#define MAX_FRAME_SIZE		267
#define METER_ACK			0x06
#define METER_BLOCK_ID		0x0C
#define METER_NACK			0x15
#define METER_ETB			0x17
#define METER_ESC			0x1B
#define METER_X_COMMAND		'X'
#define METER_Y_COMMAND		'Y'
#define METER_Z_COMMAND		'Z'

typedef struct {
	QUEUE_DEF		link;
	LEVEL1_HANDLE	lower_handle;
	MV90_CLASS		mv90_class;
	UCHAR			*read_buffer;
	size_t			read_index;
	size_t			read_buffer_size;
	int				shutdown_ready;
	char			name [MAX_STRING];
} HANDLE_DEF;

static QUEUE_DEF		root_handle;

static HANDLE_DEF	*add_handle (char *, MV90_CLASS);
static void			delete_handle (HANDLE_DEF *);
static MV90_STATUS	read_lower (HANDLE_DEF *);

/****
*
****/

MV90_STATUS mv90_init (void)
{
	TRACE ("mv90_init, stub");
	queue_init (&root_handle);
	return MV90_NORMAL;
}

/****
*
****/

MV90_STATUS mv90_publish (MV90_HANDLE *user_handle, char *name)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;

	TRACE ("mv90_publish, entry point");
	handle = add_handle (name, CLASS_PUBLISHED);
	if (!handle) {
		ERROR_TRACE ("mv90_publish, add_handle failed");
		*user_handle = (MV90_HANDLE) NULL;
		return MV90_NO_MEMORY;
	}
	status = level1_publish (&handle -> lower_handle, name);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("mv90_publish, level1_publish = %d", status);
		*user_handle = (MV90_HANDLE) NULL;
		return MV90_BAD_NAME;
	}
	*user_handle = (MV90_HANDLE) handle;
	return MV90_NORMAL;
}

/****
*
****/

MV90_STATUS mv90_accept (MV90_HANDLE	*client_handle
					   , MV90_HANDLE	server_handle)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*client;
	HANDLE_DEF		*server;

	TRACE ("mv90_accept, entry point");
	if (!server_handle) {
		ERROR_TRACE ("mv90_accept, NULL handle");
		*client_handle = (MV90_HANDLE) NULL;
		return MV90_BAD_HANDLE;
	}
	server = (HANDLE_DEF *) server_handle;
	client = add_handle ("", CLASS_ACCEPTED);
	if (!client) {
		ERROR_TRACE ("mv90_accept, add_handle failed");
		*client_handle = (MV90_HANDLE) NULL;
		return MV90_NO_MEMORY;
	}
	status = level1_accept (&client -> lower_handle, server -> lower_handle);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("mv90_accept, level1_accept = %d", status);
		delete_handle (client);
		*client_handle = (MV90_HANDLE) NULL;
		return MV90_BAD_NAME;
	}
	*client_handle = (MV90_HANDLE) client;
	return MV90_NORMAL;
}

/****
*
****/

MV90_STATUS mv90_write (MV90_HANDLE		user_handle
					  , size_t			buffer_size
					  , void 			*buffer)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;

	TRACE ("mv90_write, entry point");
	if (!user_handle) {
		ERROR_TRACE ("mv90_write, NULL handle");
		return MV90_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	TRACE_B ("mv90_write", buffer_size, buffer);
	status = level1_write (handle -> lower_handle, buffer_size, buffer);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("mv90_write, level1_write = %d", status);
		return MV90_SHUTDOWN;
	}
	return MV90_NORMAL;
}

/****
*
****/

MV90_STATUS mv90_poll (MV90_HANDLE *user_handle, size_t *bytes_ready)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;

	handle = (HANDLE_DEF *) queue_next (&root_handle);
	while (handle) {
		if (handle -> shutdown_ready) {
			TRACE ("mv90_poll, handle needs to be shutdown");
			*bytes_ready = 0;
			*user_handle = (MV90_HANDLE) handle;
			return MV90_NORMAL;
		}
		status = level1_poll (handle -> lower_handle, bytes_ready);
		if (status != LEVEL1_NO_DATA) {
			if (CLASS_PUBLISHED == handle -> mv90_class) {
				TRACE ("mv90_poll, incoming connection");
				*bytes_ready = 0;
				*user_handle = (MV90_HANDLE) handle;
				return MV90_NORMAL;
			}
			status = read_lower (handle);
			if (status != MV90_NO_DATA) {
				TRACE_I ("mv90_poll, read_lower = %d", status);
				*bytes_ready = handle -> read_index;
				*user_handle = (MV90_HANDLE) handle;
				return MV90_NORMAL;
			}
		}
		handle = (HANDLE_DEF *) queue_next (&handle -> link);
	}
	*user_handle = (MV90_HANDLE) NULL;
	return MV90_NO_DATA;
}

/****
*
****/

MV90_STATUS mv90_read (MV90_HANDLE	user_handle
					 , size_t		buffer_size
					 , void			*buffer
					 , size_t		*return_length)
{
	HANDLE_DEF		*handle;

	TRACE ("mv90_read, entry point");
	if (!user_handle) {
		ERROR_TRACE ("mv90_read, NULL handle");
		*return_length = 0;
		return MV90_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	if (handle -> shutdown_ready) {
		TRACE ("mv90_read, handle is shutdown");
		*return_length = 0;
		return LEVEL1_SHUTDOWN;
	}
	if (!handle -> read_index) {
		TRACE ("mv90_read, no data");
		*return_length = 0;
		return LEVEL1_NO_DATA;
	}
	if (handle -> read_index > buffer_size) {
		*return_length = buffer_size;
	} else {
		*return_length = handle -> read_index;
	}
	memcpy (buffer, handle -> read_buffer, *return_length);
	handle -> read_index = 0;
	TRACE_B ("mv90_read", *return_length, buffer);
	return MV90_NORMAL;
}

/****
*
****/

MV90_STATUS mv90_shutdown (MV90_HANDLE user_handle)
{
	LEVEL1_STATUS	status;
	HANDLE_DEF		*handle;

	TRACE ("mv90_shutdown, entry point");
	if (!user_handle) {
		ERROR_TRACE ("mv90_shutdown, NULL handle");
		return MV90_BAD_HANDLE;
	}
	handle = (HANDLE_DEF *) user_handle;
	status = level1_shutdown (handle -> lower_handle);
	if (LEVEL1_NORMAL != status) {
		ERROR_TRACE_I ("mv90_shutdown, level1_shutdown = %d", status);
		delete_handle (handle);
		return MV90_SHUTDOWN;
	}
	delete_handle (handle);
	return MV90_NORMAL;
}

/****
*
****/

MV90_STATUS mv90_exit (void)
{
	HANDLE_DEF		*handle;

	TRACE ("mv90_exit, stub");
	while (queue_next (&root_handle)) {
		TRACE ("mv90_exit, shutdown next handle");
		handle = (HANDLE_DEF *) queue_next (&root_handle);
		mv90_shutdown ((LEVEL1_HANDLE) handle);
	}
	return MV90_NORMAL;
}

/****
*
****/

static HANDLE_DEF *add_handle (char *name, MV90_CLASS mv90_class)
{
	HANDLE_DEF	*handle;

	handle = (HANDLE_DEF *) malloc (sizeof (HANDLE_DEF));
	if (!handle) {		
		ERROR_TRACE_I ("add_handle, malloc #1 = %d", sizeof (HANDLE_DEF));
		return (HANDLE_DEF *) NULL;
	}
	switch (mv90_class) {
	case CLASS_CONNECTED:
	case CLASS_ACCEPTED:
		handle -> read_buffer_size = ini_get_int (name
												, "MAX_FRAME"
												, MAX_FRAME_SIZE);
		handle -> read_buffer = (UCHAR *) malloc (handle -> read_buffer_size);
		if (!handle -> read_buffer) {
			ERROR_TRACE_I ("add_handle, malloc #2 = %d"
						  , handle -> read_buffer_size);
			delete_handle (handle);
			return (HANDLE_DEF *) NULL;
		}
		break;
	default:
		handle -> read_buffer_size = 0;
		handle -> read_buffer = (UCHAR *) NULL;
	}
	handle -> read_index = 0;
	handle -> shutdown_ready = FALSE;
	handle -> mv90_class = mv90_class;
	strcpy (handle -> name, name);
	handle -> mv90_class = mv90_class;
	queue_append (&root_handle, &handle -> link);
	return handle;
}

/****
*
****/

static void delete_handle (HANDLE_DEF *handle)
{
	TRACE ("delete_handle (ipc), entry point");
	if (handle -> read_buffer) {
		free (handle -> read_buffer);
	}
	queue_remove (&root_handle, &handle -> link);
	free (handle);
}

/****
*
****/

static int read_lower (HANDLE_DEF *handle)
{
	size_t		count;
	int			status;

 	status = level1_read (handle -> lower_handle
						, handle -> read_buffer_size - handle -> read_index
						, &handle -> read_buffer [handle -> read_index]
						, &count);
	if (LEVEL1_NORMAL != status) {
		TRACE_I ("read_lower, level1_read = %d", status);
		handle -> read_index = 0;
		handle -> shutdown_ready = TRUE;
		return MV90_SHUTDOWN;
	}
	if (LEVEL1_NO_DATA == status) {
		TRACE ("read_lower, no data");
		return MV90_NO_DATA;
	}
	handle -> read_index += count;
	if (handle -> read_index < 4) {
		TRACE_I ("read_lower, partial packet = %d", handle -> read_index);
		return MV90_NO_DATA;
	}
	switch (handle -> read_buffer [1]) {
	case METER_Y_COMMAND:
	case METER_X_COMMAND:
	case METER_Z_COMMAND:
		TRACE ("read_lower, METER_XYZ_COMMAND");
		if (handle -> read_index >= 15) {
			return MV90_NORMAL;
		}
		break;
	case METER_NACK:
	case METER_ACK:
	case METER_ESC:
		TRACE ("read_lower, METER_ACK/NAK/ESC");
		if (handle -> read_index >= 5) {
			return MV90_NORMAL;
		}
		break;
	case METER_BLOCK_ID:
		TRACE ("read_lower, METER_BLOCK_ID");
		if (handle -> read_index >= 267) {
			return MV90_NORMAL;
		}
		break;
	}
	return MV90_NO_DATA;
}
