/*dbScan.c,v 1.1.1.1 1999/08/03 14:42:34 tis4000 Exp */

/* dbScan.c */
/* share/src/db  @(#)dbScan.c	1.11     9/20/91 */

/* tasks and subroutines to scan the database */
/*
 *      Author:          Bob Dalesio
 *      Date:            11-30-88
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	02-24-89	lrd	modified for vxWorks 4.0
 *				changed spawn to taskSpawn and changed args
 * .02	02-24-89	lrd	moved task info into task_params.h
 *				linked to vw for vx include files
 * .03	02-28-89	lrd	added check in build_scan_lists to verify the
 *				database section is loaded
 * .04	03-14-89	lrd	change task id's to int
 * .05	03-23-89	lrd	add restart logic
 * .06	03-27-89	lrd	modified to use the dbcommon structure
 *				deleted subroutine set_alarm
 * .07  04-05-89	lrd     added flag to signal drivers when we're ready to
 *                              accept events
 * .08	05-03-89	lrd	removed process mask from dbProcess calls
 * .09	05-25-89	lrd	added support for the PID record
 * .10	06-06-89	lrd	added support for the SEL record
 * .11  06-28-89	lrd	only give undefined event msg on the first 
 *				occurance of the event - keep a list of the
 *				undefined ones
 * .10	07-21-89	lrd	added support for the COMPRESS record
 * .11	01-25-90	lrd	added support for SUB records
 * .12	04-05-90	lrd	added the callback output task
 * .13  05-22-89        mrk     periodic scan now scans at correct rate
 * .14	08-08-90	lrd	removed T_AI from initialization at start up
 *				only outputs need to be read at initialization
 * .15	08-30-90	lrd	renamed the interrupt scanner wakeup from 
 *				intr_event_poster to io_scanner_wakeup
 * .16  09-14-90	mrk	changed for new record/device support
 * .17  10-24-90	mrk	replaced momentary task by general purpose callback task
 * .18	03-30-91	joh	updated for V5 vxWorks
 * .19  03-09-94	ems	added new periodic scan times, added scan_types.h for
 *				definitions of scan types
 * .20	11-02-94	ems	1-redefined event structures
 *				2-rewrote eventTask()
 *				3-scan_init(), delete init of event lock, no such animal now
 *				4-remove_scan_tasks(), if EVENT, td the task and free
 *				  any associated memory
 *				5-rewrite add_to_event_list()
 *				6-rewrite delete_from_event_list()
 *				7-added stdioLib.h
 *				8-rewrite print_event_lists()
 */

/*
 * Code Portions:
 *
 *    periodicScanTask	Task which processes the periodic records
 *    ioEventTask	Task which processes records on I/O interrupt
 *    eventTask		Task which processes records on global events
 *    wdScanTask	Task which restarts other tasks when suspended
 *    callbackTask	General purpose callback task
 *    callbackRequest   request callback
 *	arg
 *		pointer to arbitrary structure with first element being
 *		address of callback routine. When called addr of
 *		structure is passed.
 *    scan_init		Build the scan lists and start the tasks
 *    remove_scan_tasks	Delete scan tasks, free any malloc'd memory
 *	args
 *		tasks		flags which tasks to remove
 *    build_scan_lists	Looks through database and builds scan lists
 *	args
 *		lists		flags which lists to build
 *    intialize_ring_buffers	Initializes the ring buffers for the events
 *	args
 *		lists		flags which ring buffers to initialize
 *    start_scan_tasks	Start the scan tasks
 *	args
 *		tasks		flags which scan tasks to start
 *    add_to_scan_list	Adds an entry to a scan list
 *	args
 *		paddr		pointer to this record
 *		lists		which lists to build
 *    add_to_periodic_list	Add an entry to the periodic scan list
 *	args
 *		paddr		pointer to this record
 *		phase		phase on which this record is scanned
 *		list_index	which periodic list to place this record on
 *	returns
 *		0		successful
 *		-1		failed
 *    add_to_io_event_list	Add an entry to the I/O event list
 *	args
 *		paddr		pointer to this record
 *		phase		phase on which this record is scanned
 *		io_type		I/O type of this record
 *		card_type	card type of this records input
 *		card_number	card number of this records input
 *	returns
 *		0		successful
 *		-1		failed
 *    add_to_event_list	Add an entry to the global event scanner
 *	args
 *		paddr		pointer to this record
 *		phase		phase on which this record is scanned
 *		event		event on which this record will be scanned
 *	returns
 *		0		successful
 *		-1		failed
 *    delete_from_scan_list	Delete an entry from a scan list
 *	args
 *		paddr		pointer to this record
 *	returns
 *		0		successful
 *		-1		failed
 *
 *    delete_from_periodic_list	Delete an entry from the periodic scan list
 *	args
 *		paddr		pointer to this record
 *		list_index	which periodic list to remove this record from
 *	returns
 *		0		successful
 *		-1		failed
 *    delete_from_io_event_list	Delete an entry from the I/O event list
 *	args
 *		paddr		pointer to this record
 *		phase		phase on which this record is scanned
 *		io_type		i/o type of this record
 *		card_type	type of this records io card
 *		card_number	card number of this record
 *	returns
 *		0		successful
 *		-1		failed
 *    delete_from_event_list	Delete an entry from the global event list
 *	args
 *		paddr		pointer to this record
 *		phase		phase on which this record is scanned
 *		event		event on which this record is scanned
 *	returns
 *		0		successful
 *		-1		failed
 *    get_io_info		get io_type,card_type,card_number
 *	args
 *		paddr
 *		&pio_type
 *		&card_type
 *		&card_number
 *    print_lists		Prints the periodic scan lists
 *    print_io_event_lists	Print the I/O event lists
 *    print_event_lists		Print the event lists
 *
 *    io_scanner_wakeup		Post an I/O event
 *	args
 *		io_type		io type from which the event has come
 *		card_type	card type from which the event has come
 *		card_number	card number from which the event has come
 *
 *    post_event		Post an event
 *	args
 *		event		event number
 */

#include	<vxWorks.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<sysLib.h>
#include	<semLib.h>          /* library for semaphore support */
#include 	<rngLib.h>          /* library for ring buffer support */
#include	<wdLib.h>
#include	<logLib.h>
#include	<tickLib.h>
#include	<ioLib.h>
#include	<usrLib.h>
#include	<string.h>
#include	<inetLib.h>

/* this is included, because it defines the func panic
 * used herein, it is really a system call
 * ems 3 sep 98
*/
#include	<net/unixLib.h>

#include	"dbDefs.h"
#include	"alarm.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbRecords.h"
#include	"devSup.h"
#include	"interruptEvent.h"
#include	"task_params.h"
/* #include	<module_types.h> */
#include	"fast_lock.h"
#include	"scanLock.h"
#include	"scan_types.h"
#include	"callbacks.h"
#include	"redque.h"
#include	"error.h"
#include	"dbScan.h"
#include	"rbMgt.h"
#include	"recGbl.h"

/* local routines */
void remove_scan_tasks (short tasks);
void build_scan_lists (short lists);
void initialize_ring_buffers (short lists);
void start_scan_tasks (short tasks);
void add_to_scan_list (struct dbAddr *paddr, short lists);

static int  add_to_periodic_list (struct dbAddr *paddr, short phase, short list_index);
static int  add_to_io_event_list (struct dbAddr *paddr, short phase,
				  short io_type, short card_type, short card_number);
static int  add_to_event_list (struct dbAddr *paddr, short phase, short event);
static int  delete_from_periodic_list(struct dbAddr *paddr, short list_index);
static int  delete_from_io_event_list (struct dbAddr *paddr, short phase,
				       short io_type, short card_type, short card_number);
static int  delete_from_event_list (struct dbAddr *paddr, short phase, short event);
static long get_io_info (struct dbAddr *paddr, short *pio_type,
			 short *pcard_type, short *pcard_number);
void print_list (void);
void print_io_event_lists (void);
int print_event_lists (void);

void post_event (short event_number);
void stallScanTasks (void);
void resumeScanTasks (void);

#ifndef TIME
typedef	unsigned long TIME;
#endif TIME

/* number of addr records in each set (malloc'ed at a time) */
#define	NUM_RECS_PER_SET	100
#define	NUM_SETS		40

/* define number events, used in various places */

#define	MC68000	1
#if CPU != MC68000
#define	MAX_EVENTS		100
#else
#define MAX_EVENTS		1
#endif

/************** periodic scan information *************/

/* db address struct: will contain the scan lists for all tasks */
struct scan_list{
	FAST_LOCK	lock;
	unsigned short	scan_type;
	unsigned short	num_records;
	unsigned short	num_scan;
	unsigned short	scan_index;
	unsigned short	set;
	unsigned short	index;
	unsigned int	diagnostic;
	struct scan_element	*pscan;
	struct scan_element	*psets[NUM_SETS];
};

#define SL60SECOND	0
#define SL30SECOND	1
#define SL15SECOND	2
#define SL10SECOND	3
#define SL5SECOND	4
#define	SL2SECOND	5
#define SLSECOND        6
#define SLHALFSECOND    7
#define SLFIFTHSECOND   8
#define SLTENTHSECOND   9
#define	NUM_LISTS	(SLTENTHSECOND+1)

struct scan_element{
	struct dbAddr	dbAddr;
	short		phase;
};

#define	PERIOD_BASE_RATE	6	/* 60 ticks per second */

/************** I/O event driven scanner information **********/

#define IO_EVENT_PRIORITY	99

#if CPU != MC68000
#define	MAX_IO_EVENTS		10
#define	MAX_IO_EVENT_CHANS	32
#else
#define MAX_IO_EVENTS		1
#define MAX_IO_EVENT_CHANS	32
#endif

struct io_event_list{
	FAST_LOCK	lock;
	short	defined;
	short	io_type;
	short	card_type;
	short	card_number;
	short	number_defined;
	struct scan_element	element[MAX_IO_EVENT_CHANS];
};


/************** Event driven scanner information **************/

#define	EVENTS_PER_LIST		100

struct event_scan_element
{
    struct event_scan_element	*nextScanElement;
    short			phase;
    struct dbAddr		dbAddr;
};

struct event_element
{
    short			event_number;
    struct event_scan_element	*pFirstScan;
};

struct event_list
{
    struct event_list		*nextEventList;
    struct event_element	*pElement[EVENTS_PER_LIST];
};



/*  initialization field in database records */
#define	ARCH_INIT	2		/* first archive msg sent */



#define	IO_EVENT	0x01
#define	EVENT		0x02
#define	PERIODIC	0x04
#define	WDSCAN		0x08
#define	CALLBACK	0x10

int     fd;			/* used for the print list diagnostics */

/* pointer to the interruptable card arrays for each I/O type */
/* I/O types */
#define IO_AI		0
#define IO_AO		1
#define IO_BI		2
#define IO_BO		3
#define IO_SM   	4
#define IO_WF  		5
#define	IO_TIMER	6
#define MAX_IO_TYPE	IO_TIMER

/* see epicsH/fast_lock.h, the LOCKS are init'd in src/db/iocInit.c */
extern FAST_LOCK	LOCK_PERIODIC;
extern FAST_LOCK	LOCK_EVENT;
extern FAST_LOCK	LOCK_IOINTR;

/************* PERIODIC SCANNER GLOBALS *************/

/* periodic scan lists */
static struct scan_list	lists[NUM_LISTS];

/* number of periods in each scan list */
static short	periods_to_complete[] = {600,300,150,100,50,20,10,5,2,1};

static int	periodicScanTaskId = 0;


/************ I/O EVENT GLOBALS ***********/

/* I/O event scan list */
static struct io_event_list	io_event_lists[MAX_IO_EVENTS];
static struct io_event_list	io_events_undefined[MAX_IO_EVENTS];

/* semaphore on which the I/O scan task waits */
static SEM_ID ioEventSem;

/* ring buffer into which the drivers place the events which occured */
static RING_ID ioEventQ;

static int	ioEventTaskId = 0;


/************ EVENT GLOBALS ***************/

/* Event scan list */
static struct event_list	*pFirstEventList = 0;

/* semaphore on which the EVENT scan list */
static SEM_ID eventSem;

/* ring buffer into which the drivers place the events which occured */
static RING_ID eventQ;

static int	eventTaskId = 0;


/*********** WATCHDOG GLOBALS **************/

static int	wdScanTaskId = 0;
static int	wdScanOff = 0;
int		wdRestart=0;


/************ CALLBACK GLOBALS **************/

static SEM_ID callbackSem;
static RING_ID callbackQ;
static int	callbackTaskId = 0;

/* flag to the drivers that they can start to send events to the event tasks */
short    wakeup_init;


/*
 * periodicScanTask
 *
 * scan all periodic scan lists
 */
void periodicScanTask()
{
    register short		list,scan;
    register struct scan_list	*plist;
    short			disableListScan[NUM_LISTS];
    unsigned long		nextTimeToScan[NUM_LISTS];
    unsigned long		now;
    int				i;

    now = tickGet ();
    for(i=0; i < NUM_LISTS; i++){
	nextTimeToScan[i] = now;
	disableListScan[i] = FALSE;
    }

    while(1){
	/* Obtain global "stall the scan tasks" lock */
	FASTLOCK(&LOCK_PERIODIC);

	/* get the current tick value */
	now = tickGet();

	/* do for all records */
	plist = &lists[0];
	for (list = 0; list < NUM_LISTS; list++,plist++){
		if (plist->num_records == 0){
			continue;
		}
	  
		/*enable processing  list every periods_to_complete*/
		  if(nextTimeToScan[list] <= now ){
		   disableListScan[list] = FALSE;
		   nextTimeToScan[list] += periods_to_complete[list] * 6;
		}

	     	/*after processing all records in list , wait until enabled*/
	      	 if(disableListScan[list]) continue;

		/* lock the list */
		FASTLOCK(&plist->lock);

		/* process each set in each scan list */
		for (scan=0; scan < plist->num_scan; scan++){
			struct dbAddr	*paddr=&plist->pscan->dbAddr;
			struct dbCommon *precord=
				(struct dbCommon *)(paddr->precord);
			struct scanLock *lock;

			/* process this record if it has all its resources */

			lock = precord -> lsad;
			if (lock -> resGot == lock -> resDefined)
				{
				dbScanLock((caddr_t)precord);
				dbProcess(paddr);
				dbScanUnlock((caddr_t)precord);
				}

			/* get the pointer to the next record */
			plist->scan_index++;
			plist->index++;
			if (plist->scan_index >= plist->num_records){
				/*when all records in list processed*/
				/* disable  processing  list*/
				plist->index = 0;
				plist->set = 0;
				plist->scan_index = 0;
				plist->pscan = plist->psets[0];
				disableListScan[list] = TRUE;
				break;
			}else if (plist->index >= NUM_RECS_PER_SET){
				plist->index = 0;
				plist->set++;
				if (plist->set > NUM_SETS) plist->set = 0;
				plist->pscan = plist->psets[plist->set];
			}else{
				plist->pscan++;
			}
		}
		/* unlock it */
		FASTUNLOCK(&plist->lock);

	/*
	If there are any messages in the redundant update Queue, post
	the dbserver now.
	*/

	redQPost();
	}

	/* Free the lock for everyone else */
	FASTUNLOCK(&LOCK_PERIODIC);

	/* sleep 1 tick */
	taskDelay(1);
    }
}

/*
 * ioEventTask
 *
 * interrupt event scanner
 */
void ioEventTask()
{
    struct intr_event			intr_event;
    register short			event_index;
    register short			index;
    register short			found;
    register struct io_event_list	*pio_event_list;
    register struct scan_element	*pelement;
    register struct intr_event		*pintr_event = &intr_event;

    /* forever */
    FOREVER {
        /* wait for somebody to wake us up */
#	ifdef V5_vxWorks
       		semTake(ioEventSem, WAIT_FOREVER);
#	else
       		semTake(ioEventSem);
#	endif

	/* Obtain the global "stall the scan tasks" lock */
	FASTLOCK(&LOCK_IOINTR);

       /* process requests in the command ring buffer */
        while (rngNBytes(ioEventQ)>=INTR_EVENT_SZ){
                rngBufGet(ioEventQ, (char *)pintr_event, INTR_EVENT_SZ);
		/* find the event list */
		event_index = 0;
		pio_event_list = &io_event_lists[0];
		found = FALSE;
		while ((event_index < MAX_IO_EVENTS) && (!found)){
			if ((pio_event_list->defined)
			  && ((pio_event_list->io_type == pintr_event->io_type)
			  && (pio_event_list->card_type == pintr_event->card_type)
			  && (pio_event_list->card_number == pintr_event->card_number))){
				found = TRUE;
			}else{
		  	 	pio_event_list++;
				event_index++;
			}
		}
		if (!found){
			/* only give a message the first time an undefined  */
			/* event occurs                                     */
			pio_event_list = &io_events_undefined[0];
			found = FALSE;
			while ((event_index < MAX_IO_EVENTS) && (!found) && (pio_event_list->defined)){
				if ((pio_event_list->io_type == pintr_event->io_type)
				  && (pio_event_list->card_type == pintr_event->card_type)
				  && (pio_event_list->card_number == pintr_event->card_number)){
					found = TRUE;
				}else{
		  	 		pio_event_list++;
					event_index++;
				}
			}
			if (!found){
				logMsg("no event list for iotype: %d card type: %d card number: %d\n",
				  pintr_event->io_type,
				  pintr_event->card_type,
				  pintr_event->card_number,0,0,0);

				/* add this to the list */
				if (pio_event_list->defined == 0){
					pio_event_list->defined = 1;
					pio_event_list->io_type = pintr_event->io_type;
					pio_event_list->card_type = pintr_event->card_type;
					pio_event_list->card_number = pintr_event->card_number;
				}
			}
			continue;
		}

		/* process each record in the scan list */
		FASTLOCK(&pio_event_list->lock); /* lock it */
		index = 0;
		pelement = &pio_event_list->element[0];
		while (index < pio_event_list->number_defined){
			struct dbAddr	*paddr=&pelement->dbAddr;
			struct dbCommon *precord=
				(struct dbCommon *)(paddr->precord);
			struct	scanLock *lock;

			/* process this record if it has all its resources */
			lock = precord -> lsad;
			if (lock -> resGot == lock -> resDefined)
				{
				dbScanLock((caddr_t)precord);
				dbProcess(paddr);
				dbScanUnlock((caddr_t)precord);
				}

			pelement++;
			index++;
		}
		FASTUNLOCK(&pio_event_list->lock);		/* unlock it */
		redQPost();
	}
    /* Unlock for the rest of the world */
    FASTUNLOCK(&LOCK_IOINTR);
    }
}


/***********
	eventTask
***********/

void eventTask ()
{
    register int		eventIndex;
    short			eventNumber;
    struct event_list		*currList;

    for (;;)
    {
#	ifdef V5_vxWorks
	    semTake (eventSem, WAIT_FOREVER);
#	else
	    semTake (eventSem);
#	endif

	/* globally lock event scaning */
	FASTLOCK (&LOCK_EVENT);

	/* get event numbers from ring buffer,
	   see post_event() this file
	*/
	while (rngNBytes (eventQ) >= sizeof (short))
	{
	    rngBufGet (eventQ, (char *)&eventNumber, sizeof (short));

	    currList = pFirstEventList;

	    while (currList)
	    {
		for (eventIndex = 0; eventIndex < EVENTS_PER_LIST; eventIndex++)
		{
		    if (currList -> pElement[eventIndex])
		    {
			if (currList -> pElement[eventIndex] -> event_number == 
									eventNumber)
			{
			    struct dbCommon		*precord;
			    struct scanLock		*lock;
			    struct event_scan_element	*currScan;

			    currScan = currList -> pElement[eventIndex] -> pFirstScan;
			    while (currScan)
			    {
				/* process record if it has all its resources */
				precord = (struct dbCommon*)(currScan -> dbAddr.precord);
				lock    = precord -> lsad;
				if (lock -> resGot == lock -> resDefined)
				{
				    dbScanLock ((caddr_t)precord);
				    dbProcess (&currScan -> dbAddr);
				    dbScanUnlock ((caddr_t)precord);
				}
				currScan = currScan -> nextScanElement;
			    }
			}
		    }
		}
	    	currList = currList -> nextEventList;
	    }

	} /* end while stuff in rng buffer */

	/* unlock the global lock */
	FASTUNLOCK (&LOCK_EVENT);

	/* will post a semaphore to dbserver if its msgQ is not empty
	   see src/redund/redqlib.c
	*/
	redQPost ();

    } /* end for(;;) */
}


/*
 * wdScanTask
 *
 * watch dog scan task
 */
void wdScanTask()
{
    short	lists;

    while(1){

	/* verify the watchdog is desired */
	if (wdScanOff == 0){
		lists = 0;

		/* check the I/O Event Task */
		if (taskIsSuspended(ioEventTaskId)){
			logMsg("wdScanTask: Restarting ioEventTask\n",0,0,0,0,0,0);
			lists |= IO_EVENT;
		}

		/* check the Event Task */
		if (taskIsSuspended(eventTaskId)){
			logMsg("wdScanTask: Restarting eventTask\n",0,0,0,0,0,0);
			lists |= EVENT;
		}

		/* check the Periodic Scan Task */
		if (taskIsSuspended(periodicScanTaskId)){
			logMsg("wdScanTask: Restarting periodicScanTask\n",0,0,0,0,0,0);
			lists |= PERIODIC;
		}

		/* if any are suspended - restart them */
		if (lists && wdRestart){
			/* remove scan tasks */
			remove_scan_tasks(lists);

			/* build the scan lists */
			build_scan_lists(lists);

			/* create event ring buffers */
			initialize_ring_buffers(lists);

			/* Spawn scanner tasks */
			start_scan_tasks(lists);
		}
	}

	/* sleep awhile */
	taskDelay(WDSCAN_DELAY);
    }
}

/* General purpose callback task */
void callbackTask(){
    CALLBACK_PTR	pcallback;

    FOREVER {
	/* wait for somebody to wake us up */
#	ifdef V5_vxWorks
       	 	semTake (callbackSem, WAIT_FOREVER);
#	else
       	 	semTake (callbackSem);
#	endif

	/* process requests in the command ring buffer */
	while(rngNBytes(callbackQ)>=sizeof(pcallback)) {
		rngBufGet(callbackQ, (char *)&pcallback, sizeof(pcallback));
		(*pcallback->callback)(pcallback);
	}
    }
}

/* Routine which places requests into callback queue*/
void callbackRequest(pcallback)
    CALLBACK_PTR	pcallback;
{
        rngBufPut(callbackQ, (char *)&pcallback, sizeof(pcallback));
	semGive(callbackSem);
}

/*
 * SCAN_INIT
 *
 * scanner intialization code:
 *	builds scan lists
 *	spawns the scan and watchdog tasks
 */
void scan_init()
{
	int i;

	/* Initialize all the fast locks */
	for(i=0; i<NUM_LISTS; i++) {
		FASTLOCKINIT(&lists[i].lock);
	}
	for(i=0; i<MAX_IO_EVENTS; i++) {
		FASTLOCKINIT(&io_event_lists[i].lock);
	}

	/* remove scan tasks */
	remove_scan_tasks(EVENT | IO_EVENT | PERIODIC | CALLBACK);

	/* build the scan lists */
	build_scan_lists(EVENT | IO_EVENT | PERIODIC);

	/* create event ring buffers */
	initialize_ring_buffers(EVENT | IO_EVENT | CALLBACK);

	/* Spawn scanner tasks */
/*	start_scan_tasks(EVENT | IO_EVENT | PERIODIC | WDSCAN | CALLBACK);	*/

	start_scan_tasks(EVENT | IO_EVENT | PERIODIC | CALLBACK);

	/* let drivers know we're ready to accept events */
	wakeup_init = 1;

	return;
}

/*
 * REMOVE_SCAN_TASKS
 *
 * remove any tasks which are in existance
 */
void remove_scan_tasks(tasks)
register short	tasks;
{

    register short		set;
    register short		list_index;
    struct event_list		*currList, *nextList;
    struct event_scan_element	*currScan, *nextScan;
    register int		eventIndex;

    /* delete the ioEventTask if it is running */
    if (tasks & IO_EVENT)
    {
	if (ioEventTaskId)
	    td(ioEventTaskId);
	ioEventTaskId = 0;

	/* initialize the IO event lists */
	bfill ((char *)io_event_lists,sizeof(struct event_list)*MAX_IO_EVENTS,0);
    }

    /* delete the event task if it is running, also
       free any memory associated with the event lists
    */
    if (tasks & EVENT)
    {
	if (eventTaskId)
	{
	    td (eventTaskId);
	    eventTaskId = 0;
	}

	currList = pFirstEventList;
	while (currList)
	{
	    for (eventIndex = 0; eventIndex < EVENTS_PER_LIST; eventIndex++)
	    {
		if (currList -> pElement[eventIndex])
		{
		    currScan = currList -> pElement[eventIndex] -> pFirstScan;
		    while (currScan)
		    {
			nextScan = currScan -> nextScanElement;
			free ((void*)currScan);
			currScan = nextScan;
		    }
		    free ((void*)currList -> pElement[eventIndex]);
		}
	    }
	    nextList = currList -> nextEventList;
	    free ((void*)currList);
	    currList = nextList;
	}
	pFirstEventList = 0;
    }

    /* delete the callbackTask if it is running */
    if (tasks & CALLBACK)
    {
	if (callbackTaskId)
	    td (callbackTaskId);
	callbackTaskId = 0;
    }

    /* delete the periodicScanTask if it is running */
    if (tasks & PERIODIC)
    {
	if (periodicScanTaskId)
	{
	    td (periodicScanTaskId);
	    periodicScanTaskId = 0;
	    for(list_index = 0; list_index < NUM_LISTS; list_index++)
	    {
		for (set = 0; set < NUM_SETS; set++)
		{
		    if (lists[list_index].psets[set])
		    {
		  	free(lists[list_index].psets[set]);
		    }
		}
	    }
	}
	/* initialize the scan lists */
	bfill ((char *)lists,sizeof(lists),0);
    }

    /* delete the watchdog scan task */
    if (tasks & WDSCAN)
    {
	if (wdScanTaskId)
	{ 
	    td(wdScanTaskId);
	    wdScanTaskId = 0;
	}
    }
}

/*
 * BUILD_SCAN_LISTS
 *
 * build the specified scan lists
 */
void build_scan_lists(lists)
register short	lists;
{
	struct dbAddr		dbAddr;		/* database address */
	struct recLoc		*precLoc;
	struct dbCommon		*precord;	/* pointer to record	*/
	struct recChain		*pchain;
	int			i,j;

	if(dbRecords==NULL) {
	    errMessage(S_record_noRecords,
			"Error detected in build_scan_lists");
#ifdef vxWorks
	    taskSuspend (taskIdSelf());
#else
	    exit(1);
#endif vxWorks
	}
	/* look through all of the database records and place them on lists */
	for (i=0; i<dbRecords->number; i++) {
		if((precLoc=dbRecords->papRecLoc[i])==NULL) continue;
		for(j=0, pchain = (struct recChain *) precLoc -> pFirst;
		    j<precLoc->no_records;
		    j++, pchain = pchain -> next) {
			precord = (struct dbCommon *) (pchain -> pRecord);
			if(precord->name[0]==0) continue;
			/* get database address from access routines */
			dbAddr.precord = (caddr_t) precord;
			dbAddr.record_type = i;
				/* add it to a scan list */
				add_to_scan_list(&dbAddr,lists);
		}
	}
}

/*
 * INITIALIZE_RING_BUFFERS
 *
 * initialize the ring buffers for the event tasks
 */
void initialize_ring_buffers(lists)
register short	lists;
{
	/* create the interrupt event ring buffer and semaphore */
	if (lists & IO_EVENT){
		/* clear it if it is created */
		if (ioEventQ){
			rngFlush(ioEventQ);
		/* create it if it does not exist */
		}else if ((ioEventQ = rngCreate(INTR_EVENT_SZ * MAX_EVENTS))
		  == (RING_ID)NULL){
	 	      panic ("scan_init: ioEventQ not created\n");
		}
		ioEventSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
	}

	/* create the event ring buffer and semaphore */
	if (lists & EVENT){
		/* clear it if it is created */
		if (eventQ){
			rngFlush(eventQ);
		/* create it if it does not exist */
		}else if ((eventQ = rngCreate(sizeof(short) * MAX_EVENTS))
		  == (RING_ID)NULL){
	 	      panic ("scan_init: eventQ not created\n");
		}
		eventSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
	}

	/* create the callback ring buffer and semaphore */
	if (lists & CALLBACK){
		/* clear it if it is created */
		if (callbackQ){
			rngFlush(callbackQ);
		/* create it if it does not exist */
		}else if ((callbackQ = rngCreate(sizeof(struct callback *) * MAX_EVENTS))
		  == (RING_ID)NULL){
	 	      panic ("scan_init: callbackQ not created\n");
		}
		callbackSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
	}
}

/*
 * START_SCAN_TASKS
 *
 * start the specified scan tasks
 * parameters are from task_params.h
 */
void start_scan_tasks(tasks)
register short	tasks;
{

#if CPU != MC68000
	if (tasks & IO_EVENT){
		ioEventTaskId = 
		  taskSpawn(IOEVENTSCAN_NAME,
			    IOEVENTSCAN_PRI,
			    IOEVENTSCAN_OPT,
			    IOEVENTSCAN_STACK,
			    (FUNCPTR)ioEventTask, 0,0,0,0,0,0,0,0,0,0);
	}

	if (tasks & EVENT){
		eventTaskId = 
		  taskSpawn(EVENTSCAN_NAME,
			    EVENTSCAN_PRI,
			    EVENTSCAN_OPT,
			    EVENTSCAN_STACK,
			    (FUNCPTR)eventTask, 0,0,0,0,0,0,0,0,0,0);
	}
#endif

	if (tasks & PERIODIC){
		periodicScanTaskId = 
		  taskSpawn(PERIODSCAN_NAME,
			    PERIODSCAN_PRI,
			    PERIODSCAN_OPT,
			    PERIODSCAN_STACK,
			    (FUNCPTR)periodicScanTask, 0,0,0,0,0,0,0,0,0,0);
	}

	if (tasks & WDSCAN){
		wdScanTaskId = 
		  taskSpawn(WDSCAN_NAME,
			    WDSCAN_PRI,
			    WDSCAN_OPT,
			    WDSCAN_STACK,
			    (FUNCPTR)wdScanTask, 0,0,0,0,0,0,0,0,0,0);
	}

	if (tasks & CALLBACK){
		callbackTaskId = 
		  taskSpawn(CALLBACK_NAME,
			    CALLBACK_PRI,
			    CALLBACK_OPT,
			    CALLBACK_STACK,
			    (FUNCPTR)callbackTask, 0,0,0,0,0,0,0,0,0,0);
	}
}

/*
 * ADD_TO_SCAN_LIST
 *
 * add a new record's address to a scan list
 */
void add_to_scan_list(paddr,lists)
register struct	dbAddr *paddr;
register short		lists;
{
	struct dbCommon *precord= (struct dbCommon *)(paddr->precord);
	short		scan_type;
	short		phase;
	short		event;
	long		status = -1;

	/* get the list on which this record belongs */
	scan_type = precord->scan;
	phase = precord->phas;
	event = precord->evnt;
	/* add to the periodic scan or the event list */
	switch (scan_type){
	case (PASSIVE):
		break;

	case (P60SECOND):
		if (lists & PERIODIC)
			status = 
			  add_to_periodic_list(paddr,phase,SL60SECOND);
		else
			return;
		break;
		
	case (P30SECOND):
		if (lists & PERIODIC)
			status = 
			  add_to_periodic_list(paddr,phase,SL30SECOND);
		else
			return;
		break;
		
	case (P15SECOND):
		if (lists & PERIODIC)
			status = 
			  add_to_periodic_list(paddr,phase,SL15SECOND);
		else
			return;
		break;
		
	case (P10SECOND):
		if (lists & PERIODIC)
			status = 
			  add_to_periodic_list(paddr,phase,SL10SECOND);
		else
			return;
		break;
		
	case (P5SECOND):
		if (lists & PERIODIC)
			status = 
			  add_to_periodic_list(paddr,phase,SL5SECOND);
		else
			return;
		break;
		
	case (P2SECOND):
		if (lists & PERIODIC)
			status = 
			  add_to_periodic_list(paddr,phase,SL2SECOND);
		else
			return;
		break;
		
	case (PSECOND):
		if (lists & PERIODIC)
			status =
			  add_to_periodic_list(paddr,phase,SLSECOND);
		else
			return;
		break;

	case (PHALFSECOND):
		if (lists & PERIODIC)
			status =
			  add_to_periodic_list(paddr,phase,SLHALFSECOND);
		else
			return;
		break;

	case (PFIFTHSECOND):
		if (lists & PERIODIC)
			status =
			  add_to_periodic_list(paddr,phase,SLFIFTHSECOND);
		else
			return;
		break;

	case (PTENTHSECOND):
		if (lists & PERIODIC)
			status =
			  add_to_periodic_list(paddr,phase,SLTENTHSECOND);
		else
			return;
		break;

	case (E_IO_INTERRUPT): 
		if (lists & IO_EVENT) {
			short		io_type;
			short		card_type;
			short		card_number;

			status=get_io_info(paddr,&io_type,
				&card_type,&card_number);
			if(status!=0) {
			    recGblDbaddrError(status,paddr,
				"dbScan(get_io_info)");
			    return;
			}
			status = add_to_io_event_list(paddr,phase,
				io_type,card_type,card_number);
		} else return;
		break;

	case (E_EXTERNAL):
		if (lists & EVENT)
			status =
			  add_to_event_list(paddr,phase,event);
		else
			return;
		break;

	default:
		return;
	}

	if (status < 0 && (precord->nsev<MAJOR_ALARM)){
		precord->nsta = SCAN_ALARM;
		precord->nsev = MAJOR_ALARM;
	}
	return;
}

/*
 * ADD_TO_PERIODIC_LIST
 *
 * add a new record's address to a periodic scan list
 */
static int add_to_periodic_list(paddr,phase,list_index)
register struct	dbAddr *paddr;
short			phase;
short			list_index;
{
	register short			set,index,found;
	register struct scan_element	*pspare_element;
	register struct scan_element	*ptest_element;

	/* lock the list during modification */
	FASTLOCK(&lists[list_index].lock); /* lock it */

	/* determine if we need to create a new set on this scan list */
	set = lists[list_index].num_records / NUM_RECS_PER_SET;
	if ((lists[list_index].num_records % NUM_RECS_PER_SET) == 0){
		if (set >= NUM_SETS){
			FASTUNLOCK(&lists[list_index].lock); /* unlock it */
			return(-1);
		}
		if ((lists[list_index].psets[set] = 
		  (struct scan_element *)malloc(NUM_RECS_PER_SET*sizeof(struct scan_element))) == 0){
			FASTUNLOCK(&lists[list_index].lock); /* unlock it */
			return(-1);
		}
		bfill ((char *)lists[list_index].psets[set],NUM_RECS_PER_SET*sizeof(struct scan_element),0);
	}

	/* pointer to the slot at the end of the scan list */
	index = lists[list_index].num_records % NUM_RECS_PER_SET;
	pspare_element = lists[list_index].psets[set];
	pspare_element += index;

	/* pointer to the last used slot in the scan list */
	index--;
	if (index < 0){
		index = NUM_RECS_PER_SET - 1;
		set--;
		ptest_element = lists[list_index].psets[set];
		ptest_element += index;
	}else{
		ptest_element = pspare_element - 1;
	}

	/* determine where this record belongs */
	found = FALSE;
	while(!found){
		/* beginning of list or test record has higher or equal phase */
		if ((set < 0) || (ptest_element->phase <= phase)){
			/* put the new record into the spare space */
			pspare_element->dbAddr = *paddr;
			pspare_element->phase = phase;
			found = TRUE;
		/* move the tested record into the spare record space */
		}else{
			*pspare_element = *ptest_element;
			pspare_element = ptest_element;
		}

		/* find the next test element */
		index--;
		if (index < 0){
			index = NUM_RECS_PER_SET - 1;
			set--;
			ptest_element = lists[list_index].psets[set];
			ptest_element += index;
		}else{
			ptest_element--;
		}
	}

	/* adjust the list information */
	lists[list_index].num_records++;
	lists[list_index].num_scan = 
	  ((lists[list_index].num_records - 1) / periods_to_complete[list_index]) + 1;
	lists[list_index].pscan = lists[list_index].psets[0];
	lists[list_index].scan_index = 0;
	lists[list_index].index = 0;
	lists[list_index].set = 0;

	/* unlock it */
	FASTUNLOCK(&lists[list_index].lock);

	return(0);
}

/*
 * ADD_TO_IO_EVENT_LIST
 *
 * add a new record's address to an interrupt event scan list
 */
static int add_to_io_event_list(paddr,phase,io_type,card_type,card_number)
register struct	dbAddr *paddr;
short			phase;
short			io_type;
short			card_type;
short			card_number;
{
	struct io_event_list	*pio_event_list;
	struct scan_element	*pelement;
	struct scan_element	*pspare;
	register short		event_index;
	register short		index;
	register short		found;

	/* verify this card has an interrupt */
/* DISABLED FOR NOW!!! XXXXX
	if (!pinterruptable[io_type][card_type])
		return(-1);
*/
	/* is there a list for this iotype and card type */
	pio_event_list = &io_event_lists[0];
	event_index = 0;
	found = FALSE;
	while ((event_index < MAX_IO_EVENTS) && (!found)){
		if ((pio_event_list->defined)
		  && ((pio_event_list->io_type == io_type)
		  && (pio_event_list->card_type == card_type)
		  && (pio_event_list->card_number == card_number))){
			found = TRUE;
		}else{
	  	 	pio_event_list++;
			event_index++;
		}

	}

	/* look for a spare list */
	if (!found){
		pio_event_list = &io_event_lists[0];
		event_index = 0;
		found = FALSE;
		while ((event_index < MAX_IO_EVENTS) && (!found)){
			if (!pio_event_list->defined){
				found = TRUE;
			}else{
	  	 		pio_event_list++;
				event_index++;
			}
		}
		if (found){
			pio_event_list->defined = TRUE;
			pio_event_list->io_type = io_type;
			pio_event_list->card_type = card_type;
			pio_event_list->card_number = card_number;
		}else{
			return(-1);
		}
	}

	/* check for available space */
	if (pio_event_list->number_defined >= MAX_IO_EVENT_CHANS){
		return(-1);
	}

	/* lock the list during modification */
	FASTLOCK(&pio_event_list->lock);

	/* add this record to the list */
	if (pio_event_list->number_defined == 0){
		pio_event_list->element[0].dbAddr = *paddr;
		pio_event_list->element[0].phase = phase;
	}else{
		index = pio_event_list->number_defined;
		pspare = &pio_event_list->element[index];
		pelement = &pio_event_list->element[index-1];
		while ((pelement->phase > phase) && (index > 0)){
			*pspare = *pelement;
			pelement--;
			pspare--;
			index--;
		}
		pspare->phase = phase;
		pspare->dbAddr = *paddr;
	}
	pio_event_list->number_defined++;

	/* unlock the list */
	FASTUNLOCK(&pio_event_list->lock);

	return(0);
}

/**********
	ADD_TO_EVENT_LIST

	add a new record to the event list
	called from add_to_scan_list this source and
	is called from .../src/db/dbAccess.c:dbPut()
	NOTE: stallScanTasks which locks the list is 
		called by dbPutField which calls dbPut
		it then calls resumeScanTask to unlock 
		the list
**********/

static int add_to_event_list (paddr, phase, event)
    register struct dbAddr	*paddr;
    short			phase;
    short			event;
{
    struct event_list		*currList, *newList;
    struct event_element	*newEvent;
    struct event_scan_element	*currScan, *prevScan, *newScan;
    register int		eventIndex;
    register int		found_event;
    register int		first_open_element;

    /* get 1st event list */
    currList = pFirstEventList;
    first_open_element = -1;
    found_event = -1;

    while (currList) 
    {
	first_open_element = -1;
	for (eventIndex = 0; eventIndex < EVENTS_PER_LIST; eventIndex++)
	{
	    if (currList -> pElement[eventIndex])
	    {
		if (currList -> pElement[eventIndex] -> event_number == event)
		{
		    found_event = eventIndex;
		    first_open_element = eventIndex;
		    break;
		}
	    }
	    else
	    {
		if (first_open_element < 0)
		    first_open_element = eventIndex;
	    }
	}
	/* if nextEventList is 0, break now so we have a valid
	   ptr to currList for later use
	*/
	if ( (found_event >= 0) || (currList -> nextEventList == 0) )
	    break;
	currList = currList -> nextEventList;
    }

    /* did not find the event number in the list */
    if (found_event < 0)
    {
	/* also did not have a place to put a new event
	   so allocate a new list, init 1st list if needed
	*/
	if (first_open_element < 0)
	{
	    if ((newList = (struct event_list*)malloc (sizeof (struct event_list))) == NULL)
	    {
		logMsg ("add_to_event_list, malloc event_list failed\n",0,0,0,0,0,0);
		return (-1);
	    }
	    newList -> nextEventList = 0;
	    bzero ((char *)(newList -> pElement), sizeof (newList -> pElement));
	    if (pFirstEventList == 0)
		pFirstEventList = newList;
	    else
		currList -> nextEventList = newList;
	    currList = newList;
	    first_open_element = 0;
	}

	/* make a new event */
	if ((newEvent = (struct event_element*)malloc (sizeof (struct event_element))) == NULL)
	{
	    logMsg ("add_to_event_list, malloc event_element failed\n",0,0,0,0,0,0);
	    return (-1);
	}
	newEvent -> event_number = event;
	newEvent -> pFirstScan   = 0;
	currList -> pElement[first_open_element] = newEvent;
    }

    /* make a new scan element to add to event list */
    if ((newScan = (struct event_scan_element *)malloc (sizeof (struct event_scan_element))) == NULL)
    {
	logMsg ("add_to_event_list, malloc event_scan_element failed\n",0,0,0,0,0,0);
	return (-1);
    }
    newScan -> nextScanElement = 0;
    newScan -> phase = phase;
    newScan -> dbAddr = *paddr;

    /* if no scan elements for this event yet just add it */
    if (currList -> pElement[first_open_element] -> pFirstScan == 0)
	currList -> pElement[first_open_element] -> pFirstScan = newScan;

    /* else put it in decreasing phase order */
    else
    {
	currScan = prevScan = currList -> pElement[first_open_element] -> pFirstScan;
	while (currScan)
	{
	    if (newScan -> phase >= currScan -> phase)
		break;
	    prevScan = currScan;
	    currScan = currScan -> nextScanElement;
	}
	/* adding it to end of scan list */
	if (currScan == 0)
	    prevScan -> nextScanElement = newScan;

	/* adding it to head of scan list, must reset head of list
	   with new value
	*/
	else if (prevScan == currScan)
	{
	    currList -> pElement[first_open_element] -> pFirstScan = newScan;
	    newScan -> nextScanElement = currScan;
	}

	/* add it inside list by reseting forward links */
	else
	{
	    prevScan -> nextScanElement = newScan;
	    newScan -> nextScanElement = currScan;
	}
    }

    return (0);
}

/*
 * DELETE_FROM_SCAN_LIST
 *
 * delete a record's address from a scan list
 */
int delete_from_scan_list(paddr)
register struct	dbAddr *paddr;
{
	struct dbCommon *precord= (struct dbCommon *)(paddr->precord);
	short		scan_type;
	short		phase;
	short		event;
	register short	status;

	/* get the list on which this record belongs */
	scan_type = precord->scan;
	phase = precord->phas;
	event = precord->evnt;
	/* add to the periodic scan or the event list */
	switch (scan_type){
	case (PASSIVE):
		return(-1);
	case (P60SECOND):
		return(delete_from_periodic_list(paddr,SL60SECOND));
	case (P30SECOND):
		return(delete_from_periodic_list(paddr,SL30SECOND));
	case (P15SECOND):
		return(delete_from_periodic_list(paddr,SL15SECOND));
	case (P10SECOND):
		return(delete_from_periodic_list(paddr,SL10SECOND));
	case (P5SECOND):
		return(delete_from_periodic_list(paddr,SL5SECOND));
	case (P2SECOND):
		return(delete_from_periodic_list(paddr,SL2SECOND));
	case (PSECOND):
		return(delete_from_periodic_list(paddr,SLSECOND));
	case (PHALFSECOND):
		return(delete_from_periodic_list(paddr,SLHALFSECOND));
	case (PFIFTHSECOND):
		return(delete_from_periodic_list(paddr,SLFIFTHSECOND));
	case (PTENTHSECOND):
		return(delete_from_periodic_list(paddr,SLTENTHSECOND));
	case (E_IO_INTERRUPT): {
		short		io_type;
		short		card_type;
		short		card_number;

		status=get_io_info(paddr,&io_type,
			&card_type,&card_number);
		if(status!=0) {
		    recGblDbaddrError(status,paddr,
			"dbScan(get_io_info)");
		    return(-1);
		}
		return(delete_from_io_event_list(paddr,phase,io_type,
			card_type,card_number));
	    }
	case (E_EXTERNAL):
		return(delete_from_event_list(paddr,phase,event));
	default:
		return(-1);
	}
}

/*
 * DELETE_FROM_PERIODIC_LIST
 *
 * delete a record's address from a periodic scan list
 */
static int delete_from_periodic_list(paddr,list_index)
register struct	dbAddr *paddr;
short			list_index;
{
	register short			set,index,found;
	register struct scan_element	*pelement;
	register struct scan_element	*pnext_element;
	short				end_of_list;

	/* lock the list during modification */
	FASTLOCK(&lists[list_index].lock);

	/* find this record */
	found = FALSE;
	set = 0;
	index = 0;
	end_of_list = FALSE;
	pelement = lists[list_index].psets[set];
	while((!found) && (!end_of_list)){
		if (pelement->dbAddr.precord == paddr->precord){
			found = TRUE;
		}else{
			index++;
			if (index >= NUM_RECS_PER_SET){
				index = 0;
				set++;
				pelement = lists[list_index].psets[set];
				if (set >= NUM_SETS) end_of_list = TRUE;
			}else{
				pelement++;
			}
		}
	}
	if (!found){
		FASTUNLOCK(&lists[list_index].lock);	/* unlock it */
		return(-1);
	}

	/* move each record in the scan list down */
	while (!end_of_list){
		index++;
		if (index >= NUM_RECS_PER_SET){
			index = 0;
			set++;
			if (set >= NUM_SETS){
				end_of_list = TRUE;
			}else{
				pnext_element=lists[list_index].psets[set];
				if (!pnext_element)
					end_of_list = TRUE;
			}
		}else{
			pnext_element = pelement + 1;
		}

		if (!end_of_list){
			*pelement = *pnext_element;
			pelement = pnext_element;
		}else{
			bfill ((char *)pelement,sizeof(struct scan_element),0);
		}
	}

	/* adjust the list information */
	lists[list_index].num_records--;
	lists[list_index].num_scan = 
	  (lists[list_index].num_records/periods_to_complete[list_index])+1;

	/* determine if we need to free the memory for a set */
	if ((lists[list_index].num_records % NUM_RECS_PER_SET) == 0){
		set = lists[list_index].num_records / NUM_RECS_PER_SET;
		free(lists[list_index].psets[set]);
		lists[list_index].psets[set] = 0;
	}

	/* unlock the scan list */
	FASTUNLOCK(&lists[list_index].lock);

	return(0);
}

/*
 * DELETE_FROM_IO_EVENT_LIST
 *
 * delete a record's address to an interrupt event scan list
 */
static int delete_from_io_event_list(paddr,phase,io_type,card_type,card_number)
register struct	dbAddr *paddr;
short			phase;
short			io_type;
short			card_type;
short			card_number;
{
	struct io_event_list	*pio_event_list;
	struct scan_element	*pelement;
	struct scan_element	*pnext_element;
	register short		event_index;
	register short		index;
	register short		found;

	/* is there a list for this iotype,card type and card number */
	pio_event_list = &io_event_lists[0];
	event_index = 0;
	found = FALSE;
	while ((event_index < MAX_IO_EVENTS) && (!found)){
		if ((pio_event_list->defined)
		  && ((pio_event_list->io_type == io_type)
		  && (pio_event_list->card_type == card_type)
		  && (pio_event_list->card_number == card_number))){
			found = TRUE;
		}else{
	  	 	pio_event_list++;
			event_index++;
		}

	}
	if (!found) return(-1);

	/* lock the list during modification */
	FASTLOCK(&pio_event_list->lock);

	/* find this record in the list */
	index = 0;
	pelement = &pio_event_list->element[0];
	found = FALSE;
	while ((!found)
	  && (index < pio_event_list->number_defined)){
		if (pelement->dbAddr.precord == paddr->precord){
			found = TRUE;
		}else{
			pelement++;
			index++;
		}
	}
	if (!found){
		FASTUNLOCK(&pio_event_list->lock); /* unlock it */
		return(-1);
	}

	/* delete this record from the list */
	index++;
	pnext_element = &pio_event_list->element[index];
	while (index < pio_event_list->number_defined) {
		*pelement = *pnext_element;
		pelement++;
		pnext_element++;
		index++;
	}
	bfill ((char *)pelement,sizeof(struct scan_element),0);
	pio_event_list->number_defined--;

	/* if this is the last scan element in the list - free it */
	if (pio_event_list->number_defined == 0){
		pio_event_list->defined = FALSE;
		pio_event_list->io_type = 0;
		pio_event_list->card_type = 0;
		pio_event_list->card_number = 0;
	}

	/* unlock the list */
	FASTUNLOCK(&pio_event_list->lock);

	return(0);
}


/**********
	DELETE_FROM_EVENT_LIST

	delete a record from an event list
	called from delete_from_scan_list this source
	which is called from .../src/db/dbAccess.c:dbPut()
	NOTE: stallScanTasks which locks the list is 
		called by dbPutField which calls dbPut
		it then calls resumeScanTask to unlock 
		the list
**********/

static int delete_from_event_list (paddr, phase, event)
    register struct dbAddr	*paddr;
    short			phase;
    short			event;
{
    struct event_list		*currList;
    struct event_scan_element	*currScan, *prevScan;
    register int		eventIndex;
    register int		found_event;
    register int		found;

    currList = pFirstEventList;
    found_event = -1;

    while (currList) 
    {
	for (eventIndex = 0; eventIndex < EVENTS_PER_LIST; eventIndex++)
	{
	    if (currList -> pElement[eventIndex])
	    {
		if (currList -> pElement[eventIndex] -> event_number == event)
		{
		    found_event = eventIndex;
		    break;
		}
	    }
	}
	/* if nextEventList is 0, break now so we have a valid
	   ptr to currList for later use
	*/
	if ( (found_event >= 0) || (currList -> nextEventList == 0) )
	    break;
	currList = currList -> nextEventList;
    }
    if (found_event < 0)
	return (-1);

    /* look for record in list */
    found = 0;
    prevScan = currScan = currList -> pElement[found_event] -> pFirstScan;
    while (currScan)
    {
	if (currScan -> dbAddr.precord == paddr -> precord)
	{
	    found = 1;
	    break;
	}
	prevScan = currScan;
	currScan = currScan -> nextScanElement;
    }
    if (!found)
	return (-1);

    /* we are looking at the first in the list, check if only
       1 in the list, if so 0 out pFirstScan, else update
       pFirstScan with new value
    */
    if (prevScan == currScan)
    {
	if (currScan -> nextScanElement == 0)
	    currList -> pElement[found_event] -> pFirstScan = 0;
	else
	    currList -> pElement[found_event] -> pFirstScan = 
					currScan -> nextScanElement;
    }
    else
	prevScan -> nextScanElement = currScan -> nextScanElement;
    free ((void*)currScan);

    return (0);
}

/* 
 * GET_IO INFO
 *
 * get the io_type,card_type, and card_number
 */
static long get_io_info(paddr,pio_type,pcard_type,pcard_number)
struct dbAddr	*paddr;
short		*pio_type;
short		*pcard_type;
short		*pcard_number;
{
	struct dbCommon *precord=(struct dbCommon *)(paddr->precord);
	long		status;

	if((precord->dset == NULL) ||(precord->dset->get_ioint_info == NULL ))
		return(S_dev_noDevSup);
	status=(*precord->dset->get_ioint_info)
		(precord,pio_type,pcard_type,pcard_number);
	return(status);
}

/*
 * PRINT_LIST
 *
 * print the scan lists
 */
void print_list()
{
	struct scan_element	*pelement;
	struct scan_list	*plist;
	register short		set,num_records,index;
	short			list;
	char input[5];

	fd = STD_IN;

	/* print each scan list */
	printf("\n");
	plist = &lists[0];
	for (list = 0; list < NUM_LISTS; list++,plist++){
		if (plist->num_records == 0){
			printf("list %d is empty\n",list);
			continue;
		}
		if (plist->num_records >= NUM_RECS_PER_SET * NUM_SETS){
			printf("list %d is full\n",list);
		}else{
			printf("list %d has %d entries\n",list,plist->num_records);
		}

		/* print each set in each scan list */
		num_records = 0;
		for (set=0; (set<NUM_SETS) && (num_records<plist->num_records); set++){
			printf("\tset %d\n",set);

			/* print each element in each set */
			pelement = plist->psets[set];
			for (index=0; (index<NUM_RECS_PER_SET) && (num_records<plist->num_records); index++,pelement++){
				num_records++;
				printf("\t\t%s\n",pelement->dbAddr.precord);
			}
		}

		/* wait for a character to print the next set */
		read(fd,input,1);
	}
}

/*
 * PRINT_IO_EVENT_LISTS
 *
 * print the io event lists
 */
void print_io_event_lists()
{
	struct io_event_list	*plist;
	register short		num_records,index;
	short			list;
	char input[5];
	struct scan_element	*pelement;

	fd = STD_IN;

	/* print each scan list */
	printf("\n");
	plist = &io_event_lists[0];
	for (list = 0; list < MAX_IO_EVENTS; list++,plist++){
		if (plist->defined == 0){
			printf("event list %d is unused\n",list);
			continue;
		}
		if (plist->number_defined >= MAX_IO_EVENT_CHANS){
			printf("event list %d is full\n",list);
			printf("\tio type: %d  card type: %d card number: %d\n",
			  plist->io_type,plist->card_type,plist->card_number);
		}else{
			printf("event list %d has %d entries\n",list,plist->number_defined);
			printf("\tio type: %d  card type: %d card number: %d\n",
			  plist->io_type,plist->card_type,plist->card_number);
		}

		/* print each set in each scan list */
		num_records = 0;
		pelement = &plist->element[0];
		for (index=0; (index<MAX_IO_EVENT_CHANS) && (num_records<plist->number_defined); index++,pelement++){
			num_records++;
			if ((pelement->dbAddr.precord[0] >= 'A') && (pelement->dbAddr.precord[0] <= 'z'))
				printf("\t\t%s\n",pelement->dbAddr.precord);
			else
				printf("\t\tnot printable\n");
		}

		/* wait for a character to print the next set */
		read(fd,input,1);
	}
}

/***************
	PRINT_EVENT_LISTS

***************/

int print_event_lists ()
{
    struct event_list		*currList;
    struct event_scan_element	*currScan;
    int				eventIndex;
    int				fflag;
    char			input[5];
    char			fname[80];
    FILE			*fp = 0;

    fd = STD_IN;

    printf ("PUT OUTPUT LISTING IN FILE ??? (y|n) : ");
    scanf ("%s", input);
    if (input[0] == 'y')
	fflag = 1;
    else
	fflag = 0;

    if (fflag)
    {
	printf ("ENTER FULL PATHANME OF OUTPUT LISTING FILE : ");
	scanf ("%s", fname);
	printf ("open %s for APPENDING!!!", fname);
	if ((fp = fopen (fname, "a")) == NULL)
	{
	    printf ("print_event_lists: fopen error = %d\nfile = %s\n",
			errno, fname);
	    return (-1);
	}
    }
    if (fflag)
	fprintf (fp, "\nEVENT LIST PRINT\n");
    else
	printf ("\nEVENT LIST PRINT\n");
    if (pFirstEventList == 0)
    {
	if (fflag)
	{
	    fprintf (fp, "There are no event lists\n");
	    fclose (fp);
	}
	else
	    printf ("There are no event lists\n");
	return (0);
    }
    currList = pFirstEventList;
    while (currList)
    {
	for (eventIndex = 0; eventIndex < EVENTS_PER_LIST; eventIndex++)
	{
	    if (currList -> pElement[eventIndex])
	    {
		if (fflag)
		    fprintf (fp, "Event list for event %d, index = %d\n",
				currList -> pElement[eventIndex] -> event_number,
				eventIndex);
		else
		    printf ("Event list for event %d, index = %d\n",
				currList -> pElement[eventIndex] -> event_number,
				eventIndex);
		currScan = currList -> pElement[eventIndex] -> pFirstScan;
		if (currScan)
		{
		    while (currScan)
		    {
			if ( (currScan -> dbAddr.precord[0] >= 'A') &&
			     (currScan -> dbAddr.precord[0] <= 'z') )
			{
			    if (fflag)
				fprintf (fp, "\t%.28s, phase = %d\n",
						currScan -> dbAddr.precord,
						currScan -> phase);
			    else
				printf ("\t%.28s, phase = %d\n",
						currScan -> dbAddr.precord,
						currScan -> phase);
			}
			else
			{
			    if (fflag)
				fprintf (fp, "\telement not printable\n");
			    else
				printf ("\telement not printable\n");
			}
			currScan = currScan -> nextScanElement;
		    }
		    /* wait for a char to proceed to next event list
		       if not printing to a file
		    */
		    if (!fflag)
		    {
			printf ("hit return to continue\n");
		    	read (fd, input, 1);
		    }
		}
		else
		{
		    if (fflag)
			fprintf (fp, "\tevent list for %d is empty\n",
				currList -> pElement[eventIndex] -> event_number);
		    else
			printf ("\tevent list for %d is empty\n",
				currList -> pElement[eventIndex] -> event_number);
		}
	    }
	    else
	    {
		if (fflag)
		    fprintf (fp, "no event list for index %d\n", eventIndex);
		else
		    printf ("no event list for index %d\n", eventIndex);
	    }
	}
	currList = currList -> nextEventList;
	if (currList)
	{
	    if (fflag)
		fprintf (fp, "getting next EventList\n");
	    else
		printf ("getting next EventList\n");
	}
    }
    if (fflag)
    {
	fprintf (fp, "\nEND EVENT LIST PRINT\n");
	fclose (fp);
    }
    printf ("\nEND EVENT LIST PRINT\n");
    return (0);
}


/*
 * IO_SCANNER_WAKEUP
 *
 * wake up the I/O interrupt event scanner
 */
void io_scanner_wakeup(io_type,card_type,card_number)
short	io_type,card_type,card_number;
{
	struct intr_event	intr_event;

	if (wakeup_init != 1) return;     /* not awake yet */
	intr_event.io_type = io_type;
	intr_event.card_type = card_type;
	intr_event.card_number = card_number;

	rngBufPut(ioEventQ, (char *)&intr_event, INTR_EVENT_SZ);
	semGive(ioEventSem);
}

/*
 * POST_EVENT
 *
 * wake up the event scanner
 */
void post_event(event_number)
short	event_number;
{
	if (wakeup_init != 1) return;     /* not awake yet */
	/* logMsg("Post Events\n",0,0,0,0,0,0); */
	rngBufPut (eventQ, (char *)&event_number, sizeof(short));
	semGive (eventSem);
}

/*
 * stallScanTasks
 *
 * Stop the scan tasks at a safe place
 */

void stallScanTasks()
{
	FASTLOCK(&LOCK_PERIODIC);
	FASTLOCK(&LOCK_EVENT);
	FASTLOCK(&LOCK_IOINTR);
}

/*
 * resumeScanTasks
 *
 * Resume the scan tasks after a stallScanTasks
 */

void resumeScanTasks()
{
	FASTUNLOCK(&LOCK_PERIODIC);
	FASTUNLOCK(&LOCK_EVENT);
	FASTUNLOCK(&LOCK_IOINTR);
}
