/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*		RECORD INITIALIZATION AND PERIODIC UPDATE	*/
/*								*/
/*	This task has two purposes. First, it will try to	*/
/*	init every record that hasn't already been intialized	*/
/*	every ten seconds. Then, it will make sure that we	*/
/*	get a redundant update for every point in the database	*/
/*	every five minutes.					*/
/*								*/
/****************************************************************/

/* Modifcation log
*  ---------------
*  .01	06-05-94	ems	use first_pass flag to let task traverse
*				the database before any redundant update
*				requests can be made
*				chg THROTTLE from 30 to 10 to process
*				redundant requests faster
*  .02  06-06-94	ems	while in middle of requesting redundant
*				updates from the MASTER, we are slowing
*				down inits for MASTER points in this IOC
*				therefore call init_masters() to deal 
*				with this, also this negates the need for
*				the first_pass flag so it was deleted
*  .03	02-01-95	ems	make task stack a #define in task_params.h
*				and increased stack from 4000 to 8192
*  .04	06-08-98	ems	added initCheck ()
*/


#include <vxWorks.h>
#include <sockLib.h>
#include <stdio.h>
#include <in.h>
#include <tickLib.h>
#include <string.h>
#include <fioLib.h>
#include <semLib.h>
#include <logLib.h>

#include "task_params.h"
#include "dbDefs.h"
#include "redmsgs.h"
#include "dbRecords.h"
#include "dbCommon.h"
#include "recSup.h"
#include "errno.h"
#include "restable.h"
#include "redflds.h"

extern SEM_ID	oldHoldSem;
extern int startTime;		/* defined in dbserver.c */


/******
	Tuning Constants
******/

#define INIT_DELTA	600	/* Ten seconds, in ticks */
#define IDLE_DELTA	18000	/* Five minutes, in ticks */
#define	THROTTLE	10	/* Delay after requesting an update
				   to avoid swamping the network, this
				   gives us 6/second */
#define TWO_MINUTES	(2*60*60)	/* 60 ticks/sec * 60 sec/min * 2 min */

/******
	Be careful. You only want to request updates for points
	which are connected to a resource that we DON'T own.
	Also, we will initialize in Phase 1 so that we will
	be requesting redundant updates fairly soon (i.e., within
	ten seconds) after we start and/or restart.
******/


static int	nextRequestTime = 0;	/* Causes immediate state change */
static int	requestSocket;		/* Socket for requesting update */
static unsigned long throttle_cnt = 0;	/* while THROTTLE, tell when to call init_masters() */
static int firstTime1 = 1;	       	/* these are used to gate the logMsg delay notification */
static int firstTime2 = 1;
int dboinitDelay = TWO_MINUTES;
void		dbo_init_task();
static void	init_masters();


/****************************************************************/
/*								*/
/*			INITIALIZE TASK				*/
/*								*/
/****************************************************************/

int start_output_init()
{
    if ((requestSocket = socket(AF_INET, SOCK_DGRAM, 0)) == ERROR)
    {
	printErr("Can't create redundant update request socket: %d\n", errno);
	return (-1);
    }
    if (taskSpawn(DBOINIT_NAME, DBOINIT_PRI, VX_FP_TASK, DBOINIT_STACK, (FUNCPTR)dbo_init_task, 0,0,0,0,0,0,0,0,0,0)
			== ERROR)
    {
	printErr("Can't spawn passive point redundant update task.\n");
	return (-1);
    }
    return (0);
}

/****************************************************************/
/*								*/
/*		FORCE REDUNDANT UPDATE FROM MASTER		*/
/*								*/
/****************************************************************/

int forceRedundantUpdate()
{
    nextRequestTime = 0;
    return (0);
}


/****************************************************************/
/*								*/
/*		THE PERIODIC REDUNDANT UPDATE TASK		*/
/*								*/
/****************************************************************/

void dbo_init_task()
{
    int			i;
    short		answer;
    int			clearing = 0;
    struct sockaddr_in	reqAddr;

    struct
    {
	short		msgType;
	RED_ASKUPDT	ask;
    } reqMsg;

    for (;;)
    {

	/*
	See if a phase change is appropriate.
	*/

	if (tickGet() > nextRequestTime)
	{
	    clearing = 1;
	    nextRequestTime = tickGet() + IDLE_DELTA;
	}

	/*
	Traverse the database.
	*/

	for (i = 0; i < dbRecords -> number; i++)
	{
	    struct recLoc	*recLoc;
	    struct recChain	*chain;
	    struct rset	*prset;
	    struct fldDes	**fldDes;

	    if (!(recLoc = dbRecords -> papRecLoc[i]))
		continue;

	    prset = GET_PRSET(i);
	    fldDes = papRedFldList[i];

	    for (chain = (struct recChain *) (recLoc -> pFirst);
		 chain; chain = chain -> next)
	    {
		struct dbCommon	*recHdr;

		if (!(recHdr = (struct dbCommon *) (chain -> pRecord)))
		    continue;

		if (!(recHdr -> name[0]))
		    continue;

		/* if no lset yet then initDatabase has not done it's
		   thing so no use trying to init record, ems 3/4/94
		*/
		if (!(recHdr -> lset))
		    continue;

		/* both initialization flags must be set before a file
		   is fully init'd, see each point type, ems 4/22/94
		*/

		if ((!(recHdr -> init)) || (!(recHdr -> ini2)))
		{
		    if (prset -> init_record)
		    {
			semTake(oldHoldSem, WAIT_FOREVER);
			(*(prset -> init_record))(recHdr);
			semGive(oldHoldSem);
		    }
		}

		if (!clearing)
		    continue;

		if (!fldDes)
		    continue;	/* No redundant fields to ask for! */

		/* if points resource is MASTER, no need to go on */

		resGetCaAnswer(recHdr -> radr, &answer);
		/*
		  If I am master(answer == 1) then set the redundant
		  update flag so when I become slave I wont ask for 
		  redundant updates.

		  GCT 2-Oct-1998
		*/
		if(answer)
		  {
		    recHdr->rdur = 1;
		    continue;
		  }
		/*
		  Do not request any redundant updates until 2 minutes after
		  database download completion. startTime is defined in dbserver.c
		  and is set immediately after a database download.


		  GCT 2-Oct-1998
		*/
		if(firstTime1 > 0)
		  {
		    firstTime1--;
		    logMsg("Waiting %d minutes for first redundant update\n",dboinitDelay/3600,0,0,0,0,0);
		  }
		if((tickGet() - startTime) < dboinitDelay)
		  {
		    continue;
		  }
		if(firstTime2 > 0)
		  {
		    firstTime2--;
		    logMsg("dboinit enabled after %d min. delay\n",dboinitDelay/3600,0,0,0,0,0);
		  }

		/* if we have not received a redundant update form the master in
		   the allocated amount of time, or we are still not fully init'd
		   then we need to get from the MASTER, ems 5/2/94
		*/
		if (recHdr -> rdur && recHdr -> ini2)
		    recHdr -> rdur = 0;
		else
		{
		    RES_TABLE	*resEntry;

		    if (recHdr -> rdur)
			recHdr -> rdur = 0;
		    resEntry = (RES_TABLE *)(recHdr -> radr);
		    if (!(resEntry -> dbServerConnected))
			continue;
		    reqAddr.sin_family = AF_INET;
		    reqAddr.sin_addr = resEntry -> partnerInet;
		    reqAddr.sin_port = htons(DBSERVER_UDP_PORT);
		    reqMsg.msgType = htons(RED_UPDREQ);
		    strcpy(reqMsg.ask.redPointName, recHdr -> name);
		    sendto(requestSocket, (char *) &reqMsg, sizeof(reqMsg), 0,
						(struct sockaddr *)&reqAddr, sizeof(reqAddr));
		    taskDelay(THROTTLE);
		    throttle_cnt++;

		    /* after 10 seconds, we had better make a pass over the
		       database to try and int the MASTER points
		       THROTTLE = 10 ticks or 6/second
		    */
		    if ( (throttle_cnt % 60) == 0)
		        init_masters ();
		}
	    } /* end all pts of type */

	} /* end for all pt types */
	clearing = 0;
	throttle_cnt = 0;
	taskDelay(INIT_DELTA);

    } /* end for (;;) */
}


static void init_masters ()
{
    int 	i;
    short	answer;

    for (i = 0; i < dbRecords -> number; i++)
    {
	struct recLoc	*recLoc;
	struct recChain	*chain;
	struct rset	*prset;

	if (!(recLoc = dbRecords -> papRecLoc[i]))
	    continue;

	prset = GET_PRSET(i);

	for (chain = (struct recChain *) (recLoc -> pFirst);
	     chain; chain = chain -> next)
	{
	    struct dbCommon	*recHdr;

	    if (!(recHdr = (struct dbCommon *) (chain -> pRecord)))
		continue;

	    if (!(recHdr -> name[0]))
		continue;

	    /* if no lset yet then initDatabase has not done it's
	       thing so no use trying to init record, ems 3/4/94
	    */
	    if (!(recHdr -> lset))
		continue;

	    /* see if we are MASTER */

	    resGetCaAnswer(recHdr -> radr, &answer);

	    /* if either of the init flags is not set and this points
	       resource is MASTER then try to init the point
	    */

	    if (answer)
	    {
		if ( (!(recHdr -> init)) || (!(recHdr -> ini2)) )
		{
		    if (prset -> init_record)
		    {
			semTake(oldHoldSem, WAIT_FOREVER);
			(*(prset -> init_record))(recHdr);
			semGive(oldHoldSem);
		    }
		}
	    }
	}
    }
    return;
}


int initCheck ()
{
    int 	i;
    int		numNotInited = 0;

    for (i = 0; i < dbRecords -> number; i++)
    {
	struct recLoc	*recLoc;
	struct recChain	*chain;
	struct rset	*prset;

	if (!(recLoc = dbRecords -> papRecLoc[i]))
	    continue;

	prset = GET_PRSET(i);

	for (chain = (struct recChain *) (recLoc -> pFirst);
	     chain; chain = chain -> next)
	{
	    struct dbCommon	*recHdr;

	    if (!(recHdr = (struct dbCommon *) (chain -> pRecord)))
		continue;

	    if (!(recHdr -> name[0]))
		continue;

	    if ( (!(recHdr -> lset)) || (!(recHdr -> init)) || (!(recHdr -> ini2)) )
	    {
		numNotInited++;
		if (numNotInited > 50)
		{
		    printf ("From initCheck, There are more than 50, goodbye\n");
		    return (0);
		}
		else
		{
		    printf ("%-*.*s\n", PVNAME_SZ, PVNAME_SZ, recHdr -> name);
		}
	    }
	}
    }
    printf ("From initCheck num not init'd = %d\n", numNotInited);
    return (0);
}

