/****************************************************************/
/*								*/
/*	LOAD SDR FROM TEXT FILE INTO MEMORY ARRAYS		*/
/*								*/
/*	This routine reads a text file and fills in the		*/
/*	memory arrays from the contents of the text file.	*/
/*	These text files are produced by the various "bld"	*/
/*	programs and are therefore assumed to be error 		*/
/*	free. 							*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifdef vxWorks
#include <vxWorks.h>
#include <symLib.h>
#include <logLib.h>
#endif

#if !(defined(SUN) || defined(SOLARIS) || defined(WIN32))
void	*malloc();
void	*calloc();
#else
#include <sys/types.h>
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "dbDefs.h"
#include "dbRecType.h"
#include "dbRecDes.h"
#include "choice.h"
#include "devSup.h"
#include "drvSup.h"
#include "cvtTable.h"
#include "recSup.h"
#include "dbRecords.h"
#include "sdrLoad.h"
#ifdef vxWorks
#include "dbPvd.h"
#endif vxWorks

#ifndef ERROR
#define ERROR -1
#endif

#if defined(SOLARIS) || defined(WIN32)
#define bzero(b,n) memset(b,0,n)
#endif

#ifdef vxWorks
char			*sysModel();
extern SYMTAB_ID	sysSymTbl;
static char		*sdrStart;
static char		*sdrEnd;
#endif

int	getInt (FILE *fd);
float	getFloat(FILE *fd);
int	hiafgets (char *buff, int buffSize, FILE *fd);
void	trimNewLine (char *string);

/*
Global pointers to the various sdr structures.
*/

struct recType		*dbRecType = NULL;
struct arrChoiceSet	*choiceGbl = NULL;
struct choiceRec	*choiceRec = NULL;
struct devChoiceRec	*choiceDev = NULL;
struct recDevSup	*devSup    = NULL;
struct drvSup		*drvSup    = NULL;
struct arrBrkTable	*cvtTable  = NULL;
struct choiceSet	*choiceCvt = NULL;
struct recDes		*dbRecDes  = NULL;
struct recSup		*recSup    = NULL;
struct pvd		*dbPvd     = NULL;
struct recHeader	*dbRecords = NULL;

/*
Flag, if false, means that we are on a microsol system.
*/

static short	realIOC = 1;

int sdrLoad(fileName)
char	*fileName;
{
    FILE	*sdrFile = NULL;
    char	command[128];
    int		i;
    int		j;
    int		numRecs;
    int		k;
    char	*newLine;

    /*
    Determine if we are on a Microsol machine or not.
    */

#ifdef vxWorks
    realIOC = strcmp(sysModel(), "MicroSol XCell");
    if (!realIOC)
	{
	symFindByName(sysSymTbl, "_binArrayStart", &sdrStart, (SYM_TYPE *)&i);
	symFindByName(sysSymTbl, "_binArrayEnd", &sdrEnd, (SYM_TYPE *)&i);
	}
#endif


    /*
    Open the file.
    */

    if (realIOC)
	{
        if ((sdrFile = fopen(fileName, "r")) == NULL)
	    {
	    perror("Can't open SDR file:");
	    return (-1);
	    }
	}

    /*
    This starts our main loop. Get the type of record that we are
    loading. Then, execute code specific to each type of record.
    */

    while (hiafgets(command, sizeof(command), sdrFile) != 0)
	{
	/*
	hiafgets() leaves a new line at the end of each line. Wipe
	that out.
	*/

	trimNewLine(command);

	if (!strcmp(command, "dbRecType"))
	    {

	    /*
	    List of point type names for the system -- get number of types.
	    Allocate the header structure and the array.
	    */

	    numRecs = getInt(sdrFile);
	    dbRecType = (struct recType *) malloc(sizeof(struct recType));
	    dbRecType -> number = numRecs;
	    dbRecType -> papName = (char **)calloc(numRecs, sizeof(char *));

	    /*
	    Now read the list of names.
	    */

	    for (i = 0; i < dbRecType -> number; i++)
		{
		hiafgets(command, sizeof(command), sdrFile);
		if ((numRecs = strlen(command)) > 1)
		    {
		    dbRecType -> papName[i] = malloc(numRecs);
		    trimNewLine(command);
		    strcpy(dbRecType -> papName[i], command);
		    }
		}

   	    /*
    	    contiguous space must be allocated for the header and pointer array
    	    */
 
    	    numRecs = sizeof(struct recHeader) + dbRecType->number * sizeof(void *);
    	    if ((dbRecords = (struct recHeader *) malloc((unsigned) numRecs)) == NULL)
		{
                perror("malloc error for database records\n");
                return (ERROR);
    	        }
 
    	    bzero((void *) dbRecords , (int) numRecs);
    	    (long)dbRecords->papRecLoc = (long) ((long)dbRecords + sizeof(struct recHeader));
    	    dbRecords->number = dbRecType->number;
	    }

	else if (!strcmp(command, "choiceGbl"))
	    {

	    /*
	    List of global choices for the system -- get number of choices.
	    Allocate the header structure and the array of pointers.
	    */

	    numRecs = getInt(sdrFile);
	    choiceGbl = (struct arrChoiceSet *) malloc(sizeof(struct arrChoiceSet));
	    choiceGbl -> number = numRecs;
	    choiceGbl -> papChoiceSet = (struct choiceSet **) calloc(numRecs, sizeof(struct choiceSet *));

	    /*
	    Now read in each choice set
	    */

	    for (i = 0; i < choiceGbl -> number; i++)
		{
	    	numRecs = getInt(sdrFile);
		if (!numRecs)
		    continue;
		choiceGbl -> papChoiceSet[i] = (struct choiceSet *) malloc(sizeof(struct choiceSet));
		choiceGbl -> papChoiceSet[i] -> number = numRecs;
		choiceGbl -> papChoiceSet[i] -> papChoice = (char **) calloc(numRecs, sizeof(char *));
		for (j = 0; j < choiceGbl -> papChoiceSet[i] -> number; j++)
		    {
		    hiafgets(command, sizeof(command), sdrFile);
		    if ((numRecs = strlen(command)) >= 1)
			{
			choiceGbl -> papChoiceSet[i] -> papChoice[j] = (char *) malloc(numRecs);
		    	trimNewLine(command);
			strcpy(choiceGbl -> papChoiceSet[i] -> papChoice[j], command);
			}
		    }
		}
	    }

	else if (!strcmp(command, "choiceRec"))
	    {

	    /*
	    List of choices specific to each record type -- get number of types.
	    */

	    numRecs = getInt(sdrFile);
	    choiceRec = (struct choiceRec *) malloc(sizeof(struct choiceRec));
	    choiceRec -> number = numRecs;
	    choiceRec -> papArrChoiceSet = (struct arrChoiceSet **) calloc(numRecs, sizeof(struct arrChoiceSet *));
	    for (i = 0; i < choiceRec -> number; i++)
		{
		numRecs = getInt(sdrFile);
		if (!numRecs)
		    continue;
		choiceRec -> papArrChoiceSet[i] = (struct arrChoiceSet *) calloc(1, sizeof(struct arrChoiceSet));
		choiceRec -> papArrChoiceSet[i] -> number = numRecs;
		choiceRec -> papArrChoiceSet[i] -> papChoiceSet = (struct choiceSet **) calloc(numRecs, sizeof(struct choiceSet *));
		for (j = 0; j < choiceRec -> papArrChoiceSet[i] -> number; j++)
		    {
		    numRecs = getInt(sdrFile);
		    if (!numRecs)
			continue;
		    choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] = (struct choiceSet *) calloc(1, sizeof(struct choiceSet));
		    choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> number = numRecs;
		    choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice = (char **) calloc(numRecs, sizeof(char *));
		    for (k = 0; k < choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> number; k++)
			{
			hiafgets(command, sizeof(command), sdrFile);
			if ((numRecs = strlen(command)) > 1)
			    {
			    choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice[k] = (char *) malloc(numRecs);
			    trimNewLine(command);
			    strcpy(choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice[k], command);
			    }
			}
		    }
		}
	    }

	else if (!strcmp(command, "devSup"))
	    {

	    /*
	    Device Support. For each record type we map device support names to addresses
	    of Device Support Entry Tables (DSETs).
	    */

	    numRecs = getInt(sdrFile);
	    devSup = (struct recDevSup *) malloc(sizeof(struct recDevSup));
	    devSup -> number = numRecs;
	    devSup -> papDevSup = (struct devSup **) calloc(numRecs, sizeof(struct devSup *));
	    for (i = 0; i < devSup -> number; i++)
		{
		numRecs = getInt(sdrFile);
		if (!numRecs)
		    continue;
		devSup -> papDevSup[i] = (struct devSup *) calloc(1, sizeof(struct devSup));
		devSup -> papDevSup[i] -> number = numRecs;
		devSup -> papDevSup[i] -> papDset = (struct dset **) calloc(numRecs, sizeof(struct dset *));
		devSup -> papDevSup[i] -> dsetName = (char **) calloc(numRecs, sizeof(char *));
		for (j = 0; j < devSup -> papDevSup[i] -> number; j++)
		    {
		    hiafgets(command, sizeof(command), sdrFile);
		    if ((numRecs = strlen(command)) > 1)
			{
			devSup -> papDevSup[i] -> dsetName[j] = (char *) malloc(numRecs);
			trimNewLine(command);
			strcpy(devSup -> papDevSup[i] -> dsetName[j], command);
			}
		    }
		}
	    }

	else if (!strcmp(command, "choiceDev"))
	    {

	    /*
	    Device Choice. For each record type we have the link types and device choices.
	    */

	    numRecs = getInt(sdrFile);
	    choiceDev = (struct devChoiceRec *) malloc(sizeof(struct devChoiceRec));
	    choiceDev -> number = numRecs;
	    choiceDev -> papDevChoiceSet = (struct devChoiceSet **) calloc(numRecs, sizeof(struct devChoiceSet *));
	    for (i = 0; i < choiceDev -> number; i++)
		{
		numRecs = getInt(sdrFile);
		if (!numRecs)
		    continue;
		choiceDev -> papDevChoiceSet[i] = (struct devChoiceSet *) malloc(sizeof(struct devChoiceSet));
		choiceDev -> papDevChoiceSet[i] -> number = numRecs;
		choiceDev -> papDevChoiceSet[i] -> papDevChoice = (struct devChoice **) calloc(numRecs, sizeof(struct devChoice *));
		for (j = 0; j < choiceDev -> papDevChoiceSet[i] -> number; j++)
		    {
		    numRecs = getInt(sdrFile);
		    hiafgets(command, sizeof(command), sdrFile);
		    if (numRecs < 0)
			continue;
		    choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] = (struct devChoice *) malloc(sizeof(struct devChoice));
		    choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] -> link_type = numRecs;
		    numRecs = strlen(command);
		    choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] -> pchoice = (char *) malloc(numRecs);
		    trimNewLine(command);
		    strcpy(choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] -> pchoice, command);
		    }
		}
	    }

	else if (!strcmp(command, "drvSup"))
	    {

	    /*
	    Driver Support. A list of driver entry table names and pointers for them.
	    */

	    numRecs = getInt(sdrFile);
	    drvSup = (struct drvSup *) malloc(sizeof(struct drvSup));
	    drvSup -> number = numRecs;
	    drvSup -> drvetName = (char **) calloc(numRecs, sizeof(char *));
	    drvSup -> papDrvet = (struct drvet **) calloc(numRecs, sizeof(struct drvet *));
	    for (i = 0; i < drvSup -> number; i++)
		{
		hiafgets(command, sizeof(command), sdrFile);
		if ((numRecs = strlen(command)) > 1)
		    {
		    drvSup -> drvetName[i] = (char *) malloc(numRecs);
		    trimNewLine(command);
		    strcpy(drvSup -> drvetName[i], command);
		    }
		}
	    }

	else if (!strcmp(command, "cvtTable"))
	    {

	    /*
	    Breakpoint tables known to the system.
	    */

	    numRecs = getInt(sdrFile);
	    cvtTable = (struct arrBrkTable *) malloc(sizeof(struct arrBrkTable));
	    cvtTable -> number = numRecs;
	    cvtTable -> papBrkTable = (struct brkTable **) calloc(numRecs, sizeof(struct brkTable *));
	    for (i = 0; i < cvtTable -> number; i++)
		{
		hiafgets(command, sizeof(command), sdrFile);
		if ((numRecs = strlen(command)) > 1)
		    {
		    cvtTable -> papBrkTable[i] = (struct brkTable *) malloc(sizeof(struct brkTable));
		    cvtTable -> papBrkTable[i] -> name = (char *) malloc(numRecs);
		    trimNewLine(command);
		    strcpy(cvtTable -> papBrkTable[i] -> name, command);
		    numRecs = getInt(sdrFile);
		    cvtTable -> papBrkTable[i] -> number = numRecs;
		    cvtTable -> papBrkTable[i] -> rawLow = getInt(sdrFile);
		    cvtTable -> papBrkTable[i] -> rawHigh = getInt(sdrFile);
		    cvtTable -> papBrkTable[i] -> papBrkInt = (struct brkInt **) malloc(numRecs * sizeof(struct brkInt *));
		    for (j = 0; j < numRecs; j++)
			{
			cvtTable -> papBrkTable[i] -> papBrkInt[j] = (struct brkInt *) malloc(sizeof(struct brkInt));
			cvtTable -> papBrkTable[i] -> papBrkInt[j] -> raw = getInt(sdrFile);
			cvtTable -> papBrkTable[i] -> papBrkInt[j] -> slope = getFloat(sdrFile);
			cvtTable -> papBrkTable[i] -> papBrkInt[j] -> eng = getFloat(sdrFile);
			}
		    }
		}
	    }

	else if (!strcmp(command, "choiceCvt"))
	    {

	    /*
	    Selection choices for the conversion types.
	    */

	    numRecs = getInt(sdrFile);
	    choiceCvt = (struct choiceSet *) malloc(sizeof(struct choiceSet));
	    choiceCvt -> number = numRecs;
	    choiceCvt -> papChoice = (char **) calloc(numRecs, sizeof(char *));
	    for (i = 0; i < choiceCvt -> number; i++)
		{
		hiafgets(command, sizeof(command), sdrFile);
		if ((numRecs = strlen(command)) > 1)
		    {
		    choiceCvt -> papChoice[i] = (char *) malloc(numRecs);
		    trimNewLine(command);
		    strcpy(choiceCvt -> papChoice[i], command);
		    }
		}
	    }

	else if (!strcmp(command, "dbRecDes"))
	    {

	    /*
	    Record definitions.
	    */

	    numRecs = getInt(sdrFile);
	    dbRecDes = (struct recDes *) malloc(sizeof(struct recDes));
	    dbRecDes -> number = numRecs;
	    dbRecDes -> papRecTypDes = (struct recTypDes **) calloc(numRecs, sizeof(struct recTypDes *));
	    for (i = 0; i < dbRecDes -> number; i++)
		{
		numRecs = getInt(sdrFile);
		if (!numRecs)
		    continue;		/* Unused data type */
		dbRecDes -> papRecTypDes[i] = (struct recTypDes *) malloc(sizeof(struct recTypDes));
		dbRecDes -> papRecTypDes[i] -> rec_size = numRecs;
		dbRecDes -> papRecTypDes[i] -> no_fields = getInt(sdrFile);
		dbRecDes -> papRecTypDes[i] -> no_prompt = getInt(sdrFile);
		dbRecDes -> papRecTypDes[i] -> no_links = getInt(sdrFile);
		dbRecDes -> papRecTypDes[i] -> link_ind = (short *) malloc(dbRecDes -> papRecTypDes[i] -> no_links
										* sizeof(short));
		dbRecDes -> papRecTypDes[i] -> sortFldName = (unsigned long *) malloc(dbRecDes -> papRecTypDes[i] -> no_fields 
										* sizeof(long));
		dbRecDes -> papRecTypDes[i] -> sortFldInd = (short *) malloc(dbRecDes -> papRecTypDes[i] -> no_fields 
										* sizeof(short));
		dbRecDes -> papRecTypDes[i] -> papFldDes = (struct fldDes **) malloc(dbRecDes -> papRecTypDes[i] -> no_fields
										* sizeof(struct fldDes *));
		for (j = 0; j < dbRecDes -> papRecTypDes[i] -> no_links; j++)
		    dbRecDes -> papRecTypDes[i] -> link_ind[j] = getInt(sdrFile);
		for (j = 0; j < dbRecDes -> papRecTypDes[i] -> no_fields; j++)
		    {
		    hiafgets(command, sizeof(command), sdrFile);

		    /*
		    Be careful here. If the field is less than the FLDNAME_SZ bytes that we want, then
		    we should make sure that it is padded to that length with blanks. We'll use a rather brute
		    force method to do that.
		    */

    		    if ((newLine = strchr(command, '\n')) != NULL)
		        *newLine = 0;
		    strcat(command, "               ");
		    strncpy((char *) &(dbRecDes -> papRecTypDes[i] -> sortFldName[j]), command, FLDNAME_SZ);
		    }
	 	for (j = 0; j < dbRecDes -> papRecTypDes[i] -> no_fields; j++)
		    dbRecDes -> papRecTypDes[i] -> sortFldInd[j] = getInt(sdrFile);
		for (j = 0; j < dbRecDes -> papRecTypDes[i] -> no_fields; j++)
		    {
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] = (struct fldDes *) malloc(sizeof(struct fldDes));
		    hiafgets(command, sizeof(command), sdrFile);
		    if ((newLine = strchr(command, '\n')) != NULL)
			*newLine = 0;
		    strcat(command, "               ");
		    strncpy(dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> fldname, command, FLDNAME_SZ);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> offset = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> size = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> special = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> field_type = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> dbr_field_type = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> process_passive = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> choice_set = getInt(sdrFile);
		    dbRecDes -> papRecTypDes[i] -> papFldDes[j] -> promptflag = getInt(sdrFile);
		    }
		}
	    }

		
	else
	    {
	    /*
	    None of the above! This should never happen.
	    */
#ifdef vxWorks
	    logMsg("Invalid sdr file contents in %s!\n", (int)fileName,0,0,0,0,0);
#else
	    printf("Invalid sdr file contents in %s!\n", fileName);
#endif vxWorks
	    return (-1);
	    }
	}
    
    /*
    Done deal.
    */

    if (realIOC)
        fclose(sdrFile);
    return (0);
}

/****************************************************************/
/*								*/
/*		TRIM TRAILING NEW LINE IF THERE IS ONE		*/
/*								*/
/****************************************************************/

void trimNewLine(string)
char	*string;
{
    char	*newLine;


    if ((newLine = strchr(string, '\n')) != NULL)
	*newLine = 0;
}

/****************************************************************/
/*								*/
/*		GET AN INTEGER FROM THE INPUT STREAM		*/
/*								*/
/*	Read a line from the input stream and convert it	*/
/*	to an integer. Don't be tempted to use fscanf		*/
/*	because it has strange behavior when mixed with		*/
/*	hiafgets(). You end up swallowing the new line at the	*/
/*	end of the integer.					*/
/*								*/
/****************************************************************/

int getInt(fd)
FILE	*fd;
{
    char	buff[100];

    hiafgets(buff, sizeof(buff), fd);
    return (atoi(buff));
}

/****************************************************************/
/*								*/
/*		GET A FLOAT FROM THE INPUT STREAM		*/
/*								*/
/*	Read a line from the input stream and convert it	*/
/*	to a float. Don't be tempted to use fscanf		*/
/*	because it has strange behavior when mixed with		*/
/*	hiafgets(). You end up swallowing the new line at the	*/
/*	end of the float.					*/
/*								*/
/****************************************************************/

float getFloat(FILE *fd)
{
    char	buff[100];

    hiafgets(buff, sizeof(buff), fd);
    return ((float)atof(buff));
}


/****************************************************************/
/*								*/
/*		HIA ALTERNATIVE TO FGETS			*/
/*								*/
/*	The things we do! This function is the same as fgets()	*/
/*	when run in a real computer. When running in a Micro-	*/
/*	Sol, it pulls the data out of an array that has been	*/
/*	linked into the PROM. YES, I know it's an editorial	*/
/*	comment!						*/
/*								*/
/****************************************************************/

int hiafgets(buff, buffSize, fd)
char	*buff;
int	buffSize;
FILE	*fd;
{
#ifdef vxWorks
    char	byteMoved;
    int		saveBuff;
#endif

    if (realIOC)
	return ((int) fgets(buff, buffSize, fd));

#ifdef vxWorks
    saveBuff = (int) buff;
    while (buffSize > 0)
	{
	if (sdrStart >= sdrEnd)
	    return (0);
	byteMoved = *sdrStart++;
	*buff++ = byteMoved;
	if (byteMoved == '\n')
	    {
	    *buff = 0;
	    break;
	    }
	}
    return (saveBuff);
#endif
}
	


/****************************************************************/
/*								*/
/*		FREE ALL DATABASE MEMORY			*/
/*								*/
/*	This routine frees all of the memory that has been	*/
/*	allocated to store the database. This routine is 	*/
/*	called when it has been decided to abort a download.	*/
/*	By freeing all of the memory before starting over,	*/
/*	we end up with a clean slate.				*/
/*								*/
/****************************************************************/

int deleteDatabase()
{
    int		i;
    int		j;
    int		k;

    /*
    Record support.
    */

    if (recSup)
	{
	free(recSup);
	recSup = 0;
	}

    /*
    Driver support.
    */

    if (drvSup)
	{
	free(drvSup);
	if (drvSup -> drvetName)
	    {
	    for (i = 0; i < drvSup -> number; i++)
	        if (drvSup -> drvetName[i])
		    free(drvSup -> drvetName[i]);
	    free (drvSup -> drvetName);
	    }
	drvSup = 0;
	}

    /*
    Conversion Table
    */

    if (cvtTable)
	{
	if (cvtTable -> papBrkTable)
	    {
	    for (i = 0; i < cvtTable -> number; i++)
	        {
	        if (cvtTable -> papBrkTable[i])
		    {
		    for (j = 0; j < cvtTable -> papBrkTable[i] -> number; j++)
		        free(cvtTable -> papBrkTable[i] -> papBrkInt[j]);
		    free(cvtTable -> papBrkTable[i] -> papBrkInt);
		    free(cvtTable -> papBrkTable[i]);
		    }
		}
	    free(cvtTable -> papBrkTable);
	    }
	free(cvtTable);
	cvtTable = 0;
	}

    /*
    Device Support
    */

    if (devSup)
	{
	if (devSup -> papDevSup)
	    {
	    for (i = 0; i < devSup -> number; i++)
		{
		if (devSup -> papDevSup[i])
		    {
		    if (devSup -> papDevSup[i] -> dsetName)
			{
			for (j = 0; j < devSup -> papDevSup[i] -> number; j++)
			    free(devSup -> papDevSup[i] -> dsetName[j]);
			free(devSup -> papDevSup[i] -> dsetName);
			}
		    free(devSup -> papDevSup[i]);
		    }
		}
	    free(devSup -> papDevSup);
	    }
	free(devSup);
	devSup = 0;
	}

    /*
    Device Choice
    */

    if (choiceDev)
	{
	if (choiceDev -> papDevChoiceSet)
	    {
	    for (i = 0; i < choiceDev -> number; i++)
		{
		if (choiceDev -> papDevChoiceSet[i])
		    {
		    if (choiceDev -> papDevChoiceSet[i] -> papDevChoice)
			{
			for (j = 0; j < choiceDev -> papDevChoiceSet[i] -> number; j++)
			    {
			    if (choiceDev -> papDevChoiceSet[i] -> papDevChoice[j])
				free(choiceDev -> papDevChoiceSet[i] -> papDevChoice[j]);
			    }
			free(choiceDev -> papDevChoiceSet[i] -> papDevChoice);
			}
		    free(choiceDev -> papDevChoiceSet[i]);
		    }
		}
	    free(choiceDev -> papDevChoiceSet);
	    }
	free(choiceDev);
	choiceDev = 0;
	}

    /*
    Record Choice
    */

    if (choiceRec)
	{
	if (choiceRec -> papArrChoiceSet)
	    {
	    for (i = 0; i < choiceRec -> number; i++)
		{
		if (choiceRec -> papArrChoiceSet[i])
		    {
		    if (choiceRec -> papArrChoiceSet[i] -> papChoiceSet)
			{
		        for (j = 0; j < choiceRec -> papArrChoiceSet[i] -> number; j++)
			    {
			    if (choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j])
			        {
			        if (choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice)
				    {
			            for (k = 0; k < choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> number; k++)
				        {
				        if (choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice[k])
					    free(choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice[k]);
				        }
				    free(choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j] -> papChoice);
				    }
			        free(choiceRec -> papArrChoiceSet[i] -> papChoiceSet[j]);
			        }
			    }
			free(choiceRec -> papArrChoiceSet[i] -> papChoiceSet);
			}
		    free(choiceRec -> papArrChoiceSet[i]);
		    }
		}
	    free(choiceRec -> papArrChoiceSet);
	    }
	free(choiceRec);
	choiceRec = 0;
	}

    /*
    Conversion Choice
    */

    if (choiceCvt)
	{
	if (choiceCvt -> papChoice)
	    {
	    for (i = 0; i < choiceCvt -> number; i++)
		{
		if (choiceCvt -> papChoice[i])
		    free(choiceCvt -> papChoice[i]);
		}
	    free(choiceCvt -> papChoice);
	    }
	free(choiceCvt);
	choiceCvt = 0;
	}

    /*
    Global Choice
    */

    if (choiceGbl)
	{
	if (choiceGbl -> papChoiceSet)
	    {
	    for (i = 0; i < choiceGbl -> number; i++)
		{
		if (choiceGbl -> papChoiceSet[i])
		    {
		    if (choiceGbl -> papChoiceSet[i] -> papChoice)
			{
			for (j = 0; j < choiceGbl -> papChoiceSet[i] -> number; j++)
			    {
			    if (choiceGbl -> papChoiceSet[i] -> papChoice[j])
				free(choiceGbl -> papChoiceSet[i] -> papChoice[j]);
			    }
			free(choiceGbl -> papChoiceSet[i] -> papChoice);
			}
		    free(choiceGbl -> papChoiceSet[i]);
		    }
		}
	    free(choiceGbl -> papChoiceSet);
	    }
	free(choiceGbl);
	choiceGbl = 0;
	}

    /*
    Record Description
    */

    if (dbRecDes)
	{
	if (dbRecDes -> papRecTypDes)
	    {
	    for (i = 0; i < dbRecDes -> number; i++)
		{
		if (dbRecDes -> papRecTypDes[i])
		    {
		    if (dbRecDes -> papRecTypDes[i] -> link_ind)
			free(dbRecDes -> papRecTypDes[i] -> link_ind);
		    if (dbRecDes -> papRecTypDes[i] -> sortFldInd)
			free(dbRecDes -> papRecTypDes[i] -> sortFldInd);
		    if (dbRecDes -> papRecTypDes[i] -> sortFldName)
			free(dbRecDes -> papRecTypDes[i] -> sortFldName);
		    if (dbRecDes -> papRecTypDes[i] -> papFldDes)
			{
			for (j = 0; j < dbRecDes -> papRecTypDes[i] -> no_fields; j++)
			    {
			    if (dbRecDes -> papRecTypDes[i] -> papFldDes[j])
				free(dbRecDes -> papRecTypDes[i] -> papFldDes[j]);
			    }
			free(dbRecDes -> papRecTypDes[i] -> papFldDes);
			}
		    free(dbRecDes -> papRecTypDes[i]);
		    }
		}
	    free(dbRecDes -> papRecTypDes);
	    }
	free(dbRecDes);
	dbRecDes = 0;
	}

    /*
    Record Types
    */

    if (dbRecType)
	{
	if (dbRecType -> papName)
	    {
	    for (i = 0; i < dbRecType -> number; i++)
		{
		if (dbRecType -> papName[i])
		    free(dbRecType -> papName[i]);
		}
	    free(dbRecType -> papName);
	    }
	free(dbRecType);
	dbRecType = 0;
	}

    /*
    The database records themselves
    */

    if (dbRecords)	
	{
	int	i;

	for (i = 0; i < dbRecords -> number; i++)
	    {
	    if (dbRecords -> papRecLoc[i])
		{
		struct recChain	*currChain;
		struct recChain	*nextChain;

		currChain = (struct recChain *)(dbRecords -> papRecLoc[i] -> pFirst);

		while (currChain)
		    {
		    nextChain = currChain -> next;
		    free(currChain);
		    currChain = nextChain;
		    }
		free(dbRecords -> papRecLoc[i]);
		}
	    }
	free(dbRecords);
	dbRecords = 0;
	}

    /*
    And finally, the pvd
    */

    if (dbPvd)
	{
	free(dbPvd);
	dbPvd = 0;
#ifdef vxWorks
	pvd_create();
#endif
	}
    return (0);
}
