/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	U N D O . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	maintain the undo queue
 *
 *
 *	History
 *	-------
 *
 * 	Improvements
 * 	------------
 *
 */

#include <edd.h>
#include <dlinfo.h>
#include <dlcvrt.h>

#ifdef AXP
#include <string.h>
#endif

static enum undo_type 	generic_type();

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

static void 		undo_field_change();
static void 		perform_undo();
static void 		trim_undo_queue();
static void 		free_undo_op();
static void 		pop_undo();

/*
 * 	A Header for undo queue entries
 *
 *	peue will be NULL if we are not modifying
 *	an element.
 *
 *
 */
enum undo_type{
	undo_type_double, 
	undo_type_long, 
	undo_type_string
};

enum undoable_op {
	undo_none, 
	undo_create, 
	undo_delete,
	undo_front_back, 	/* pos changed in the dl */
	undo_field
};

struct undo_field_info{
	struct sinfo		*psinfo;	/* struct mem to undo	*/
	void			*pstruct;	/* structure pointer 	*/
	unsigned		elementno;	/* array index		*/
	enum undo_type		type;		/* type of old value	*/
};

struct undo_delete_info{
	unsigned		nbytes;		/* nbytes cb following	*/
	eue			*peue_above;	/* elem that was above 	*/
};

struct undo_front_back_info{
	eue			*peue_above;	/* elem that was above 	*/
};

union undo_op_specific{
	struct undo_field_info		ufi;
	struct undo_delete_info		udi;
	struct undo_front_back_info	ufbi;
};

typedef struct undo_operation_header{
	NODE			node;		/* list header		*/
	enum undoable_op	op;		/* undo operation	*/
	eue			*peue;		/* element to undo 	*/
	unsigned		sequence;	/* undo sequence number */
	union undo_op_specific	uos;		/* operation specific 	*/
}undo_op;


/*
 *	PUSH_DELETE_UNDO
 *	(only one element per undo)
 */
void 
push_delete_undo(pdue, peue)
due		*pdue;
eue		*peue;
{
	undo_op		*pundo_op;

	pdue->dl_mdt = TRUE;

	close_element(pdue, peue, TRUE, TRUE);

	pundo_op = (undo_op *) tisCalloc(__FILE__, __LINE__, 1, sizeof(*pundo_op));
	if(!pundo_op)
		return;

	pundo_op->op = undo_delete;

	/*
	 * element will be deleted
	 */
	pundo_op->peue = 		peue;
	pundo_op->sequence =		pdue->undo_sequence;
	pundo_op->uos.udi.peue_above = 	(eue *)peue->node.previous;

	lstAdd(&pdue->undo, pundo_op);

	trim_undo_queue(pdue);
}


/*
 *	PUSH_FRONT_BACK_UNDO
 */
void 
push_front_back_undo(pdue, peue, peue_above)
due		*pdue;
eue		*peue;
eue		*peue_above;
{
	undo_op		*pundo_op;

	pdue->dl_mdt = TRUE;

	pundo_op = (undo_op *) 
		tisCalloc(__FILE__, __LINE__, 1, sizeof(*pundo_op));
	if(!pundo_op)
		return;

	pundo_op->op = undo_front_back;
	/*
	 * This is a pointer to where the object used to be
	 * (this is used to fix decapitated undo items)
	 */
	pundo_op->peue = 		peue;
	pundo_op->sequence =		pdue->undo_sequence;
	pundo_op->uos.ufbi.peue_above = peue_above;

	lstAdd(&pdue->undo, pundo_op);

	trim_undo_queue(pdue);
}



/*
 *	PUSH_CREATE_UNDO
 *	(also paste undo)
 */
void 
push_create_undo(pdue, peue)
due			*pdue;
eue			*peue;
{
	undo_op		*pundo_op;

	pdue->dl_mdt = TRUE;

	pundo_op = (undo_op *) 
		tisCalloc(__FILE__, __LINE__, 1, sizeof(*pundo_op));
	if(!pundo_op)
		return;

	pundo_op->op = undo_create;
	pundo_op->peue = peue;
	pundo_op->sequence = pdue->undo_sequence;

	lstAdd(&pdue->undo, pundo_op);

	trim_undo_queue(pdue);
}




/*
 *	PUSH_FIELD_UNDO
 */
void 
push_field_undo(pdue, peue, psinfo, pstruct, elementno)
due			*pdue;
eue			*peue;
struct sinfo		*psinfo;	/* struct mem to undo	*/
void			*pstruct;	/* structure pointer 	*/
unsigned		elementno;	/* array index		*/
{
	enum undo_type 	type;
	unsigned	size;
	undo_op		*pundo_op;
	int 		status;

	pdue->dl_mdt = TRUE;

	pundo_op = (undo_op *) pdue->undo.node.previous;
	if(pundo_op){
		/*
 		 * check to see if we are just repeating
		 * a previous operation
		 */
		if(pundo_op->op==undo_field && pundo_op->peue==peue){
			if(	pundo_op->uos.ufi.psinfo == psinfo &&
				pundo_op->uos.ufi.pstruct == pstruct &&
				pundo_op->uos.ufi.elementno == elementno){

				return;
			}
		}
	}

	type = generic_type(psinfo);
	switch(type){
	case	undo_type_double:
	{
		double		undo_double;

		size = sizeof(undo_double);
		undo_double = dl_to_double(
				pstruct, 
				psinfo, 
				elementno);
		pundo_op = (undo_op *) 
			tisCalloc(__FILE__, __LINE__, 1, size+sizeof(*pundo_op));
		if(!pundo_op)
			return;
		*(double *)(pundo_op+1) = undo_double;
		break;
	}
	case	undo_type_long:
	{
		long		undo_long;

		size = sizeof(undo_long);
		undo_long = dl_to_long(
				pstruct, 
				psinfo, 
				elementno);
		pundo_op = (undo_op *) 
			tisCalloc(__FILE__, __LINE__, 1, size+sizeof(*pundo_op));
		if(!pundo_op)
			return;
		*(long *)(pundo_op+1) = undo_long;
		break;
	}
	case	undo_type_string:
	{
		char		*pundo_string;

		pundo_string = dl_to_pstring(
				pstruct, 
				psinfo, 
				&pdue->ct, 
				&pdue->st,
				elementno);
		size = strlen(pundo_string)+1;
		pundo_op = (undo_op *) 
			tisCalloc(__FILE__, __LINE__, 1, size+sizeof(*pundo_op));
		if(!pundo_op)
			return;
		strcpy((char *)(pundo_op+1), pundo_string);
		break;
	}
	default:
		assert(FALSE);
	}

	pundo_op->op = undo_field;
	pundo_op->peue = peue;
	pundo_op->sequence = pdue->undo_sequence;
	pundo_op->uos.ufi.psinfo = psinfo;
	pundo_op->uos.ufi.pstruct = pstruct;
	pundo_op->uos.ufi.elementno = elementno;
	pundo_op->uos.ufi.type = type;

	lstAdd(&pdue->undo, pundo_op);

	trim_undo_queue(pdue);
}


/*
 *
 *	TRIM_UNDO_QUEUE
 *
 * 	put some limit on the size of the undo
 * 	queue (dont break up the current seq however)
 */
static void
trim_undo_queue(pdue)
due	*pdue;
{
	undo_op		*pundo_op;
	int		status;
	unsigned	this_seq;

	if(pdue->undo.count < MAXUNDO)
		return;

	pundo_op = (undo_op *) pdue->undo.node.next;

	if(pundo_op->sequence == pdue->undo_sequence)
		return;
	
	this_seq = pundo_op->sequence;

	/*
	 * dont leave partial undo seq
	 * on the queue.
	 */
	while(pundo_op){
		if(this_seq != pundo_op->sequence)
			break;

		lstDelete(&pdue->undo, pundo_op);

		free_undo_op(pdue, pundo_op);

		pundo_op = (undo_op *) pdue->undo.node.next;
	}
}


/*
 *
 *	GENERIC_TYPE
 *
 */
static enum undo_type 	generic_type(psinfo)
struct sinfo	*psinfo;
{
	enum undo_type 	generic_type;

	switch(psinfo->type){
	case dl_string_index:
	case dl_pc_index:
		generic_type = undo_type_string;
		break;

	case dl_color_index:
	case dl_bckgnd_color_index:
	case dl_char:
	case dl_uchar:
	case dl_short:
	case dl_ushort:
	case dl_int:
	case dl_uint:
	case dl_field:
	case dl_ufield:
	case dl_pchar:
	case dl_pshort:
	case dl_pint:
		generic_type = undo_type_long;
		break;

	case dl_double:
	case dl_float:
	case dl_pdouble:
	case dl_pfloat:
		generic_type = undo_type_double;
		break;

	case dl_struct:
		assert(FALSE);
		break;
	default:
		printf("~edd: strange undo type saved as string\n");
		generic_type = undo_type_string;
		break;
	}

	return generic_type;
}


/*
 *	revert_current_sequence
 *
 *	remove changes made so far on this commmand
 *	(dont redraw)
 */
void 
revert_current_sequence(pdue)
due		*pdue;
{
	undo_op		*pundo_op;

	pundo_op = (undo_op *) pdue->undo.node.previous;
	if(pundo_op){
		if(pundo_op->sequence == pdue->undo_sequence){
			pop_undo(pdue, FALSE);
		}
	}
}


/*
 *	POP_UNDO_FROM_MENU
 */
void 
pop_undo_from_menu(param, pdue, pbe)
void		*param;
due		*pdue;
XButtonEvent	*pbe;
{
	pop_undo(pdue, TRUE);
}


/*
 *	POP_UNDO
 */
static void 
pop_undo(pdue, redraw)
due	*pdue;
int	redraw;
{
	undo_op		*pundo_op;
	unsigned	seq;
	XRectangle	bounds;

	pundo_op = (undo_op *) pdue->undo.node.previous;
	if(!pundo_op){
		sprintf(pdue->pmsg, "~Nothing left to undo\n");
		redraw_msg(pdue);
		return;
	}

	seq = pundo_op->sequence;

	bounds.width = bounds.height = 0;
	while(pundo_op){
		if(pundo_op->sequence != seq)
			break;

		perform_undo(pdue, pundo_op, &bounds);
		pundo_op = (undo_op *) pdue->undo.node.previous;
	}

	/*
	 * redraw only items which changed
	 */
	if( (bounds.width || bounds.height) && redraw)
       		redraw_dl(pdue, &bounds);
}

/*
 *	PERFORM_UNDO
 */
static void 
perform_undo(pdue, pundo_op, pbounds)
due		*pdue;
undo_op		*pundo_op;
XRectangle	*pbounds;
{
	int		status;
	XRectangle	rec;

	pdue->dl_mdt = TRUE;

	lstDelete(&pdue->undo, pundo_op);

	/*
	 * save a copy for computation of the 
	 * redraw region
	 */
	if(pundo_op->peue){
		compute_bounds(
			pundo_op->peue, 
			&rec);

		if(pbounds->width==0 && pbounds->height==0){
			*pbounds = rec;
		}
		else{
			ADD_TO_MAXBOUNDS(*pbounds, rec);
		}
	}

	switch(pundo_op->op){
	case undo_create:
		close_element(pdue, pundo_op->peue, TRUE, FALSE);
		tisFree(PEUE_TO_PSEL(pundo_op->peue));
		pundo_op->peue = NULL;
		break;

	case undo_front_back:
		lstDelete(	&pdue->dl, 
				pundo_op->peue);
		lstInsert(	&pdue->dl, 
				pundo_op->uos.ufbi.peue_above, 
				pundo_op->peue);
		break;

	case undo_delete:
		/*
		 * force strings back int the dl
		 * tables
		 */
		reopen_element(pdue, pundo_op->peue);

		/*
		 * place it back into the dl at
		 * the coreect position
		 */
		lstInsert(	&pdue->dl, 
				pundo_op->uos.udi.peue_above, 
				pundo_op->peue);
		break;
	
	case undo_field:
		undo_field_change(pdue, pundo_op);
		break;

	case undo_none:
	default:
		assert(FALSE);
	}

	if(pundo_op->peue){
		compute_bounds(
			pundo_op->peue, 
			&rec);

		if(pbounds->width==0 && pbounds->height==0){
			*pbounds = rec;
		}
		else{
			ADD_TO_MAXBOUNDS(*pbounds, rec);
		}
	}

	tisFree(pundo_op);
}


/*
 *
 *	purge_undo_list
 *
 *
 */
void 
purge_undo_list(pdue)
due	*pdue;
{
	undo_op	*pundo_op;

	while(pundo_op = (undo_op *) lstGet(&pdue->undo)){
		free_undo_op(pdue, pundo_op);
	}
}


/*
 *
 *	free_undo_op
 *
 *
 */
static void 
free_undo_op(pdue, pundo_op)
due	*pdue;
undo_op	*pundo_op;
{
	int status;

	if(pundo_op->op == undo_delete){

		close_element(pdue, pundo_op->peue, FALSE, FALSE);
		tisFree(PEUE_TO_PSEL(pundo_op->peue));
	}

	tisFree(pundo_op);
}





/*
 *
 *	UNDO_FIELD_CHANGE
 *
 *
 */
static void undo_field_change(pdue, pundo_op)
due	*pdue;
undo_op	*pundo_op;
{
	eue		neweue;
	void		*pstruct;
	unsigned	offset;

	if(pundo_op->peue){
		neweue = *pundo_op->peue;

		offset = ((char *)pundo_op->uos.ufi.pstruct-
				(char *)pundo_op->peue);
		pstruct =  (void *) (offset + (char *)&neweue); 
	}
	else{
		pstruct = pundo_op->uos.ufi.pstruct;
	}

	switch(pundo_op->uos.ufi.type){
	case	undo_type_double:
		double_to_dl(
			pstruct, 
			pundo_op->uos.ufi.psinfo, 
			*(double *)(pundo_op+1), 
			pundo_op->uos.ufi.elementno);
		break;

	case	undo_type_long:
		long_to_dl(
			pstruct, 
			pundo_op->uos.ufi.psinfo, 
			*(long *)(pundo_op+1), 
			pundo_op->uos.ufi.elementno);
		break;

	case	undo_type_string:
		pstring_to_dl(
			pstruct, 
			pundo_op->uos.ufi.psinfo, 
			pundo_op+1, 
			&pdue->ct, 
			&pdue->st,
			pundo_op->uos.ufi.elementno);
		break;

	default:
		assert(FALSE);
	}

	/*
	 * This takes care of element type changes
	 * and other field changes with side effects
	 */
	if(pundo_op->peue){
		change_element(pdue, pundo_op->peue, &neweue);
	}
	else{
		configure_due(pdue, pundo_op->uos.ufi.psinfo->name);
	}
}


/*
 *	discard_current_sequence
 *
 *	forget the undo info collected so far on this commmand
 */
void 
discard_current_sequence(pdue)
due		*pdue;
{
	int	status;
	undo_op	*pundo_op	= (undo_op *) pdue->undo.node.previous;

	for( ; pundo_op && pundo_op->sequence == pdue->undo_sequence; pundo_op = (undo_op *) pdue->undo.node.previous){
		pdue->dl_mdt = TRUE;
		lstDelete(&pdue->undo, pundo_op);
		tisFree(pundo_op);
	}
}
