/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		TASK MONITOR FOR CRITICAL SYSTEM TASKS		*/
/*								*/
/*	This is a task that watches over a list of tasks that	*/
/*	are considered critical to the system. If a task that	*/
/*	we are watching either disappears or becomes suspended,	*/
/*	then we write a message on the console and then reboot	*/
/*	the machine. This is done to make sure that this IOCs	*/
/*	redundant partner will take over. Convenience routines	*/
/*	are provided to allow users to manipulate the list.	*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <wdLib.h>
#include <logLib.h>
#include <taskLib.h>
#include <stdlib.h>
#include <string.h>
#include <tickLib.h>
#include <rebootLib.h>
#include <sysLib.h>
#include <stdio.h>
#include <ioLib.h>
#include <netShow.h>
#include <ifLib.h>
#include <usrLib.h>
#include <dbgLib.h>
#include <symLib.h>
#include <sysSymTbl.h>
#include <envLib.h>

#include "tsDefs.h"
#include "platform.h"

#if CPU==I80486
/* FPGA I/O */
#define	FPGA_BASE    0x50
#define	FPGA_WDSTROBEPORT  (FPGA_BASE+0x0B)
#define	FPGA_INDEXPORT  (FPGA_BASE+0x0D)
#define FPGA_REGPORT	(FPGA_BASE+0x0F)

/* FPGA registers */
#define	FPGA_WDTREGISTER 3

/* Watchdog control definitions */
#define	WDT_CLEAR   0x80 /* clear WD signal (strobe) */
#define	WDT_DISABLE 0x00 /* mode values */
#define WDT_REBOOT  0x18
#define WDT_17_8MS  0x00 /* strobeRate timeouts */
#define WDT_71_1MS  0x01
#define WDT_284MS   0x02
#define WDT_1_14S   0x03
#define WDT_4_55S   0x04
#define WDT_18_22S  0x05
#define WDT_72_8S   0x06
#define WDT_291S    0x07

static short pv5000WDT = 0;
#else CPU== I80486
#define MC68000 1
#define MV147WDOG_CSR	((char *)0xfffe101d)
static short mv147WDT = 0;
static short mv167WDT = 0;
static short mv162WDT = 0;
#endif CPU == I80486
struct taskList
    {
    char		taskName[32];		/* Task name */
    int			taskId;			/* Task ID, -1 = never seen yet */
    int			holdOff;		/* If TRUE, hold off checking */
    struct taskList	*nextTask;		/* Forward Link */
    };

struct taskList	*taskMonList = 0;
static short wdtCountDown = 24;
static short slowDog;
static WDOG_ID wdogId;
static short taskMonEnabled = 1;
static short wdtEnabled = 1;
static int pokeVal;

int tisCrashDump (void);

/****************************************************************/
/*								*/
/*		ADD A NAME TO THE TASK MONITOR LIST		*/
/*								*/
/****************************************************************/

int addTask(char *name)
{
    int			taskId;
    struct taskList	*currTask;

    if ((taskId = taskNameToId(name)) < 0)
	logMsg("Warning: Task %s not found yet.\n", (int)name,0,0,0,0,0);

    currTask = (struct taskList *) calloc(1, sizeof(struct taskList));
    if (!taskMonList)
	taskMonList = currTask;
    else
	{
	currTask -> nextTask = taskMonList;
	taskMonList = currTask;
	}
    strncpy(currTask -> taskName, name, sizeof(currTask -> taskName));
    currTask -> taskId = taskId;
    return (0);
}

/****************************************************************/
/*								*/
/*		DELETE A TASK FROM THE MONITOR LIST		*/
/*								*/
/****************************************************************/

int deleteTask(char *name)
{
    struct taskList	*currTask = NULL;
    struct taskList	*prevTask = NULL;

    currTask = taskMonList;
    while (currTask)
	{
	if (!strcmp(currTask -> taskName, name))
	    {
	    if (prevTask)
		prevTask -> nextTask = currTask -> nextTask;
	    else
		taskMonList = currTask -> nextTask;
	    free(currTask);
	    return (1);
	    }
	prevTask = currTask;
	currTask = currTask -> nextTask;
	}
    logMsg ("Task %s not found in current monitor list.\n", (int)name,0,0,0,0,0);
    return (-1);
}

/****************************************************************/
/*								*/
/*		SET AND CLEAR THE HOLDOFF FLAG			*/
/*								*/
/*	There is a flag in each member of the list that tells	*/
/*	us that it's OK if the target task gets suspended.	*/
/*	This flag should be set prior to attaching to a task	*/
/*	with the debugger. If you don't, then taskMon will see	*/
/*	that task is suspended and then reboot the IOC out	*/
/*	from underneath you. Bad form.				*/
/*								*/
/****************************************************************/

int setTaskHoldOff(char *name)
{
    struct taskList	*currTask;

    currTask = taskMonList;
    while (currTask)
	{
	if (!strcmp(name, currTask -> taskName))
	    break;
	currTask = currTask -> nextTask;
	}
    if (!currTask)
	{
	logMsg ("Task %s not found in taskMon list. Ignored.\n", (int)name,0,0,0,0,0);
	return (-1);
	}
    currTask -> holdOff = 1;
    return (1);
}

int clearTaskHoldOff(char *name)
{
    struct taskList	*currTask;

    currTask = taskMonList;
    while (currTask)
	{
	if (!strcmp(name, currTask -> taskName))
	    break;
	currTask = currTask -> nextTask;
	}
    if (!currTask)
	{
	logMsg ("Task %s not found in taskMon list. Ignored.\n", (int)name,0,0,0,0,0);
	return (-1);
	}
    currTask -> holdOff = 0;
    return (1);
}

int taskMonShow (void)
{
    struct taskList	*currTask;

    currTask = taskMonList;
    while (currTask)
	{
	printf ("%s\n", currTask -> taskName);
	currTask = currTask -> nextTask;
	}
    return (1);
}

/****************************************************************/
/*								*/
/*		THE TASK MONITOR ROUTINE, ITSELF		*/
/*								*/
/****************************************************************/

#if CPU!=I80486
static int	gmuDog;
#endif CPU!=I80486

void taskMon(void)
{
    struct taskList	*currTask;
    int			options;
    int			lastTable = 0;

    for (;;)
	{
	/*
	Refresh the countdown to indicate that we are running.
	*/

	if (slowDog)
	    {
	    wdtCountDown = pokeVal;
	    }


	/*
	See if taskMon has been disabled. If so, don't bother checking.
	*/

	if (!taskMonEnabled)
	    {
	    taskDelay(8);
	    continue;
	    }
	/*
	Now, check the list of tasks once a second.
	*/

	if ((tickGet() - lastTable) < 60)
	    {
	    taskDelay(8);
	    continue;
	    }
	if (slowDog)
	    slowDog--;
	lastTable = tickGet();
	currTask = taskMonList;
	while (currTask)
	    {
	    if (currTask -> taskId == ERROR)
		{
    		if ((currTask -> taskId = taskNameToId(currTask -> taskName)) != ERROR)
		    logMsg ("taskMon: Task %s now added to monitor list.\n",
			     (int)currTask -> taskName,0,0,0,0,0);
		}
	    else
		{
		if (taskOptionsGet(currTask -> taskId, &options) == ERROR)
		    {
		    wdtCountDown = pokeVal << 4;
		    printf("taskMon: Task %s has disappeared! REBOOTING!\n",
			    currTask -> taskName);
		    tisCrashDump ();
		    taskDelay(60);
#if CPU!=I80486
		    if (mv162WDT)
#if CPU==MC68000
			;
#else
			tisReboot162();
#endif
		    else
		    	reboot(0);
#else
		    /* disable stay alive kick */
		    wdtCountDown = 0;
		    /* delay for hardware reset (2 sec) */
		    taskDelay (120);
		    /* soft reset if all else fails */
		    sysReboot ();
#endif CPU!=I80486
		    }
		else if (taskIsSuspended(currTask -> taskId))
		    {
		    if (!(currTask -> holdOff))
			{
			wdtCountDown = pokeVal << 4;
			printf("taskMon: Task %s has been suspended! REBOOTING!\n",
				currTask -> taskName);
			tisCrashDump ();
			taskDelay(60);
#if CPU!=I80486
			if (mv162WDT)
#if CPU==MC68000
			    ;
#else
			    tisReboot162();
#endif
			else
			    reboot(0);
#else
			/* disable stay alive kick */
			wdtCountDown = 0;
			/* delay for hardware reset (2 sec) */
			taskDelay (120);
			/* soft reset if all else fails */
			sysReboot ();
#endif CPU!=I80486
			}
		    }
		}
	    currTask = currTask -> nextTask;
	    }
	taskDelay(8);
	}
}

/****************************************************************/
/*								*/
/*	    PET THE WATCH DOG TO KEEP IT FROM BARKING		*/
/*								*/
/****************************************************************/

void wdogPet(void)
{
    /*
    If wdt isn't enabled, do nothing.
    */

    if (!wdtEnabled)
	return;

    /*
    Schedule our next execution so we can keep on petting.
    */

    wdStart(wdogId, 8, (FUNCPTR)wdogPet, 0);

    /*
    Check the countdown timer. If this decrements through zero,
    then we have a problem because our application a priority
    1 (equal to the shell) hasn't had a chance to run. In that
    case, don't pet the dog. Let him bark to give us a new life.
    */

    if (--wdtCountDown < 0)
	return;

    /*
    Pet the hardware watch dog so he doesn't bark.
    */

#if CPU==I80486
    sysOutByte (FPGA_WDSTROBEPORT, 0);
#else CPU== I80486
    if (mv147WDT)
        {
        gmuDog = *MV147WDOG_CSR;
        *MV147WDOG_CSR = 0xfd;
        }
    else if (mv167WDT || mv162WDT)	/* Both use VME Chip 2 */
        {
        gmuDog = *(int *) 0xfff40060;
        gmuDog &= 0xff00ffff;
        gmuDog |= 0x006f0000;
        *(int *) 0xfff40060 = gmuDog;
        }
#endif CPU == I80486
}

/****************************************************************/
/*								*/
/*		THE LOWEST PRIORITY TASK FOR CPU TIME		*/
/*								*/
/****************************************************************/

void prio255Mon(void)
{

    /*
    Just set a variable to 120 once a second. The high speed
    end of this task will count that down once a second. If
    it ever gets to zero, it means that this task hasn't had
    a chance to run for two minutes. Therefore, some other
    task is hogging the CPU and that is wrong. Reboot this
    machine. 

    Why do this once a second rather than just once every 2 mins?
    We want to give the system a lot of opportunities to let this
    guy run.
    */

    for (;;)
	{
	slowDog = 120;
	taskDelay(60);
	}
}

/****************************************************************/
/*								*/
/*			INITIALIZE taskMon			*/
/*								*/
/****************************************************************/

int initTaskMon(int usrPokeVal)
{
    char	*model;

    /*
    Start off by adding the standard system stuff to the list.
    */

    addTask("tNetTask");
#if CPU==I80486
    addTask("GPSTask");
    addTask("tsslTelnetd");
    addTask("dbEmulator");
#else CPU==I80486
    addTask("resMgr");
    addTask("timeStamp");
#endif CPU!= 80486
    addTask("redundIO");
    addTask("dbServer");
    addTask("callbackTask");
    addTask("ioEventScanner");
    addTask("periodicScanner");
    addTask("datamove");
    addTask("rsrv I/O");
    addTask("CA Server");
    addTask("chanAcq");
    addTask("sockMnt");
    addTask("initUpdate");

    /*
    Determine if we are running on a platform with a known WDT
    */

#if CPU==I80486
    model = (char *)sysModel();
    if (!strcmp(model, "PC 486"))
    {
	pv5000WDT = 1;
	wdogId = wdCreate();
	wdStart(wdogId, 8, (FUNCPTR)wdogPet, 0);  /* pet every 133 msecs */
	wdKick();		/* routine sets wd timeout at 284 msec */
    }
#else CPU== I80486
    model = (char *)sysModel();
    if (!strcmp(model, "Motorola MVME147"))
	mv147WDT = 1;
    else if (!strcmp(model, "Motorola MVME167"))
	mv167WDT = 1;
    else if (!strcmp(model, "Motorola MVME162"))
	mv162WDT = 1;

    if (mv147WDT || mv167WDT || mv162WDT)
	{
	wdogId = wdCreate();
	wdStart(wdogId, 8, (FUNCPTR)wdogPet, 0);
	wdKick();
	}
#endif CPU == I80486

    /*
    See if the user specified a poke value. If not, guess at some
    defaults.
    */

    if (!usrPokeVal)

#if CPU == MC68000
	pokeVal = 200;
#else
	pokeVal = 24;
#endif

    else
	pokeVal = usrPokeVal;

    /*
    Launch the slow version of the countdown task. This determines that there
    is at least minimal CPU time available to a task at the lowest priority
    in the system.
    */

    taskSpawn("prio255Mon", 255, 8, 2048, (FUNCPTR)prio255Mon, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    /*
    Now, launch the task monitor itself.
    */

    taskSpawn("taskMon", 1, 8, 8192, (FUNCPTR)taskMon, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    return (1);
}

/****************************************************************/
/*								*/
/*		KICK THE WATCHDOG TO GET IT GOING		*/
/*								*/
/****************************************************************/

void wdKick(void)
{
    /*
    Kick the WDT to get it going.
    */

#if CPU==I80486
    /* clear old WD state */
    sysOutByte (FPGA_INDEXPORT, FPGA_WDTREGISTER);
    sysOutByte (FPGA_REGPORT, WDT_CLEAR+WDT_DISABLE);
    /* enable WD at desired rate */
    sysOutByte (FPGA_INDEXPORT, FPGA_WDTREGISTER);
    sysOutByte (FPGA_REGPORT, WDT_REBOOT+WDT_284MS);
#else CPU== I80486
    if (mv147WDT)
    	*MV147WDOG_CSR = 0xfd;
    else if (mv167WDT || mv162WDT)	/* Both use VME Chip 2 */
	{
	gmuDog = *(int *) 0xfff4004c;
	gmuDog &= 0xfff0ff;
	gmuDog |= 0x900;		/* 256 msec */
	*(int *) 0xfff4004c = gmuDog;
	}
#endif CPU == I80486
}

/****************************************************************/
/*								*/
/*		ENABLE & DISABLE THE TASK MONITOR		*/
/*								*/
/****************************************************************/

int enableTaskMon(void)
{
    taskMonEnabled = 1;
    printf("taskMon now enabled.\n");
    return (0);
}

int disableTaskMon(void)
{
    taskMonEnabled = 0;
    printf("taskMon now disabled.\n");
    return (0);
}

/****************************************************************/
/*								*/
/*		ENABLE AND DISABLE THE WATCHDOG			*/
/*								*/
/****************************************************************/

int disableWdt()
{
#if CPU==MC68000
    printf("Wdt can't be disabled on this machine.\n");
#else
#if CPU==I80486
    /* clear old WD state */
    sysOutByte (FPGA_INDEXPORT, FPGA_WDTREGISTER);
    sysOutByte (FPGA_REGPORT, WDT_CLEAR+WDT_DISABLE);
#else CPU== I80486
    if (mv147WDT)
	*MV147WDOG_CSR = 0;
    else if (mv167WDT || mv162WDT)	/* Both use VME Chip 2 */
	{
	gmuDog = *(int *) 0xfff40060;
	gmuDog &= 0xff00ffff;
	*(int *) 0xfff40060 = gmuDog;
	}
    wdtEnabled = 0;
    printf("Hardware WDT now disabled.\n");
#endif CPU==80486
#endif
    return (0);
}

int enableWdt(void)
{
    if (wdtEnabled)
	{
	printf("WDT already enabled!\n");
	return (-1);
	}
    wdtEnabled = 1;
    wdStart(wdogId, 8, (FUNCPTR)wdogPet, 0);
    wdKick();
    return (0);
}

/****************************************************************/
/*								*/
/*	Dumpa a record of the crash to a storage device		*/
/*								*/
/****************************************************************/

int tisCrashDump (void)
{
    static char cmdtext [200];
    static char *argp;
    static char *cmdPtr;
    static int	argc;
    static int	argv[10];
    static int	(*function)();
    static char type;
    static int	cmdix;
    static int	tempfd;
    static FILE *crashfd;
    static int activeTasks;
    static int idList [200];
    static char c;
    static char filename[20];
    static TS_STAMP	now;
    static char	nowText[32];
    static char *cp;

    if ((cp = getenv ("SDR:")) != NULL)
        sprintf (filename, "%s%8.8ld.txt", cp, tickGet ());
    else
        sprintf (filename, "SDR:%8.8ld.txt", tickGet ());
    tempfd = open (filename, O_CREAT|O_RDWR, 0x666);
    if (tempfd != ERROR)
    {    
	ioTaskStdSet (0, 0, tempfd);
	ioTaskStdSet (0, 1, tempfd);
	ioTaskStdSet (0, 2, tempfd);
	printf ("Crash text file ");
	/* standard commands */

        tsLocalTime(&now);
        printf("%s\n", tsStampToText(&now, TS_TEXT_MONDDYYYY, nowText));
	i (0);
	checkStack (0);
	ifShow (0);
	inetstatShow ();
	mbufShow ();
	close (tempfd);
    }

    strcpy (strchr (filename, '.'), ".tt");
    tempfd = open (filename, O_CREAT|O_RDWR, 0x666);
    if (tempfd != ERROR)
    {    
	ioTaskStdSet (0, 0, tempfd);
	ioTaskStdSet (0, 1, tempfd);
	ioTaskStdSet (0, 2, tempfd);
	/* dump supended task info */
	activeTasks = taskIdListGet (idList, 200);
	for (cmdix = 0; cmdix < activeTasks; cmdix++)
	{
	    if (taskIsSuspended (idList[cmdix]))
	    {
		tt (idList[cmdix]);
		ti (idList[cmdix]);
	    }
	}
	close (tempfd);
    }

    if ((cp = getenv ("SDR:")) != NULL)
    {
	sprintf (cmdtext, "%scrash.cmd", cp);
        crashfd = fopen (cmdtext, O_RDONLY);
    }
    else
    {
        crashfd = fopen ("SDR:crash.cmd", O_RDONLY);
    }
    if (crashfd != 0)
    {
	strcpy (strchr (filename, '.'), ".usr");
	tempfd = open (filename, O_CREAT|O_RDWR, 0x666);
	if (tempfd == ERROR)
	{
	    fclose (crashfd);
	    return -1;
	}    
	ioTaskStdSet (0, 0, tempfd);
	ioTaskStdSet (0, 1, tempfd);
	ioTaskStdSet (0, 2, tempfd);
	/* execute additional commands from crash.cmd */
	while ((crashfd != 0) && !feof(crashfd))
	{
	    cmdix = 1;	/* start at 1 to allow '_' addition to routine name */
	    if (fread ((void *)&c, 1, 1, crashfd) != 1)
	    {
		fclose (crashfd);
		crashfd = 0;
		break;
	    }
	    while ((c != 0x0d) && (c != 0x0a))
	    {
		if ((c == 0x08) || (c == 0x7f))
		{
		    if (cmdix)
			cmdix --;
		    if (c == 0x7f)
			c = 0x8;
		}
		else
		{
		    if (cmdix < 199)
			cmdtext[cmdix++] = c;
		    write (tempfd, &c, 1);
		}
		if (fread((void *)&c, 1, 1, crashfd) != 1)
		{
		    fclose (crashfd);
		    crashfd = 0;
		    break;
		}
	    }
	    write (tempfd, "\n\r", 2);
	    cmdtext[cmdix] = 0;

	    if (strlen(&cmdtext[1]))
	    {
		/* execute a function */
		printf ("emergency shell - executing %s\n", &cmdtext[1]);

		/* break command into pieces */
		argp = &cmdtext[1];

		/* skip over white space */
		while (*argp && (*argp == ' ' || *argp == '\t'))
		{
		    argp++;
		}
		/* process command text - get routine name, and look it up */
		cmdPtr = argp;
		if ((argp = strchr (argp, ' ')) || (argp = strchr (argp, '\t')))
		{
		    *argp = 0;
		    argp++;
		}
		else
		{
		    argp = argp + strlen(cmdPtr);
		}
		if (strcmp ("exit", cmdPtr) == 0)
		{
		    break;
		}
		else if (symFindByName (sysSymTbl, cmdPtr, (char **)&function, &type) != OK)
		{
		    cmdPtr--;
		    *cmdPtr = '_';
		    if (symFindByName (sysSymTbl, cmdPtr, (char **)&function, &type) != OK)
		    {
			printf ("Unknown function %s cannot execute\n", cmdPtr);
			function = NULL;
		    }

		}

		/* skip to first arg */
		while (*argp && (*argp == ' ' || *argp == '\t'))
		    argp++;

		/* get arguments */
		for (argc = 0; argp && *argp && argc < 10; argc++)
		{
		    /* handle quoted argument (present as string) */
		    if (*argp == '\"')
		    {
			argp++;
			argv [argc] = (int)argp;
			/* go to for terminating '"' */
			argp = strchr (argp, '"');
			if (argp)
			{
			    *argp = 0;
			    argp++;
			    while (*argp && (*argp == ' ' || *argp == '\t'))
				argp++;
		        }
		        else /* missing trailing '"' */
			    break;
		    }
		    else
		    {
			/* set value arg */
			if (*argp == '0')
			{
			    argp++;
			    if ((*argp == 'x') || (*argp == 'X'))
			    {
				argp++; argv[argc] = 0;
				while (*argp && ((('0' <= *argp) && (*argp <= '9'))
				  || (('a' <= *argp) && (*argp <= 'f'))
				  || (('A' <= *argp) && (*argp <= 'F'))))
				{
				    if (('0' <= *argp) && (*argp <= '9'))
					 argv[argc] = (argv[argc] * 16) + *argp - '0';
				    if (('a' <= *argp) && (*argp <= 'f'))
					 argv[argc] = (argv[argc] * 16) + (*argp - 'a') + 10;
				    if (('A' <= *argp) && (*argp <= 'F'))
					 argv[argc] = (argv[argc] * 16) + (*argp - 'A') + 10;
				    argp++;
				}
			    }
			    else
			    {
				argv[argc] = 0;
				while (('0' <= *argp) && (*argp <= '7'))
				{
				    argv[argc] = (argv[argc] * 8) + *argp - '0';
				    argp++;
				}
			    }
			}
			else
		            argv [argc] = atol(argp);
			/* go to next white space - check for blank */
			if ((argp = strchr (argp, ' ')) || (argp = strchr (argp, '\t')))
			{
			    argp++;
			    /* skip to next arg */
			    while (*argp && (*argp == ' ' || *argp == '\t'))
				argp++;
			}
		    }
	    	}
		/* pad argument list to 10 */
		while (argc < 10)
		    argv [argc++] = 0;

		/* submit command */
		if (function)
		{
		    printf ("%s returned %x\n", cmdPtr,
			(*function)(argv[0], argv[1], argv[2], argv[3], argv[4],
			           argv[5], argv[6], argv[7], argv[8], argv[9]));
		}
	    }
	}
	if (crashfd)
	    fclose (crashfd);
	close (tempfd);
    }
    return 0;
}
