/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	ABORT SWITCH HANDLERS FOR MV147 and MV167 SBCs		*/
/*								*/
/*	These routines establish an abort handler for the	*/
/*	MV147 and MV167 single board computers. When the	*/
/*	user presses the black ABORT button (NOT the red	*/
/*	RESET button), then all of the tis4000 tasks will	*/
/*	be suspended. They can be manually resumed one at	*/
/*	a time. This is a useful facility when you are 		*/
/*	trying to get a global picture of the IOC when		*/
/*	something strange happens.  The user can add		*/
/*	to the list of tasks that are suspended by calling	*/
/*	the addAbortTask() routine in his startup.cmd prior	*/
/*	to iocInit (because it may be too late after that).	*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/


#include <vxWorks.h>
#include <semLib.h>
#include <taskLib.h>
#include <logLib.h>
#include <intLib.h>
#include <usrLib.h>
#include <string.h>

typedef struct zombie
	{
	struct zombie	*next;
	char		*name;
	} ZOMBIE;

static ZOMBIE	*firstZombie = 0;
static SEM_ID	bsem;

void mv167abort(void *par);
void mv147abort(void *par);
void tisAbortHdlr(void);


/****************************************************************/
/*								*/
/*		INIT ROUTINE FOR MV167 SBC			*/
/*								*/
/****************************************************************/

int mv167AbortHandler(void)
{
	bsem = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
	if ((int) bsem == ERROR)
		{
		logMsg("Can't create semaphore, %d\n", errno,0,0,0,0,0);
		return(-1);
		}

	if (intConnect((VOIDFUNCPTR *)(0x5e << 2), mv167abort, 0) == ERROR)
		{
		logMsg("Can't hook up autovector, %d\n", errno,0,0,0,0,0);
		return (-1);
		}
	
	if (taskSpawn("tisAbortHdlr", 1, VX_FP_TASK, 2048, (FUNCPTR) tisAbortHdlr, 
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR)
		{
		logMsg("Can't launch tisAbortHdlr task, %d\n", errno,0,0,0,0,0);
		return (-1);
		}

	return (1);
}


/****************************************************************/
/*								*/
/*	ABORT SWITCH INTERRUPT HANDLER FOR THE MV167 SBC	*/
/*								*/
/****************************************************************/

void mv167abort(void *par)
{
	*(int *) 0xfff40074 = (1 << 30);
	semGive(bsem);
}


/****************************************************************/
/*								*/
/*		INIT ROUTINE FOR THE MV147 SBC			*/
/*								*/
/****************************************************************/

int mv147AbortHandler(void)
{
	bsem = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
	if ((int) bsem == ERROR)
		{
		logMsg("Can't create semaphore, %d\n",errno,0,0,0,0,0);
		return(-1);
		}

	if (intConnect((VOIDFUNCPTR *)(0x42 << 2), mv147abort, 0) == ERROR)
		{
		logMsg("Can't hook up autovector, %d\n",errno,0,0,0,0,0);
		return (-1);
		}
	
	if (taskSpawn("tisAbortHdlr", 1, VX_FP_TASK, 2048, (FUNCPTR) tisAbortHdlr, 
			0, 0, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR)
		{
		logMsg("Can't launch tisAbortHdlr task, %d\n",errno,0,0,0,0,0);
		return (-1);
		}

	return (1);
}


/****************************************************************/
/*								*/
/*	ABORT SWITCH INTERRUPT HANDLER FOR THE MV147 SBC	*/
/*								*/
/****************************************************************/

void mv147abort(void *par)
{
	*(char *) 0xfffe1024 = 0x88;
	semGive(bsem);
}


/****************************************************************/
/*								*/
/*			WATCH DOG TASK				*/
/*								*/
/*	This task waits for a semaphore that is given by the	*/
/*	appropriate abort switch interrupt handler. When it	*/
/*	runs, it first suspends all the standard tasks and	*/
/*	then suspends all the tasks that the user has added.	*/
/*								*/
/****************************************************************/

void tisAbortHdlr(void)
{
	ZOMBIE	*currZombie;
	for (;;)
		{
		semTake(bsem, WAIT_FOREVER);
		ts((int)"resMgr");
		ts((int)"timeStamp");
		ts((int)"redundIO");
		ts((int)"dbServer");
		ts((int)"callbackTask");
		ts((int)"ioEventScanner");
		ts((int)"eventScanner");
		ts((int)"periodicScanner");
		ts((int)"datamove");
		ts((int)"GEN Timers");
		ts((int)"rsrv I/O");
		ts((int)"CA Server");
		ts((int)"chanAcq");
		ts((int)"sockMnt");
		ts((int)"initUpdate");
		ts((int)"snlMon");

		currZombie = firstZombie;
		while (currZombie)
			{
			ts((int)(currZombie -> name));
			currZombie = (ZOMBIE *) (currZombie -> next);
			}
		}
}

/****************************************************************/
/*								*/
/*			ADD A TASK TO THE LIST			*/
/*								*/
/*	This routine can be called any number of times. It	*/
/*	should be called prior to iocInit because the set	*/
/*	of conditions that you are trying to trap may occur	*/
/*	during initialization.					*/
/*								*/
/****************************************************************/

int addAbortTask(name)
char	*name;
{
	ZOMBIE	*newZombie;

	if (!name)
		return (1);
	newZombie = (ZOMBIE *) malloc(sizeof(ZOMBIE));
	newZombie -> name = malloc(strlen(name));
	strcpy(newZombie -> name, name);
	newZombie -> next = firstZombie;
	firstZombie = newZombie;
	return (1);
}
