/****************************************************************
 *                                                              *
 *      Name:    ProdInv.c                                      *
 *      Date:    September 10, 1993                             *
 *      Author:  Donald E. Kyne, Jr.                            *
 *      Purpose: Handles reporting of product inventories.      *
 *               First screen displays a total summary of the   *
 *               selected product(s).  Second screen displays a *
 *               breakdown of the pipe OR tank points which     *
 *               comprise the totals that are summarized on the *
 *               first page.					*
 *                                                              *
 ****************************************************************/
/****************************************************************/
/*								*/
/*	This software was developed for DLA by Tate Integrated  */
/*	Systems, L.P. via a contract with Vitro Corporation.	*/
/*	It is designed to work with TIS4000 Release 4.2.	*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>

#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/Text.h>
#include <Xm/SelectioB.h>
#include <Xm/List.h>
#include <Xm/MwmUtil.h>

#include <Mrm/MrmPublic.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

/* Channel access Includes */
#include "genDefs.h"
#include "db_access.h"
#include "sydDefs.h"
#include "cadef.h"
#include "msgMon.h"
#include "ProdInv.h"

#ifdef MIPSEL
#include	<error.h>	/* DEC Stations - gotta love 'em!! */
#endif



/*****************************************************************************
 *                                                                           *
 *      static functions                                                     *
 *                                                                           *
 *****************************************************************************/

static void		OpenDatabase ();
static void		ScanDatabase ();
static void		updatePipeInv ();
static void		updateTankInv ();
static void		newProdEntry ();
static void		RegisterWidgets ();
static void		ButtonCallback ();
static void		ProdListCallback ();
static void		DetailListCallback ();
static void		DialogCallback ();
static char		*StripWhitespace ();
static void		getLevelUnits();
static void		getVolUnits();
static void		buildTankList();
static void		buildPipeList();
static void		PrintProdDet();
static void		PrintProdInv();
static void		StoreProdInv();
static void		InvFileInit();
static void		AppendProdRecord();
static void		AppendTankRecord();
static void		AppendOtherRecord();
static void		piForceUpdate();
static void		UserMessage();
static void		ListProdTotals();
static void		PipeChanAccess();
static void		TankChanAccess();
static void		RefreshProdInv();
static void		CASearch();

#ifdef MIPSEL
ERRSYMTAB_ID	errSymTbl;				/* defined in error.h */
#endif


/*****************************************************************************
 *                                                                           *
 *      static variables                                                     *
 *                                                                           *
 *****************************************************************************/

static XtAppContext	applContext;
static Display		*display;
static Widget		appShell;
static Widget		productsWindow;
static MrmType		productsWindowClass;
static Widget		detailsWindow;
static MrmType		detailsWindowClass;
static Widget		gravityDialog;
static MrmType		gravityDialogClass;
static MrmHierarchy	mrmHierarchy;
static String		uidFilenames[] =
{
	"ProdInv.uid"
};
#define	NUM_UID_FILES	(sizeof (uidFilenames) / sizeof (uidFilenames[0]))

static MrmRegisterArg	namesRegister[] =
{
	{ (String) "RegisterWidgets",		(caddr_t) RegisterWidgets },
	{ (String) "ButtonCallback",		(caddr_t) ButtonCallback },
	{ (String) "ProdListCallback",		(caddr_t) ProdListCallback },
	{ (String) "DetailListCallback",	(caddr_t) DetailListCallback },
	{ (String) "DialogCallback",		(caddr_t) DialogCallback }
};
#define	NUM_NAME_REGISTER	(sizeof (namesRegister) / sizeof (namesRegister[0]))

static int		selectedProd;
static int		selectedPoint;
static ProdRecord	*prods;
static PipeDetails	*pipes;
static TankDetails	*tanks;

static double		totalBBLS;
static double		ullage;
static double		netVolume;
static double		apiGravity;

static int		numProds;
static int		currentList;

static XmString		nullString;

static XmString		errorMessageString;
static Widget		errorMessage;


static Widget		prodList;
static Widget		tankInvButton;
static Widget		otherInvButton;
static Widget		printInvButton;
static Widget		exitProdButton;
static Widget		detailList;
static Widget		detailTotals;
static Widget		exitDetailsButton;
static Widget		printDetailsButton;
static Widget		editGravityButton;
static Widget		saveInvButton;
static Widget		rfrshInvButton;

#ifdef	WIN32
static char		*pvmapName = 		"\\tis4000db\\pvmap.dbf";
static char		*pipeName = 		"\\tis4000db\\pipe.dbf";
static char		*pipeVarName = 		"\\tis4000db\\pipevars.dbf";
static char		*tankName =	 	"\\tis4000db\\tank1.dbf";
static char		*tankPrbName =	 	"\\tis4000db\\tankprbs.dbf";
static char		*tankVarName =	 	"\\tis4000db\\tankvars.dbf";
#else
static char		*pvmapName = 		"/tis4000db/pvmap.dbf";
static char		*pipeName = 		"/tis4000db/pipe.dbf";
static char		*pipeVarName = 		"/tis4000db/pipevars.dbf";
static char		*tankName =	 	"/tis4000db/tank1.dbf";
static char		*tankPrbName =	 	"/tis4000db/tankprbs.dbf";
static char		*tankVarName =	 	"/tis4000db/tankvars.dbf";
#endif	WIN32

	/*
	 *	The next three files will be prefixed with "/tis4000db" 
	 *	in the Store Product Inventory button callback.
	 */
static char		*ProdInvName = 			"prodinv.dbf";
static char		*TankInvName = 			"tankinv.dbf";
static char		*OtherInvName =			"otherinv.dbf";


static C4CODE		codeBaseDesc;
static D4DATA		*pvmapDesc;
static D4DATA		*pipeDesc;
static D4DATA		*pipeVarDesc;
static D4DATA		*tankDesc;
static D4DATA		*tankPrbDesc;
static D4DATA		*tankVarDesc;

static D4DATA		*ProdInvDesc;
static D4DATA		*TankInvDesc;
static D4DATA		*OtherInvDesc;

static T4TAG		*pvmapTag;


static	char		productName[PRODUCT_ENTRY_SIZE+1];
static	char		pointName[POINTNAME_SIZE+1];
static	char		pointDescr[DESCRIPTOR_SIZE+1];
static	char		caChan[POINTNAME_SIZE+6];	/* Allow ".NNNN" field specifier */
static	char		auxChan[POINTNAME_SIZE+1];
static	chid		ChanId;
ChanDetails		TopChan, *currentChan;


static	PipeDetails	*currentPipe;
static	PipeDetails	*printPipe;
static	TankDetails	*currentTank;
static	TankDetails	*printTank;
static	ProdRecord	*currentProd;

time_t			CurTime;
struct	tm		CurTmElements;

char			txtBuf[BUFSIZ];
char			printCmd[BUFSIZ];


/*****************************************************************************
 *                                                                           *
 *      main prgram for lineup interface                                     *
 *                                                                           *
 *****************************************************************************/

void main (int argc, char *argv[])
{
	int			n;
	Arg			args[MAX_ARGS];

	tisOpenLog ("ProdInv", 6, 2, NULL, NULL);

	/*
	 * check program usage
	 */
	if (argc != 1)
	{
	    tisSysLog (0,0, "Usage: ProdInv");
	    exit (1);
	}

	/*
	 * initialize variables
	 */
	selectedProd = 0;
	currentChan = &TopChan;
	memset( &TopChan, 0, CHAN_DET_SIZE );
	errorMessage = NULL;
	errorMessageString = NULL;
	nullString = XmStringCreate ("", XmSTRING_DEFAULT_CHARSET);
	XmStringFree( nullString );

	/*
	 * initialize and open display
	 */
	MrmInitialize ();
	XtToolkitInitialize ();
	applContext = XtCreateApplicationContext ();
	display = XtOpenDisplay (applContext, NULL, argv[0], 
					"ProdInv", NULL, 0, &argc, argv);
	if ( !display)
	{
	    tisSysLog (0,0,"ProdInv: Cannot open display");
	    exit (1);
	}

	n = 0;
	XtSetArg( args[n], XmNallowShellResize, True ); n++;
	appShell = XtAppCreateShell ("Product Inventory", "ProdInv", 
				applicationShellWidgetClass, display, args, n);

	/*
	 * fetch main window widget and a couple of its children
	 */
	if (MrmOpenHierarchy (NUM_UID_FILES, uidFilenames, NULL, &mrmHierarchy) != MrmSUCCESS)
	{
	    tisSysLog (0,0, "Cannot open UID file %s", uidFilenames[0]);
	    exit (1);
	}
	MrmRegisterNames (namesRegister, NUM_NAME_REGISTER);

	if (MrmFetchWidget (mrmHierarchy, "productsWindow", appShell, 
			&productsWindow, &productsWindowClass) != MrmSUCCESS)
	{
	    tisSysLog (0,0, "Cannot fetch Product Inventory window widget");
	    exit (1);
	}

	if (MrmFetchWidget (mrmHierarchy, "detailsWindow", appShell, 
			&detailsWindow, &detailsWindowClass) != MrmSUCCESS)
	{
	    tisSysLog (0,0, "Cannot fetch Product Inventory Details window widget");
	    exit (1);
	}
	XtSetArg (args[n], XmNx, -5); n++;
	XtSetArg (args[n], XmNy, 80); n++;
	XtSetValues (detailsWindow, args, n);

	if (MrmFetchWidget (mrmHierarchy, "gravityDialog", appShell, 
			&gravityDialog, &gravityDialogClass) != MrmSUCCESS)
	{
	    tisSysLog (0,0, "Cannot fetch Product Inventory Dialog window widget");
	    exit (1);
	}
	XtUnmanageChild( XmSelectionBoxGetChild( gravityDialog, XmDIALOG_APPLY_BUTTON ));
	XtUnmanageChild( XmSelectionBoxGetChild( gravityDialog, XmDIALOG_HELP_BUTTON ));

	/*
	 * realize main window
	 */
	XtManageChild (productsWindow);
	XtRealizeWidget (appShell);
	UserMessage();			/* Please wait... */


	/*
	 * Initialize Channel Access 
	 */
	if( ca_task_initialize() != ECA_NORMAL )  {
		tisSysLog(0,0, "Can't initialize link to Channel Access." );
		exit(0);
	}

	/*
	 * open and read database
	 */
	d4init (&codeBaseDesc);
#ifdef	USECB63
	codeBaseDesc.readLock = 0;
	codeBaseDesc.errGo = 0;
	codeBaseDesc.errOpen = 0;
#else
	codeBaseDesc.read_lock = 0;
	codeBaseDesc.go_error = 0;
	codeBaseDesc.open_error = 0;
#endif	USECB63

	OpenDatabase ();
	ScanDatabase();
	ListProdTotals();

	/*
	 * get and dispatch events
	 */
	XtAppMainLoop (applContext);
}

static void	UserMessage()
{
	sprintf( txtBuf, "Gathering Product Inventory data.  Please wait..." );
	errorMessageString = XmStringCreate( txtBuf, XmSTRING_DEFAULT_CHARSET );
	XmListAddItem( prodList, errorMessageString, 0);
	piForceUpdate(productsWindow);
	return;
}

static void	piForceUpdate( Widget w)
{
	Widget	diashell,topshell;
	Window	diawindow,topwindow;
	XtAppContext cxt = XtWidgetToApplicationContext(w);
	Display	*dpy;
	XWindowAttributes	xwa;
	XEvent	event;

	for(diashell = w;!XtIsShell(diashell);diashell = XtParent(diashell));
	for(topshell = diashell;!XtIsTopLevelShell(topshell);topshell = XtParent(topshell));
	if(XtIsRealized(diashell) && XtIsRealized(topshell))  {
	    dpy = XtDisplay(topshell);
	    diawindow = XtWindow(diashell);
	    topwindow  = XtWindow(topshell);
	    while(XGetWindowAttributes(dpy,diawindow,&xwa) && xwa.map_state != IsViewable)  {
		if(XGetWindowAttributes(dpy,topwindow,&xwa) && xwa.map_state != IsViewable)
			break;
		XtAppNextEvent(cxt,&event);
		XtDispatchEvent(&event);
	    }
	}
	XmUpdateDisplay(topshell);
}

static void	ListProdTotals()
{
	numProds = 0;
	prods = ProdInv.prodList;

	/* Build display list of inventory totals  */
	XmListDeleteAllItems( prodList );	/* Remove the kludge from above. */
	while( prods )  {
	    prods -> nonIssuable = prods -> grossVolume - prods -> issuable;
	    sprintf (txtBuf, "%-16.16s %-3.3s    %8.1f      %8.1f       %8.1f      %8.1f       %8.1f     %8.1f   %10.3f",
			prods -> product, prods -> onOff, prods -> capacity,
			prods -> grossVolume, prods -> bswVolume, prods -> ullage,
			prods -> nonIssuable, prods -> issuable,
			prods -> netVolume );

	    prods->prodStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
	    numProds++;
	    XmListAddItem (prodList, prods->prodStr, 0);
	    XmStringFree( prods->prodStr );
/*	    printf( "Product # %d in inventory -> %s.\n", numProds, prods->product );	*/
	    prods = prods -> NextProd;
	}


	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      OpenDatabase                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          open dBase database files                                        *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void OpenDatabase ()
{

	if ( !(pvmapDesc = d4open (&codeBaseDesc, pvmapName)))
	{
	    tisSysLog (0,0, "Cannot open TIS-4000 database file \"%s\".", 
								pvmapName);
	    exit (0);
	}
	d4unlock_all(pvmapDesc);

	if ( !(pipeDesc = d4open (&codeBaseDesc, pipeName)))
	{
	    tisSysLog (0,0, "Cannot open TIS-4000 database file \"%s\".", pipeName);
	    exit (0);
	}
	d4unlock_all(pipeDesc);

	if ( !(pipeVarDesc = d4open (&codeBaseDesc, pipeVarName)))
	{
	    tisSysLog (0,0, "Cannot open TIS-4000 database file \"%s\".", 
								pipeVarName);
	    exit (0);
	}
	d4unlock_all(pipeVarDesc);

	if ( !(tankDesc = d4open (&codeBaseDesc, tankName)))
	{
	    tisSysLog (0,0, "Cannot open TIS-4000 database file \"%s\".", tankName);
	    exit (0);
	}
	d4unlock_all(tankDesc);

	if ( !(tankPrbDesc = d4open (&codeBaseDesc, tankPrbName)))
	{
	    tisSysLog (0,0, "Cannot open TIS-4000 database file \"%s\".", 
								tankPrbName);
	    exit (0);
	}
	d4unlock_all(tankPrbDesc);

	if ( !(tankVarDesc = d4open (&codeBaseDesc, tankVarName)))
	{
	    tisSysLog (0,0, "Cannot open TIS-4000 database file \"%s\".", 
								tankVarName);
	    exit (0);
	}
	d4unlock_all(tankVarDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ScanDatabase                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Scan pvmap for tank & pipe pts. then retrieve the product 	     *
 *	    data that has to be presented on the form.			     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void ScanDatabase ()
{
	int	rv;
	char	ptype[16];
	long	numRecs, curRec;

	/* Roll through pvmap.dbf, noting each and every pipe & tank point,
	 * as well as the product and certain channel access data that is
	 * to be reported on the screen.
	 */

	time( &CurTime );	/* Record time of data fetching */
	CurTmElements = *(localtime( &CurTime ));


	pvmapTag = d4tag( pvmapDesc, "NAME" );
	d4tag_select( pvmapDesc, pvmapTag );
	d4unlock_all(pvmapDesc);

	numRecs = d4reccount( pipeDesc );
	for( curRec = 1; curRec <= numRecs; curRec++ )  {
	    if( d4go( pipeDesc, curRec ) < 0 )  {
		printf( "ScanDatabase: Error accessing \"pipe.dbf\".  Code = %d\n", 
#ifdef	USECB63
							codeBaseDesc.errorCode );
#else
							codeBaseDesc.error_code );
#endif	USECB63
		exit(0);
	    }

	    /* Ignore deleted records. */
	    if( d4deleted( pipeDesc ))		continue;

	    /* Get name, but leave intact - it's the seek string. */
	    memset( pointName, 0, sizeof pointName );
	    strncpy( pointName, f4str( d4field( pipeDesc, "NAME" )),
				f4len( d4field( pipeDesc, "NAME" )));

	    /* Lookup point in pvmap.dbf before continuing. */
	    d4top( pvmapDesc );
	    rv = d4seek( pvmapDesc, pointName );
	    d4unlock_all(pvmapDesc);
	    if( rv )	continue;		/* Point isn't in pvmap.dbf */
	    if( d4deleted( pvmapDesc ))		continue;

	    memset( ptype, 0, sizeof ptype );
	    strncpy( ptype, f4str( d4field( pvmapDesc, "TYPE" )),
			    f4len( d4field( pvmapDesc, "TYPE" )));
	    d4unlock_all(pvmapDesc);

	    StripWhitespace( ptype );

	    /* Get descriptor, but leave intact. */
	    memset( pointDescr, 0, sizeof pointDescr );
	    strncpy( pointDescr, f4str( d4field( pvmapDesc, "DESC" )),
				f4len( d4field( pvmapDesc, "DESC" )));
	    d4unlock_all(pvmapDesc);

	    updatePipeInv();

 	}  /* End for */


	numRecs = d4reccount( tankDesc );
	for( curRec = 1; curRec <= numRecs; curRec++ )  {
	    if( d4go( tankDesc, curRec ) < 0 )  {
		printf( "ScanDatabase: Error accessing \"tank.dbf\".  Code = %d\n",
#ifdef	USECB63
							codeBaseDesc.errorCode );
#else
							codeBaseDesc.error_code );
#endif	USECB63
		exit(0);
	    }

	    /* Ignore deleted records. */
	    if( d4deleted( tankDesc ))		continue;

	    /* Get name, but leave intact - it's the seek string. */
	    memset( pointName, 0, sizeof pointName );
	    strncpy( pointName, f4str( d4field( tankDesc, "NAME" )),
				f4len( d4field( tankDesc, "NAME" )));

	    /* Lookup point in pvmap.dbf before continuing. */
	    d4top( pvmapDesc );
	    rv = d4seek( pvmapDesc, pointName );
	    d4unlock_all(pvmapDesc);
	    if( rv )	continue;		/* Point isn't in pvmap.dbf */
	    if( d4deleted( pvmapDesc ))		continue;

	    memset( ptype, 0, sizeof ptype );
	    strncpy( ptype, f4str( d4field( pvmapDesc, "TYPE" )),
			    f4len( d4field( pvmapDesc, "TYPE" )));
	    d4unlock_all(pvmapDesc);

	    StripWhitespace( ptype );

	    /* Get descriptor, but leave intact. */
	    memset( pointDescr, 0, sizeof pointDescr );
	    strncpy( pointDescr, f4str( d4field( pvmapDesc, "DESC" )),
				f4len( d4field( pvmapDesc, "DESC" )));
	    d4unlock_all(pvmapDesc);

	    updateTankInv();

 	}  /* End for */


	d4close( pvmapDesc );
	d4close( pipeDesc );
	d4close( pipeVarDesc );
	d4close( tankDesc );
	d4close( tankPrbDesc );
	d4close( tankVarDesc );

	return;

}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      updatePipeInv                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Obtain the product the product for this point from pipe.dbf.     *
 *	    Locate this product's record in the product inventor list -      *
 *	    create it if it doesn't exist.  Finally, update the pipe list    *
 *	    for this product by adding an entry for this point at the end of *
 *	    the list.							     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void updatePipeInv ()
{
	T4TAG		*pipeVarTag;
	int		i;

	/* Set up for pipevars.dbf search */
	d4top( pipeVarDesc );
	pipeVarTag = d4tag( pipeVarDesc, "NAMEID" );
	d4tag_select( pipeVarDesc, pipeVarTag );

	memset( productName, 0, sizeof productName );
	strncpy( productName, f4str( d4field( pipeDesc, "PRODUCT" )),
			      f4len( d4field( pipeDesc, "PRODUCT" )));
	StripWhitespace( productName );

	/* Locate product in inventory list. */
	if( ProdInv.numProds )  {
	    for( i = 0, currentProd = ProdInv.prodList; i < ProdInv.numProds; i++ )  {
		if( !strcmp( productName, currentProd->product ))
			break;

		if( !currentProd -> NextProd )
			break;

		currentProd = currentProd -> NextProd;
	    }
	}

	if(( !ProdInv.numProds ) || ( i == ProdInv.numProds ) || (( !currentProd -> NextProd ) && strcmp(productName, currentProd->product)))  {
	    /* Product is not in inventory list.  Make new entry. */
	    newProdEntry();
	}

	/* Add Pipe Detail record to current product's list and fetch
	 * the necessary values for the point using channel access.
	 */
	currentPipe = currentProd -> pipeList;

	while( currentPipe -> NextPipe )  {
	    currentPipe = currentPipe -> NextPipe;
	}

	if( !currentPipe -> NextPipe )  {
	    currentPipe -> NextPipe = (PipeDetails *)calloc( 1, sizeof( PipeDetails ));
	    if( !currentPipe -> NextPipe )  {
		tisSysLog( 0,0, "Not enough memory to store Product information for pipe pt., %s\n",
					pointName );
		exit(0);
	    }
	    ++currentProd -> numPipes;
	}

	strcpy( currentPipe -> pointName, pointName );
	strcpy( currentPipe -> pointDescr, pointDescr );

	PipeChanAccess();	/* Perform channel access for pipe elements. */

	return;

}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      updateTankInv                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Obtain the product the product for this point from tank.dbf.     *
 *	    Locate this product's record in the product inventor list -      *
 *	    create it if it doesn't exist.  Finally, update the tank list    *
 *	    for this product by adding an entry for this point at the end of *
 *	    the list.							     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void updateTankInv ()
{
	T4TAG		*tankPrbTag, *tankVarTag;
	int		i;


	/* Set up for tankprbs.dbf search */
	d4top( tankPrbDesc );
	tankPrbTag = d4tag( tankPrbDesc, "NAMEID" );
	d4tag_select( tankPrbDesc, tankPrbTag );

	/* Set up for tankvars.dbf search */
	d4top( tankVarDesc );
	tankVarTag = d4tag( tankVarDesc, "NAMEID" );
	d4tag_select( tankVarDesc, tankVarTag );

	memset( productName, 0, sizeof productName );
	strncpy( productName, f4str( d4field( tankDesc, "PRODUCT" )),
			      f4len( d4field( tankDesc, "PRODUCT" )));
	StripWhitespace( productName );

	/* Locate product in inventory list. */
	if( ProdInv.numProds )  {
	    for( i = 0, currentProd = ProdInv.prodList; i < ProdInv.numProds; i++ )  {
		if( !strcmp( productName, currentProd->product ))
			break;

		if( !currentProd -> NextProd )
			break;

		currentProd = currentProd -> NextProd;
	    }
	}

	if(( !ProdInv.numProds ) || ( i == ProdInv.numProds ) || (( !currentProd -> NextProd ) && strcmp(productName, currentProd->product)))  {
	    /* Product is not in inventory list.  Make new entry. */
	    newProdEntry();
	}

	/* Add Tank Detail record to current product's list and fetch
	 * the necessary values for the point using channel access.
	 */
	currentTank = currentProd -> tankList;

	while( currentTank -> NextTank )  {
	    currentTank = currentTank -> NextTank;
	}

	if( !currentTank -> NextTank )  {
	    currentTank -> NextTank = (TankDetails *)calloc( 1, sizeof( TankDetails ));
	    if( !currentTank -> NextTank )  {
		tisSysLog(0,0, "Not enough memory to store Product information for tank pt., %s\n",
					pointName );
		exit(0);
	    }
	    ++currentProd -> numTanks;
	}

	strcpy( currentTank -> pointName, pointName );
	strcpy( currentTank -> pointDescr, pointDescr );

	TankChanAccess();

	return;

}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      newProdEntry                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Create new product list item 				     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void newProdEntry ()
{

	if( !ProdInv.prodList ) {
	    /* First product list entry */
	    ProdInv.prodList = (ProdRecord *)calloc( 1, sizeof( ProdRecord ));
	    currentProd = ProdInv.prodList;
	    if( !currentProd )  {
		tisSysLog(0,0, "newProdEntry:  Not enough memory to to create product entry for %s",
				productName );
		exit(0);
	    }
	}
	else  {
	    /* All subsequent product list entries */
	    currentProd -> NextProd = (ProdRecord *)calloc( 1, sizeof( ProdRecord ));
	    if( !currentProd -> NextProd )  {
		tisSysLog(0,0, "newProdEntry:  Not enough memory to to create product entry for %s",
				productName );
		exit(0);
	    }
	    currentProd = currentProd -> NextProd;
	}

	strcpy( currentProd -> product, productName );

	/* Allocate first pipe & tank list entries */
	currentProd -> pipeList = (PipeDetails *)calloc( 1, sizeof( PipeDetails ));
	if( !currentProd -> pipeList )  {
		tisSysLog(0,0, "newProdEntry:  Not enough memory to initialize product entry for %s",
					productName );
		exit(0);
	}

	currentProd -> tankList = (TankDetails *)calloc( 1, sizeof( TankDetails ));
	if( !currentProd -> tankList )  {
		tisSysLog(0,0, "newProdEntry:  Not enough memory to initialize product entry for %s",
					productName );
		exit(0);
	}


	ProdInv.numProds++;
	strcpy( currentProd -> onOff, "On" );		/* It should stay that way, unless ... */

	/* Just in case ... */
	currentProd -> numPipes = 0;
	currentProd -> numTanks = 0;

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      getVolUnits                                                          *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Get enum value of volume units 				     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void getVolUnits ( short	*v_index)	/* Buffer for returned value */
{
	/* Fetch the VOLU field for the PV */
	strcpy( caChan, pointName );
	StripWhitespace( caChan );
	strcat( caChan, ".VOLU" );

	CASearch( caChan );	/* Set ChanId for this channel. */

	if( ca_get( OLD_DBR_ENUM, ChanId, (void *)v_index ) != ECA_NORMAL )  {
		tisSysLog(0,0, "getVolUnits: Lost connection to volume measurement units for %s",
							pointName );
		exit(0);
	}

	if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
		tisSysLog(0,0, "getVolUnits: Timeout waiting for volume measurement units storage for %s",

						caChan );
		exit(0);
	}


	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      getLevelUnits                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none		                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Retrieves units for level value  				     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void getLevelUnits ( short *LevUnitsVal)
{
	char		LevChan[POINTNAME_SIZE];

	strcpy( LevChan, pointName );
	StripWhitespace( LevChan );
	strcat( LevChan, ".LEVU" );


	CASearch( LevChan );	/* Set ChanId for this channel. */

	if( ca_get( OLD_DBR_ENUM, ChanId, (void *)LevUnitsVal ) != ECA_NORMAL )  {
		tisSysLog(0,0, "getLevelUnits: Lost connection to Level Units for %s",
					pointName );
		exit(0);
	}

	if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
		tisSysLog(0,0, "getLevelUnits: Timeout waiting for auxiliary data storage for %s",
					caChan );
		exit(0);
	}

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      PipeChanAccess                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none		                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Performs channel access for the current pipe point's data.	     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void PipeChanAccess()
{
	int		rv;
	int		i;
	char		OnOffStates[2];		/* MODE state of Pressure & Temperature PVs */
	char		seekString[50];		/* Should be plenty */
	short		e_val;			/* enum value of mode */
	static char	*(pipeFields[])  = {
		".VAL",		/* Current total - accum for product also */
		".ULLG",		/* Gross Ullage - accum for product also */
		".CTEM",		/* Temperature */
		".SGRV",		/* API Specific Gravity */
		".TEXP",		/* Factor */
		".NETV",		/* Net BBLs - accumulate for product also */
		".PRPV",		/* Need to fetch MODE for Pressure PV */
		".TEPV",		/* Need to fetch MODE for Temperature PV */
		".GCAP",		/* Just accumulat for product */
		   0
	};
	double		runningTot, cnvFactor;

	/* Determine conversion factor for volumes */
	getVolUnits( &e_val );
	cnvFactor = CnvTable[e_val].CnvFactor;

	/* Fetch the fields necessary for display. */
	i = 0;
	OnOffStates[0] = OnOffStates[1] = 0;
	do {
	    if(( i < 6 ) || ( i > 7 )) {
		strcpy( caChan, pointName );
		StripWhitespace( caChan );
		strcat( caChan, pipeFields[i] );
		CASearch( caChan );		/* SetChanId for this caChan. */
	    } /* End if */

	    /* Channel connection has been established.  Get the data. */
	    switch( i )  {
		case 0:		/* .VAL */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> totalBBLS )
						!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
		    	exit(0);
		    }

		    /* Update the Gross Volume for the PRODUCT */
		    currentPipe -> totalBBLS *= cnvFactor;
		    currentProd -> grossVolume += currentPipe -> totalBBLS;

		    break;


		case 1:		/* .ULLG */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> ullage )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", caChan );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
		    	exit(0);
		    }

		    /* Update the Ullage for the PRODUCT */
		    currentPipe -> ullage *= cnvFactor;
		    currentProd -> ullage += currentPipe -> ullage;

		    break;


		case 2:		/* .CTEM */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> temp )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", caChan );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
						caChan );
		    	exit(0);
		    }

		    break;


		case 3:		/* .SGRV */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> apiGravity )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", caChan );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
		    	exit(0);
		    }

		    break;


		case 4:		/* .TEXP */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> factor )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", ChanId );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
		    	exit(0);
		    }

		    break;


		case 5:		/* .NETV */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> netVolume )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", ChanId );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
		    	exit(0);
		    }

		    /* Update the Net Volume for the PRODUCT */
		    currentPipe -> netVolume *= cnvFactor;
		    currentProd -> netVolume += currentPipe -> netVolume;

		    break;


		case 6:		/* .PRPV */
		case 7:		/* .TEPV */
		    /* Fetch, then fetch again */
		    memset( seekString, 0, sizeof seekString );
		    if( i == 6 )  {
			sprintf( seekString, "%-*.*sPressure PV     ", 
					POINTNAME_SIZE, POINTNAME_SIZE, pointName );
		    }
		    else {
			sprintf( seekString, "%-*.*sTemp PV         ", 
					POINTNAME_SIZE, POINTNAME_SIZE, pointName );
		    }
		    d4top( pipeVarDesc );
		    rv = d4seek( pipeVarDesc, seekString );
		    if( rv )  {
			tisSysLog(0,0, "PipeChanAccess: Unable to obtain auxiliary PV for %s",
						pointName );
		 	exit(0);
		    }

		    /* Extract the name - make sure it's nonBlank */
		    strncpy( auxChan, f4str( d4field( pipeVarDesc, "VARNAME" )),
				      f4len( d4field( pipeVarDesc, "VARNAME" )));

		    if( auxChan[0] <= ' ' )  {
			/* Nothing spec'd */
			OnOffStates[i-6] = TRUE;
			break;
		    }

		    /* Fetch the OSCN field for the auxiliary PV */
		    StripWhitespace( auxChan );
		    strcat( auxChan, ".OSCN" );

		    CASearch( auxChan );	/* Set ChanId for auxChan. */

		    if( ca_get( OLD_DBR_ENUM, ChanId, (void *)&e_val ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Lost connection to auxiliary data for %s",
							pointName );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
						caChan );
			exit(0);
		    }

		    /* Set PrOn or TmpOn (depending on 'i') based on return value */
		    OnOffStates[i-6] = (e_val == YES) ? TRUE : FALSE;
		    break;

		case 8:		/* .GCAP */
		    runningTot = 0;
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&runningTot )
						!= ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Can't fetch pipe data, %s", caChan );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "PipeChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Update the Gross Capacity for the PRODUCT */
		    currentProd -> capacity += ( runningTot * cnvFactor );
		    break;

	    }  /* End switch */

	    i++;
	} while( pipeFields[i] );

	/* Determine On/Off state for the point ( and, implicitly, the product) */
	for( i = 0; (i < sizeof OnOffStates) && (OnOffStates[i] == TRUE); i++ );
	if( i < sizeof OnOffStates )  {
		strcpy( currentPipe -> onOff, "Off" );
		strcpy( currentProd -> onOff, "Off" );
	}
	else  {
		strcpy( currentPipe -> onOff, "On" );
	}


	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      TankChanAccess                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none		                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Performs channel access for the current tank point's data.	     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void TankChanAccess()
{
	int		rv;
	int		i;
	short		e_val;		/* Buffer for enum data */
	char		dateString[MAX_STRING_SIZE+1];
	char		seekString[50];
	char		OnOffStates[3];	/* MODE state of auxiliary PVs */

	static char	*(tankFields[]) = {
	/* 0 */		".PLEV",	/* Unconverted level */
	/* 1 */		".GVOL",	/* Total BBLS - accumulate for product also */
	/* 2 */		".BSWL",	/* Unconverted BSW level - use LEVU for conversion */
	/* 3 */		".ULLG",	/* Gross Ullage - accumulate for product also */
	/* 4 */		".TPAV",	/* Temperature */
	/* 5 */		".SGRV",	/* API Specific Gravity */
	/* 6 */		".APID",	/* API Date */
	/* 7 */		".TEXP",	/* Factor */
	/* 8 */		".NETV",	/* Net BBLs - accumulate for product also */
	/* 9 */	".GCAP",	/* Just accumulate for product */
	/* 10 */	".BSWV",	/* Just accumulate for product */
	/* 11 */	".AVAP",	/* Just accumulate for product - to be subtracted from total GVOL */
	/* 12 */	".LEVP",	/* Need MODE fetched for point */
	/* 13 */	".BSWP",	/* Need MODE fetched for point */
	/* 14 */	".TMPP",	/* Need MODE fetched for Average Temperature Probe */
			    0
	};

	double		runningTotal, cnvFactor;	/* Used for product inventory - not tank */


	/* Determine conversion factor for volumes */
	getVolUnits( &e_val );
	cnvFactor = CnvTable[e_val].CnvFactor;

	/* Fetch the fields necessary for display. */
	i = 0;
	memset( OnOffStates, 0, sizeof OnOffStates );
	do {
	    if( i < 12 )  {
		strcpy( caChan, pointName );
		StripWhitespace( caChan );
		strcat( caChan, tankFields[i] );
		CASearch( caChan );		/* Set ChanId for caChan. */
	    } /* End if */

	    /* Channel connection has been established.  Get the data. */
	    switch( i )  {
		case 0:		/* .PLEV */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&runningTotal )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Check level units and convert accordingly */
		    getLevelUnits( &e_val );
		    switch( e_val )  {
	    		case METERS:
			    runningTotal *= 39.37;	/* Inches first */
	    		case INCHES:
			    currentTank -> levelFt = runningTotal / 12;
			    currentTank -> levelIn = (runningTotal - (currentTank -> levelFt * 12));
			    currentTank -> level8ths = (runningTotal - (currentTank -> levelFt * 12 ) - currentTank -> levelIn ) * 8;
			    break;

	    		case FEET:
			    currentTank -> levelFt = (long)runningTotal;
			    currentTank -> levelIn = (runningTotal - currentTank -> levelFt) * 12;
			    currentTank -> level8ths = (((runningTotal - currentTank -> levelFt) * 12) - currentTank -> levelIn ) * 8;
			    break;

		    }  /* End switch */

		    if( currentTank -> level8ths >= 8.0 )  {
			/* Due to integer rounding, correct the level's parts. */
			currentTank -> level8ths -= 8;	/* In case it went > 8 */
			++currentTank -> levelIn;
		    }
		    else if( currentTank -> level8ths >= 7.95 )  {
			    /* Due to display rounding, correct the level's parts. */
			    currentTank -> level8ths = 0.0;
			    ++currentTank -> levelIn;
			}

		    if( currentTank -> levelIn >= 12 )  {
			/* Due to rounding, correct the level's parts. */
			currentTank -> levelIn -= 12;	/* In case it went > 12 */
			++currentTank -> levelFt;
		    }

		    break;


		case 1:		/* .GVOL */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> totalBBLS )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Update product's total */
		    currentTank -> totalBBLS *= cnvFactor;
		    currentProd -> grossVolume += currentTank -> totalBBLS;
		    break;

		case 2:		/* .BSWL */
		    runningTotal = 0;
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&runningTotal  )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Check level units and convert accordingly */
		    getLevelUnits( &e_val );
		    switch( e_val )  {
	    		case METERS:
			    runningTotal *= 39.37;	/* Inches first */
	    		case INCHES:
			    currentTank -> BSWInches = runningTotal;
			    currentTank -> BSW8ths = 
				    (runningTotal-currentTank->BSWInches)*8;
			    break;

	    		case FEET:
			    runningTotal *= 12;
			    currentTank -> BSWInches = runningTotal;
			    currentTank -> BSW8ths = (runningTotal - currentTank -> BSWInches) * 8;
			    break;

		    }  /* End switch */

		    if( currentTank -> BSW8ths >= 8.0 )  {
			currentTank -> BSW8ths -= 8;
			++currentTank -> BSWInches;
		    }
		    else if( currentTank -> BSW8ths >= 7.95 )  {
			    /* Due to display rounding, correct the level's parts. */
			    currentTank -> BSW8ths = 0.0;
			    ++currentTank -> BSWInches;
			}

		    break;

		case 3:		/* .ULLG */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> ullage )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
						caChan );
			exit(0);
		    }

		    /* Update product's total */
		    currentTank -> ullage *= cnvFactor;
		    currentProd -> ullage += currentTank -> ullage;
		    break;

		case 4:		/* .TPAV */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> temp )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    break;

		case 5:		/* .SGRV */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> apiGravity )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    break;

		case 6:		/* .APID */
		    memset( dateString, ' ', sizeof dateString );
		    if( ca_get( OLD_DBR_STRING, ChanId, (void *)dateString )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    strcpy( currentTank -> apiDate, dateString );
		    break;

		case 7:		/* .TEXP */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> factor )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    break;

		case 8:		/* .NETV */
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> netVolume )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Update product's total */
		    currentTank -> netVolume *= cnvFactor;
		    currentProd -> netVolume += currentTank -> netVolume;
		    break;

		case 9:	/* .GCAP */
		    runningTotal = 0;
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&runningTotal )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Update product's total */
		    currentProd -> capacity += ( runningTotal * cnvFactor );
		    break;

		case 10:	/* .BSWV */
		    runningTotal = 0;
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&runningTotal )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Update product's total */
		    currentProd -> bswVolume += ( runningTotal * cnvFactor );
		    break;

		case 11:	/* .AVAP */
		    runningTotal = 0;
		    if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&runningTotal )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Can't fetch tank data, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Update product's total */
		    currentProd -> issuable += ( runningTotal * cnvFactor );
		    break;

		case 12:	/* .LEVP */
		case 13:
		    memset( seekString, 0, sizeof seekString );
		    if( i == 12 )  {
			sprintf( seekString, "%-*.*sLevel PV        ", 
					POINTNAME_SIZE, POINTNAME_SIZE, pointName );
		    }
		    else {
			sprintf( seekString, "%-*.*sBSW PV          ", 
					POINTNAME_SIZE, POINTNAME_SIZE, pointName );
		    }
		    d4top( tankVarDesc );
		    rv = d4seek( tankVarDesc, seekString );
		    if( rv )  {
			tisSysLog(0,0, "TankChanAccess: Unable to obtain auxiliary PV for %s",
							pointName );
		 	exit(0);
		    }

		    /* Extract the name - make sure it's nonBlank */
		    strncpy( auxChan, f4str( d4field( tankVarDesc, "VARNAME" )),
				      f4len( d4field( tankVarDesc, "VARNAME" )));

		    if( auxChan[0] <= ' ' )  {
			/* Nothing spec'd */
			OnOffStates[i-12] = TRUE;
			break;
		    }

		    /* Fetch the OSCN field for the auxiliary PV */
		    StripWhitespace( auxChan );
		    strcat( auxChan, ".OSCN" );

		    CASearch( auxChan );	/* Set ChanId for auxChan. */

		    if( ca_get( OLD_DBR_ENUM, ChanId, (void *)&e_val ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Lost connection to auxiliary data for %s",
							pointName );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Set flag (depending on 'i') based on return value */
		    OnOffStates[i-12] = (e_val == YES) ? TRUE : FALSE;
		    break;

		case 14:	/* .TMPP */
		    memset( seekString, 0, sizeof seekString );
		    sprintf( seekString, "%-*.*s%-16.16s", 
					POINTNAME_SIZE, POINTNAME_SIZE, pointName, "Avg Temp PV" );
		    d4top( tankVarDesc );
		    rv = d4seek( tankVarDesc, seekString );
		    if( rv )  {
			tisSysLog(0,0, "Search string %s", seekString);
			tisSysLog(0,0, "Length of search string %d", strlen(seekString));
			tisSysLog(0,0, "TankChanAccess: Unable to obtain auxiliary PV for %s",
						pointName );
		 	exit(0);
		    }

		    /* Extract the name - make sure it's nonBlank */
		    strncpy( auxChan, f4str( d4field( tankVarDesc, "VARNAME" )),
				      f4len( d4field( tankVarDesc, "VARNAME" )));

		    if( auxChan[0] <= ' ' )  {
			/* Nothing spec'd */
			OnOffStates[i-12] = TRUE;
			break;
		    }

		    /* Fetch the OSCN field for the auxiliary PV */
		    StripWhitespace( auxChan );
		    strcat( auxChan, ".OSCN" );

		    CASearch( auxChan );	/* Set ChanId for auxChan. */

		    if( ca_get( OLD_DBR_ENUM, ChanId, (void *)&e_val ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Lost connection to auxiliary data for %s",
						pointName );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "TankChanAccess: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* Set on/off state (depending on 'i') based on return value */
		    OnOffStates[i-12] = (e_val == YES) ? TRUE : FALSE;
		    break;

	    }  /* End switch */

	    i++;
	} while( tankFields[i] );

	/* Determine On/Off state for the point ( and, implicitly, the product) */
	for( i = 0; (i < sizeof OnOffStates) && (OnOffStates[i] == TRUE); i++ );
	if( i < sizeof OnOffStates )  {
		strcpy( currentTank -> onOff, "Off" );
		strcpy( currentProd -> onOff, "Off" );
	}
	else  {
		strcpy( currentTank -> onOff, "On" );
	}

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      RegisterWidgets                                                      *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          register a widget that was created from the UIL.                 *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void RegisterWidgets (Widget widget, 
			     int *widgetID, 
			     XmAnyCallbackStruct *widgetData)
{
	int		n;
	Arg		args[MAX_ARGS];

	switch (*widgetID)
	{
	    case PROD_COLUMNS1:
		n = 0;
		XtSetArg (args[n], XmNlabelString, XmStringCreate ("Product          On/    Capacity      Gross Volume   BS&W Volume    Ullage     Non-Issuable      Issuable  Net Volume", XmSTRING_DEFAULT_CHARSET));
		n++;
		XtSetValues (widget, args, n);
		break;

	    case PROD_COLUMNS2:
		n = 0;
		XtSetArg (args[n], XmNlabelString, XmStringCreate ("                 Off     BBLS            BBLS           BBLS         BBLS          BBLS            BBLS      BBLS", XmSTRING_DEFAULT_CHARSET)); 
		n++;
		XtSetValues (widget, args, n);
		break;

	    case DETAIL_COLUMNS1:
		n = 0;
		XtSetArg (args[n], XmNlabelString, XmStringCreate ("                             On/       Level          Total      BS&W        Gross                 API        API", XmSTRING_DEFAULT_CHARSET)); 
		n++;
		XtSetValues (widget, args, n);
		break;

	    case DETAIL_COLUMNS2:
		n = 0;
		XtSetArg (args[n], XmNlabelString, XmStringCreate ("                             Off    Ft.  In.  8ths     BBLS     In.  8ths    Ullage     Temp     Gravity      Date       Factor   Net BBLS", XmSTRING_DEFAULT_CHARSET)); 
		n++;
		XtSetValues (widget, args, n);
		break;

	    case PROD_LIST:
		prodList = widget;
		break;

	    case DETAIL_LIST:
		detailList = widget;
		break;

	    case DETAIL_TOTALS:
		detailTotals = widget;
		break;

	    case TANK_BUTTON:
		tankInvButton = widget;
		break;

	    case PIPE_BUTTON:
		otherInvButton = widget;
		break;

	    case EXIT_BUTTON:
		exitProdButton = widget;
		break;

	    case SAVE_BUTTON:
		saveInvButton = widget;
		break;

	    case DETAIL_EXIT:
		exitDetailsButton = widget;
		break;

	    case DETAIL_EDIT:
		editGravityButton = widget;
		break;

	    case DETAIL_DIALOG:
/*		gravityDialog = widget;	 */
		break;

	    case PRINT_INV:
		printInvButton = widget;
		break;

	    case PRINT_DET:
		printDetailsButton = widget;
		break;

	    case REFRESH_BUTTON:
		rfrshInvButton = widget;
		break;

	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ButtonCallback                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for button presses                                      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void ButtonCallback (Widget widget, 
			    int *widgetID,
			    XmAnyCallbackStruct *widgetData)
{
	int			n;
	int			numItems;
	Arg			args[MAX_ARGS];
	char			txtBuf[BUFSIZ];


	numItems = 0;
#ifdef  REMOVED
	if(( !currentProd ) && (( *widgetID == PIPE_BUTTON )
				|| ( *widgetID == TANK_BUTTON )))
		return;		  /* There aren't any product items yet. */
#endif
	totalBBLS = 0.0;
	ullage = 0.0;
	netVolume = 0.0;

	switch (*widgetID)
	{
	    case TANK_BUTTON:
		buildTankList();
	        memset( txtBuf, 0, sizeof txtBuf );
	        sprintf( txtBuf, "Tank Inventory  -  %-s", currentProd -> product );
	        n = 0;
	        XtSetArg (args[n], XmNdialogTitle, XmStringCreate( txtBuf, XmSTRING_DEFAULT_CHARSET )); 
	        n++;
	        XtSetValues (detailsWindow, args, n);
		currentList = TANK_LIST;
		break;

	    case PIPE_BUTTON:
		buildPipeList();
	        memset( txtBuf, 0, sizeof txtBuf );
	        sprintf( txtBuf, "Other Inventory  -  %-s", currentProd -> product );
	        n = 0;
	        XtSetArg (args[n], XmNdialogTitle, XmStringCreate( txtBuf, XmSTRING_DEFAULT_CHARSET )); 
	        n++;
	        XtSetValues (detailsWindow, args, n);
		currentList = OTHER_LIST;
		break;

	    case EXIT_BUTTON:
		MrmCloseHierarchy( mrmHierarchy );
		d4close_all(&codeBaseDesc);
		ca_task_exit();
		currentProd = ProdInv.prodList;

		/* Free up allocated memory. */
		while( currentProd )  {
		    currentPipe = currentProd -> pipeList;
		    while( currentPipe -> NextPipe )  {
			free( currentPipe );
			currentPipe = currentPipe -> NextPipe;
		    }
		    free( currentPipe );	/* Get the last one. */

		    currentTank = currentProd -> tankList;
		    while( currentTank -> NextTank )  {
			free( currentTank );
			currentTank = currentTank -> NextTank;
		    }
		    free( currentTank );	/* Get the last one. */

		    free( currentProd );
		    currentProd = currentProd -> NextProd;
		}

		currentChan = &TopChan;
		while( currentChan -> nextChan )  {
		    free( currentChan );
		    currentChan = currentChan -> nextChan;
		}
		free( currentChan );	/* Get the last one. */

		exit (1);
		break;

	    case SAVE_BUTTON:
	        n = 0;
	        XtSetArg (args[n], XmNsensitive, False ); 
	        n++;
	        XtSetValues (saveInvButton, args, n);
		piForceUpdate(productsWindow);
		StoreProdInv();
		break;

	    case DETAIL_EDIT:
		XtManageChild (gravityDialog);
		break;

	    case DETAIL_EXIT:
		XtUnmanageChild( detailsWindow );
		break;

	    case PRINT_INV:
		PrintProdInv();
		break;

	    case PRINT_DET:
		PrintProdDet();
		break;

	    case REFRESH_BUTTON:
		XmListDeleteAllItems( prodList );	/* Clear the list. */
		UserMessage();
		time( &CurTime );	/* Record time of data fetching */
		CurTmElements = *(localtime( &CurTime ));

		RefreshProdInv();
		ListProdTotals();
	        XtVaSetValues (saveInvButton, XmNsensitive, True, NULL );
	        XtVaSetValues (tankInvButton, XmNsensitive, False, NULL );
	        XtVaSetValues (otherInvButton, XmNsensitive, False, NULL );
		piForceUpdate(productsWindow);
		break;

	}

	/* Open the details list */
	if(( *widgetID == TANK_BUTTON ) || ( *widgetID == PIPE_BUTTON ))  {
	    n = 0;
	    memset( txtBuf, 0, sizeof txtBuf );
	    sprintf( txtBuf, " Totals                                            %8.1f                %8.1f                                              %10.3f", 
				totalBBLS, ullage, netVolume );
	    XtSetArg (args[n], XmNlabelString, XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET)); 
	    n++;
	    XtSetValues (detailTotals, args, n);
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (editGravityButton, args, n);
	    n = 0;
	    XtManageChild (detailsWindow);
	}
	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      RefreshProdInv                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          routine to regather product inventory data                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void RefreshProdInv()
{
	numProds = ProdInv.numProds;
	currentProd = ProdInv.prodList;

	OpenDatabase();		/* Open files long enough to get data. */

	/* Use channel access to get inventory data and inventory totals  */
	while( currentProd )  {
	    /* Zero out this product's totals. */
	    currentProd -> capacity = 0.0;
	    currentProd -> grossVolume = 0.0;
	    currentProd -> bswVolume = 0.0;
	    currentProd -> ullage = 0.0;
	    currentProd -> nonIssuable = 0.0;
	    currentProd -> issuable = 0.0;
	    currentProd -> netVolume = 0.0;

	    currentPipe = currentProd -> pipeList;

	    while( currentPipe -> NextPipe )  {
		strcpy( pointName, currentPipe -> pointName );
		strcpy( pointDescr, currentPipe -> pointDescr );

		/* Zero out the details. */
		currentPipe -> totalBBLS = 0.0;
		currentPipe -> ullage = 0.0;
		currentPipe -> temp = 0.0;
		currentPipe -> apiGravity = 0.0;
		currentPipe -> factor = 0.0;
		currentPipe -> netVolume = 0.0;

		PipeChanAccess();	/* Get latest IOC data for point. */
		currentPipe = currentPipe -> NextPipe;
	    }

	    currentTank = currentProd -> tankList;

	    while( currentTank -> NextTank )  {
		strcpy( pointName, currentTank -> pointName );
		strcpy( pointDescr, currentTank -> pointDescr );

		/* Zero out the details. */
		currentTank -> levelFt = 0;
		currentTank -> levelIn = 0;
		currentTank -> level8ths = 0.0;
		currentTank -> totalBBLS = 0.0;
		currentTank -> BSWInches = 0;
		currentTank -> BSW8ths = 0.0;
		currentTank -> ullage = 0.0;
		currentTank -> temp = 0.0;
		currentTank -> apiGravity = 0.0;
		currentTank -> factor = 0.0;
		currentTank -> netVolume = 0.0;

		TankChanAccess();	/* Get latest IOC data for point. */
		currentTank = currentTank -> NextTank;
	    }

	    currentProd = currentProd -> NextProd;
	}

	d4close( pvmapDesc );
	d4close( pipeDesc );
	d4close( pipeVarDesc );
	d4close( tankDesc );
	d4close( tankPrbDesc );
	d4close( tankVarDesc );


	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DialogCallback                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for dialog button presses                               *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void DialogCallback (Widget widget,
			    int *widgetID,
			    XmSelectionBoxCallbackStruct *widgetData)
{
	char			txtBuf[BUFSIZ], *pTxt;
	char			dateBuf[APIDATE_SIZE];
	int			n;
	int			numItems;


	numItems = 0;
	n = 0;
	switch (*widgetID)
	{
	    case DIALOG_OK:
		/* Accept new API Gravity value and ca_put the sucker where it belongs */
	        memset( txtBuf, 0, sizeof txtBuf );
		XmStringGetLtoR( widgetData -> value, XmSTRING_DEFAULT_CHARSET, &pTxt );
		strcpy( txtBuf, pTxt );
		XtFree( pTxt );		/* Minimize memory "holes" */

		/* Use global 'apiGravity' for put verification. */
		if( currentList == TANK_LIST )  {
		    sprintf( caChan, "%s.SGRV", StripWhitespace( currentTank -> pointName ));
		    currentTank -> apiGravity = atof( txtBuf );
		}
		else  {
		    sprintf( caChan, "%s.SGRV", StripWhitespace( currentPipe -> pointName ));
		    currentPipe -> apiGravity = atof( txtBuf );
		}

		/* Use channel access to put the new API Gravity value. */
		CASearch( caChan );		/* Set ChanId for caChan. */

		if( currentList == TANK_LIST )  {
		    sprintf( currentTank -> apiDate, "%2d/%02d/%02d", (CurTmElements.tm_mon+1), 
						   CurTmElements.tm_mday, 
						   CurTmElements.tm_year % 100);
		    if( ca_put( OLD_DBR_DOUBLE, ChanId, (void *)&currentTank -> apiGravity )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Can't fetch tank data, %s", caChan );
			exit(0);
		    }
		}
		else  {
		    if( ca_put( OLD_DBR_DOUBLE, ChanId, (void *)&currentPipe -> apiGravity )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Can't fetch tank data, %s", caChan );
			exit(0);
		    }
		}

#if 0
		if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
#endif
		if( ca_flush_io( ) != ECA_NORMAL )  {
			tisSysLog(0,0,  "Callback: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		}

		/* See if the value made it.  Retrieval is the surest proof. */
		if( ca_get( OLD_DBR_DOUBLE, ChanId, (void *)&apiGravity )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Can't fetch tank data, %s", caChan );
			exit(0);
		}

		if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		}

		if( currentList == TANK_LIST )  {
		    /* Compose API Date (gravity timestamp) & update for point. */
		    memset( dateBuf, 0, sizeof dateBuf );	/* Verify put by fetching into here. */

		    sprintf( caChan, "%s.APID", StripWhitespace( currentTank -> pointName ));
		    CASearch( caChan );		/* Set ChanId for caChan. */

		    if( ca_put( OLD_DBR_STRING, ChanId, (void *)currentTank -> apiDate )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Can't update API Date data, %s", caChan );
			exit(0);
		    }

#if 0
		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
#endif
		    if( ca_flush_io( ) != ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		    /* See if the value made it.  Retrieval is the surest proof. */
		    if( ca_get( OLD_DBR_STRING, ChanId, (void *)dateBuf )
							!= ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Can't fetch API date, %s", caChan );
			exit(0);
		    }

		    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
			tisSysLog(0,0, "Callback: Timeout waiting for auxiliary data storage for %s",
							caChan );
			exit(0);
		    }

		}

		/* Rebuild the details list item so that the new value gets picked up. */
		if( currentList == TANK_LIST )  {
		    if( apiGravity != currentTank -> apiGravity )  {
			tisSysLog(0,0, "Callback: API Gravity change failed update verification." );
			exit(0);
		    }

		    if( strcmp( currentTank ->apiDate, dateBuf ))  {
			tisSysLog(0,0, "Callback: API Date change failed update verification." );
			exit(0);
		    }

    		    sprintf (txtBuf, "%-*.*s %-3.3s %6d   %2d  %3.1f  %8.1f %6d   %3.1f   %8.1f %8.1f %11.3f  %-8.8s  %10.4f  %10.3f",
			DESCRIPTOR_SIZE, DESCRIPTOR_SIZE, 
			currentTank -> pointDescr, currentTank -> onOff, currentTank -> levelFt,
			currentTank -> levelIn, currentTank -> level8ths, currentTank -> totalBBLS,
			currentTank -> BSWInches, currentTank -> BSW8ths, currentTank -> ullage,
			currentTank -> temp, currentTank -> apiGravity, currentTank -> apiDate,
			currentTank -> factor, currentTank -> netVolume );

    		    currentTank->tankStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
		    XmListReplaceItemsPos( detailList, &currentTank -> tankStr, 1, selectedPoint );
		    XmStringFree( currentTank -> tankStr );
		}
		else  {
		    if( apiGravity != currentPipe -> apiGravity )  {
			tisSysLog(0,0, "Callback: API Gravity change failed update verification." );
			exit(0);
		    }
		    sprintf (txtBuf, "%-*.*s %-3.3s                   %8.1f                %8.1f %8.1f %11.3f             %10.4f  %10.3f",
			DESCRIPTOR_SIZE, DESCRIPTOR_SIZE, 
			currentPipe -> pointDescr, currentPipe -> onOff, currentPipe -> totalBBLS,
			currentPipe -> ullage, currentPipe -> temp, currentPipe -> apiGravity,
			currentPipe -> factor, currentPipe -> netVolume );

		    currentPipe->pipeStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
		    XmListReplaceItemsPos( detailList, &currentPipe -> pipeStr, 1, selectedPoint );
		    XmStringFree( currentPipe -> pipeStr );
		}
		/* Now redisplay the detail list with the new value. */
		XtUnmanageChild (gravityDialog);
		XmListSelectPos( detailList, selectedPoint, 1 );	/* Keep item selected */
		break;

	    case DIALOG_CANCEL:
		/* Nothing really to do.  Just unmanage the dialog box. */
	        memset( txtBuf, 0, sizeof txtBuf );
		XtUnmanageChild (gravityDialog);
		break;

	}
	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      CASearch	                                                     *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          maintain list of chids for channel access                        *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void CASearch( char ChanStr[])
{
	/* Do channel lookup in list. */
	currentChan = &TopChan;
	while( currentChan -> nextChan )  {
		if( !(strcmp( ChanStr, currentChan -> chanName )))
			break;

		currentChan = currentChan -> nextChan;
	}

	if(( !currentChan -> nextChan ) && ( strcmp( ChanStr, currentChan -> chanName )))  {
	    /* Setup a new channel connection. */
	    currentChan -> nextChan = (ChanDetails *)calloc( 1, CHAN_DET_SIZE );
	    currentChan = currentChan -> nextChan;
	    strcpy( currentChan -> chanName, ChanStr );

	    memset( &ChanId, 0, sizeof ChanId );
	    if( ca_search( ChanStr, &ChanId ) != ECA_NORMAL )  {
		tisSysLog(0,0, "CASearch: Can't find %s", ChanStr );
		exit(0);
	    }

	    if( ca_pend_io( 20.0 ) != ECA_NORMAL )  {
		tisSysLog(0,0, "CASearch: Timeout waiting for %s data", ChanStr );
		exit(0);
	    }

	    currentChan -> ChanId = ChanId;
	}
	else  {
		ChanId = currentChan -> ChanId;
	}
	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ProdListCallback                                                     *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for the product totals list                             *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void ProdListCallback (Widget widget,
			      int *widgetID,
			      XmListCallbackStruct *widgetData)
{
	int		*positionList;
	int		positionCount;
	int		i;
	int		n;
	Arg		args[MAX_ARGS];
	char		*SelectionText;

	n = 0;
	if (XmListGetSelectedPos (prodList, &positionList, &positionCount))
	{
	    XtSetArg (args[n], XmNsensitive, True); n++;
	    XtSetValues (tankInvButton, args, n);
	    XtSetArg (args[n], XmNsensitive, True); n++;
	    XtSetValues (otherInvButton, args, n);
	    selectedProd = positionList[0];
	    XtFree ((char *)positionList);

	    /* Scan the list item for the product name so that the
	     * "currentProd" pointer can be properly set.
	     */
	    XmStringGetLtoR( widgetData -> item, XmSTRING_DEFAULT_CHARSET, &SelectionText );
	    strcpy( txtBuf, SelectionText );
	    XtFree( SelectionText );		/* Minimize memory "holes" */

	    sscanf( txtBuf, "%16s", productName );
	    for( i=0, currentProd = ProdInv.prodList; i < ProdInv.numProds; i++ )  {
		if( !strcmp( productName, currentProd -> product ))
		    break;

		currentProd = currentProd -> NextProd;
	    }

	}
	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DetailListCallback                                                   *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for the product details list                            *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void DetailListCallback (Widget widget,
				int *widgetID,
				XmListCallbackStruct *widgetData)
{
	int		i;
	int		n;
	Arg		args[MAX_ARGS];
	char		*SelectionText;

	n = 0;
	XtSetArg (args[n], XmNsensitive, True); n++;
	XtSetValues (editGravityButton, args, n);
	selectedPoint = widgetData->item_position;

	/* Scan the list item for the product name so that the
	 * "currentProd" pointer can be properly set.
	 */
	XmStringGetLtoR( widgetData -> item, XmSTRING_DEFAULT_CHARSET, &SelectionText );
	strcpy( txtBuf, SelectionText );
	XtFree( SelectionText );		/* Minimize memory "holes" */

	strncpy( pointDescr, txtBuf, DESCRIPTOR_SIZE );


	/* Locate the point in the appropriate list, based
	 * on the value of currentList.
	 */
	if( currentList == TANK_LIST )  {
	    /* Point can be found in the current product's tank list */
	    for( i=0, currentTank = currentProd -> tankList; i < currentProd -> numTanks; i++ )  {
		if( !strncmp( pointDescr, currentTank -> pointDescr, strlen( pointDescr )))
		    break;

		currentTank = currentTank -> NextTank;
	    }
	    apiGravity = currentTank -> apiGravity;
	}
	else  {
	    /* Point can be found in the current product's pipe list */
	    for( i=0, currentPipe = currentProd -> pipeList; i < currentProd -> numPipes; i++ )  {
		if( !strncmp( pointDescr, currentPipe -> pointDescr, strlen( pointDescr )))
		    break;

		currentPipe = currentPipe -> NextPipe;
	    }
	    apiGravity = currentPipe -> apiGravity;
	}

	/* Update the gravity value that should be displayed. */
	memset( txtBuf, 0, sizeof txtBuf );
	sprintf( txtBuf, "%10.3f", apiGravity );
	n = 0;
	XtSetArg (args[n], XmNtextString, XmStringCreate( txtBuf, XmSTRING_DEFAULT_CHARSET )); 
	n++;
	XtSetValues (gravityDialog, args, n);

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      StripWhitespace                                                      *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          string      pointer to character string                          *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          pointer to first non-space character in string                   *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          strip leading and trailing whitespace from character string      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static char *StripWhitespace ( char *string)
{
	int	i;

	for (i=strlen (string)-1; i >= 0 && isspace(string[i]); string[i--] = 0)
	    ;
	for (i=0; string[i] && isspace(string[i]); i++)
	    ;
	if (i)
	{
	    strcpy (string, &string[i]);
	}
	return (string);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      buildTankList                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Build details list for tank inventory			     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void buildTankList ()
{
	int			numItems;
	char			txtBuf[BUFSIZ];

	XmListDeleteAllItems( detailList );
	/* Setup "currentTank" pointer & build the detail list */
	currentTank = currentProd -> tankList;
	while( currentTank )  {
	    if( currentTank -> pointName[0] <= ' ' )  {
		/* This better be the end of the list, so get out. */
		break;
	    }
	    memset( txtBuf, 0, sizeof txtBuf );

	    /* Add the tank details record to the details list widget */
    	    sprintf (txtBuf, "%-*.*s %-3.3s %6d   %2d  %3.1f  %8.1f %6d   %3.1f   %8.1f %8.1f %11.3f  %-8.8s  %10.4f  %10.3f",
			DESCRIPTOR_SIZE, DESCRIPTOR_SIZE, 
			currentTank -> pointDescr, currentTank -> onOff, currentTank -> levelFt,
			currentTank -> levelIn, currentTank -> level8ths, currentTank -> totalBBLS,
			currentTank -> BSWInches, currentTank -> BSW8ths, currentTank -> ullage,
			currentTank -> temp, currentTank -> apiGravity, currentTank -> apiDate,
			currentTank -> factor, currentTank -> netVolume );

    	    currentTank->tankStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
	    numItems++;
    	    XmListAddItem (detailList, currentTank->tankStr, numItems);
	    XmStringFree( currentTank -> tankStr );

	    /* Update display totals */
	    totalBBLS += currentTank -> totalBBLS;
	    ullage += currentTank -> ullage;
	    netVolume += currentTank -> netVolume;

	    currentTank = currentTank -> NextTank;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      buildPipeList                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Build details list for pipe inventory			     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void buildPipeList()
{
	int			numItems;
	char			txtBuf[BUFSIZ];

	XmListDeleteAllItems( detailList );
	/* Setup "currentPipe" pointer & build the detail list */
	currentPipe = currentProd -> pipeList;
	while( currentPipe )  {
	    if( currentPipe -> pointName[0] <= ' ' )  {
		/* This better be the end of the list, so get out. */
		break;
	    }

	    memset( txtBuf, 0, sizeof txtBuf );
	    /* Add the pipe details record to the details list widget */
	    sprintf (txtBuf, "%-*.*s %-3.3s                   %8.1f                %8.1f %8.1f %11.3f             %10.4f  %10.3f",
			DESCRIPTOR_SIZE, DESCRIPTOR_SIZE, 
			currentPipe -> pointDescr, currentPipe -> onOff, currentPipe -> totalBBLS,
			currentPipe -> ullage, currentPipe -> temp, currentPipe -> apiGravity,
			currentPipe -> factor, currentPipe -> netVolume );

	    currentPipe->pipeStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
	    numItems++;
	    XmListAddItem (detailList, currentPipe->pipeStr, numItems);
	    XmStringFree( currentPipe -> pipeStr );

	    /* Update display totals */
	    totalBBLS += currentPipe -> totalBBLS;
	    ullage += currentPipe -> ullage;
	    netVolume += currentPipe -> netVolume;

	    currentPipe = currentPipe -> NextPipe;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      PrintProdDet                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Print product details list for current pipe or tank list.	     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void PrintProdDet()
{
	char			detailsOut[100];
	FILE			*fp;



	/* Setup current list pointer & build the detail list */
	if( currentList == TANK_LIST )  {
	    /* Open the output file */
#ifdef	WIN32
	    sprintf( detailsOut, "\\tis4000db\\%s.tank", productName );
#else
	    sprintf( detailsOut, "/tis4000db/%s.tank", productName );
#endif	WIN32
	    if(( fp = fopen( detailsOut, "w" )) == NULL )  {
		/* Major problem.  Cant't print anything.  Probably can't even do
	 	* anything.  Get out of here after reporting exit.
	 	*/
		tisSysLog(0,0, "Can't open print file.  Exiting.\nCheck to see if directory is protected." );
		exit( 1 );
	    }
	    fprintf( fp, "\t\t\t\t\t\t\t\tTANK INVENTORY - %s\n", productName );
	    fprintf( fp, "\t\t\t\t\t\t\t      %s\n", ctime( &CurTime ));
	    fprintf( fp, "                             On/       Level          Total       BS&W        Gross                API       API\n" ); 
	    fprintf( fp, "                             Off    Ft.  In.  8ths     BBLS     In.  8ths    Ullage     Temp     Gravity    Date         Factor   Net BBLS\n\n" ); 

	    printTank = currentProd -> tankList;
	    while( printTank )  {
		if( printTank -> pointName[0] <= ' ' )  {
		    /* This better be the end of the list, so get out. */
		    break;
		}

		/* Add the tank details record to the output file */
    		fprintf (fp, "%-*.*s %-3.3s %6d   %2d  %3.1f  %8.1f %6d   %3.1f   %8.1f %8.1f  %10.3f  %-8.8s  %10.4f  %10.3f\n",
			DESCRIPTOR_SIZE, DESCRIPTOR_SIZE, 
			printTank -> pointDescr, printTank -> onOff, printTank -> levelFt,
			printTank -> levelIn, printTank -> level8ths, printTank -> totalBBLS,
			printTank -> BSWInches, printTank -> BSW8ths, printTank -> ullage,
			printTank -> temp, printTank -> apiGravity, printTank -> apiDate,
			printTank -> factor, printTank -> netVolume );

		/* Update display totals */
		totalBBLS += printTank -> totalBBLS;
		ullage += printTank -> ullage;
		netVolume += printTank -> netVolume;

		printTank = printTank -> NextTank;
	    }
	}
	else  {
	    /* Open the output file */
#ifdef	WIN32
	    sprintf( detailsOut, "\\tis4000db\\%s.other", productName );
#else
	    sprintf( detailsOut, "/tis4000db/%s.other", productName );
#endif	WIN32
	    if(( fp = fopen( detailsOut, "w" )) == NULL )  {
		/* Major problem.  Cant't print anything.  Probably can't even do
	 	* anything.  Get out of here after reporting exit.
	 	*/
		tisSysLog(0,0, "Can't open print file.  Exiting.\nCheck to see if directory is protected." );
		exit( 1 );
	    }
	    fprintf( fp, "\t\t\t\t\t\t\t\tOTHER INVENTORY - %s\n", productName );
	    fprintf( fp, "\t\t\t\t\t\t\t      %s\n", ctime( &CurTime ));
	    fprintf( fp, "                             On/       Level          Total       BS&W        Gross                API       API\n" ); 
	    fprintf( fp, "                             Off    Ft.  In.  8ths     BBLS     In.  8ths    Ullage     Temp     Gravity    Date         Factor   Net BBLS\n\n" ); 

	    printPipe = currentProd -> pipeList;

	    while( printPipe )  {
		if( printPipe -> pointName[0] <= ' ' )  {
		    /* This better be the end of the list, so get out. */
		    break;
		}

		/* Add the pipe details record to the output file */
		fprintf (fp, "%-*.*s %-3.3s                   %8.1f                %8.1f %8.1f %10.3f             %10.4f  %10.3f\n",
			DESCRIPTOR_SIZE, DESCRIPTOR_SIZE, 
			printPipe -> pointDescr, printPipe -> onOff, printPipe -> totalBBLS,
			printPipe -> ullage, printPipe -> temp, printPipe -> apiGravity,
			printPipe -> factor, printPipe -> netVolume );

		/* Update display totals */
		totalBBLS += printPipe -> totalBBLS;
		ullage += printPipe -> ullage;
		netVolume += printPipe -> netVolume;

		printPipe = printPipe -> NextPipe;
	    }
	}
	fprintf( fp, "\n Totals                                            %8.1f                %8.1f                                             %10.3f\n", 
			totalBBLS, ullage, netVolume );

	fclose( fp );
	chmod( detailsOut, 0666 );
	sprintf( printCmd, "lpr -h %s", detailsOut );
	system( printCmd );
	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      PrintProdInv                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Print current product list.					     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void PrintProdInv()
{
	char			productsOut[100];
	FILE			*fp;


	/* Open the output file */
#ifdef	WIN32
	sprintf( productsOut, "\\tis4000db\\ProdInv.lis" );
#else
	sprintf( productsOut, "/tis4000db/ProdInv.lis" );
#endif	WIN32
	if(( fp = fopen( productsOut, "w" )) == NULL )  {
	    /* Major problem.  Cant't print anything.  Probably can't even do
	     * anything.  Get out of here after reporting exit.
	     */
	    tisSysLog(0,0, "Can't open print file.  Exiting.\nCheck to see if directory is protected." );
	    exit( 1 );
	}
	fprintf( fp, "\t\t\t\t\t\tPRODUCT INVENTORY - %s\n", ctime( &CurTime ));
	fprintf( fp, "Product          On/    Capacity      Gross Volume   BS&W Volume    Ullage     Non-Issuable      Issuable  Net Volume\n" );
	fprintf( fp, "                 Off     BBLS            BBLS           BBLS         BBLS          BBLS            BBLS      BBLS\n\n" ); 

	prods = ProdInv.prodList;

	/* Build display list of inventory totals  */
	while( prods )  {
	    prods -> nonIssuable = prods -> grossVolume - prods -> issuable;
	    fprintf (fp, "%-16.16s %-3.3s    %8.1f      %8.1f       %8.1f      %8.1f       %8.1f     %8.1f   %10.3f\n",
			prods -> product, prods -> onOff, prods -> capacity,
			prods -> grossVolume, prods -> bswVolume, prods -> ullage,
			prods -> nonIssuable, prods -> issuable,
			prods -> netVolume );

	    prods = prods -> NextProd;
	}

	fprintf( fp, "\n" );	/* Terminate with newline */
	fclose( fp );
	chmod( productsOut, 0666 );
	sprintf( printCmd, "lpr -h %s", productsOut );
	system( printCmd );

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      StoreProdInv                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Store current product inventory.				     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void StoreProdInv()
{
/*
 *	NOTE:	The issue of file locking is NOT being addressed here,
 *		since the files in question are opened/created by this
 *		program (this routine, even).  Too little time will have
 *		elapsed in the creation/updating of these files to make 
 *		locking of the files an issue.  What this means, though,
 *		is that nobody should be accessing the dBase files affected
 *		by this routine while it is executing.
 */
	InvFileInit( ProdInvName, &ProdInvDesc, &ProdInvFlds[0], &ProdInvTag[0] );
	InvFileInit( TankInvName, &TankInvDesc, &TankInvFlds[0], &TankInvTag[0] );
	InvFileInit( OtherInvName, &OtherInvDesc, &OtherInvFlds[0], &OtherInvTag[0] );

	prods = ProdInv.prodList;

	/* Store product inventory that was gathered to dBase IV files. */
	while( prods )  {
	    prods -> nonIssuable = prods -> grossVolume - prods -> issuable;
	    AppendProdRecord();

	    /* Process this product's Tank & Other inventory
	     * lists before going to next product.
	     */
	    tanks = prods -> tankList;
	    while( tanks )  {
		/* Append record for this tank's details. */
		if( tanks -> pointName[0] <= ' ' )
		    break;		/* This is the last list item. */

		AppendTankRecord();

		/* Get next tank */
		tanks = tanks -> NextTank;
	    }

	    pipes = prods -> pipeList;
	    while( pipes )  {
		/* Append record for this pipe's details. */
		if( pipes -> pointName[0] <= ' ' )
		    break;		/* This is the last list item. */

		AppendOtherRecord();

		/* Get next pipe */
		pipes = pipes -> NextPipe;
	    }

	    /* Get next product. */
	    prods = prods -> NextProd;
	}

	d4close( ProdInvDesc );
	d4close( TankInvDesc );
	d4close( OtherInvDesc );

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      InvFileInit                                                          *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Open/Create storage inventory file.				     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void InvFileInit(char *InvFileName,
			D4DATA **InvFileDesc,
			F4FIELD_INFO *InvFlds,
			T4TAG_INFO *InvTag)
{
	int			sc;
	struct stat		statInfo;
	char			InventoryFile[100];
	char			pathData[100];	/* Buffer for translating file link(s). */
	char			fileData[200];	/* Buffer for translating file link(s). */
	char			fileToOpen[200];/* Even though CodeBase allows 258 chars. */


	/* Initialize working buffers. */
	memset( pathData, 0, sizeof pathData );
	memset( fileData, 0, sizeof fileData );
	memset( fileToOpen, 0, sizeof fileToOpen );
	memset( InventoryFile, 0, sizeof InventoryFile );

	/* Translate /tis4000db */
#ifdef	WIN32
	sc = readlink( "\\tis4000db", pathData, sizeof pathData );
#else
	sc = readlink( "/tis4000db", pathData, sizeof pathData );
#endif	WIN32

	/* Open the output files - create if necessary. */
#ifdef	WIN32
	sprintf( InventoryFile, "\\tis4000db\\%s", InvFileName );
#else
	sprintf( InventoryFile, "/tis4000db/%s", InvFileName );
#endif	WIN32
	sc = lstat( InventoryFile, &statInfo );
#ifdef MIPSEL
	sc = ( statInfo.st_mode & S_IFLNK );		/* st_mode is bit-encoded */
#else
	sc = S_ISLNK( statInfo.st_mode );
#endif
	if(sc)  {
	    sc = readlink( InventoryFile, fileData, sizeof fileData );
	    if( fileData[0] == '/' )  {
		/* File is a link to a different file system.  Don't concatenate. */
		strcpy( fileToOpen, fileData );
	    }
	    else  {
		sprintf( fileToOpen, "%s/%s",  pathData, fileData );
	    }
	}
	else  {
	    sprintf( fileToOpen, "%s/%s", pathData, InvFileName );
	}

	fileToOpen[strcspn( fileToOpen, "." )] = 0;
	if( !( *InvFileDesc = d4open( &codeBaseDesc, fileToOpen )))  {
		/* Inventory storage file must be created. */
#ifdef	USECB63
		codeBaseDesc.errorCode = 0;	/* Clear out the open error */
#else
		codeBaseDesc.error_code = 0;	/* Clear out the open error */
#endif	USEcb63
		*InvFileDesc = d4create( &codeBaseDesc, fileToOpen, InvFlds, InvTag );
		e4exit_test( &codeBaseDesc );		/* Make sure it was created okay. */
	}


	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      AppendProdRecord                                                     *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Create and append new product inventory record.		     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void AppendProdRecord()
{
	char		fieldData[40];	/* Intermediate buffer for writing data. */


	d4append_start( ProdInvDesc, 0 );	/* '0' => no memo fields */
	d4blank( ProdInvDesc );

	f4assign_n( d4field( ProdInvDesc, "PRODUCT" ), prods -> product, strlen( prods -> product ));

	/* The twenty years is what was subtracted to make time in synch with sample times */
	memset( fieldData, 0, sizeof fieldData );
	sprintf( fieldData, "%2d/%02d/%02d", (CurTmElements.tm_mon+1), 
						   CurTmElements.tm_mday, 
						   CurTmElements.tm_year % 100 );
	f4assign_n( d4field( ProdInvDesc, "DATE" ), fieldData, strlen( fieldData ));

	memset( fieldData, 0, sizeof fieldData );
	sprintf( fieldData, "%2d:%02d:%02d", (CurTmElements.tm_hour), 
						   CurTmElements.tm_min, 
						   CurTmElements.tm_sec );
	f4assign_n( d4field( ProdInvDesc, "TIME" ), fieldData, strlen( fieldData ));

	f4assign_n( d4field( ProdInvDesc, "ON_OFF" ), prods -> onOff, strlen( prods -> onOff ));

	f4assign_double( d4field( ProdInvDesc, "CAPACITY" ), prods -> capacity );
	f4assign_double( d4field( ProdInvDesc, "GROSS_VOL" ), prods -> grossVolume );
	f4assign_double( d4field( ProdInvDesc, "BSW_VOLUME" ), prods -> bswVolume );
	f4assign_double( d4field( ProdInvDesc, "ULLAGE" ), prods -> ullage );
	f4assign_double( d4field( ProdInvDesc, "NOT_ISSUE" ), prods -> nonIssuable);
	f4assign_double( d4field( ProdInvDesc, "ISSUABLE" ), prods -> issuable );
	f4assign_double( d4field( ProdInvDesc, "NETVOLUME" ), prods -> netVolume );

	f4assign_int( d4field( ProdInvDesc, "NUM_OTHER" ), prods -> numPipes);
	f4assign_int( d4field( ProdInvDesc, "NUM_TANKS" ), prods -> numTanks);

	d4append( ProdInvDesc );
	d4flush_all( ProdInvDesc );
/*	d4update_header( ProdInvDesc );	*/

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      AppendTankRecord                                                     *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Create and append new tank inventory record.		     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void AppendTankRecord()
{
	char		fieldData[40];	/* Intermediate buffer for writing data. */


	d4append_start( TankInvDesc, 0 );	/* '0' => no memo fields */
	d4blank( TankInvDesc );

	f4assign_n( d4field( TankInvDesc, "PRODUCT" ), prods -> product, strlen( prods -> product ));
	f4assign_n( d4field( TankInvDesc, "POINTNAME" ), tanks -> pointName, strlen( tanks -> pointName ));
	f4assign_n( d4field( TankInvDesc, "POINTDESC" ), tanks -> pointDescr, strlen( tanks -> pointDescr ));

	/* The twenty years is what was subtracted to make time in synch with sample times */
	memset( fieldData, 0, sizeof fieldData );
	sprintf( fieldData, "%2d/%02d/%02d", (CurTmElements.tm_mon+1), 
						   CurTmElements.tm_mday, 
						   CurTmElements.tm_year % 100);
	f4assign_n( d4field( TankInvDesc, "DATE" ), fieldData, strlen( fieldData ));

	memset( fieldData, 0, sizeof fieldData );
	sprintf( fieldData, "%2d:%02d:%02d", (CurTmElements.tm_hour), 
						   CurTmElements.tm_min, 
						   CurTmElements.tm_sec );
	f4assign_n( d4field( TankInvDesc, "TIME" ), fieldData, strlen( fieldData ));

	f4assign_n( d4field( TankInvDesc, "ON_OFF" ), tanks -> onOff, strlen( tanks -> onOff ));

	f4assign_int( d4field( TankInvDesc, "LEVEL_FT" ), tanks -> levelFt);
	f4assign_int( d4field( TankInvDesc, "LEVEL_IN" ), tanks -> levelIn);

	f4assign_double( d4field( TankInvDesc, "LEVEL_8THS" ), tanks -> level8ths );
	f4assign_double( d4field( TankInvDesc, "TOTAL_BBLS" ), tanks -> totalBBLS );
	f4assign_double( d4field( TankInvDesc, "BSW_INCHES" ), tanks -> BSWInches );
	f4assign_double( d4field( TankInvDesc, "BSW_8THS" ), tanks -> BSW8ths );
	f4assign_double( d4field( TankInvDesc, "ULLAGE" ), tanks -> ullage );
	f4assign_double( d4field( TankInvDesc, "TEMP" ), tanks -> temp );
	f4assign_double( d4field( TankInvDesc, "APIGRAVITY" ), tanks -> apiGravity );

	f4assign_n( d4field( TankInvDesc, "API_DATE" ), tanks -> apiDate, strlen( tanks -> apiDate ));

	f4assign_double( d4field( TankInvDesc, "FACTOR" ), tanks -> factor );
	f4assign_double( d4field( TankInvDesc, "NETVOLUME" ), tanks -> netVolume );

	d4append( TankInvDesc );
	d4flush_all( TankInvDesc );
/*	d4update_header( TankInvDesc );	*/

	return;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      AppendOtherRecord                                                    *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          Create and append new "other" inventory record.		     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void AppendOtherRecord()
{
	char		fieldData[40];	/* Intermediate buffer for writing data. */


	d4append_start( OtherInvDesc, 0 );	/* '0' => no memo fields */
	d4blank( OtherInvDesc );

	f4assign_n( d4field( OtherInvDesc, "PRODUCT" ), prods -> product, strlen( prods -> product ));
	f4assign_n( d4field( OtherInvDesc, "POINTNAME" ), pipes -> pointName, strlen( pipes -> pointName ));
	f4assign_n( d4field( OtherInvDesc, "POINTDESC" ), pipes -> pointDescr, strlen( pipes -> pointDescr ));

	/* The twenty years is what was subtracted to make time in synch with sample times */
	memset( fieldData, 0, sizeof fieldData );
	sprintf( fieldData, "%2d/%02d/%02d", (CurTmElements.tm_mon+1), 
						   CurTmElements.tm_mday, 
						   CurTmElements.tm_year % 100);
	f4assign_n( d4field( OtherInvDesc, "DATE" ), fieldData, strlen( fieldData ));

	memset( fieldData, 0, sizeof fieldData );
	sprintf( fieldData, "%2d:%02d:%02d", (CurTmElements.tm_hour), 
						   CurTmElements.tm_min, 
						   CurTmElements.tm_sec );
	f4assign_n( d4field( OtherInvDesc, "TIME" ), fieldData, strlen( fieldData ));

	f4assign_n( d4field( OtherInvDesc, "ON_OFF" ), pipes -> onOff, strlen( pipes -> onOff ));

	f4assign_double( d4field( OtherInvDesc, "TOTAL_BBLS" ), pipes -> totalBBLS );
	f4assign_double( d4field( OtherInvDesc, "ULLAGE" ), pipes -> ullage );
	f4assign_double( d4field( OtherInvDesc, "TEMP" ), pipes -> temp );
	f4assign_double( d4field( OtherInvDesc, "APIGRAVITY" ), pipes -> apiGravity );
	f4assign_double( d4field( OtherInvDesc, "FACTOR" ), pipes -> factor );
	f4assign_double( d4field( OtherInvDesc, "NETVOLUME" ), pipes -> netVolume );

	d4append( OtherInvDesc );
	d4flush_all( OtherInvDesc );
/*	d4update_header( OtherInvDesc );	*/

	return;
}

