/****
*
*	File name : 	names.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		ini_init
*		ini_exit
*		ini_get_str
*		ini_get_array
*		ini_get_int
*		ini_get_hex
*		ini_get_long
*
*	Returns the equivalent name for the given environment name
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <ssqueues.h>
#include <sslib.h>

#define IS_WHITE_SPACE(c)	(((c) == ' ') || ((c) == '\t'))
#define	IS_EOL(c)			(((c) == '\0') || ((c) == '\n'))
#define IS_COMMENT(c)		((c=='*') || (c==';') || (c=='!') || (c=='#'))
#define IS_MACRO(c)			((c) == '@')

typedef struct {
	QUEUE_DEF	root_macro;
	QUEUE_DEF	root_group;
	char		*group_name;
} INI_CONTEXT;

typedef struct {
	QUEUE_DEF	link;
	int			array_index;
	char		*name;
	char		*value;
} ENTRY_DEF;

typedef struct {
	QUEUE_DEF	link;
	char		*name;
	char		*value;
} MACRO_DEF;

typedef struct {
	QUEUE_DEF	link;
	char		*name;
	QUEUE_DEF	root_entry;
} GROUP_DEF;

static char			*skip_space (char *);
static char			*get_name (char *, char *);
static void			get_group (char *, char *);
static void			get_file_name (char *, char *);
static char			*get_macro_name (char *, char *);
static GROUP_DEF	*search_to_group (INI_CONTEXT *, char *);
static MACRO_DEF	*search_to_macro (INI_CONTEXT *, char *);
static ENTRY_DEF	*search_to_name (GROUP_DEF *, char *, int);
static int			load_file (INI_CONTEXT *, char *, char *);
static void			load_table (INI_CONTEXT *, FILE *, char *);
static GROUP_DEF	*add_group (INI_CONTEXT *, char *);
static void			add_chain (INI_CONTEXT *, GROUP_DEF *, char *, char *);
static void			add_macro (INI_CONTEXT *, GROUP_DEF *, char *, char *);
static void			add_name (INI_CONTEXT *, GROUP_DEF *, char *);
static void			add_one_entry (GROUP_DEF *, char *, int, char *);
static void			add_one_macro (INI_CONTEXT *, char *, char *);
static void			add_array (INI_CONTEXT *, GROUP_DEF *, char *, char *);
static void			copy_value (INI_CONTEXT *, char *, char *);
static void			copy_macro (INI_CONTEXT *, char **, char **);
static char			*copy_element (char *, char *);
static long			hex_to_long (char *);

INI_HANDLE ini_init (char *file_name, char *user_group)
{
	INI_CONTEXT		*context;

	context = (INI_CONTEXT *) malloc (sizeof (INI_CONTEXT));
	if (!context) {
		return (INI_HANDLE) NULL;
	}
	queue_init (&context -> root_group);
	queue_init (&context -> root_macro);
	context -> group_name = (char *) malloc (1 + strlen (user_group));
	if (!context -> group_name) {
		return (INI_HANDLE) NULL;
	}
	strcpy (context -> group_name, user_group);
	if (!load_file (context, file_name, context -> group_name)) {
		return (INI_HANDLE) NULL;
	}
	return (INI_HANDLE) context;
}

void ini_exit (INI_HANDLE handle)
{
	GROUP_DEF		*group;
	ENTRY_DEF		*entry;
	INI_CONTEXT		*context;

	context = (INI_CONTEXT *) handle;
	while (queue_next (&context -> root_group)) {
		group = (GROUP_DEF *) queue_get (&context -> root_group);
		while (queue_next (&group -> root_entry)) {
			entry = (ENTRY_DEF *) queue_get (&group-> root_entry);
			free (entry -> name);
			free (entry);
		}
		free (group -> name);
		free (group);
	}
	free (context);
	return;
}

/****
*
****/

int ini_get_str (char		*ini_group
			   , char		*ini_name
			   , char		*ini_value)
{
	GROUP_DEF		*group;
	ENTRY_DEF		*entry;
	INI_CONTEXT		*context;

	context = (INI_CONTEXT *) sslib_get_inihandle (GET_TASK_ID ());
	if (!context) {
		return FALSE;
	}
	group = search_to_group (context, ini_group);
	if (group) {
		entry = search_to_name (group, ini_name, 0);
		if (entry) {
			strcpy (ini_value, entry -> value);
			TRACE_S ("ini_get_str, ini_group (%s)", ini_group);
			TRACE_S ("ini_get_str, ini_name (%s)", ini_name);
			TRACE_S ("ini_get_str, ini_value (%s)", ini_value);
			return TRUE;
		} else {
			TRACE_S ("ini_get_str, no such name = %s", ini_name);
			TRACE_S ("ini_get_str, in group = %s", ini_group);
		}
	} else {
		TRACE_S ("ini_get_str, no such group = %s", ini_group);
	}
	return FALSE;
}

/****
*
****/

int ini_get_array (char			*ini_group
				 , char			*ini_name
				 , int			array_index
				 , char			*ini_value)
{
	GROUP_DEF		*group;
	ENTRY_DEF		*entry;
	INI_CONTEXT		*context;

	context = (INI_CONTEXT *) sslib_get_inihandle (GET_TASK_ID ());
	group = search_to_group (context, ini_group);
	if (group) {
		entry = search_to_name (group, ini_name, array_index);
		if (entry) {
			strcpy (ini_value, entry -> value);
			TRACE_S ("ini_get_str, ini_group (%s)", ini_group);
			TRACE_S ("ini_get_str, ini_name (%s)", ini_name);
			TRACE_S ("ini_get_str, ini_value (%s)", ini_value);
			return TRUE;
		} else {
			TRACE_S ("ini_get_str, no such name = %s", ini_name);
			TRACE_S ("ini_get_str, in group = %s", ini_group);
		}
	} else {
		TRACE_S ("ini_get_str, no such group = %s", ini_group);
	}
	return FALSE;
}

/****
*
****/

int ini_get_int (char			*ini_group
			   , char			*ini_name
			   , int			default_value)
{
	char		user_string [MAX_STRING];

	if (ini_get_str (ini_group, ini_name, user_string)) {
		return atoi (user_string);
	} else {
		return default_value;
	}
}

/****
*
****/

long ini_get_hex (char			*ini_group
				, char			*ini_name
				, long			default_value)
{
	char		user_string [MAX_STRING];

	if (ini_get_str (ini_group, ini_name, user_string)) {
		return hex_to_long (user_string);
	} else {
		return default_value;
	}
}

/****
*
****/

long ini_get_long (char			*ini_group
				 , char			*ini_name
				 , long			default_value)
{
	char	user_string [MAX_STRING];

	if (ini_get_str (ini_group, ini_name, user_string)) {
		return atol (user_string);
	} else {
		return default_value;
	}
}

static int load_file (INI_CONTEXT *context, char *file_name, char *group)
{
	FILE		*file;

	file = fopen (file_name, "r");
	if (file) {
		load_table (context, file, group);
		fclose (file);
		return TRUE;
	} else {
		ERROR_TRACE_I ("load_file, fopen = %d", GetLastError ());
		ERROR_TRACE_S ("load_file, file_name = %s", file_name);
		return FALSE;
	}
}

static void load_table (INI_CONTEXT *context, FILE *file, char *group_name)
{
	char		*line;
	char		buffer [MAX_STRING];
	GROUP_DEF	*group;

	group = (GROUP_DEF *) NULL;
	while (fgets (buffer, MAX_STRING, file)) {
		line = skip_space (buffer);
		if (IS_COMMENT (*line)) {
		} else if (IS_EOL (*line)) {
		} else if (*line == '[') {
			++line;
			group = add_group (context, line);
			if (!group) {
				return;
			}
		} else if (group) {
			if (*line == '{') {
				add_chain (context, group, group_name, line);
			} else if (IS_MACRO (*line)) {
				add_macro (context, group, group_name, line);
			} else {
				add_name (context, group, line);
			}
		}
	}
}

static GROUP_DEF *add_group (INI_CONTEXT *context, char *line)
{
	GROUP_DEF	*group;
	char		token [MAX_STRING];

	get_group (line, token);
	if (!strlen (token)) {
		return (GROUP_DEF *) NULL;
	}
	group = (GROUP_DEF *) malloc (sizeof (GROUP_DEF));
	if (!group) {
		return (GROUP_DEF *) NULL;
	}
	group -> name = (char *) malloc (1 + strlen (token));
	if (!group -> name) {
		free (group);
		return (GROUP_DEF *) NULL;
	}
	strcpy (group -> name, token);
	queue_init (&group -> root_entry);
	queue_append (&context -> root_group, &group -> link);
	return group;
}

static void add_chain (INI_CONTEXT		*context
					 , GROUP_DEF		*group
					 , char				*group_name
					 , char				*line)
{
	char		token [MAX_STRING];

	if (case_compare (group_name, group -> name)) {
		++line;
		get_file_name (line, token);
		if (strlen (token)) {
			load_file (context, token, group_name);
		}
	}
}

static void add_macro (INI_CONTEXT		*context
					 , GROUP_DEF	*group
					 , char			*group_name
					 , char			*line)
{
	char		macro_name [MAX_STRING];
	char		macro_value [MAX_STRING];

	if (!case_compare (group_name, group -> name)) {
		return;
	}
	line = get_macro_name (line, macro_name);
	if (!strlen (macro_name)) {
		return;
	}
	line = skip_space (line);
	if (*line != '=') {
		return;
	}
	++line;
	line = skip_space (line);
	copy_value (context, line, macro_value);
	add_one_macro (context, macro_name, macro_value);
}


static void add_name (INI_CONTEXT *context, GROUP_DEF *group, char *line)
{
	char	name [MAX_STRING];
	char	value [MAX_STRING];

	line = get_name (line, name);
	if (line) {
		line = skip_space (line);
		if (*line == '=') {
			++line;
			line = skip_space (line);
			if (*line == '{') {
				++line;
				line = skip_space (line);
				add_array (context, group, name, line);
			} else {
				copy_value (context, line, value);
				add_one_entry (group, name, 0, value);
			}
		}
	}
}

static void add_array (INI_CONTEXT		*context
					 , GROUP_DEF	*group
					 , char			*name
					 , char			*line)
{
	int		array_index;
	char	value [MAX_STRING];

	array_index = 0;
	for (;;) {
		line = copy_element (line, value);
		if (!strlen (value)) {
			return;
		}
		line = skip_space (line);
		add_one_entry (group, name, array_index, value);
		++array_index;
	}
}

static void add_one_entry (GROUP_DEF	*group
						 , char			*name
						 , int			array_index
						 , char			*value)
{
	ENTRY_DEF	*entry;

	entry = (ENTRY_DEF *) malloc (sizeof (ENTRY_DEF));
	if (!entry) {
		return;
	}
	entry -> name = (char *) malloc (1 + strlen (name));
	if (!entry -> name) {
		free (entry);
		return;
	}
	entry -> value = (char *) malloc (1 + strlen (value));
	if (!entry -> name) {
		free (entry -> name);
		free (entry);
		return;
	}
	strcpy (entry -> name, name);
	strcpy (entry -> value, value);
	entry -> array_index = array_index;
	queue_append (&group -> root_entry, &entry -> link);
}

static void add_one_macro (INI_CONTEXT *context, char *name, char *value)
{
	MACRO_DEF	*macro;

	macro = (MACRO_DEF *) malloc (sizeof (MACRO_DEF));
	if (!macro) {
		return;
	}
	macro -> name = (char *) malloc (1 + strlen (name));
	if (!macro -> name) {
		free (macro);
		return;
	}
	macro -> value = (char *) malloc (1 + strlen (value));
	if (!macro -> name) {
		free (macro -> name);
		free (macro);
		return;
	}
	strcpy (macro -> name, name);
	strcpy (macro -> value, value);
	queue_append (&context -> root_macro, &macro -> link);
}

static void get_file_name (char *line, char *token)
{
	while ((!IS_EOL (*line))
	&&     (!IS_WHITE_SPACE (*line))
	&&     (*line != '}')) {
		*token = *line;
		++line;
		++token;
	}
	*token = '\0';
}

static char *get_macro_name (char *line, char *token)
{
	if (!IS_MACRO (*line)) {
		*token = '\0';
		return line;
	}
	++line;
	while (!IS_MACRO (*line)) {
		if (IS_EOL (*line)) {
			*token = '\0';
			return line;
		}
		*token = *line;
		++line;
		++token;
	}
	if (!IS_MACRO (*line)) {
		*token = '\0';
		return line;
	}
	++line;
	*token = '\0';
	return line;
}

static void get_group (char *line, char *token)
{
	while ((!IS_EOL (*line))
	&&     (!IS_WHITE_SPACE (*line))
	&&     (*line != ']')) {
		*token = *line;
		++line;
		++token;
	}
	*token = '\0';
}

static char *get_name (char *line, char *token)
{
	line = skip_space (line);
	while ((!IS_EOL (*line))
	&&     (!IS_WHITE_SPACE (*line))
	&&     (*line != '=')) {
		*token = *line;
		++line;
		++token;
	}
	*token = '\0';
	return line;
}

static char *skip_space (char *line)
{
	while ((!IS_EOL (*line)) && (IS_WHITE_SPACE (*line))) {
		++line;
	}
	return line;
}

static void copy_value (INI_CONTEXT *context, char *source, char *target)
{
	source = skip_space (source);
	if (*source == '"') {
		++source;
		while ((!IS_EOL (*source)) && (*source != '"')) {
			*target = *source;
			++target;
			++source;
		}
	} else {
		while ((!IS_EOL (*source)) && (!IS_WHITE_SPACE (*source))) {
			if (IS_MACRO (*source)) {
				copy_macro (context, &source, &target);
			} else {
				*target = *source;
				++target;
				++source;
			}
		}
	}
	*target = '\0';
}

static void copy_macro (INI_CONTEXT *context, char **source, char **target)
{
	char		*line;
	MACRO_DEF	*macro;
	char		macro_name [MAX_STRING];

	line = get_macro_name (*source, macro_name);
	if (!strlen (macro_name)) {
		return;
	}
	macro = search_to_macro (context, macro_name);
	if (!macro) {
		*source = line;
		return;
	}
	*source = line;
	strcpy (*target, macro -> value);
	*target += strlen (macro -> value);
}

static char *copy_element (char *source, char *target)
{
	while ((!IS_EOL (*source))
	&&     (*source != ',')
	&&     (*source != '}')
	&&     (!IS_WHITE_SPACE (*source))) {
		*target = *source;
		++target;
		++source;
	}
	source = skip_space (source);
	if ((*source == ',') || (*source == '}')) {
		++source;
	}
	*target = '\0';
	return source;
}

/****
*
****/

static GROUP_DEF *search_to_group (INI_CONTEXT *context, char *ini_group)
{
	GROUP_DEF	*this_group;

	this_group = (GROUP_DEF *) queue_next (&context -> root_group);
	while (this_group) {
		if (case_compare (this_group -> name, ini_group)) {
			return this_group;
		}
		this_group = (GROUP_DEF *) queue_next (&this_group -> link);
	}
	return (GROUP_DEF *) NULL;
}

/****
*
****/

static ENTRY_DEF *search_to_name (GROUP_DEF	*group
								, char		*ini_name
								, int		array_index)
{
	ENTRY_DEF	*this_entry;

	this_entry = (ENTRY_DEF *) queue_next (&group -> root_entry);
	while (this_entry) {
		if (case_compare (this_entry -> name, ini_name)) {
			if (this_entry -> array_index == array_index) {
				return this_entry;
			}
		}
		this_entry = (ENTRY_DEF *) queue_next (&this_entry -> link);
	}
	return (ENTRY_DEF *) NULL;
}

/****
*
****/

static MACRO_DEF *search_to_macro (INI_CONTEXT *context, char *macro_name)
{
	MACRO_DEF	*this_macro;

	this_macro = (MACRO_DEF *) queue_next (&context -> root_macro);
	while (this_macro) {
		if (case_compare (this_macro -> name, macro_name)) {
			return this_macro;
		}
		this_macro = (MACRO_DEF *) queue_next (&this_macro -> link);
	}
	return (MACRO_DEF *) NULL;
}

static long hex_to_long (char *hex_string)
{
	char		one_char;
	long		value;

	value = 0L;
	for (;;) {
		one_char = *hex_string;
		++hex_string;
		if ((toupper (one_char) >= 'A') && (toupper (one_char) <= 'F')) {
			value = (value << 4) + 10 + (toupper (one_char) - 'A');
		} else if ((one_char >= '0') && (one_char <= '9')) {
			value = (value << 4) + (toupper (one_char) - '0');
		} else {
			return value;
		}
	}
}

