/****
*
*	File name :		timers.c
*	Author :		Steve Johnson
*	Creation date :	September 10, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		time_get
*		time_expired
*		time_diff
*		time_max
*		time_compare
*		time_date_to_time
*		time_time_to_date
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <sstimer.h>

/***
*
*	delta times are in milliseconds
*
****/

#define DELTA_UNIT		1000L
#define DELTA_UNIT_USEC	1000L
#define USEC_UNIT		1000000L
#define BASE_YEAR		1992

static long		secs_year (int);
static long		secs_month (int, int);
static long		secs_day (int);
static long		secs_hour (int);
static long		secs_minute (int);
static int		leap_year (int);
static long		get_year (long, long *);
static long		get_month (long, long, long *);
static long		get_day (long, long *);
static long		get_hour (long, long *);
static long		get_minute (long, long *);

int time_expired (TIME_DEF *start_time, TIME_DEF *this_time, long delta)
{
	TIME_DEF	end_time;

	end_time . tv_sec = start_time -> tv_sec;
	end_time . tv_sec += (long) delta / DELTA_UNIT;
	end_time . tv_usec = start_time -> tv_usec;
	end_time . tv_usec += ((long) delta % DELTA_UNIT) * DELTA_UNIT_USEC;
	if (end_time . tv_usec >= USEC_UNIT) {
		++end_time . tv_sec;
		end_time . tv_usec -= USEC_UNIT;
	}
	if (end_time . tv_sec < this_time -> tv_sec) {
		return TRUE;
	} else if (end_time . tv_sec == this_time -> tv_sec) {
		if (end_time . tv_usec <= this_time -> tv_usec) {
			return TRUE;
		}
	}
	return FALSE;
}

void time_diff (TIME_DEF *start_time, TIME_DEF *end_time, TIME_DEF *diff)
{
	if (end_time -> tv_usec < start_time -> tv_usec) {
		end_time -> tv_usec += USEC_UNIT;
		--end_time -> tv_sec;
	}
	diff -> tv_usec = end_time -> tv_usec - start_time -> tv_usec;
	diff -> tv_sec = end_time -> tv_sec - start_time -> tv_sec;
}

void time_max (TIME_DEF *time1, TIME_DEF *time2, TIME_DEF *max_time)
{
	if (time1 -> tv_sec > time1 -> tv_sec) {
		max_time -> tv_sec = time1 -> tv_sec;
		max_time -> tv_usec = time1 -> tv_usec;
	} else if (time1 -> tv_sec < time2 -> tv_sec) {
		max_time -> tv_sec = time2 -> tv_sec;
		max_time -> tv_usec = time2 -> tv_usec;
	} else if (time1 -> tv_usec > time2 -> tv_usec) {
		max_time -> tv_sec = time1 -> tv_sec;
		max_time -> tv_usec = time1 -> tv_usec;
	} else if (time1 -> tv_usec < time2 -> tv_usec) {
		max_time -> tv_sec = time2 -> tv_sec;
		max_time -> tv_usec = time2 -> tv_usec;
	} else {
		max_time -> tv_sec = time1 -> tv_sec;
		max_time -> tv_usec = time1 -> tv_usec;
	}
}

/****
*
*	-1 if first < second
*	0 if first = second
*	1 if first > second
*
****/

int time_compare (TIME_DEF *first_time, TIME_DEF *second_time)
{
	if (first_time -> tv_sec > second_time -> tv_sec) {
		return 1;
	} else if (first_time -> tv_sec < second_time -> tv_sec) {
		return -1;
	} else {
		if (first_time -> tv_usec > second_time -> tv_usec) {
			return 1;
		} else if (first_time -> tv_usec < second_time -> tv_usec) {
			return -1;
		} else {
			return 0;
		}
	}
}

void time_date_to_time (DATE_TIME_DEF *date_time, TIME_DEF *time_only)
{
	time_only -> tv_usec = date_time -> milliseconds * 1000;
	time_only -> tv_sec = secs_year (date_time -> year)
					    + secs_month (date_time -> month, date_time -> year)
					    + secs_day (date_time -> day)
					    + secs_hour (date_time -> hour)
					    + secs_minute (date_time -> minute)
					    + date_time -> second;
}

void time_time_to_date (TIME_DEF *time_only, DATE_TIME_DEF *date_time)
{
	long		remainder;

	date_time -> year = get_year (time_only -> tv_sec, &remainder);
	date_time -> month = get_month (date_time -> year, remainder, &remainder);
	date_time -> day = get_day (remainder, &remainder);
	date_time -> hour = get_hour (remainder, &remainder);
	date_time -> minute = get_minute (remainder, &remainder);
	date_time -> second = remainder;
	date_time -> milliseconds = time_only -> tv_usec / 1000;
}

static long secs_year (int year)
{
	int		count;
	long	seconds;

	seconds = 0L;
	for (count = BASE_YEAR; count < year; ++ count) {
		if (leap_year (count)) {
			seconds += (366L * 24L * 60L * 60L);
		} else {
			seconds += (365L * 24L * 60L * 60L);
		}
	}
	return seconds;
}

static long get_year (long seconds, long *remainder)
{
	long	year;
	long	test_seconds;
	long	last_seconds;

	test_seconds = 0L;
	last_seconds = 0L;
	for (year = BASE_YEAR; year < 2099; ++ year) {
		if (leap_year (year)) {
			test_seconds += (366L * 24L * 60L * 60L);
		} else {
			test_seconds += (365L * 24L * 60L * 60L);
		}
		if (test_seconds > seconds) {
			*remainder = seconds - last_seconds;
			return year;
		}
		last_seconds = test_seconds;
	}
	*remainder = 0;
	return year;
}

static long secs_month (int month, int year)
{
	int		count;
	long	seconds;

	seconds = 0L;
	for (count = 1; count < month; ++count) {
		switch (count) {
		case 1 :
		case 3 :
		case 5 :
		case 7 :
		case 8 :
		case 10 :
		case 12 :
			seconds += (31L * 24L * 60L * 60L);
			break;
		case 4 :
		case 6 :
		case 9 :
		case 11 :
			seconds += (30L * 24L * 60L * 60L);
			break;
		case 2 :
			if (leap_year (year)) {
				seconds += (29L * 24L * 60L * 60L);
			} else {
				seconds += (28L * 24L * 60L * 60L);
			}
			break;
		}
	}
	return seconds;
}

static long get_month (long year, long seconds, long *remainder)
{
	long	month;
	long	test_seconds;
	long	last_seconds;

	test_seconds = 0L;
	last_seconds = 0L;
	for (month = 1; month <= 12; ++month) {
		switch (month) {
		case 1 :
		case 3 :
		case 5 :
		case 7 :
		case 8 :
		case 10 :
		case 12 :
			test_seconds += (31L * 24L * 60L * 60L);
			break;
		case 4 :
		case 6 :
		case 9 :
		case 11 :
			test_seconds += (30L * 24L * 60L * 60L);
			break;
		case 2 :
			if (leap_year (year)) {
				test_seconds += (29L * 24L * 60L * 60L);
			} else {
				test_seconds += (28L * 24L * 60L * 60L);
			}
			break;
		}
		if (test_seconds > seconds) {
			*remainder = seconds - last_seconds;
			return month;
		}
		last_seconds = test_seconds;
	}
	*remainder = 0;
	return month;
}

static long secs_day (int day)
{
	return ((day - 1) * (24L * 60L * 60L));
}

static long get_day (long seconds, long *remainder)
{
	long	day;
	long	test_seconds;
	long	last_seconds;

	test_seconds = 0L;
	last_seconds = 0L;
	for (day = 1; day <= 31; ++day) {
		test_seconds += (24L * 60L * 60L);
		if (test_seconds > seconds) {
			*remainder = seconds - last_seconds;
			return day;
		}
		last_seconds = test_seconds;
	}
	*remainder = 0;
	return day;
}

static long secs_hour (int hour)
{
	return (hour * (60L * 60L));
}

static long get_hour (long seconds, long *remainder)
{
	long	hour;
	long	test_seconds;
	long	last_seconds;

	test_seconds = 0L;
	last_seconds = 0L;
	for (hour = 0; hour < 24; ++hour) {
		test_seconds += (60L * 60L);
		if (test_seconds > seconds) {
			*remainder = seconds - last_seconds;
			return hour;
		}
		last_seconds = test_seconds;
	}
	*remainder = 0;
	return hour;
}

static long secs_minute (int minute)
{
	return (minute * 60L);
}

static long get_minute (long seconds, long *remainder)
{
	long	minute;
	long	test_seconds;
	long	last_seconds;

	test_seconds = 0L;
	last_seconds = 0L;
	for (minute = 0; minute < 60; ++minute) {
		test_seconds += 60L;
		if (test_seconds > seconds) {
			*remainder = seconds - last_seconds;
			return minute;
		}
		last_seconds = test_seconds;
	}
	*remainder = 0;
	return minute;
}

static int leap_year (int year)
{
	if (year & 0x03) {
		return FALSE;
	}
	if (0 == (year % 400)) {
		return TRUE;
	}
	if (0 == (year % 100)) {
		return FALSE;
	}
	return TRUE;
}

