/****
*
*	File name :		trace.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		trace_init (...)
*		trace_normal (...)
*		trace_message (...)
*		trace_int (...)
*		trace_long (...)
*		trace_string (...)
*		trace_buffer (...)
*		trace_error_message (...)
*		trace_error_int (...)
*		trace_error_long (...)
*		trace_error_string (...)
*		trace_error_buffer (...)
*		trace_error_system (...)
*		trace_exit (...)
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <ssnames.h>
#include <sslib.h>

#define TRACE_GROUP		"TRACES"
#define WRAP_MESSAGE	"\n\n#### Wrapping Trace File ####\n\n"

typedef struct {
	int			console_on;
	int			traces_on;
	int			errors_on;
	TFILE		trace_file;
	int			line_count;
	char		*prefix;
} TRACE_CONTEXT;

static void make_prefix (TRACE_CONTEXT *, char *);
static void	get_time_string (char *);
static void create_file (char *);
static void	write_trace (TRACE_CONTEXT *, char *);

/****
*
****/

TRACE_HANDLE trace_init (char *this_prefix)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			trace_name [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) malloc (sizeof (TRACE_CONTEXT));
	if (!context) {
		return (TRACE_HANDLE) NULL;
	}
	context -> traces_on = FALSE;
	context -> errors_on = FALSE;
	context -> console_on = FALSE;
	context -> prefix = (char *) malloc (1 + strlen (this_prefix));
	if (!context -> prefix) {
		free (context);
		return (TRACE_HANDLE) NULL;
	}
	strcpy (context -> prefix, this_prefix);
	context -> console_on = ini_get_int (TRACE_GROUP, "CONSOLE_ON", FALSE);
	if (ini_get_str (TRACE_GROUP, "TRACE_LEVEL", trace_data)) {
		if (!ini_get_str (TRACE_GROUP, "FILE_NAME", trace_name)) {
			strcpy (trace_name, "trace.err");
		}
		if (!strcmp (trace_data, "FULL")) {
			context -> traces_on = TRUE;
			context -> errors_on = TRUE;
		} else if (!strcmp (trace_data, "ERRORS")) {
			context -> traces_on = FALSE;
			context -> errors_on = TRUE;
		}
	} else {
		context -> traces_on = FALSE;
		context -> errors_on = FALSE;
	}
	if (context -> errors_on) {
		if (strcmp (trace_name, "STDOUT")) {
			create_file (trace_name);
			context -> trace_file = TOPENW (trace_name);
			if (TERROR (context -> trace_file)) {
				sprintf (trace_line, "trace file name = (%s),", trace_name);
				sprintf (trace_data, "Error = 0x%lX\n", (long) GetLastError ());
				strcat (trace_line, trace_data);
				fwrite (trace_line, 1, strlen (trace_line), stderr);
				context -> traces_on = FALSE;
				context -> errors_on = FALSE;
				return FALSE;
			}
		} else {
			context -> trace_file = TFILENO (stdout);
		}
	}
	context -> line_count = 0;
	return (TRACE_HANDLE) context;
}

void trace_exit (TRACE_HANDLE handle)
{
	TRACE_CONTEXT		*context;

	ERROR_TRACE ("LAST TRACE BEFORE NORMAL SHUTDOWN");
	context = (TRACE_CONTEXT *) handle;
	if (!context) {
		return;
	}
	if (context -> trace_file) {
		TFLUSH (context -> trace_file);
		if (context -> trace_file != TFILENO (stdout)) {
			TCLOSE (context -> trace_file);
		} else {
			TFLUSH (context -> trace_file);
		}
	}
	free (context -> prefix);
	free (context);
}

void trace_console_on (TRACE_HANDLE handle)
{
	TRACE_CONTEXT		*context;

	context = (TRACE_CONTEXT *) handle;
	if (!context) {
		return;
	}
	context -> console_on = TRUE;
}

void trace_console_off (TRACE_HANDLE handle)
{
	TRACE_CONTEXT		*context;

	context = (TRACE_CONTEXT *) handle;
	if (!context) {
		return;
	}
	context -> console_on = FALSE;
}

void trace_debug_on (TRACE_HANDLE handle)
{
	TRACE_CONTEXT		*context;

	context = (TRACE_CONTEXT *) handle;
	if (!context) {
		return;
	}
	context -> traces_on = TRUE;
}

void trace_debug_off (TRACE_HANDLE handle)
{
	TRACE_CONTEXT		*context;

	context = (TRACE_CONTEXT *) handle;
	if (!context) {
		return;
	}
	context -> traces_on = FALSE;
}

void trace_message (char *message)
{
	char			trace_line [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> traces_on) {
		make_prefix (context, trace_line);
		strcat (trace_line, message);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_int (char *format, int value)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> traces_on) {
		make_prefix (context, trace_line);
		sprintf (trace_data, format, value);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_long (char *format, long value)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> traces_on) {
		make_prefix (context, trace_line);
		sprintf (trace_data, format, value);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_string (char *format, char *string)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> traces_on) {
		make_prefix (context, trace_line);
		sprintf (trace_data, format, string);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_buffer (char *message, int size, void *buffer)
{
	UCHAR			*char_buffer;
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			header [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> traces_on) {
		make_prefix (context, header);
		sprintf (trace_data, "%s,(%d) = ", message, size);
		strcat (header, trace_data);
		strcpy (trace_line, header);
		char_buffer = (UCHAR *) buffer;
		while (size) {
			if (strlen (trace_line) > 75) {
				strcat (trace_line, NEW_LINE);
				write_trace (context, trace_line);
				strcpy (trace_line, header);
			}
			sprintf (trace_data, "%X,", *char_buffer);
			strcat (trace_line, trace_data);
			++char_buffer;
			--size;
		}
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_system (char *format)
{
	long			last_error;
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			system_message [MAX_STRING];
	char			time_string [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> traces_on) {
		last_error = GetLastError ();
		get_time_string (time_string);
		make_prefix (context, trace_line);
		strcat (trace_line, " - ");
		FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
					   NULL,
					   last_error,
					   0,
					   system_message,
					   MAX_STRING,
					   NULL);
		sprintf (trace_data, format, last_error, system_message);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_error_message (char *message)
{
	char			trace_line [MAX_STRING];
	char			time_string [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> errors_on) {
		get_time_string (time_string);
		make_prefix (context, trace_line);
		strcat (trace_line, "ERROR (");
		strcat (trace_line, time_string);
		strcat (trace_line, "): ");
		strcat (trace_line, message);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_error_int (char *format, int value)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			time_string [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> errors_on) {
		get_time_string (time_string);
		make_prefix (context, trace_line);
		strcat (trace_line, "ERROR (");
		strcat (trace_line, time_string);
		strcat (trace_line, "): ");
		sprintf (trace_data, format, value);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_error_long (char *format, long value)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			time_string [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> errors_on) {
		get_time_string (time_string);
		make_prefix (context, trace_line);
		strcat (trace_line, "ERROR (");
		strcat (trace_line, time_string);
		strcat (trace_line, "): ");
		sprintf (trace_data, format, value);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}


void trace_error_string (char *format, char *string)
{
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			time_string [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> errors_on) {
		get_time_string (time_string);
		make_prefix (context, trace_line);
		strcat (trace_line, "ERROR (");
		strcat (trace_line, time_string);
		strcat (trace_line, "): ");
		sprintf (trace_data, format, string);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_error_buffer (char *message, int size, void *buffer)
{
	UCHAR			*char_buffer;
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			header [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> errors_on) {
		make_prefix (context, header);
		sprintf (trace_data, "%s,E(%d) = ", message, size);
		strcat (header, trace_data);
		strcpy (trace_line, header);
		char_buffer = (UCHAR *) buffer;
		while (size) {
			if (strlen (trace_line) > 75) {
				strcat (trace_line, NEW_LINE);
				write_trace (context, trace_line);
				strcpy (trace_line, header);
			}
			sprintf (trace_data, "%X,", *char_buffer);
			strcat (trace_line, trace_data);
			++char_buffer;
			--size;
		}
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

void trace_error_system (char *format)
{
	long			last_error;
	char			trace_data [MAX_STRING];
	char			trace_line [MAX_STRING];
	char			system_message [MAX_STRING];
	char			time_string [MAX_STRING];
	TRACE_CONTEXT	*context;

	context = (TRACE_CONTEXT *) sslib_get_tracehandle (GET_TASK_ID ());
	if (!context) {
		return;
	}
	if (context -> errors_on) {
		last_error = GetLastError ();
		get_time_string (time_string);
		make_prefix (context, trace_line);
		strcat (trace_line, "ERROR (");
		strcat (trace_line, time_string);
		strcat (trace_line, "):");
		strcat (trace_line, NEW_LINE);
		strcat (trace_line, " - ");
		FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
					   NULL,
					   last_error,
					   0,
					   system_message,
					   MAX_STRING,
					   NULL);
		sprintf (trace_data, format, last_error, system_message);
		strcat (trace_line, trace_data);
		strcat (trace_line, NEW_LINE);
		write_trace (context, trace_line);
	}
}

static void create_file (char *file_name)
{
	FILE	*stupid_systems;

	stupid_systems = fopen (file_name, "w");
	fclose (stupid_systems);
}

static void make_prefix (TRACE_CONTEXT *context, char *trace_line)
{
	sprintf (trace_line, "%s(%ld): ", context -> prefix, GetTickCount());
}

static void get_time_string (char *time_string)
{
	time_t	time_stamp;
	char	*c_string;

	time (&time_stamp);
	c_string = ctime (&time_stamp);
	if (c_string) {
		strcpy (time_string, c_string);
		time_string [strlen (time_string) - 1] = 0;
	} else {
		strcpy (time_string, "*ctime failed*");
	}
}

static void write_trace (TRACE_CONTEXT *context, char *trace_line)
{
	++context -> line_count;
	if (context -> line_count > 32000) {
		context -> line_count = 0;
		TSEEK (context -> trace_file, 5000L);
		TWRITE (context -> trace_file, WRAP_MESSAGE, strlen (WRAP_MESSAGE));
	}
	TWRITE (context -> trace_file, trace_line, strlen (trace_line));
	TFLUSH (context -> trace_file);
	if (context -> console_on) {
		printf (trace_line);
	}
}

