/**************************************************************/
/*                                                            */
/*    This software was developed for DLA by Tate Integrated  */
/*    Systems, L.P. via a contract with Vitro Corporation.    */
/*    It is designed to work with TIS4000 Release 4.2.        */
/*                                                            */
/*    The information in this software is subject to          */
/*    change without notice and should not be construed       */
/*    as a commitment by Tate Integrated Systems L.P. or      */
/*    its third party suppliers.                              */
/*                                                            */
/*    Tate Integrated Systems, L.P. (TIS) and its third       */
/*    party suppliers assume no responsibility for the        */
/*    use or inability to use any of its software. TIS        */
/*    software is provided "as is" without warranty of        */
/*    any kind and TIS expressly disclaims all implied        */
/*    warranties, including but not limited to the implied    */
/*    warranties of merchantability and fitness for a         */
/*    particular purpose.                                     */
/*                                                            */
/*    Notice: Notwithstanding any other lease or license      */
/*    that may pertain to or accompany the delivery of this   */
/*    computer software, the rights of the Government         */
/*    regarding its use, reproduction, and disclosure are     */
/*    as set forth in Section 52.227-19 of the FARS           */
/*    Computer Software-Restricted Rights clause.             */
/*                                                            */
/**************************************************************/
#undef DEBUG

/* The Evolution Processor schedules CSL script execution to implement Tank Farm Evolutions. */

#include	<stdio.h>
#include	<syslog.h>
#include	<setjmp.h>
#include	<sys/signal.h>
#include	<stdlib.h>
#include	<malloc.h>
#include	<ctype.h>
#include	<lstLib.h>
#include	<cadef.h>
#include	<db_access.h>
#include	<CSLMonitor.h>
#include	<EVOLManager.h>
#include	<d4all.h>
#include	<TFdBASE.h>
#include	<msapp.h>
#include 	<sys/wait.h>

void	Exit ();
void	killHandler ();
void	childHandler ();

#define	REFRESH_RATE				5	/* in seconds */
#define	CHECK_FOR_CHANGE_TO_MASTER_RATE		2
#define	DEFAULT_CA_LOOP_COUNT			20	/* 1 second delays */

char	*localhostname = "localhost";
char	*completionLogicPath = NULL;
char	*errorLogicPath = NULL;
char	*environmentString = "";
jmp_buf	error_env;
char	errorTextBuffer [200];
int	Redundancy, Verbose, Debug, Trace;
int	maxCaLoopCount = DEFAULT_CA_LOOP_COUNT;
int	arbiterPresent, prevState;
int	arbiterHandle;

double	latestDoubleValue;		/* Values fetched from channel access */
int	latestIntegerValue;
int	fetchComplete;

/*-----------------------------------------------------------------------------------------------
 * Program errors are trapped by assertions: i.e. assert (p != NULL, FATAL, "can't get memory");
 * See the loop in main () for the receiving end of non-FATAL assertions.
 *
 * NOTE: This is modeled after the UNIX assert () macro, but an extended version is defined here.
 *
 * The assertion macro tests the given condition and if FALSE takes the given action.
 * The given text is logged via syslog, which the user can configure (man syslogd).
 */

#define	IGNORE		1
#define	PUT_INTO_ERROR	2
#define	FATAL		3

#define	assert(cond, action, text) \
    do \
	if (! (cond)) \
	    { \
	    if (Debug) \
		printf ("EVOLProcessor Error:%s.  [false assertion on line %d]\n", text, __LINE__); \
	    else \
		syslog (LOG_ALERT|LOG_LOCAL1, "EVOLProcessor Error:%s. [false assertion on line %d]\n", text, __LINE__);\
	    switch (action) \
		{ \
		case FATAL: \
		    Exit (1); \
		case PUT_INTO_ERROR: \
		    if (errorLogicPath) \
			{ \
			char	command [BUFSIZ]; \
			sprintf (&command [0], "%s%s \"%s\" line=%d", environmentString, errorLogicPath, text, __LINE__); \
			if (fork () == 0) \
			    /* child becomes error logic (exit (1) in case execlp () fails) */ \
			    execlp ("/bin/sh", "sh", "-c", command, 0), exit (1); \
			 }	/* no break */ \
		case IGNORE: \
		    longjmp (error_env, (action)); \
		} \
	    } \
	while (0)

/*---------------------------------------------------------------------------------------------------------------------
 * TEMPORARY termcap functions which assist debugging:
 */
char	termcapEntry [1024];
char	termcapData [1024];
char	*pClearScreen;
char	*pClearEOL;
char	*pClearEOS;
char	*pHomeCursor;

/*----------*/
initTermcap ()

    {

	char	*tgetstr ();
	char	*area = &termcapData [0];
	
	tgetent (&termcapEntry [0], getenv ("TERM"));
	pClearScreen = tgetstr ("cl", &area);
	pClearEOL = tgetstr ("ce", &area);
	pClearEOS = tgetstr ("cd", &area);
	pHomeCursor = tgetstr ("ho", &area);

    }

/*----------*/
outc (c)

    char	c;

    {

	if (Debug)
	    putc (c, stdout);
    }

#define	clearScreen()	tputs (pClearScreen, 1, outc)
#define	clearEOL()	tputs (pClearEOL, 1, outc)
#define	clearEOS()	tputs (pClearEOS, 1, outc)
#define	homeCursor()	tputs (pHomeCursor, 1, outc)


/*---------------------------------- Utility functions -------------------------------*/

/* A signal handler for CHILD process completions to keep zombies from lasting forever */

/*----------*/
void childHandler(sig)
          int sig;
	{
#ifdef SOLARIS
	int	val;
	while (waitpid ((pid_t)0, &val, WNOHANG|WUNTRACED) > 0)
#else
	while (wait3((union wait *)NULL, WNOHANG|WUNTRACED, (struct rusage *)NULL) > 0)
#endif
		;
	}         

/*----------*/
EvolutionState findStatus (statusStr)

    char	*statusStr;

    {

	EvolutionState	status;

	for (status = Pending ; status < BadState ; status++)
	    if (strncmp (statusStr, EvolutionStateText [status], STATUS_ENTRY_SIZE) == 0)
		break;
	return status;

    }


/*----------*/
EvolutionProcessorRequestType findRequestType (typeStr)

    char	*typeStr;

    {

	EvolutionProcessorRequestType	type;

	for (type = NONE; type < BADREQ; type++)
	    if (strncmp (typeStr, EvolutionProcessorRequestText [type], REQTYPE_ENTRY_SIZE) == 0)
		break;
	return type;

    }

/*----------*/
char *itoa (ival)

    int	ival;

    {

	static char	buf [50];
	
	sprintf (buf, "%d", ival);
	return &buf [0];

    }
	
/*--------------------------------- Channel access functions ---------------------------------*/

typedef struct {
	NODE	node;
	chid	chid;
} Channel;

LIST	channelList;

#define foreachChannel(x)	for (x = (Channel *)lstFirst (&channelList); x; x = (Channel *)lstNext (x))
#define	ReadTimeout		5.0		/* timeout for data reception in float secs */

/*----------*/
char *buildChannelName (name, field)

    char	*name;
    char	*field;

    {

	static char	fullname [CHANNEL_NAME_ENTRY_SIZE+1];	/* to build the returned string in */
	char		*p = &fullname [0];
	char		*end = &fullname [sizeof fullname - 2];
	
	while (p < end && *name && *name != ' ' && *name != '.') /* copy the non-blank name chars, up to any '.' */
	    *p++ = *name++;
	while (p < end && *field)				/* append the field name */
	    *p++ = *field++;
	*p = 0;
	
	return fullname;		/* note: it only exists until the next time this is called */

    }
	
/*----------*/
Channel *getChannel (name)

    char	*name;

    {

	Channel		*p;
	char		*q;

	q = strchr (name, ' ');
	if (q)
	    *q = '\0';
		
	foreachChannel (p)
	    if (strcmp (ca_name (p->chid), name) == 0)   /* Already in the channel list */
		return p;

	/* Not yet in the list, add it */

	p = (Channel *)calloc (1, sizeof (Channel));
	assert (p != NULL, FATAL, "can't get memory");
	
	sprintf (errorTextBuffer, "ca_search () failed for channel: (%s)", name);
	assert (ca_search (name, &p->chid) == ECA_NORMAL, PUT_INTO_ERROR, errorTextBuffer);

	ca_pend_io (ReadTimeout);
	while (ca_state (p->chid) != currOpen) {
	    printf ("EVOLProcessor: Unable to connect to %s -- retrying.\n", name);
	    ca_pend_io (ReadTimeout);
	}
	lstAdd (&channelList, &p->node);
	return p;

    }

/*----------*/
void closeAllChannels ()

    {

	Channel	*p;

	foreachChannel (p)
	    if (p->chid)
		ca_clear_channel (p->chid);
	ca_task_exit ();

    }

/*----------*/	
double writeDoubleToChannel (name, value)

    char	*name;
    double	value;

    {

	int	status;
	Channel	*p;
	int	Count = 0;

	p = getChannel (name);
	do {
	    status = ca_put (DBR_DOUBLE, p->chid, &value);
	    if (Verbose)
		printf ("EVOLProcessor: Writing (%s) Count:%d Value:%10.2lf Status:%d\n",
				ca_name (p->chid), Count, value, status);
	    if (status == ECA_NORMAL)
		ca_flush_io ();
	} while (status == ECA_BADCHID && Count++ < maxCaLoopCount);
	
	sprintf (errorTextBuffer, "EVOLProcessor: Can't put a value to channel:%s status:%d",
			ca_name (p->chid), status);
	assert (status == ECA_NORMAL && ca_state (p->chid) == currOpen, PUT_INTO_ERROR, errorTextBuffer);

	return value;

    }

/*----------*/
double readChannelAsDouble (name)

    char	*name;

    {

	int	status;
	Channel	*p = getChannel (name);
	int	count;
	void	getCallback ();

	sprintf (errorTextBuffer, "EVOLProcessor: Channel %s not connected", ca_name (p->chid));
	assert (ca_state (p->chid) == currOpen, PUT_INTO_ERROR, errorTextBuffer);

	status = ca_get_callback (DBR_DOUBLE, p->chid, getCallback, 0);
	fetchComplete = 0;
	count = 0;
	do {

	    if (Verbose)
		syslog (LOG_ALERT|LOG_LOCAL1, "EVOLProcessor: Reading %s, count = %d ",
				ca_name (p->chid), count);

	    status = ca_pend_io (1.0);

	} while (!fetchComplete && (count++ < maxCaLoopCount));

	if (!fetchComplete) {
	    sprintf (errorTextBuffer, "EVOLProcessor: Error while reading %s", ca_name (p->chid));
	    assert (status == ECA_NORMAL && ca_state (p->chid) == currOpen,
			PUT_INTO_ERROR, errorTextBuffer);
	    return -999999.99;
	}

	if (Verbose)
	    syslog (LOG_ALERT|LOG_LOCAL1, "EVOLProcessor: Read %s = %10.2lf\n",
			ca_name (p->chid), latestDoubleValue);

	return latestDoubleValue;
	    	    
    }

/*----------*/
int readChannelAsInt (name)
    char	*name;

    {

	int	status;
	Channel	*p = getChannel (name);
	int	count;
	void	getCallback ();
    	    
	sprintf (errorTextBuffer, "EVOLProcessor: Channel %s not connected", ca_name (p->chid));
	assert (ca_state (p->chid) == currOpen, PUT_INTO_ERROR, errorTextBuffer);

	status = ca_get_callback (DBR_INT, p->chid, getCallback, 1);
	fetchComplete = 0;
	count = 0;
	do {
	    if (Verbose)
		syslog (LOG_ALERT|LOG_LOCAL1, "EVOLProcessor: Reading %s, count = %d ",
				ca_name (p->chid), count);
	    status = ca_pend_io (1.0);
	} while (!fetchComplete && (count++ < maxCaLoopCount));

	if (!fetchComplete) {
	    sprintf (errorTextBuffer, "EVOLProcessor: Error while reading %s", ca_name (p->chid));
	    assert (status == ECA_NORMAL && ca_state (p->chid) == currOpen,
			PUT_INTO_ERROR, errorTextBuffer);
	    return -999999.99;
	}

	if (Verbose)
	    syslog (LOG_ALERT|LOG_LOCAL1, "EVOLProcessor: Read %s = %d\n",
			ca_name (p->chid), latestIntegerValue);

	return latestIntegerValue;

    }

/*----------*/
void getCallback (args)

    struct	event_handler_args	args;

    {

	if (args.usr == 0) {
	    latestDoubleValue = * ( (double *)args.dbr);
	} else {
	    latestIntegerValue = * ( (int *)args.dbr);
	}
	fetchComplete = 1;

    }

/*------------------------------------- Database access functions ------------------------------*/

C4CODE	codeBaseDesc;

/*----------*/
dBASEopen ()

    {

	d4init (&codeBaseDesc);
	codeBaseDesc.read_lock = FALSE;
	codeBaseDesc.create_error = FALSE;
	codeBaseDesc.field_name_error = FALSE;
	codeBaseDesc.go_error = FALSE;
	codeBaseDesc.open_error = FALSE;
	codeBaseDesc.skip_error = FALSE;
	codeBaseDesc.tag_name_error = FALSE;

    }

/*----------*/
dBASEupdateString (desc, field, string)

    D4DATA	*desc;
    F4FIELD	*field;
    char	*string;

    {

	if (strncmp (&desc->record [field->offset], string, field->len) != 0) {
	    f4assign (field, string);
	}

    }

/*----------*/
dBASEupdateInt (desc, field, value)

    D4DATA	*desc;
    F4FIELD	*field;
    int		value;

    {

	char	buf [BUFSIZ];
	
	sprintf (buf, "%*d", field->len, value);
	if (strncmp (&desc->record [field->offset], buf, field->len) != 0) {
	    f4assign_int (field, value);
	}

    }

/*----------*/	
dBASEupdateDouble (desc, field, value)

    D4DATA	*desc;
    F4FIELD	*field;
    double	value;

    {

	char	buf [BUFSIZ];

	sprintf (buf, "%*.*f", field->len, field->dec, value);
	if (strncmp (&desc->record [field->offset], buf, field->len) != 0) {
	    f4assign_double (field, value);
	}

    }

/*----------*/
char *dbType (name)

    char 	*name;

    {

	static D4DATA	*PointDB = NULL;	/* Point Database file descriptor */
	char		*p;
	
	if (!PointDB) {
	    red_RedundancyOff ();
	    assert (PointDB = d4open (&codeBaseDesc, pointName), FATAL, "cannot open point database file");
	    red_RedundancyOn ();
	}

	d4top (PointDB);
	d4tag_select (PointDB, d4tag (PointDB, "POINTNAME"));
	sprintf (errorTextBuffer, "didn't find point: (%s) in point database", name);
	assert (d4seek (PointDB, name) != 0, PUT_INTO_ERROR, errorTextBuffer);

	p = f4str (d4field (PointDB, "TYPE"));
	d4unlock_all (PointDB);
	return p;

    }

D4DATA		*evolDesc = NULL;
F4FIELD		*evolEvolNumberField;
F4FIELD		*evolProductField;
F4FIELD		*evolQuantityField;
F4FIELD		*evolUnitsField;
F4FIELD		*evolMovetypeField;
F4FIELD		*evolAutoSwngField;
F4FIELD		*evolClientField;
F4FIELD		*evolClientidField;
F4FIELD		*evolDateField;
F4FIELD		*evolTimeField;
F4FIELD		*evolTimestampField;
F4FIELD		*evolStatusField;
F4FIELD		*evolEvolModeField;
F4FIELD		*evolWeightedAvgField;
T4TAG		*evolTag;

typedef struct {
	int	evolNumber;
	char	product [PRODUCT_ENTRY_SIZE+1];
	double	quantity;
	char	units [UNITS_ENTRY_SIZE+1];
	char	movetype [MOVE_TYPE_ENTRY_SIZE+1];
	char	autoSwing [AUTO_SWING_ENTRY_SIZE+1];
	char	client [CLIENT_ENTRY_SIZE+1];
	char	clientid [CLIENT_ID_ENTRY_SIZE+1];
	char	date [DATE_STRING_ENTRY_SIZE+1];
	char	time [TIME_STRING_ENTRY_SIZE+1];
	char	timestamp [TIMESTAMP_ENTRY_SIZE+1];
	char	status [STATUS_ENTRY_SIZE+1];
	char	evolMode [EVOLMODE_ENTRY_SIZE+1];
	double	weightedAvg;
} EvolutionRecord;

/*----------*/
EvolutionRecord *getEvolutionRecord (evolution)

    {

	static EvolutionRecord	evol;
	char	searchString [EVOL_NUMBER_ENTRY_SIZE+1];

	memset ( (char *) &evol, 0, sizeof evol);

	sprintf (searchString, "%*d", EVOL_NUMBER_ENTRY_SIZE, evolution);
	d4tag_select (evolDesc, evolTag);
	d4seek (evolDesc, searchString);
	d4unlock_all (evolDesc);

	evol.evolNumber = f4int (evolEvolNumberField);
	sprintf (errorTextBuffer, "didn't find evolution:%d", evolution);
	assert (evol.evolNumber == evolution, FATAL, errorTextBuffer);

	f4ncpy (evolProductField, evol.product, PRODUCT_ENTRY_SIZE);
	f4ncpy (evolUnitsField, evol.units, UNITS_ENTRY_SIZE);
	f4ncpy (evolMovetypeField, evol.movetype, MOVE_TYPE_ENTRY_SIZE);
	f4ncpy (evolAutoSwngField, evol.autoSwing, AUTO_SWING_ENTRY_SIZE);
	f4ncpy (evolClientField, evol.client, CLIENT_ENTRY_SIZE);
	f4ncpy (evolClientidField, evol.clientid, CLIENT_ID_ENTRY_SIZE);
	f4ncpy (evolDateField, evol.date, DATE_STRING_ENTRY_SIZE);
	f4ncpy (evolTimeField, evol.time, TIME_STRING_ENTRY_SIZE);
	f4ncpy (evolTimestampField, evol.timestamp, TIMESTAMP_ENTRY_SIZE);
	f4ncpy (evolStatusField, evol.status, STATUS_ENTRY_SIZE);
	f4ncpy (evolEvolModeField, evol.evolMode, EVOLMODE_ENTRY_SIZE);
	evol.quantity = f4double (evolQuantityField);
	evol.weightedAvg = f4double (evolWeightedAvgField);
	
	return &evol;

    }

/*----------*/	
evolutionMode (evolution)

    {

	return strcmp (getEvolutionRecord (evolution)->evolMode, EvolutionModeText [EVOLUTION]) == 0;

    }

/*----------*/	
segmentMode (evolution)

    {

	return strcmp (getEvolutionRecord (evolution)->evolMode, EvolutionModeText [SEGMENT]) == 0;

    }
	
/*----------*/
evolSwingEnabled (evolution)

    {

	return strcmp (getEvolutionRecord (evolution)->autoSwing, YES_STRING) == 0;

    }

/*----------*/	
evolIsAReceive (pEvol)

    EvolutionRecord	*pEvol;

    {

	return (strncmp (pEvol->movetype, PIER_RECEIVE_STRING, strlen (PIER_RECEIVE_STRING)) == 0) ||
		   (strncmp (pEvol->movetype, TRUCK_RECEIVE_STRING, strlen (TRUCK_RECEIVE_STRING)) == 0);

    }
	
/*----------*/
evolIsAnIssue (pEvol)

    EvolutionRecord	*pEvol;

    {

	return (strncmp (pEvol->movetype, PIER_ISSUE_STRING, strlen (PIER_ISSUE_STRING)) == 0) ||
		   (strncmp (pEvol->movetype, TRUCK_ISSUE_STRING, strlen (TRUCK_ISSUE_STRING)) == 0);

    }
	
/*----------*/
evolIsATransfer (pEvol)

    EvolutionRecord	*pEvol;

    {

	return (strncmp (pEvol->movetype, TRANSFER_STRING, strlen (TRANSFER_STRING)) == 0);

    }
	
D4DATA		*evolSegsDesc = NULL;
F4FIELD		*segEvolNumField;
F4FIELD		*segSegNumField;
F4FIELD		*segOrderNumField;
F4FIELD		*segQuantityField;
F4FIELD		*segUnitsField;
F4FIELD		*segLineupField;
F4FIELD		*segPauseField;
F4FIELD		*segStatusField;
F4FIELD		*segQuantityXferField;
F4FIELD		*segActFlowRateField;
F4FIELD		*segActFlowTimeField;
F4FIELD		*segLastStartActFlowTimeField;
F4FIELD		*segLastStartTimeField;
F4FIELD		*segSwingEvolField;
F4FIELD		*segSwingEvolSegField;
F4FIELD		*segSwingTimeField;
F4FIELD		*segRequestTypeField;
F4FIELD		*segEnvironmentStringField;
F4FIELD		*segScriptIdField;
F4FIELD		*segAvgApiField;
T4TAG		*evolSegsTag;

typedef struct {
	int	evolNumber;
	int	segNumber;
	int	orderNumber;
	double	quantity;
	char	units [UNITS_ENTRY_SIZE+1];
	char	lineup [LINEUP_NAME_ENTRY_SIZE+1];
	char	pause [PAUSE_ENTRY_SIZE+1];
	char	status [STATUS_ENTRY_SIZE+1];
	double	quantityXfer;
	double	actFlowRate;
	double	actFlowTime;
	double	lastStartActFlowTime;
	int	lastStartTime;
	int	swingEvol;
	int	swingEvolSeg;
	char	swingTime [TIME_STRING_ENTRY_SIZE+1];
	char	requestType [REQTYPE_ENTRY_SIZE+1];
	int	scriptId;
	double	avgApi;
	char	environmentString [ENVIRONMENTSTRING_ENTRY_SIZE+1];
} SegmentRecord;

/*----------*/
SegmentRecord *getSegmentRecord (evolution, segment)

    {

	static SegmentRecord	seg;

	memset ( (char *) &seg, 0, sizeof seg);

	d4tag_select (evolSegsDesc, evolSegsTag);
	if (d4top (evolSegsDesc) == 0)
	    do {
		seg.evolNumber	= f4int (segEvolNumField);
		seg.segNumber	= f4int (segSegNumField);
		d4unlock_all (evolSegsDesc);
	    } while ( (seg.evolNumber != evolution || seg.segNumber != segment) && d4skip (evolSegsDesc, 1) == 0);

	sprintf (errorTextBuffer, "didn't find evolution:%d segment:%d", evolution, segment);
	assert (seg.evolNumber == evolution && seg.segNumber == segment, PUT_INTO_ERROR, errorTextBuffer);

	f4ncpy (segUnitsField, seg.units, UNITS_ENTRY_SIZE);
	f4ncpy (segLineupField, seg.lineup, LINEUP_NAME_ENTRY_SIZE);
	f4ncpy (segPauseField, seg.pause, PAUSE_ENTRY_SIZE);
	f4ncpy (segStatusField, seg.status, STATUS_ENTRY_SIZE);
	f4ncpy (segSwingTimeField, seg.swingTime, TIME_STRING_ENTRY_SIZE);
	f4ncpy (segRequestTypeField, seg.requestType, REQTYPE_ENTRY_SIZE);
	f4ncpy (segEnvironmentStringField, seg.environmentString, ENVIRONMENTSTRING_ENTRY_SIZE);
	seg.quantity = f4double (segQuantityField);
	seg.quantityXfer = f4double (segQuantityXferField);
	seg.actFlowRate = f4double (segActFlowRateField);
	seg.actFlowTime = f4double (segActFlowTimeField);
	seg.lastStartActFlowTime = f4double (segLastStartActFlowTimeField);
	seg.lastStartTime = f4int (segLastStartTimeField);
	seg.orderNumber = f4int (segOrderNumField);
	seg.swingEvol = f4int (segSwingEvolField);
	seg.swingEvolSeg = f4int (segSwingEvolSegField);
	seg.scriptId = f4int (segScriptIdField);
	seg.avgApi = f4double (segAvgApiField);
	
	return &seg;

    }

/*----------*/
SegmentRecord *getNextSegmentRecord (evolution, orderNumber)

    int	evolution;
    int	orderNumber;

    {

	static SegmentRecord	seg;
	EvolutionState		status;
	char			searchString [EVOL_NUMBER_ENTRY_SIZE+ORDER_NUMBER_ENTRY_SIZE+1];

	memset ( (char *) &seg, 0, sizeof seg);

	d4tag_select (evolSegsDesc, d4tag (evolSegsDesc, "ORDERTAG"));
	sprintf (searchString, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, evolution, ORDER_NUMBER_ENTRY_SIZE, orderNumber);
	d4seek (evolSegsDesc, searchString);
	d4unlock_all (evolSegsDesc);

	sprintf (errorTextBuffer, "didn't find the current evolution:%d orderNumber:%d", evolution, orderNumber);
	assert (f4int (segEvolNumField) == evolution && f4int (segOrderNumField) == orderNumber, PUT_INTO_ERROR, errorTextBuffer);

	if (d4skip (evolSegsDesc, 1) == 0);
	    do {		/* now find the next non-complete one */
		seg.evolNumber	= f4int (segEvolNumField);
		f4ncpy (segStatusField, seg.status, STATUS_ENTRY_SIZE);
		status = findStatus (seg.status);
		d4unlock_all (evolSegsDesc);
	    } while (seg.evolNumber == evolution && (status == Completed || status == MCompleted) && d4skip (evolSegsDesc, 1) == 0);

	if (seg.evolNumber != evolution)	/* There is no next, return the current one */
	    d4seek (evolSegsDesc, searchString);
	d4unlock_all (evolSegsDesc);

	seg.evolNumber = f4int (segEvolNumField);
	seg.segNumber = f4int (segSegNumField);
	seg.orderNumber = f4int (segOrderNumField);
	f4ncpy (segUnitsField, seg.units, UNITS_ENTRY_SIZE);
	f4ncpy (segLineupField, seg.lineup, LINEUP_NAME_ENTRY_SIZE);
	f4ncpy (segPauseField, seg.pause, PAUSE_ENTRY_SIZE);
	f4ncpy (segStatusField, seg.status, STATUS_ENTRY_SIZE);
	f4ncpy (segSwingTimeField, seg.swingTime, TIME_STRING_ENTRY_SIZE);
	f4ncpy (segRequestTypeField, seg.requestType, REQTYPE_ENTRY_SIZE);
	f4ncpy (segEnvironmentStringField, seg.environmentString, ENVIRONMENTSTRING_ENTRY_SIZE);
	seg.quantity = f4double (segQuantityField);
	seg.quantityXfer = f4double (segQuantityXferField);
	seg.actFlowRate = f4double (segActFlowRateField);
	seg.actFlowTime	= f4double (segActFlowTimeField);
	seg.lastStartActFlowTime = f4double (segLastStartActFlowTimeField);
	seg.lastStartTime = f4int (segLastStartTimeField);
	seg.orderNumber = f4int (segOrderNumField);
	seg.swingEvol = f4int (segSwingEvolField);
	seg.swingEvolSeg = f4int (segSwingEvolSegField);
	seg.scriptId = f4int (segScriptIdField);
	seg.avgApi = f4double (segAvgApiField);
	
	return &seg;

    }

typedef struct {
	NODE	node;
	int	evolNumber;
	int	segNumber;
	int	orderNumber;
} EvolSegRecord;

LIST	evolSegList;

#define foreachEvolSegRecord(x)	for (x = (EvolSegRecord *)lstFirst (&evolSegList); x; x = (EvolSegRecord *)lstNext (x))

/*----------*/
void getEvolSegList ()

    {

	static int	firstTime = 1;

	if (firstTime) {
	    lstInit (&evolSegList);
	    firstTime = 0;
	} else
	    lstFree (&evolSegList);

	d4tag_select (evolSegsDesc, evolSegsTag);
	if (d4top (evolSegsDesc) == 0)
	    do {
		EvolSegRecord *p = (EvolSegRecord *)calloc (1, sizeof (EvolSegRecord));
		assert (p != NULL, FATAL, "can't get memory");
		p->evolNumber = f4int (segEvolNumField);
		p->segNumber = f4int (segSegNumField);
		p->orderNumber = f4int (segOrderNumField);
		lstAdd (&evolSegList, &p->node);
		d4unlock_all (evolSegsDesc);
	    } while (d4skip (evolSegsDesc, 1) == 0);
	d4unlock_all (evolSegsDesc);

    }

D4DATA		*lineupMsrmntVolsDesc = NULL;
F4FIELD		*lineupMsrmntVolsLineupNameField;
F4FIELD		*lineupMsrmntVolsLineupMsrmntVolsField;
T4TAG		*lineupMsrmntVolsTag;

typedef struct {
	NODE	node;
	char	lineupName [LINEUP_NAME_ENTRY_SIZE+1];
	char	msrmntVol [CHANNEL_NAME_ENTRY_SIZE+1];
} LineupMsrmntVolsRecord;

LIST	lineupMsrmntVolsList;

#define foreachLineupMsrmntVolsRecord(x)	for (x = (LineupMsrmntVolsRecord *)lstFirst (&lineupMsrmntVolsList); x; x = (LineupMsrmntVolsRecord *)lstNext (x))

/*----------*/
void getLineupMsrmntVolsList (lineupName)

    char	*lineupName;

    {

	static int	firstTime = 1;
	char		searchString [LINEUP_NAME_ENTRY_SIZE+1];

	if (firstTime) {
	    lstInit (&lineupMsrmntVolsList);
	    firstTime = 0;
	} else
	    lstFree (&lineupMsrmntVolsList);
			
	sprintf (searchString, "%-*s", LINEUP_NAME_ENTRY_SIZE, lineupName);
	d4tag_select (lineupMsrmntVolsDesc, lineupMsrmntVolsTag);
	if (d4seek (lineupMsrmntVolsDesc, searchString) == 0)
	    do {
		char	buf [100];

		f4ncpy (lineupMsrmntVolsLineupNameField, buf, LINEUP_NAME_ENTRY_SIZE);
		if (strncmp (buf, lineupName, LINEUP_NAME_ENTRY_SIZE) == 0) {
		    LineupMsrmntVolsRecord *p = (LineupMsrmntVolsRecord *)calloc (1, sizeof (LineupMsrmntVolsRecord));
		    assert (p != NULL, FATAL, "can't get memory");
		    f4ncpy (lineupMsrmntVolsLineupNameField, p->lineupName, LINEUP_NAME_ENTRY_SIZE);
		    f4ncpy (lineupMsrmntVolsLineupMsrmntVolsField, p->msrmntVol, CHANNEL_NAME_ENTRY_SIZE);
		    lstAdd (&lineupMsrmntVolsList, &p->node);
		} else
		    break;
		d4unlock_all (lineupMsrmntVolsDesc);
	    } while (d4skip (lineupMsrmntVolsDesc, 1) == 0);
	d4unlock_all (lineupMsrmntVolsDesc);

    }

D4DATA		*lineupXrefVolsDesc = NULL;
F4FIELD		*lineupXrefVolsLineupNameField;
F4FIELD		*lineupXrefVolsLineupXrefVolsField;
T4TAG		*lineupXrefVolsTag;

typedef struct {
	NODE	node;
	char	lineupName [LINEUP_NAME_ENTRY_SIZE+1];
	char	xrefVol [CHANNEL_NAME_ENTRY_SIZE+1];
} LineupXrefVolsRecord;

LIST	lineupXrefVolsList;

#define foreachLineupXrefVolsRecord(x)	for (x = (LineupXrefVolsRecord *)lstFirst (&lineupXrefVolsList); x; x = (LineupXrefVolsRecord *)lstNext (x))

/*----------*/
void getLineupXrefVolsList (lineupName)

    char	*lineupName;

    {

	static int	firstTime = 1;
	char		searchString [LINEUP_NAME_ENTRY_SIZE+1];

	if (firstTime) {
	    lstInit (&lineupXrefVolsList);
	    firstTime = 0;
	} else
	    lstFree (&lineupXrefVolsList);
			
	sprintf (searchString, "%-*s", LINEUP_NAME_ENTRY_SIZE, lineupName);
	d4tag_select (lineupXrefVolsDesc, lineupXrefVolsTag);
	if (d4seek (lineupXrefVolsDesc, searchString) == 0)
	    do {
		char	buf [100];
		f4ncpy (lineupXrefVolsLineupNameField, buf, LINEUP_NAME_ENTRY_SIZE);
		if (strncmp (buf, lineupName, LINEUP_NAME_ENTRY_SIZE) == 0) {
		    LineupXrefVolsRecord *p = (LineupXrefVolsRecord *)calloc (1, sizeof (LineupXrefVolsRecord));
		    assert (p != NULL, FATAL, "can't get memory");
		    f4ncpy (lineupXrefVolsLineupNameField, p->lineupName, LINEUP_NAME_ENTRY_SIZE);
		    f4ncpy (lineupXrefVolsLineupXrefVolsField, p->xrefVol, CHANNEL_NAME_ENTRY_SIZE);
		    lstAdd (&lineupXrefVolsList, &p->node);
		} else
		    break;
		d4unlock_all (lineupXrefVolsDesc);
	    } while (d4skip (lineupXrefVolsDesc, 1) == 0);
	d4unlock_all (lineupXrefVolsDesc);

    }

D4DATA		*msrsDesc = NULL;
F4FIELD		*msrsEvolNumberField;
F4FIELD		*msrsSegNumberField;
F4FIELD		*msrsChannelField;
F4FIELD		*msrsStartVolField;
F4FIELD		*msrsStartTimeField;
F4FIELD		*msrsStopVolField;
F4FIELD		*msrsStopTimeField;

/*----------*/
void saveStartMsrsRecord (evolution, segment)

    int	evolution;
    int	segment;

    {

	SegmentRecord		*pseg = getSegmentRecord (evolution, segment);
	LineupMsrmntVolsRecord	*pmv;
	long			timestamp = time (0);

	/* Save the timestamp and the elapsed time values into this evol,seg record */

	dBASEupdateString (evolSegsDesc, segLastStartTimeField, itoa (timestamp));
	dBASEupdateDouble (evolSegsDesc, segLastStartActFlowTimeField, pseg->actFlowTime);
	updateDBASEfile (evolSegsDesc);

	getLineupMsrmntVolsList (getSegmentRecord (evolution, segment)->lineup);
	foreachLineupMsrmntVolsRecord (pmv) {
	    double vol = readChannelAsDouble (pmv->msrmntVol);		
	    d4tag_select (msrsDesc, d4tag (msrsDesc, "MSRSTAG"));
	    d4append_blank (msrsDesc);
	    d4go (msrsDesc, d4recno (msrsDesc));
	    dBASEupdateString (msrsDesc, msrsEvolNumberField, itoa (evolution));
	    dBASEupdateString (msrsDesc, msrsSegNumberField, itoa (segment));
	    dBASEupdateString (msrsDesc, msrsChannelField, pmv->msrmntVol);
	    dBASEupdateString (msrsDesc, msrsStartTimeField, itoa (timestamp));
	    dBASEupdateDouble (msrsDesc, msrsStartVolField, vol);
	    dBASEupdateString (msrsDesc, msrsStopTimeField, itoa (timestamp));	/* 0 means hasn't stopped yet */
	    dBASEupdateDouble (msrsDesc, msrsStopVolField, vol);
	    updateDBASEfile (msrsDesc);
	}
	d4unlock_all (msrsDesc);

    }

/*----------*/
void saveStopMsrsRecord (evolution, segment)

    int		evolution;
    int		segment;

    {

	LineupMsrmntVolsRecord	*pmv;
	double			stopVolume;

	getLineupMsrmntVolsList (getSegmentRecord (evolution, segment)->lineup);
	foreachLineupMsrmntVolsRecord (pmv) {
	    char searchString [EVOL_NUMBER_ENTRY_SIZE+SEG_NUMBER_ENTRY_SIZE+CHANNEL_NAME_ENTRY_SIZE+1];
	    sprintf (searchString, "%*d%*d%-*s", EVOL_NUMBER_ENTRY_SIZE, evolution, SEG_NUMBER_ENTRY_SIZE,
				segment, CHANNEL_NAME_ENTRY_SIZE, pmv->msrmntVol);
	    d4tag_select (msrsDesc, d4tag (msrsDesc, "REVMSRSTAG"));
	    if (d4seek (msrsDesc, searchString) == 0) {
		dBASEupdateString (msrsDesc, msrsStopTimeField, itoa (time (0)));
		stopVolume = readChannelAsDouble (pmv->msrmntVol);
		dBASEupdateDouble (msrsDesc, msrsStopVolField, stopVolume);
		updateDBASEfile (msrsDesc);
	    }
	}
	d4unlock_all (msrsDesc);

    }
	
typedef struct {
	NODE	node;
	char	channel   [CHANNEL_NAME_ENTRY_SIZE+1];
	double	startVolume;
	long	startTime;
	double	stopVolume;
	long	stopTime;
} MsrsRecord;

LIST	msrsList;

/*----------*/
void getMsrsList (evolution, segment)

    int	evolution;
    int	segment;

    {
	static int	firstTime = 1;
	char		searchString   [EVOL_NUMBER_ENTRY_SIZE+SEG_NUMBER_ENTRY_SIZE+1];

	if (firstTime) {
	    lstInit (&msrsList);
	    firstTime = 0;
	} else
	    lstFree (&msrsList);

	sprintf (searchString, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, evolution, SEG_NUMBER_ENTRY_SIZE, segment);
	d4tag_select (msrsDesc, d4tag (msrsDesc, "REVMSRSTAG"));
	if (d4seek (msrsDesc, searchString) == 0) {
	    do {
		int e = f4int (msrsEvolNumberField);
		int s = f4int (msrsSegNumberField);

		if (e != evolution || s != segment)
		    break;
		else {
		    MsrsRecord *p = (MsrsRecord *) calloc (1, sizeof (MsrsRecord));
		    assert (p != NULL, FATAL, "can't get memory");
		    f4ncpy (msrsChannelField, p->channel, CHANNEL_NAME_ENTRY_SIZE);
		    p->startVolume = f4double (msrsStartVolField);
		    p->startTime = f4int (msrsStartTimeField);
		    p->stopVolume = f4double (msrsStopVolField);
		    p->stopTime = f4int (msrsStopTimeField);
		    lstAdd (&msrsList, &p->node);
		}
		d4unlock_all (msrsDesc);
	    } while (d4skip (msrsDesc, 1) == 0);
	}
	d4unlock_all (msrsDesc);

    }
	
D4DATA		*xrefDesc = NULL;
F4FIELD		*xrefEvolNumberField;
F4FIELD		*xrefSegNumberField;
F4FIELD		*xrefChannelField;
F4FIELD		*xrefStartVolField;
F4FIELD		*xrefStartTimeField;
F4FIELD		*xrefStopVolField;
F4FIELD		*xrefStopTimeField;

/*----------*/
void saveStartXrefRecord (evolution, segment)

    int	evolution;
    int	segment;

    {

	LineupXrefVolsRecord	*pmv;
	long			timestamp = time (0);

	getLineupXrefVolsList (getSegmentRecord (evolution, segment)->lineup);
	foreachLineupXrefVolsRecord (pmv) {
	    double vol = readChannelAsDouble (pmv->xrefVol);
	    d4tag_select (xrefDesc, d4tag (xrefDesc, "XREFTAG"));
	    d4append_blank (xrefDesc);
	    d4go (xrefDesc, d4recno (xrefDesc));
	    dBASEupdateString (xrefDesc, xrefEvolNumberField, itoa (evolution));
	    dBASEupdateString (xrefDesc, xrefSegNumberField, itoa (segment));
	    dBASEupdateString (xrefDesc, xrefChannelField, pmv->xrefVol);
	    dBASEupdateString (xrefDesc, xrefStartTimeField, itoa (timestamp));
	    dBASEupdateDouble (xrefDesc, xrefStartVolField, vol);
	    dBASEupdateString (xrefDesc, xrefStopTimeField, itoa (timestamp));	/* 0 means hasn't stopped yet */
	    dBASEupdateDouble (xrefDesc, xrefStopVolField, vol);
	    updateDBASEfile (xrefDesc);
	}
	d4unlock_all (xrefDesc);

    }

/*----------*/
void saveStopXrefRecord (evolution, segment)

    int		evolution;
    int		segment;

    {

	LineupXrefVolsRecord	*pmv;
	double			stopVolume;

	getLineupXrefVolsList (getSegmentRecord (evolution, segment)->lineup);
	foreachLineupXrefVolsRecord (pmv) {
	    char searchString   [EVOL_NUMBER_ENTRY_SIZE+SEG_NUMBER_ENTRY_SIZE+CHANNEL_NAME_ENTRY_SIZE+1];
	    sprintf (searchString, "%*d%*d%-*s", EVOL_NUMBER_ENTRY_SIZE, evolution,
			SEG_NUMBER_ENTRY_SIZE, segment, CHANNEL_NAME_ENTRY_SIZE, pmv->xrefVol);
	    d4tag_select (xrefDesc, d4tag (xrefDesc, "REVXREFTAG"));
	    if (d4seek (xrefDesc, searchString) == 0) {
		dBASEupdateString (xrefDesc, xrefStopTimeField, itoa (time (0)));
		stopVolume = readChannelAsDouble (pmv->xrefVol);
		dBASEupdateDouble (xrefDesc, xrefStopVolField, stopVolume);
		updateDBASEfile (xrefDesc);
	    }
	}
	d4unlock_all (xrefDesc);

    }

D4DATA		*lineupDesc = NULL;
F4FIELD		*lineupNameField;
F4FIELD		*lineupSetupTimeField;
F4FIELD		*lineupShutdownTimeField;
F4FIELD		*lineupFlowRateField;
F4FIELD		*lineupRelatedDisplayField;
F4FIELD		*lineupManualScriptField;
F4FIELD		*lineupStartScriptField;
F4FIELD		*lineupStopScriptField;
F4FIELD		*lineupPauseScriptField;
F4FIELD		*lineupResumeScriptField;
F4FIELD		*lineupMonitorScriptField;
T4TAG		*lineupTag;

typedef struct {
	char	name [LINEUP_NAME_ENTRY_SIZE+1];
	double	setupTime;
	double	shutdownTime;
	double	flowRate;
	char	relatedDisplay [REL_DSPLY_ENTRY_SIZE+1];
	char	manualScript [CSL_SCRIPT_ENTRY_SIZE+1];
	char	startScript [CSL_SCRIPT_ENTRY_SIZE+1];
	char	stopScript [CSL_SCRIPT_ENTRY_SIZE+1];
	char	pauseScript [CSL_SCRIPT_ENTRY_SIZE+1];
	char	resumeScript [CSL_SCRIPT_ENTRY_SIZE+1];
	char	monitorScript [CSL_SCRIPT_ENTRY_SIZE+1];
} LineupRecord;

/*----------*/
LineupRecord *getLineupRecord (name)		/* from dBASE file */

    char	*name;
	
    {

	static LineupRecord	lineup;
	char			searchString [LINEUP_NAME_ENTRY_SIZE+1];

	memset ( (char *) &lineup, 0, sizeof lineup);

	sprintf (searchString, "%-*s", LINEUP_NAME_ENTRY_SIZE, name);
	d4tag_select (lineupDesc, lineupTag);
	d4seek (lineupDesc, searchString);
	d4unlock_all (lineupDesc);

	f4ncpy (lineupNameField, lineup.name, LINEUP_NAME_ENTRY_SIZE);
	sprintf (errorTextBuffer, "didn't find lineup: (%s)", name);
	assert (strncmp (lineup.name, name, LINEUP_NAME_ENTRY_SIZE) == 0, PUT_INTO_ERROR, errorTextBuffer);

	f4ncpy (lineupRelatedDisplayField, lineup.relatedDisplay, REL_DSPLY_ENTRY_SIZE);
	f4ncpy (lineupManualScriptField, lineup.manualScript, CSL_SCRIPT_ENTRY_SIZE);
	f4ncpy (lineupStartScriptField, lineup.startScript, CSL_SCRIPT_ENTRY_SIZE);
	f4ncpy (lineupStopScriptField, lineup.stopScript, CSL_SCRIPT_ENTRY_SIZE);
	f4ncpy (lineupPauseScriptField, lineup.pauseScript, CSL_SCRIPT_ENTRY_SIZE);
	f4ncpy (lineupResumeScriptField, lineup.resumeScript, CSL_SCRIPT_ENTRY_SIZE);
	f4ncpy (lineupMonitorScriptField, lineup.monitorScript, CSL_SCRIPT_ENTRY_SIZE);
	lineup.setupTime = f4double (lineupSetupTimeField);
	lineup.shutdownTime = f4double (lineupShutdownTimeField);
	lineup.flowRate = f4double (lineupFlowRateField);

	return &lineup;

    }

D4DATA		*lineupSwingsDesc = NULL;
I4INDEX		*lineupSwingsTagDesc;
F4FIELD		*lineupSwingsFromLineupField;
F4FIELD		*lineupSwingsToLineupField;
F4FIELD		*lineupSwingsSwingTimeField;
F4FIELD		*lineupSwingsCSLScriptField;
T4TAG		*lineupSwingsTag;

typedef struct {
	char	fromLineup [FROM_LINEUP_SIZE+1];
	char	toLineup [TO_LINEUP_SIZE+1];
	double	swingTime;
	char	cslScript [CSL_SCRIPT_ENTRY_SIZE+1];
} LineupSwingsRecord;

/*----------*/
LineupSwingsRecord *getLineupSwingsRecord (evolution, segment)

    int	evolution;
    int	segment;

    {

	char			searchString [FROM_LINEUP_SIZE+1+TO_LINEUP_SIZE+1];
	static LineupSwingsRecord	lineupSwings;
	SegmentRecord		*pseg = getSegmentRecord (evolution, segment);
	char			*fromLineup = pseg->lineup;
	char			*toLineup = getNextSegmentRecord (evolution, pseg->orderNumber)->lineup;

	memset ( (char *) &lineupSwings, 0, sizeof lineupSwings);

	sprintf (searchString, "%-*s%-*s", FROM_LINEUP_SIZE, fromLineup, TO_LINEUP_SIZE, toLineup);
	searchString [I4MAX_KEY_SIZE] = 0;
	d4tag_select (lineupSwingsDesc, lineupSwingsTag);
	d4seek (lineupSwingsDesc, searchString);
	d4unlock_all (lineupSwingsDesc);

	f4ncpy (lineupSwingsFromLineupField, lineupSwings.fromLineup, FROM_LINEUP_SIZE);
	sprintf (errorTextBuffer, "didn't find from lineup swing from: (%s) to: (%s)", fromLineup, toLineup);
	assert (strncmp (lineupSwings.fromLineup, fromLineup, FROM_LINEUP_SIZE) == 0,
			PUT_INTO_ERROR, errorTextBuffer);

	f4ncpy (lineupSwingsToLineupField, lineupSwings.toLineup, TO_LINEUP_SIZE);
	assert (strncmp (lineupSwings.toLineup, toLineup, TO_LINEUP_SIZE) == 0,
			PUT_INTO_ERROR, errorTextBuffer);

	f4ncpy (lineupSwingsCSLScriptField, lineupSwings.cslScript, CSL_SCRIPT_ENTRY_SIZE);
	lineupSwings.swingTime = f4double (lineupSwingsSwingTimeField);

	return &lineupSwings;

    }

D4DATA		*lineupCmpntsDesc = NULL;
F4FIELD		*lineupCmpntsLineupNameField;
F4FIELD		*lineupCmpntsComponetNameField;
F4FIELD		*lineupCmpntsStateField;

typedef struct {
	NODE	node;
	char	lineupName [LINEUP_NAME_ENTRY_SIZE+1];
	char	componentName [CHANNEL_NAME_ENTRY_SIZE+1];
	char	state [STATE_ENTRY_SIZE+1];
	char	*faultField;
	int	faultValue;
} LineupCmpntsRecord;

LIST	componentList;

#define foreachLineupCmpntsRecord(x)	for (x = (LineupCmpntsRecord *)lstFirst (&componentList); x; x = (LineupCmpntsRecord *)lstNext (x))

/*----------*/
void getLineupCmpntsList (lineupName)

    char	*lineupName;

    {

	static int	firstTime = 1;

	if (firstTime) {
	    lstInit (&componentList);
	    firstTime = 0;
	} else
	    lstFree (&componentList);
		
	d4tag_select (lineupCmpntsDesc, NULL);
	if (d4top (lineupCmpntsDesc) == 0)
	    do {
		char	buf [100];
		f4ncpy (lineupCmpntsLineupNameField, buf, LINEUP_NAME_ENTRY_SIZE);
		if (strncmp (buf, lineupName, LINEUP_NAME_ENTRY_SIZE) == 0) {
		    LineupCmpntsRecord *p = (LineupCmpntsRecord *)calloc (1, sizeof (LineupCmpntsRecord));
		    assert (p != NULL, FATAL, "can't get memory");
		    f4ncpy (lineupCmpntsLineupNameField,   p->lineupName, LINEUP_NAME_ENTRY_SIZE);
		    f4ncpy (lineupCmpntsComponetNameField, p->componentName, CHANNEL_NAME_ENTRY_SIZE);
		    f4ncpy (lineupCmpntsStateField, p->state, STATE_ENTRY_SIZE);
		    if (strcmp (dbType (p->componentName), "valve") == 0) {
			p->faultField = ".VASV";
			p->faultValue = 8;
		    } else if (strcmp (dbType (p->componentName), "pipe") == 0) {
			p->faultField = ".PASV";
			p->faultValue = 4;
		    } else {
			p->faultField = NULL;
			p->faultValue = 0;
		    }
		    lstAdd (&componentList, &p->node);
		}
		d4unlock_all (lineupCmpntsDesc);
	    } while (d4skip (lineupCmpntsDesc, 1) == 0);
	d4unlock_all (lineupCmpntsDesc);

    }

D4DATA		*lineupSrcsDesc = NULL;
F4FIELD		*lineupSrcsLineupNameField;
F4FIELD		*lineupSrcsSourceField;

typedef struct {
	NODE	node;
	char	lineupName [LINEUP_NAME_ENTRY_SIZE+1];
	char	source [CHANNEL_NAME_ENTRY_SIZE+1];
} LineupSrcsRecord;

LIST	sourceList;

#define foreachLineupSrcsRecord(x)	for (x = (LineupSrcsRecord *)lstFirst (&sourceList); x; x = (LineupSrcsRecord *)lstNext (x))

/*----------*/
void getLineupSrcsList (lineupName)

    char	*lineupName;

    {
	
	static int	firstTime = 1;

	if (firstTime) {
	    lstInit (&sourceList);
	    firstTime = 0;
	} else
	    lstFree (&sourceList);
		
	d4tag_select (lineupSrcsDesc, NULL);
	if (d4top (lineupSrcsDesc) == 0)
	    do {
		char	buf [100];
		f4ncpy (lineupSrcsLineupNameField, buf, LINEUP_NAME_ENTRY_SIZE);
		if (strncmp (buf, lineupName, LINEUP_NAME_ENTRY_SIZE) == 0) {
		    LineupSrcsRecord *p = (LineupSrcsRecord *)calloc (1, sizeof (LineupSrcsRecord));
		    assert (p != NULL, FATAL, "can't get memory");
		    f4ncpy (lineupSrcsLineupNameField, p->lineupName, LINEUP_NAME_ENTRY_SIZE);
		    f4ncpy (lineupSrcsSourceField, p->source, CHANNEL_NAME_ENTRY_SIZE);
		    lstAdd (&sourceList, &p->node);
		}
		d4unlock_all (lineupSrcsDesc);
	    } while (d4skip (lineupSrcsDesc, 1) == 0);
	d4unlock_all (lineupSrcsDesc);

    }

D4DATA		*lineupDestsDesc = NULL;
F4FIELD		*lineupDestsLineupNameField;
F4FIELD		*lineupDestsDestinationField;

typedef struct {
	NODE	node;
	char	lineupName [LINEUP_NAME_ENTRY_SIZE+1];
	char	destination [CHANNEL_NAME_ENTRY_SIZE+1];
} LineupDestsRecord;

LIST	destinationList;

#define foreachLineupDestsRecord(x)	for (x = (LineupDestsRecord *)lstFirst (&destinationList); x; x = (LineupDestsRecord *)lstNext (x))

/*----------*/
void getLineupDestsList (lineupName)

    char	*lineupName;

    {

	static int	firstTime = 1;

	if (firstTime) {
	    lstInit (&destinationList);
	    firstTime = 0;
	} else
	    lstFree (&destinationList);
		
	d4tag_select (lineupDestsDesc, NULL);
	if (d4top (lineupDestsDesc) == 0)
	    do {
		char	buf [100];
		f4ncpy (lineupDestsLineupNameField, buf, LINEUP_NAME_ENTRY_SIZE);
		if (strncmp (buf, lineupName, LINEUP_NAME_ENTRY_SIZE) == 0) {
		    LineupDestsRecord *p = (LineupDestsRecord *)calloc (1, sizeof (LineupDestsRecord));
		    assert (p != NULL, FATAL, "can't get memory");
		    f4ncpy (lineupDestsLineupNameField, p->lineupName, LINEUP_NAME_ENTRY_SIZE);
		    f4ncpy (lineupDestsDestinationField, p->destination, CHANNEL_NAME_ENTRY_SIZE);
		    lstAdd (&destinationList, &p->node);
		}
		d4unlock_all (lineupDestsDesc);
	    } while (d4skip (lineupDestsDesc, 1) == 0);
	d4unlock_all (lineupDestsDesc);

    }

D4DATA		*segSrcsDesc = NULL;
F4FIELD		*segSrcsEvolNumberField;
F4FIELD		*segSrcsSegNumberField;
F4FIELD		*segSrcsSourceField;
F4FIELD		*segSrcsApiField;
T4TAG		*segSrcsTag;

typedef struct {
	int	evolNumber;
	int	segNumber;
	char	source [CHANNEL_NAME_ENTRY_SIZE+1];
	double	api;
} SegSrcsRecord;

/*----------*/
SegSrcsRecord *getSegSrcsRecord (evolution, segment)

    {

	static SegSrcsRecord	segSrc;
	char			searchString [EVOL_NUMBER_ENTRY_SIZE+SEG_NUMBER_ENTRY_SIZE+1];

	memset ( (char *) &segSrc, 0, sizeof segSrc);

	sprintf (searchString, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, evolution,
			SEG_NUMBER_ENTRY_SIZE, segment);
	d4tag_select (segSrcsDesc, segSrcsTag);
	d4seek (segSrcsDesc, searchString);
	d4unlock_all (segSrcsDesc);

	segSrc.evolNumber = f4int (segSrcsEvolNumberField);
	sprintf (errorTextBuffer, "didn't find evolution:%d", evolution);
	assert (segSrc.evolNumber == evolution, IGNORE, errorTextBuffer);

	segSrc.segNumber = f4int (segSrcsSegNumberField);
	sprintf (errorTextBuffer, "didn't find segment:%d", segment);
	assert (segSrc.segNumber == segment, IGNORE, errorTextBuffer);

	f4ncpy (segSrcsSourceField, segSrc.source, CHANNEL_NAME_ENTRY_SIZE);
	segSrc.api = f4double (segSrcsApiField);
	
	return &segSrc;

    }

/*----------*/	
openTFdBASEfiles ()

    {

	evolDesc = d4open (&codeBaseDesc, evolName);
	if (!evolDesc)
	    evolDesc = d4create (&codeBaseDesc, evolName, evolFields, evolTagInfo);
	assert (evolDesc != NULL, FATAL, "cannot open or create evolution database file");

	evolEvolNumberField = d4field (evolDesc, "EVOLNUMBER");
	evolProductField = d4field (evolDesc, "PRODUCT");
	evolQuantityField = d4field (evolDesc, "QUANTITY");
	evolUnitsField = d4field (evolDesc, "UNITS");
	evolMovetypeField = d4field (evolDesc, "MOVETYPE");
	evolAutoSwngField = d4field (evolDesc, "AUTOSWNG");
	evolClientField = d4field (evolDesc, "CLIENT");
	evolClientidField = d4field (evolDesc, "CLIENTID");
	evolDateField = d4field (evolDesc, "DATE");
	evolTimeField = d4field (evolDesc, "TIME");
	evolTimestampField = d4field (evolDesc, "TIMESTAMP");
	evolStatusField	= d4field (evolDesc, "STATUS");
	evolEvolModeField = d4field (evolDesc, "EVOLMODE");
	evolWeightedAvgField = d4field (evolDesc, "WEIGHTAVG");
	evolTag	= d4tag (evolDesc, "EVOLTAG");
	d4unlock_all (evolDesc);

	evolSegsDesc = d4open (&codeBaseDesc, evolSegsName);
	if (!evolSegsDesc)
	    evolSegsDesc = d4create (&codeBaseDesc, evolSegsName, evolSegsFields, evolSegsTagInfo);
	assert (evolSegsDesc != NULL, FATAL, "cannot open evolution segments database file");

	segEvolNumField	= d4field (evolSegsDesc, "EVOLNUMBER");
	segSegNumField = d4field (evolSegsDesc, "SEGNUMBER");
	segOrderNumField = d4field (evolSegsDesc, "ORDERNUM");
	segQuantityField = d4field (evolSegsDesc, "QUANTITY");
	segUnitsField = d4field (evolSegsDesc, "UNITS");
	segLineupField = d4field (evolSegsDesc, "LINEUPNAME");
	segPauseField = d4field (evolSegsDesc, "PAUSE");
	segStatusField = d4field (evolSegsDesc, "STATUS");
	segQuantityXferField = d4field (evolSegsDesc, "VOLXFERED");
	segActFlowRateField = d4field (evolSegsDesc, "ACTFLWRT");
	segActFlowTimeField = d4field (evolSegsDesc, "ACTFLWTIME");
	segLastStartActFlowTimeField = d4field (evolSegsDesc, "LSTRAFTIME");
	segLastStartTimeField = d4field (evolSegsDesc, "LSTRTIME");
	segSwingEvolField = d4field (evolSegsDesc, "SWNGEVOLNM");
	segSwingEvolSegField = d4field (evolSegsDesc, "SWNGEVOLSG");
	segSwingTimeField = d4field (evolSegsDesc, "SWNGTIME");
	segRequestTypeField = d4field (evolSegsDesc, "REQTYPE");
	segScriptIdField = d4field (evolSegsDesc, "SCRIPTID");
	segAvgApiField = d4field (evolSegsDesc, "AVGAPI");
	segEnvironmentStringField = d4field (evolSegsDesc, "ENVSTRING");
	evolSegsTag = d4tag (evolSegsDesc, "EVOLSEGTAG");
	d4unlock_all (evolSegsDesc);

	segSrcsDesc = d4open (&codeBaseDesc, segSrcsName);
	if (!segSrcsDesc)
	    segSrcsDesc = d4create (&codeBaseDesc, segSrcsName, segSrcsFields, segSrcsTagInfo);
	assert (segSrcsDesc != NULL, FATAL, "cannot open segment sources database file");

	segSrcsEvolNumberField = d4field (segSrcsDesc, "EVOLNUMBER");
	segSrcsSegNumberField = d4field (segSrcsDesc, "SEGNUMBER");
	segSrcsSourceField = d4field (segSrcsDesc, "SOURCE");
	segSrcsApiField = d4field (segSrcsDesc, "API");
	segSrcsTag = d4tag (segSrcsDesc, "SEGSRCSTAG");
	d4unlock_all (segSrcsDesc);

	{
	    D4DATA	*segDestsDesc = NULL;
	    segDestsDesc = d4open (&codeBaseDesc, segDestsName);
	    if (!segDestsDesc)
		segDestsDesc = d4create (&codeBaseDesc, segDestsName, segDestsFields, segDestsTagInfo);
	    assert (segDestsDesc != NULL, FATAL, "cannot open sources database file");
	    d4unlock_all (segDestsDesc);
	    d4close (segDestsDesc);
	}

	msrsDesc = d4open (&codeBaseDesc, msrsName);
	if (!msrsDesc)
	    msrsDesc = d4create (&codeBaseDesc, msrsName, msrsFields, MsrsTagInfo);
	assert (msrsDesc != NULL, FATAL, "cannot open measurement database file");

	msrsEvolNumberField = d4field (msrsDesc, "EVOLNUMBER");
	msrsSegNumberField = d4field (msrsDesc, "SEGNUMBER");
	msrsChannelField = d4field (msrsDesc, "CHANNEL");
	msrsStartVolField = d4field (msrsDesc, "STARTVOL");
	msrsStartTimeField = d4field (msrsDesc, "STARTTIME");
	msrsStopVolField = d4field (msrsDesc, "STOPVOL");
	msrsStopTimeField = d4field (msrsDesc, "STOPTIME");
	d4unlock_all (msrsDesc);

	xrefDesc = d4open (&codeBaseDesc, xrefName);
	if (!xrefDesc)
	    xrefDesc = d4create (&codeBaseDesc, xrefName, xrefFields, XrefTagInfo);
	assert (xrefDesc != NULL, FATAL, "cannot open xref database file");

	xrefEvolNumberField = d4field (xrefDesc, "EVOLNUMBER");
	xrefSegNumberField = d4field (xrefDesc, "SEGNUMBER");
	xrefChannelField = d4field (xrefDesc, "CHANNEL");
	xrefStartVolField = d4field (xrefDesc, "STARTVOL");
	xrefStartTimeField = d4field (xrefDesc, "STARTTIME");
	xrefStopVolField = d4field (xrefDesc, "STOPVOL");
	xrefStopTimeField = d4field (xrefDesc, "STOPTIME");
	d4unlock_all (xrefDesc);

	red_RedundancyOff ();
	lineupDesc = d4open (&codeBaseDesc, lineupName);
	if (!lineupDesc)
	    lineupDesc = d4create (&codeBaseDesc, lineupName, lineupFields, lineupTagInfo);
	red_RedundancyOn ();
	assert (lineupDesc != NULL, FATAL, "cannot open lineup database file");

	lineupNameField = d4field (lineupDesc, "LINEUPNAME");
	lineupSetupTimeField = d4field (lineupDesc, "SETUPTIME");
	lineupShutdownTimeField = d4field (lineupDesc, "SHTDWNTIME");
	lineupFlowRateField = d4field (lineupDesc, "FLOWRATE");
	lineupRelatedDisplayField = d4field (lineupDesc, "RELDSPLY");
	lineupManualScriptField	= d4field (lineupDesc, "MANCSL");
	lineupStartScriptField = d4field (lineupDesc, "STARTSCRPT");
	lineupStopScriptField = d4field (lineupDesc, "STOPSCRPT");
	lineupPauseScriptField = d4field (lineupDesc, "PAUSESCRPT");
	lineupResumeScriptField	= d4field (lineupDesc, "RESUMSCRPT");
	lineupMonitorScriptField = d4field (lineupDesc, "MONSCRPT");
	lineupTag = d4tag (lineupDesc, "LINEUPTAG");
	d4unlock_all (lineupDesc);

	red_RedundancyOff ();
	lineupSwingsDesc = d4open (&codeBaseDesc, lineupSwingsName);
	if (!lineupSwingsDesc)
	    lineupSwingsDesc = d4create (&codeBaseDesc, lineupSwingsName, lineupSwingsFields, 
						lineupSwingsTagsInfo);
	red_RedundancyOn ();
	assert (lineupSwingsDesc != NULL, FATAL, "cannot open lineup switch database file");

	lineupSwingsFromLineupField = d4field (lineupSwingsDesc, "FROMLINEUP");
	lineupSwingsToLineupField = d4field (lineupSwingsDesc, "TOLINEUP");
	lineupSwingsSwingTimeField = d4field (lineupSwingsDesc, "SWINGTIME");
	lineupSwingsCSLScriptField = d4field (lineupSwingsDesc, "CSLSCRIPT");
	lineupSwingsTag = d4tag (lineupSwingsDesc, "LINEUPTAG");
	d4unlock_all (lineupSwingsDesc);

	red_RedundancyOff ();
	lineupCmpntsDesc = d4open (&codeBaseDesc, lineupCmpntsName);
	if (!lineupCmpntsDesc)
	    lineupCmpntsDesc = d4create (&codeBaseDesc, lineupCmpntsName, lineupCmpntsFields, 
						lineupCmpntsTagsInfo);
	red_RedundancyOn ();
	assert (lineupCmpntsDesc != NULL, FATAL, "cannot open lineup components database file");

	lineupCmpntsLineupNameField = d4field (lineupCmpntsDesc, "LINEUPNAME");
	lineupCmpntsComponetNameField = d4field (lineupCmpntsDesc, "CMPNTS");
	lineupCmpntsStateField = d4field (lineupCmpntsDesc, "STATE");
	d4unlock_all (lineupCmpntsDesc);
	
	red_RedundancyOff ();
	lineupSrcsDesc = d4open (&codeBaseDesc, lineupSrcsName);
	if (!lineupSrcsDesc)
	    lineupSrcsDesc = d4create (&codeBaseDesc, lineupSrcsName, lineupSrcsFields, lineupTagInfo);
	red_RedundancyOn ();
	assert (lineupSrcsDesc != NULL, FATAL, "cannot open lineup sources database file");
	
	lineupSrcsLineupNameField = d4field (lineupSrcsDesc, "LINEUPNAME");
	lineupSrcsSourceField = d4field (lineupSrcsDesc, "SRCS");
	d4unlock_all (lineupSrcsDesc);

	red_RedundancyOff ();
	lineupDestsDesc = d4open (&codeBaseDesc, lineupDestsName);
	if (!lineupDestsDesc)
	    lineupDestsDesc = d4create (&codeBaseDesc, lineupDestsName, lineupDestsFields, lineupTagInfo);
	red_RedundancyOn ();
	assert (lineupDestsDesc != NULL, FATAL, "cannot open lineup destinations database file");
	
	lineupDestsLineupNameField = d4field (lineupDestsDesc, "LINEUPNAME");
	lineupDestsDestinationField = d4field (lineupDestsDesc, "DESTS");
	d4unlock_all (lineupDestsDesc);

	red_RedundancyOff ();
	lineupMsrmntVolsDesc = d4open (&codeBaseDesc, lineupMsrmntVolsName);
	if (!lineupMsrmntVolsDesc)
	    lineupMsrmntVolsDesc = d4create (&codeBaseDesc, lineupMsrmntVolsName, lineupMsrmntVolsFields, 
					lineupMsrmntVolsTagInfo);
	red_RedundancyOn ();
	assert (lineupMsrmntVolsDesc != NULL, FATAL, "cannot open measurement volumes database file");

	lineupMsrmntVolsLineupNameField = d4field (lineupMsrmntVolsDesc, "LINEUPNAME");
	lineupMsrmntVolsLineupMsrmntVolsField = d4field (lineupMsrmntVolsDesc, "MSRMNTVOLS");
	lineupMsrmntVolsTag = d4tag (lineupMsrmntVolsDesc, "LINEUPTAG");
	d4unlock_all (lineupMsrmntVolsDesc);

	red_RedundancyOff ();
	lineupXrefVolsDesc = d4open (&codeBaseDesc, lineupXrefVolsName);
	if (!lineupXrefVolsDesc)
	    lineupXrefVolsDesc = d4create (&codeBaseDesc, lineupXrefVolsName, lineupXrefVolsFields, 
					lineupXrefVolsTagInfo);
	red_RedundancyOn ();
	assert (lineupXrefVolsDesc != NULL, FATAL, "cannot open xref volumes database file");

	lineupXrefVolsLineupNameField = d4field (lineupXrefVolsDesc, "LINEUPNAME");
	lineupXrefVolsLineupXrefVolsField = d4field (lineupXrefVolsDesc, "XREFVOLS");
	lineupXrefVolsTag = d4tag (lineupXrefVolsDesc, "LINEUPTAG");
	d4unlock_all (lineupXrefVolsDesc);

    }

/*----------*/
void CloseAllFiles ()

    {
	d4close (evolDesc);
	d4close (evolSegsDesc);
	d4close (segSrcsDesc);
	d4close (msrsDesc);
	d4close (xrefDesc);
	d4close (lineupDesc);
	d4close (lineupSwingsDesc);
	d4close (lineupCmpntsDesc);
	d4close (lineupSrcsDesc);
	d4close (lineupDestsDesc);
	d4close (lineupMsrmntVolsDesc);
	d4close (lineupXrefVolsDesc);

    }

/*------------------------------------- Operator request function ---------------------------------*/

/*----------*/
addRequest (evolution, segment, requestType, environmentString)

    int	evolution;
    int	segment;
    EvolutionProcessorRequestType	requestType;
    char	*environmentString;	/* don't change it if NULL */

    {

	if (requestType == ESTOP) {    /* ignore evolution and segment, distribute ESTOP to all of them */
	    d4tag_select (evolSegsDesc, evolSegsTag);
	    if (d4top (evolSegsDesc) == 0)
		do {
		    dBASEupdateString (evolSegsDesc, segRequestTypeField, EvolutionProcessorRequestText [ESTOP]);
		    if (environmentString)
			dBASEupdateString (evolSegsDesc, segEnvironmentStringField, environmentString);
		    updateDBASEfile (evolSegsDesc);
		    d4unlock_all (evolSegsDesc);
		} while (d4skip (evolSegsDesc, 1) == 0);
	} else {
	    if (f4int (segEvolNumField) != evolution || f4int (segSegNumField) != segment)
		getSegmentRecord (evolution, segment);
	    if (f4int (segEvolNumField) == evolution && f4int (segSegNumField) == segment) {
		dBASEupdateString (evolSegsDesc, segRequestTypeField, EvolutionProcessorRequestText [requestType]);
		if (environmentString)
		    dBASEupdateString (evolSegsDesc, segEnvironmentStringField, environmentString);
		updateDBASEfile (evolSegsDesc);
	    }
	}
	d4unlock_all (evolSegsDesc);

    }


/*---------------------------------- Evolution segment update functions -----------------------------*/

/*----------*/
EvolutionState determineEvolutionState (evol)

    int	evol;

    {

	EvolSegRecord	*pseg;
	int		pendingSeen = 0;
	int		completedSeen = 0;
	int		mCompletedSeen = 0;
	int		otherSeen = 0;
	double		qtyXfered = 0.0;
	double		evolQuantity = 0.99 * getEvolutionRecord (evol)->quantity;
	
	if (evolutionMode (evol) && !evolSwingEnabled (evol)) {		/* check for Switching */
	    int i = 0;	
	    foreachEvolSegRecord (pseg) {
		EvolutionState	status = findStatus (getSegmentRecord (pseg->evolNumber, pseg->segNumber)->status);
		if (pseg->evolNumber == evol) {				
		    if (status == Starting)
			return (i == 0) ? Starting : Switching;
		    else if (status == Stopping)
			return (getNextSegmentRecord (evol,
					pseg->orderNumber)->segNumber == pseg->segNumber) ?
					Stopping : Switching;
		    i++;
		}
	    }
	}

	foreachEvolSegRecord (pseg) {
	    if (pseg->evolNumber == evol) {
		SegmentRecord	*pSeg	= getSegmentRecord (pseg->evolNumber, pseg->segNumber);
		EvolutionState	status	= findStatus (pSeg->status);	
		qtyXfered += pSeg->quantityXfer;  /* sum the segment quantities for possible completion test below */
		switch (status) {
		    case Pending:
			pendingSeen++;
			break;
		    case Completed:
			completedSeen++;
			break;
		    case MCompleted:
			mCompletedSeen++;
			break;
		    default:
			otherSeen++;
			break;	
		}
		if (otherSeen)
		    return status;
	    }
	}

	if (pendingSeen && ! (completedSeen || mCompletedSeen))
	    return Pending;
	else if (!pendingSeen && completedSeen)
	    return (qtyXfered < evolQuantity) ? Idle : Completed;
	else if (!pendingSeen && mCompletedSeen)
	    return (qtyXfered < evolQuantity) ? Idle : MCompleted;
	else if (pendingSeen && (completedSeen || mCompletedSeen))
	    return Idle;
	else
	    return BadState;	

    }

/*----------*/
void saveStartingMeasurementsAndAPIs (evolution, segment)

    int	evolution;
    int	segment;

    {

	EvolutionRecord	*pevol = getEvolutionRecord (evolution);
        SegmentRecord   *pseg = getSegmentRecord (evolution, segment);
        LineupMsrmntVolsRecord  *pmv;

        /* Write avgApi (or weightedAvg) to api_xxx blocks */

        getLineupMsrmntVolsList (pseg->lineup);
        foreachLineupMsrmntVolsRecord (pmv) {
	    char	channelName [CHANNEL_NAME_ENTRY_SIZE+1];
	    char	*p;
	    strcat (strcpy (channelName, "API_"), pmv->msrmntVol);
	    p = strchr (channelName,'.');
	    if (p)
		*p = '\0';
	    buildChannelName (channelName, ".VAL");
	    writeDoubleToChannel (channelName, evolutionMode (evolution) ? pevol->weightedAvg : pseg->avgApi);
	}
	
	/* Update lineup sources APIs for Receive movement types */

	if (evolIsAReceive (pevol)) {
	    LineupSrcsRecord	*p;
	    getLineupSrcsList (pseg->lineup);
	    foreachLineupSrcsRecord (p)
		writeDoubleToChannel (buildChannelName (p->source, ".SGRV"), pevol->weightedAvg);
	}

	/* Save starting measuments and xrefs */

	saveStartMsrsRecord (evolution, segment);
	saveStartXrefRecord (evolution, segment);

    }

/*----------*/
void saveFinalVols (evolution, segment)

    int		evolution;
    int		segment;

    {

	saveStopMsrsRecord (evolution, segment);
	saveStopXrefRecord (evolution, segment);

    }

/*----------*/
checkSegmentSrcQtysAndDestRoom (evolution, segment)

    int	evolution;
    int	segment;

    {

	EvolutionRecord	*pevol = getEvolutionRecord (evolution);
	SegmentRecord	*pSeg = getSegmentRecord (evolution, segment);

	/* Check to be sure that there is enough product to transfer */
	/* and that there is enough empty space to receive it. */

	if (evolIsAnIssue (pevol) || evolIsATransfer (pevol)) {
	    LineupSrcsRecord	*p;
	    double		qtyOnHand = 0.0;		
	    getLineupSrcsList (pSeg->lineup);
	    foreachLineupSrcsRecord (p)
		qtyOnHand += readChannelAsDouble (buildChannelName (p->source, ".AVAP"));
	    sprintf (errorTextBuffer, "insufficient product to begin evolution %d segment %d",
				evolution, segment);
	    assert (qtyOnHand > pSeg->quantity, PUT_INTO_ERROR, errorTextBuffer);
	}

	if (evolIsAReceive (pevol) || evolIsATransfer (pevol)) {
	    LineupDestsRecord	*p;
	    double		spaceOnHand = 0.0;		
	    getLineupDestsList (pSeg->lineup);
	    foreachLineupDestsRecord (p)
		spaceOnHand += readChannelAsDouble (buildChannelName (p->destination, ".ULLG"));
	    sprintf (errorTextBuffer, "insufficient room to begin evolution %d segment %d",
					evolution, segment);
	    assert (spaceOnHand > pSeg->quantity, PUT_INTO_ERROR, errorTextBuffer);
	}

    }

/*----------*/
performPrestartTest (newState, evolution, segment)

    EvolutionState	newState;
    int			evolution;
    int			segment;

    {
	/* Check that there is a reasonable chance to complete the transfer */
	/* (if SwingTo check all remaining segments else check only the given segment) */

	if (newState == SwingTo && evolutionMode (evolution)) {
	    d4tag_select (evolSegsDesc, evolSegsTag);
	    if (d4top (evolSegsDesc) == 0)
		do {
		    if (f4int (segEvolNumField) == evolution) {
			SegmentRecord *pSeg = getSegmentRecord (evolution, f4int (segSegNumField));
			EvolutionState currentState = findStatus (pSeg->status);
			if (currentState != Completed && currentState != MCompleted)
			    checkSegmentSrcQtysAndDestRoom (evolution, f4int (segSegNumField));
		    }
		    d4unlock_all (evolSegsDesc);
		} while (d4skip (evolSegsDesc, 1) == 0);
		d4unlock_all (evolSegsDesc);
	} else
	    checkSegmentSrcQtysAndDestRoom (evolution, segment);

    }

/*----------*/
nextSegmentNumber (evolution, orderNumber)	/* finds next segment by order number but returns the segment number */

    int		evolution;
    int		orderNumber;

    {

	SegmentRecord	*pNewSeg = getNextSegmentRecord (evolution, orderNumber);
	return (pNewSeg->orderNumber != orderNumber) ? pNewSeg->segNumber : 0;

    }

/*----------*/
changeSegmentToNewState (evolution, segment, currentState, newState)

    int			evolution;
    int			segment;
    EvolutionState	currentState;
    EvolutionState	newState;

    {

	EvolutionRecord	*pevol;
	SegmentRecord	*pseg;
	LineupRecord	*lineupRecord;
	ScriptId	scriptId;
	char		*scriptName;
	int		nextSeg;

	/* Do a bit more checking before commiting to change the state */

	switch (newState) {
	    case Starting:
	    case Manual:
	    case SwingTo:
		performPrestartTest (newState, evolution, segment);
		break;
	}

	/* Change the segments state, then this evolutions state */

	pevol = getEvolutionRecord (evolution);
	pseg = getSegmentRecord (evolution, segment);
	lineupRecord = getLineupRecord (pseg->lineup);
	if (Verbose)
	    printf (" going to:%s", EvolutionStateText [newState]);
	dBASEupdateString (evolSegsDesc, segStatusField, EvolutionStateText [newState]);
	updateDBASEfile (evolSegsDesc);
	dBASEupdateString (evolDesc, evolStatusField, EvolutionStateText [determineEvolutionState (evolution)]);
	updateDBASEfile (evolDesc);
	pseg = getSegmentRecord (evolution, segment);
	if (pseg->scriptId != SCRIPTID_UNKNOWN)
	    CSLMonitorKill (localhostname, pseg->scriptId);

	/* Perform any pre-entry logic */

	switch (newState) {
	    case Error:					/* general stopping logic */
	    case Stopped:
	    case Completed:
	    case MCompleted:
	    case EStopped:
	  	if (newState != EStopped || (newState == EStopped && currentState == EStopping))
		  	saveFinalVols (evolution, segment);
	    case OnHold:
	    case MOnHold:
	    case Paused:
		scriptId = SCRIPTID_UNKNOWN;
		break;

	    case Starting:				/* general starting logic */
	    case Manual:
	    case SwingTo:
	  	saveStartingMeasurementsAndAPIs (evolution, segment);
	    case EStopping:
	    case Stopping:
	    case Pausing:
	    case Resuming:
	    case Active:
	    case SwingFrom:
		switch (newState) {
		    case Starting:
			scriptName = lineupRecord->startScript;
			break;
		    case Manual:
			scriptName = lineupRecord->manualScript;
			break;
		    case EStopping:
		    case Stopping:
			scriptName = lineupRecord->stopScript;
			break;
		    case Pausing:
			scriptName = lineupRecord->pauseScript;
			break;
		    case Resuming:
			scriptName = lineupRecord->resumeScript;
			break;
		    case Active:
			scriptName = lineupRecord->monitorScript;
			break;
		    case SwingFrom:	
			scriptName = getLineupSwingsRecord (evolution, segment)->cslScript;
			break;
		    case SwingTo:
			scriptName = NULL;
			break;
		}
		scriptId = (scriptName) ? CSLMonitorExecute (localhostname, CSLscript, scriptName, 0L,
					pseg->environmentString) : SCRIPTID_UNKNOWN;
		break;
	  }
	
	if (evolutionMode (evolution) && (nextSeg = nextSegmentNumber (evolution, pseg->orderNumber)) != 0)
	    /* Generate Swinging and/or Switching logic */
	    if (evolSwingEnabled (evolution)) {
		if (currentState == SwingFrom && newState == Completed)
		    addRequest (evolution, nextSeg, SWINGTO, NULL);
		else if (currentState == SwingFrom && newState == Error)	
		    addRequest (evolution, nextSeg, SWINGFROM, NULL);
		else if (currentState == Active    && newState == SwingFrom)
		    addRequest (evolution, nextSeg, SWINGTO, NULL);
	    } else
		switch (newState) {
		    case Completed:
			addRequest (evolution, nextSeg, START, NULL);
			break;
		    case MCompleted:
			addRequest (evolution, nextSeg, MANUAL, NULL);	break;
		}

	if (f4int (segEvolNumField) != evolution || f4int (segSegNumField) != segment)
	    /* move the segment record into view (again, since the do funcs above have moved it) */
	    getSegmentRecord (evolution, segment);

	if (f4int (segEvolNumField) == evolution && f4int (segSegNumField) == segment) {
	    dBASEupdateInt (evolSegsDesc, segScriptIdField, scriptId);
	    dBASEupdateString (evolSegsDesc, segRequestTypeField, EvolutionProcessorRequestText [NONE]);
	    updateDBASEfile (evolSegsDesc);
	}

	if (completionLogicPath && (newState == Completed || newState == MCompleted)) {
	    char	command [BUFSIZ];
	    sprintf (&command [0], "%s %d %d", completionLogicPath, evolution, segment);
	    if (fork () == 0)
		/* child becomes completion logic (exit (1) in case execlp () fails) */
		execlp ("/bin/sh", "sh", "-c", command, 0), exit (1);
	}

    }

/*----------*/
evolSegIsPauseOnComplete (evolution, segment)

    {

	return strcmp (getSegmentRecord (evolution, segment)->pause, YES_STRING) == 0;

    }

/*----------*/
haveAllComponets (evolution, segment)

    {

	LineupCmpntsRecord	*p, *needed, *inUse;
	EvolSegRecord		*pseg;
	LIST			needList;
	LIST			inUseList;
	int			returnStatus;

#define foreachNeedListRecord(x)	for (x = (LineupCmpntsRecord *)lstFirst (&needList);  x; x = (LineupCmpntsRecord *)lstNext (x))
#define foreachInUseRecord(x)		for (x = (LineupCmpntsRecord *)lstFirst (&inUseList); x; x = (LineupCmpntsRecord *)lstNext (x))

	/* Check for component conflicts make a needList and an inUseList and compare them */

	lstInit (&needList);
	if (evolutionMode (evolution)) {
	    foreachEvolSegRecord (pseg) {
		EvolutionState status = findStatus (getSegmentRecord (pseg->evolNumber, pseg->segNumber)->status);
		if (pseg->evolNumber == evolution && status != Completed && status != MCompleted) {
		    getLineupCmpntsList (getSegmentRecord (pseg->evolNumber, pseg->segNumber)->lineup);
		    foreachLineupCmpntsRecord (p) {
			LineupCmpntsRecord *np = (LineupCmpntsRecord *)calloc (1, sizeof (LineupCmpntsRecord));
			assert (np != NULL, FATAL, "can't get memory");
			*np = *p;
			lstAdd (&needList, &np->node);
		    }
		}
	    }
	} else {
	    getLineupCmpntsList (getSegmentRecord (evolution, segment)->lineup);
	    foreachLineupCmpntsRecord (p) {
		LineupCmpntsRecord *np = (LineupCmpntsRecord *)calloc (1, sizeof (LineupCmpntsRecord));
		assert (np != NULL, FATAL, "can't get memory");
		*np = *p;
		lstAdd (&needList, &np->node);
	    }
	}

	lstInit (&inUseList);
	foreachEvolSegRecord (pseg) {
	    if (pseg->evolNumber == evolution && pseg->segNumber == segment)
		continue;	/* don't add in this segments components */

	    switch (findStatus (getSegmentRecord (pseg->evolNumber, pseg->segNumber)->status)) {
		case Starting:
		case Active:
		case Manual:
		case Pausing:
		case Paused:
		case Resuming:
		case SwingFrom:
		case SwingTo:
		case Switching:
		case Stopping:
		    getLineupCmpntsList (getSegmentRecord (pseg->evolNumber, pseg->segNumber)->lineup);
		    foreachLineupCmpntsRecord (p) {
			LineupCmpntsRecord *np = (LineupCmpntsRecord *)calloc (1, sizeof (LineupCmpntsRecord));
			assert (np != NULL, FATAL, "can't get memory");
			*np = *p;
			lstAdd (&inUseList, &np->node);
		    }
		    break;
	    }
	}

	foreachNeedListRecord (needed) {  /* Compare the needed componentName to each of the componentNames in use */
	    foreachInUseRecord (inUse) {
		if (strncmp (needed->componentName, inUse->componentName, sizeof inUse->componentName) == 0) {
		    /* We need a component that is currently in use */
		    /* now check the state because it is ok if they match, unless the in use components */
		    /* state is reserved. */
		    if (strncasecmp (needed->state, inUse->state, sizeof inUse->state) ||
					strncasecmp (inUse->state, "reserved", 8) == 0) {
			lstFree (&needList);
			lstFree (&inUseList);
			return 0;
		    }
		}
	    }
	}		
	lstFree (&inUseList);

	/* Check that the components in the needList are ok */

	returnStatus = 1;
	foreachNeedListRecord (p) {
	    if (readChannelAsInt (buildChannelName (p->componentName, ".OSCN")) == 0) {
		returnStatus = 0;		/* it's off scan */
		break;
	    }
	    if (p->faultField) {	/* only for pipe or valve types */
		if (readChannelAsInt (buildChannelName (p->componentName, p->faultField)) == p->faultValue) {
		    returnStatus = 0;	/* it's faulted */
		    break;
		} else if (evolutionMode (evolution) &&
					 (readChannelAsInt (buildChannelName (p->componentName, ".SWSV")) == 0)) {
		    returnStatus = 0;	/* it's in local */
		    break;
		}
	    }
	}
	lstFree (&needList);
	return returnStatus;

    }

/*----------*/
writeAveFlow (evolution, segment)	/* Update average flow value for a segment */

    int		evolution;
    int		segment;

    {

	MsrsRecord	*pmsrs;
	double		flowRate;

	/* Get the list of segment measurements */

	getMsrsList (evolution, segment);
	pmsrs = (MsrsRecord *)lstFirst (&msrsList);

	flowRate = (pmsrs->stopVolume - pmsrs->startVolume) / (pmsrs->stopTime - pmsrs->startTime);

	dBASEupdateDouble (evolSegsDesc, segActFlowRateField, flowRate);
	updateDBASEfile (evolSegsDesc);

   }

/*----------*/
qtyCommon (evolution, segment, target)

    int		evolution;
    int		segment;
    double	target;

    {

	double		volumeNow, timeNow, quantityXfered, thisXfered, timeForXfer, flowRate;
	double		startTime;
	SegmentRecord	*pseg;
	MsrsRecord	*pmsrs;

	/* Get the list of segment measurements */

	getMsrsList (evolution, segment);
	pmsrs = (MsrsRecord *)lstFirst (&msrsList);

	/* Sample the volume channel and save the time */

	volumeNow = readChannelAsDouble (pmsrs->channel);
	timeNow = time (0);
	thisXfered = volumeNow - pmsrs->stopVolume;
	timeForXfer = timeNow - pmsrs->stopTime;

	/* The flow rate based upon the previous volume sample and time. */
 
	flowRate = thisXfered / timeForXfer;

	/* Put the final stop value and time in the measurements record to make tallying easier. */

	pmsrs->stopTime = timeNow;
	pmsrs->stopVolume = volumeNow;

	/* Save the start time to calculate the actual flow time */

	startTime = pmsrs->startTime;

	/* Tally the quantity transferred */

	quantityXfered = 0.0;
	for (; pmsrs; pmsrs = (MsrsRecord *)lstNext (pmsrs)) {
	    quantityXfered += (pmsrs->stopVolume - pmsrs->startVolume);
	    if (Verbose)
		printf ("channel (%s) stop:%f start:%f qty:%f\n",
				pmsrs->channel,pmsrs->stopVolume,pmsrs->startVolume,quantityXfered);
	}

	/* Update the database with the new values */

	getSegmentRecord(evolution, segment);
	dBASEupdateDouble (evolSegsDesc, segQuantityXferField, quantityXfered);
	timeForXfer = timeNow - startTime;
	dBASEupdateDouble (evolSegsDesc, segActFlowTimeField, timeForXfer);
	dBASEupdateDouble (evolSegsDesc, segActFlowRateField, flowRate);
	updateDBASEfile (evolSegsDesc);
        sprintf (errorTextBuffer, "negative flow rate for evolution %d segment %d", evolution, segment);
        assert ( ( (thisXfered / timeForXfer) >= -1.0), PUT_INTO_ERROR, errorTextBuffer);

	saveFinalVols (evolution, segment);

	if (Verbose)
	    printf ("qty:%f target:%f\n", quantityXfered, target);

	return (quantityXfered >= target);

    }

/*----------*/
qtyXfered (evolution, segment)

    int	evolution;
    int	segment;

    {

	SegmentRecord	*segmentRecord = getSegmentRecord (evolution, segment);
	LineupRecord	*lineupRecord = getLineupRecord (segmentRecord->lineup);
	double		correction = (lineupRecord->shutdownTime * segmentRecord->actFlowRate)/2.0;
	
	return qtyCommon (evolution, segment, (0.99 * segmentRecord->quantity) - correction);

    }

/*----------*/
swingTol (evolution, segment)

    int	evolution;
    int	segment;

    {

	SegmentRecord	*segmentRecord = getSegmentRecord (evolution, segment);
	LineupRecord	*lineupRecord = getLineupRecord (segmentRecord->lineup);
	double		correction = (getLineupSwingsRecord (evolution, segment)->swingTime *
				segmentRecord->actFlowRate)/2.0;

	return qtyCommon (evolution, segment, (0.99 * segmentRecord->quantity) - correction);

    }


typedef enum {
	ScriptError,
	ScriptOk,
	ScriptActive
} ScriptProgress;

/*----------*/
ScriptProgress scriptStatus (evolution, segment, scriptId)

    int		evolution;
    int		segment;
    ScriptId	scriptId;

    {

	ScriptProgress		progress;
	ExtendedScriptStatus	extendedScriptStatus;

	if (scriptId == SCRIPTID_UNKNOWN)
	    return ScriptError;

	switch (CSLMonitorGetScriptStatus (localhostname, scriptId, &extendedScriptStatus)) {
	    case Unknown:
		progress = ScriptError;
		break;
	    case Scheduled:
		progress = ScriptActive;
		break;
	    case Running:
		progress = ScriptActive;
		break;
	    case Killed:
		progress = ScriptError;
		break;
	    case Aborted:
		progress = ScriptError;
		break;
	    case Exited:
		progress = (extendedScriptStatus.completionCode == 0.0) ? ScriptOk : ScriptError;
		break;
	}

	if ( (progress != ScriptActive)) {
	    CSLMonitorRelease (localhostname, scriptId);
	    if (f4int (segEvolNumField) != evolution || f4int (segSegNumField) != segment)
			getSegmentRecord (evolution, segment);
	    if (f4int (segEvolNumField) == evolution && f4int (segSegNumField) == segment) {
		dBASEupdateInt (evolSegsDesc, segScriptIdField, SCRIPTID_UNKNOWN);	/* erase it */
		updateDBASEfile (evolSegsDesc);
	    }
	}
	return progress;

    }

/*----------*/
UpdateEvolutionSegment (e, s)

    int	e;
    int	s;

    {

	SegmentRecord	*pSeg = getSegmentRecord (e, s);
	int		o = pSeg->orderNumber;
	EvolutionProcessorRequestType	req = findRequestType (pSeg->requestType);
	ScriptProgress	scriptExitStatus = scriptStatus (e, s, pSeg->scriptId);
	EvolutionState	currentState = findStatus (pSeg->status);
	EvolutionState	nextState = BadState;

	environmentString = pSeg->environmentString;	/* set this up for any error notifications */
	if (Verbose)
	    printf ("e:%2d s:%2d o:%2d r:%6s es:%10s ss:%10s", e, s, pSeg->orderNumber,
			EvolutionProcessorRequestText [req],
			getEvolutionRecord (e)->status, pSeg->status), clearEOL ();
	switch (currentState) {
	    case Pending:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == SWINGFROM)
		    nextState = SwingFrom;
		else if (req == SWINGTO)
		    nextState = SwingTo;
		else if (req == START  &&  haveAllComponets (e, s))
		    nextState = Starting;
		else if (req == START  && !haveAllComponets (e, s))
		    nextState = OnHold;
		else if (req == MANUAL &&  haveAllComponets (e, s))
		    nextState = Manual;
		else if (req == MANUAL && !haveAllComponets (e, s))
		    nextState = MOnHold;
		break;

	    case OnHold:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopped;
		else if (req == STOP)
		    nextState = Stopping;
		else if (haveAllComponets (e, s))
		    nextState = Starting;
		qtyXfered (e, s);
		break;

	    case MOnHold:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopped;
		else if (haveAllComponets (e, s))
		    nextState = Manual;
		qtyXfered (e, s);
		break;

	    case Starting:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk)
		    nextState = Active;
		qtyXfered (e, s);
		break;

	    case Manual:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk) {
		    writeAveFlow (e, s);
		    nextState = MCompleted;
		}
		if (nextState != MCompleted)
		    qtyXfered (e, s);
		break;

	    case Stopped:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == START  &&  haveAllComponets (e, s))
		    nextState = Starting;
		else if (req == START  && !haveAllComponets (e, s))
		    nextState = OnHold;
		else if (req == MANUAL &&  haveAllComponets (e, s))
		    nextState = Manual;
		else if (req == MANUAL && !haveAllComponets (e, s))
		    nextState = MOnHold;
		break;

	    case EStopped:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopped;
		else if (req == START  &&  haveAllComponets (e, s))
		    nextState = Starting;
		else if (req == START  && !haveAllComponets (e, s))
		    nextState = OnHold;
		else if (req == MANUAL &&  haveAllComponets (e, s))
		    nextState = Manual;
		else if (req == MANUAL && !haveAllComponets (e, s))
		    nextState = MOnHold;
		break;

	    case Error:
		if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (req == SWINGTO)
		    nextState = SwingTo;
		else if (req == START  &&  haveAllComponets (e, s))
		    nextState = Starting;
		else if (req == START  && !haveAllComponets (e, s))
		    nextState = OnHold;
		else if (req == MANUAL &&  haveAllComponets (e, s))
		    nextState = Manual;
		else if (req == MANUAL && !haveAllComponets (e, s))
		    nextState = MOnHold;
		break;

	    case Active:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (req == PAUSE)
		    nextState = Pausing;
		else if (scriptExitStatus == ScriptError)
		    nextState = Stopping;
		else if (evolutionMode (e) && evolSwingEnabled (e) && nextSegmentNumber (e, o))
		    nextState = swingTol (e, s) ? SwingFrom : nextState;
		else if (qtyXfered (e, s))
		    nextState = evolSegIsPauseOnComplete (e, s) ? Pausing : Stopping;
		break;

	    case Pausing:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk)
		    nextState = Paused;
		qtyXfered (e, s);
		break;

	    case Paused:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (req == RESUME)
		    nextState = Resuming;
		qtyXfered (e, s);
		break;

	    case Resuming:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk)
		    nextState = Active;
		qtyXfered (e, s);
		break;

	    case Stopping:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk &&  qtyXfered (e, s)) {
		    writeAveFlow (e, s);
		    nextState = Completed;
		} else if (scriptExitStatus == ScriptOk && !qtyXfered (e, s))
		    nextState = Stopped;
		if (nextState != Completed)
		    qtyXfered (e, s);
		break;
		
	    case EStopping:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk)
		    nextState = EStopped;
		qtyXfered (e, s);
		break;			
		
	    case SwingFrom:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (scriptExitStatus == ScriptError)
		    nextState = Error;
		else if (scriptExitStatus == ScriptOk && swingTol (e, s)) {
		    writeAveFlow (e, s);
		    nextState = Completed;
		}
		if (nextState != Completed)
		    qtyXfered (e, s);
		break;

	    case SwingTo:
		if (req == GOTOERROR)
		    nextState = Error;
		else if (req == ESTOP)
		    nextState = EStopping;
		else if (req == STOP)
		    nextState = Stopping;
		else if (req == SWINGFROM)
		    nextState = Pending;
		else if (req == SWINGTO)
		    nextState = Active;
		qtyXfered (e, s);
		break;	
	}

	if (nextState != BadState)
	    changeSegmentToNewState (e, s, currentState, nextState);
	if (Verbose)
	    printf ("\n");

    }

/*----------*/
updateAPIs ()

    {

	EvolutionRecord	*pEvol;
	SegmentRecord	*pSeg;
	EvolutionState	currentState;
	double		weightedAvg, quantity;
	double		api, avgApi;
	int		n;
	char		name [CHANNEL_NAME_ENTRY_SIZE+1];

	d4tag_select (evolDesc, evolTag);
	if (d4top (evolDesc) == 0) {
	    do {
		pEvol = getEvolutionRecord (f4int (evolEvolNumberField));
		currentState = findStatus (pEvol->status);
		if (currentState != Completed && currentState != MCompleted && currentState != Error) {

		    /* update the evol,seg avgapi field */

		    d4tag_select (evolSegsDesc, evolSegsTag);
		    if (d4top (evolSegsDesc) == 0) {
			do {
			    if (f4int (segEvolNumField) == pEvol->evolNumber) {
				pSeg = getSegmentRecord (pEvol->evolNumber, f4int (segSegNumField));
				currentState = findStatus (pSeg->status);
				if (currentState != Completed && currentState != MCompleted && currentState != Error) {
				    if (evolIsAReceive (pEvol)) {
					avgApi = getSegSrcsRecord (pEvol->evolNumber, f4int (segSegNumField))->api;
				    } else { /* evol is an issue or a transfer */
					n = 0;
					avgApi = 0.0;
					d4tag_select (segSrcsDesc, segSrcsTag);
					if (d4top (segSrcsDesc) == 0) {
					    do {
						if (pEvol->evolNumber == f4int (segSrcsEvolNumberField) &&
								f4int (segSegNumField) == f4int (segSrcsSegNumberField)) {
													
						    f4ncpy (segSrcsSourceField, name, CHANNEL_NAME_ENTRY_SIZE);
						    api = readChannelAsDouble (buildChannelName (name, ".SGRV"));
						    avgApi = avgApi + api;
						    dBASEupdateDouble (segSrcsDesc, segSrcsApiField, api);
						    updateDBASEfile (segSrcsDesc);
						    n++;
						}
						d4unlock_all (segSrcsDesc);
					    } while (d4skip (segSrcsDesc, 1) == 0);
					}
					d4unlock_all (segSrcsDesc);
					if (n)
					    avgApi /= (double)n;
				    }
				    dBASEupdateDouble (evolSegsDesc, segAvgApiField, avgApi);
				    updateDBASEfile (evolSegsDesc);
				}
			    }
			    d4unlock_all (evolSegsDesc);
			} while (d4skip (evolSegsDesc, 1) == 0);
		    }
		    d4unlock_all (evolSegsDesc);
				
		    /* update the evol weightedavg field */
		    weightedAvg = 0.0;
		    quantity = 0.0;
		    d4tag_select (evolSegsDesc, evolSegsTag);
		    if (d4top (evolSegsDesc) == 0) {
			do {
			    if (f4int (segEvolNumField) == pEvol->evolNumber) {
				pSeg = getSegmentRecord (pEvol->evolNumber, f4int (segSegNumField));
				if (pEvol->quantity)
				    weightedAvg += (pSeg->quantity * pSeg->avgApi);
				    quantity += pSeg->quantity;
			    }
			    d4unlock_all (evolSegsDesc);
			} while (d4skip (evolSegsDesc, 1) == 0);
			weightedAvg /= quantity;
		    }
		    d4unlock_all (evolSegsDesc);
		    dBASEupdateDouble (evolDesc, evolWeightedAvgField, weightedAvg);
		    updateDBASEfile (evolDesc);
		}
		d4unlock_all (evolDesc);
	    } while (d4skip (evolDesc, 1) == 0);
	}
	d4unlock_all (evolDesc);

    }

/*----------------------------------------- Main functions -----------------------------------*/

#include <sys/wait.h>

/*----------*/
void Exit (code)

    {

	if (Debug)
	    abort ();
	EVOLManagerDetach ();
	closeAllChannels ();
	d4close_all (&codeBaseDesc);
	
	exit (code);

    }

/*----------*/
ImMaster ()

    {

	int		State;

	/* If no arbiter then act as master */

	if (!Redundancy)		/* Act as a master if no redundancy */ 
	    return 1;

	/* If arbiter has dissappeared, terminate program.  Otherwise return current status. */

	if (!appGetStatus (arbiterHandle, &State)) {
	    printf ("EVOLProcessor: Lost contact with arbiter -- program terminated!");
	    Exit (-1);
	}

	if (Verbose) {
	    if (State != prevState) {
		switch (State) {
		    case SLAVE:
			printf ("EVOLProcessor: I am slave.\n");
			break;
		    case MASTER:
			printf ("EVOLProcessor: I am master.\n");
			break;
		    case MASTER_NOPARTNER:
			printf ("EVOLProcessor: I am master without a partner.\n");
			break;
		}
		prevState = State;
	    }
	}

	return State != SLAVE;
	}
	
/*----------*/
static void RegisterWithArbiter ()

    {

	char	Dummy   [2];

	if (!Redundancy)
	    return;

	/* Keep trying to register with arbiter until successful. */

	for (;;) {
	    int Status;
	    if (appRegister ("EVOLProcessor", &arbiterHandle)) {
		appGetStatus (arbiterHandle, &Status);
		switch (Status) {
		    case MASTER:
			if (Verbose)
			    printf ("EVOLProcessor: Running as master.\n");
			break;
		    case SLAVE:
			if (Verbose)
			    printf ("EVOLProcessor: Running as slave.\n");
			break;
		    case MASTER_NOPARTNER:
			if (Verbose)
			    printf ("EVOLProcessor: Running as master with no partner.\n");
			break;
		}
		return;
	    } else {
		if (Verbose)
		    printf ("EVOLProcessor: Unable to connect with arbiter -- retrying.\n");
		sleep (5);
	    }
	}

    }

/*----------*/
void killHandler (sig)	/* Terminate things in an orderly manner */

    int		sig;

    {

	if (Verbose)
	    printf ("EVOLProcessor: TERM received -- terminating\n");

	d4close_all (&codeBaseDesc);
	if (Redundancy)
	    appUnRegister (arbiterHandle);
	red_Terminate ();
	sleep (10);
	exit (0);

    }

/*----------*/
main (argc, argv)

    int		argc;
    char	*argv [];

    {

	int	Cmd, Status, initialized = 0;
	char	dummy [2];

	signal (SIGTERM, killHandler);	/* Intercept kill messages */
	signal (SIGPIPE, killHandler);	/* Intercept broken pipe messages */

	Redundancy = No;
	Verbose = No;
	Debug = No;
	Trace = No;
	while ( (Cmd = getopt (argc,argv,"rvdt")) != -1) {
	    switch (Cmd) {
		case 'r':
		    Redundancy = Yes;
		    break;
		case 'v':
		    Verbose = Yes;
		    break;
		case 'd':
		    Debug = Yes;
		    Verbose = Yes;
		    break;
		case 't':
		    Trace = Yes;
		    break;
			
	    }
	}

	if (Verbose)
	    printf ("EVOLProcessor: Running.\n");

	/* Do some global variable initialization */

	completionLogicPath = getenv ("EVOL_COMPLETION_LOGIC");
	errorLogicPath = getenv ("EVOL_ERROR_LOGIC");

	/* Register with the arbiter (if necessary). */

	RegisterWithArbiter ();

	prevState = -1;		/* Force initial display of role */

	/* Initialize all channel access stuff */

	ca_task_initialize ();
	lstInit (&channelList);

	EVOLManagerAttach (localhostname);
	initTermcap ();
	clearScreen ();
	signal (SIGCHLD, childHandler);	/* instate a SIGCHLD catcher to wait3 () for dying children */
	
	while (1) {

	    if (ImMaster ()) {

		GetNextRequestResult *preq;
		EvolSegRecord *pseg = NULL;
		int errorSeen = FALSE;

		if (!initialized) {

		    /* I am now a master.  Delete the servers.enabled flag file so that the slave */
		    /* cannot created servers on the master. */

		    remove ("/tis4000db/servers.enabled");		

		    /* Open up the dBase files */

		    dBASEopen ();
		    openTFdBASEfiles ();

		    initialized = 1;

		}

		/* set up the setjmp ()/longjmp () environment for the !FATAL assert ()s */
		if (setjmp (error_env) == PUT_INTO_ERROR) {
		    errorSeen = TRUE;
		    if (pseg) {
			/* we got an error on (pseg->evolNumber, pseg->segNumber), */
			/* put the evolution into Error */
			int	evolution = pseg->evolNumber;
			foreachEvolSegRecord (pseg)
			    if ( pseg->evolNumber == evolution )
				addRequest (pseg->evolNumber, pseg->segNumber, GOTOERROR, NULL);
		    }
		}

		/* Update the APIs get all of the requests that the EVOLManager has queued up */

		if (!errorSeen)
		    updateAPIs ();
		while ( (preq = EVOLManagerGetNextRequest ())->status == TF_OK)
		    addRequest (preq->evolution, preq->segment, preq->type,
				preq->environmentString);

		/* Flush the ca requests to check for channel access i/o and */
		/* update the segments in the list */

		homeCursor ();
		ca_flush_io ();
		getEvolSegList ();
		foreachEvolSegRecord (pseg)
		    UpdateEvolutionSegment (pseg->evolNumber, pseg->segNumber);
		clearEOS ();
		updateDBASEfile (evolDesc);
		updateDBASEfile (evolSegsDesc);
		updateDBASEfile (segSrcsDesc);
		updateDBASEfile (msrsDesc);
		updateDBASEfile (xrefDesc);
		updateDBASEfile (lineupDesc);
		updateDBASEfile (lineupSwingsDesc);
		updateDBASEfile (lineupCmpntsDesc);
		updateDBASEfile (lineupSrcsDesc);
		updateDBASEfile (lineupDestsDesc);
		updateDBASEfile (lineupMsrmntVolsDesc);
		updateDBASEfile (lineupXrefVolsDesc);
		sleep (REFRESH_RATE);

	    } else {

		char	*buf;

		if (initialized) {	/* We just got busted back to slave */

		    CloseAllFiles ();

		    initialized = 0;

		}

		/* We do a wait for message to force any status change to be processed */

		appRecvData (arbiterHandle, buf, CHECK_FOR_CHANGE_TO_MASTER_RATE);

	    }
	    
	}

    }

/*----------*/
updateDBASEfile (data)

    D4DATA	*data;

    {

	d4flush_all (data);
#ifndef CB51
	data->file_lock = TRUE;		/* to force d4update_header () to write the num_recs field */
	d4update_header (data, 1, 1);
	data->file_lock = FALSE;	/* to force d4unlock_all () to write the num_recs field */
#endif
	d4unlock_all (data);

    }

