#include <vport.h>
#include vdebugHEADER
#include vexHEADER
#include vobjectHEADER
#include vresourceHEADER
#include vloadableHEADER
#include veventHEADER
#include vtextHEADER
#include vtextviewHEADER
#include vtextitemHEADER

#include "textFilters.hxx"

extern "C"
{
#include "dbDefs.h"
}

static vchar *hhmmTimeComplete    = (vchar *)"00:00";
static vchar *hhmmssTimeComplete  = (vchar *)"00:00:00";
static vchar *hhmmsstTimeComplete = (vchar *)"00:00:00.0";

// predefs of some functions
static int timeFilter (filteredText *text, vtextSelection *sel,
		       const vchar **string, long *len, int timeMask);
static int hhmmFilter (filteredText *text, vtextSelection *sel,
		       const vchar **string, long *len);
static int hhmmssFilter (filteredText *text, vtextSelection *sel,
			 const vchar **string, long *len);
static int hhmmsstFilter (filteredText *text, vtextSelection *sel,
			  const vchar **string, long *len);


//*** FILTERED TEXT:

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(filteredText, vtextitemText, "filteredText");

/*****************************************
*
*    filteredText constructors
*	init the filter function to NULL
*
*****************************************/

filteredText::filteredText(vtextitemText &from)
: vtextitemText(vsuppressDONT_INIT)
{
    vtextitemText::CopyInit(&from);
    filterfunc = NULL;
    modified = NULL;
    userPtr = NULL;
    localModified = vFALSE;
}

void filteredText::Init()
{
    vtextitemText::Init();
    filterfunc = NULL;
    modified = NULL;
    userPtr = NULL;
    localModified = vFALSE;
}


void filteredText::LoadInit(vresource resource)
{
    vtextitemText::LoadInit(resource);
    filterfunc = NULL;
    modified = NULL;
    userPtr = NULL;
    localModified = vFALSE;
}

void filteredText::CopyInit(vloadable *orig)
{
    vtextitemText::CopyInit(orig);
    filterfunc = ((filteredText*)orig)->GetFilterFunc();
    modified = ((filteredText *)orig) -> GetModifiedPtr ();
    userPtr = ((filteredText *)orig) -> GetUserPtr ();
    localModified = ((filteredText *)orig) -> GetLocalModifiedFlag ();
}


void filteredText::Copy(vloadable *orig)
{
    vtextitemText::Copy(orig);
    filterfunc = ((filteredText*)orig)->GetFilterFunc();
    modified = ((filteredText *)orig) -> GetModifiedPtr ();
    userPtr = ((filteredText *)orig) -> GetUserPtr ();
    localModified = ((filteredText *)orig) -> GetLocalModifiedFlag ();
}


/*****************************************
*
*    filteredText destructor
*	reset the filter function to NULL
*
*****************************************/

void filteredText::Destroy(void)
{
    filterfunc = NULL;
    modified = NULL;
    userPtr = NULL;
    localModified = vFALSE;
    vtextitemText::Destroy();
}

/****************************************************
*
*    filteredText insert
*	pass the text through the filter function,
*	if there is a filter and if there is text
*	before giving it to the Text manager, if ok	
*
****************************************************/

void filteredText::InsertText(vtextSelection *selection, const vchar *s, long len)
{
    long nuLen = len;

    if ( (len == 0) ||
	 (len > 1)  ||
	 (filterfunc == NULL) || 
	 (filterfunc(this, selection, &s, &nuLen)) ) 
    {
	vtextitemText::InsertText(selection, s, nuLen);
    }
}


/****************************************************
*
*    filteredText ObserveDelete
*	if modified flag != NULL, set it to vTRUE
*	and call the base class func
*
****************************************************/

void filteredText::ObserveDelete(long pos, long delta)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
    SetLocalModifiedFlag (vTRUE);
    vtext::ObserveDelete (pos, delta);
}

/****************************************************
*
*    filteredText ObserveInsert
*	if modified flag != NULL, set it to vTRUE
*	and call the base class func
*
****************************************************/

void filteredText::ObserveInsert(long pos, long delta)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
    SetLocalModifiedFlag (vTRUE);
    vtext::ObserveInsert (pos, delta);
}


// *** FILTERED TEXTITEMS:

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(filteredTextitem, vtextitem, "filteredTextitem");

/*************************************************
*
*    filteredTextitem
*	make a filteredText object, not a generic
*	vtext object
*
*************************************************/

vtext *filteredTextitem::CreateTextData()
{
    return new filteredText;
}

/*******************************************************************
*
*    filteredTextitem
*	associate an extern filter function with this filtered text.
*	NOTE: the function can be NULL, in which case no filtering
*	      is performed
*
*******************************************************************/

void filteredTextitem::SetItemFilterFunc(filterFunc filter)
{
    filteredText  *text;

    text = (filteredText *) GetTextData();

    if (!text || !text->IsKindOf(&filteredText::Kind))
	vexGenerate(vexGetValueClass(), vexMESSAGE, "invalid text object",
		    vexERRNO, 0, vexNULL);
    else
	text->SetFilterFunc(filter);
}


/*************************************************
*
*    filteredTextitem
*	make sure text item is a filterable
*	text objext
*
*************************************************/

void filteredTextitem::Init()
{
    filteredText	*text;

    vtextitem::Init();

    text = (filteredText *) GetTextData();
    if (!text || !text->IsKindOf(&filteredText::Kind))
	vexGenerate(vexGetValueClass(), vexMESSAGE, "invalid text object",
		    vexERRNO, 0, vexNULL);
}


/******************************************************************************
*
* filteredTextitem
*
* Since our superclass (and the Galaxy Visual Resource Builder) can only
* store generic vtextitemTexts with our persistent form, we must explicitly
* zap the old (vtextitemText) text data and replace it w/ one of our own.
* Memory based object are properly filtered through the CreateTextData member
* function w/ produced the correct text object.
******************************************************************************/

void filteredTextitem::LoadInit(vresource resource)
{
    vtextitemText	*oldText;
    filteredText	*newText;

    vtextitem::LoadInit(resource);

    oldText = (vtextitemText *) GetTextData();
    newText = new filteredText(*oldText);

    SetTextData(newText);
    SetSelection(newText->CreateSelection());

    delete oldText;
}


// *** TIME SUB-CLASSES OF FILTERED TEXTITEMS:

/*************************************************
*
*    hhmmTimeClass
*	override of ObserveView to capture loss
*	of focus
*
*************************************************/

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(hhmmTimeClass, filteredTextitem, "hhmmTimeClass");

void hhmmTimeClass::ObserveView (vevent *event, int message)
{
    if (message == vtextviewACCEPT_NOTIFY)
    {
	vdialogItemList	*list = this -> GetItemListIn ();
	vobjectAttributed *obj = list -> GetFocusItem ();
	vdialog *dialog = list -> DetermineDialog ();
	vdialogItem *item = dialog -> GetFocusItem ();
	if ( ((vobjectAttributed *)item) != obj)
	{
	    vchar textTime[HHMMTIME_SZ + 2];
	    GetText (textTime, HHMMTIME_SZ + 1);
	    int textLen = vcharLength (textTime);
	    if ( (textLen > 0) && (textLen < HHMMTIME_SZ) )
	    {
		vcharAppend (textTime, &hhmmTimeComplete [textLen]);
		textTime[HHMMTIME_SZ] = 0;
		SetText (textTime);
	    }
	}
    }
}

/*************************************************
*
*    hhmmssTimeClass
*	override of ObserveView to capture loss
*	of focus
*
*************************************************/

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(hhmmssTimeClass, filteredTextitem, "hhmmssTimeClass");

void hhmmssTimeClass::ObserveView (vevent *event, int message)
{
    if (message == vtextviewACCEPT_NOTIFY)
    {
	vdialogItemList	*list = this -> GetItemListIn ();
	vobjectAttributed *obj = list -> GetFocusItem ();
	vdialog *dialog = list -> DetermineDialog ();
	vdialogItem *item = dialog -> GetFocusItem ();
	if ( ((vobjectAttributed *)item) != obj)
	{
	    vchar textTime[HHMMSSTIME_SZ + 2];
	    GetText (textTime, HHMMSSTIME_SZ + 1);
	    int textLen = vcharLength (textTime);
	    if ( (textLen > 0) && (textLen < HHMMSSTIME_SZ) )
	    {
		vcharAppend (textTime, &hhmmssTimeComplete [textLen]);
		textTime[HHMMSSTIME_SZ] = 0;
		SetText (textTime);
	    }
	}
    }
}

/*************************************************
*
*    hhmmsstTimeClass
*	override of ObserveView to capture loss
*	of focus
*
*************************************************/

// macro of vkindtyped overrides
vkindTYPED_FULL_DEFINITION(hhmmsstTimeClass, filteredTextitem, "hhmmsstTimeClass");

void hhmmsstTimeClass::ObserveView (vevent *event, int message)
{
    if (message == vtextviewACCEPT_NOTIFY)
    {
	vdialogItemList	*list = this -> GetItemListIn ();
	vobjectAttributed *obj = list -> GetFocusItem ();
	vdialog *dialog = list -> DetermineDialog ();
	vdialogItem *item = dialog -> GetFocusItem ();
	if ( ((vobjectAttributed *)item) != obj)
	{
	    vchar textTime[HHMMSSTTIME_SZ + 2];
	    GetText (textTime, HHMMSSTTIME_SZ + 1);
	    int textLen = vcharLength (textTime);
	    if ( (textLen > 0) && (textLen < HHMMSSTTIME_SZ) )
	    {
		vcharAppend (textTime, &hhmmsstTimeComplete [textLen]);
		textTime[HHMMSSTTIME_SZ] = 0;
		SetText (textTime);
	    }
	}
    }
}


/**************************************************************
*
*    Filter Functions
*
**************************************************************/

/****************************************************************
*
*    Different character filters, all call NCharFilter
*
*****************************************************************/

int alarmGroupFilter (filteredText *text, vtextSelection *sel,
		      const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, ALARMGRP_SZ) );
}

int calcExpressionFilter (filteredText *text, vtextSelection *sel,
			  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, CALCEXPR_SZ) );
}

int descriptorFilter (filteredText *text, vtextSelection *sel,
		      const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, DESCRIPTOR_SZ) );
}

int digitalStateFilter (filteredText *text, vtextSelection *sel,
		        const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, DIGITALSTATE_SZ) );
}

int egutagFilter (filteredText *text, vtextSelection *sel,
		  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, EGUTAG_SZ) );
}

int iocnameFilter (filteredText *text, vtextSelection *sel,
		  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, IOCNAME_SZ) );
}

int locationFilter (filteredText *text, vtextSelection *sel,
		    const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, LOCATION_SZ) );
}

int multiDigitalStateFilter (filteredText *text, vtextSelection *sel,
			     const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, MULTIDIGITALSTATE_SZ) );
}

int operatorNameFilter (filteredText *text, vtextSelection *sel,
			const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, OPERATOR_SZ) );
}

int protectionGroupFilter (filteredText *text, vtextSelection *sel,
			   const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, PROTECTIONGRP_SZ) );
}

int pvnameFilter (filteredText *text, vtextSelection *sel,
		  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, PVNAME_SZ) );
}

int resourceNameFilter (filteredText *text, vtextSelection *sel,
		        const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, RESOURCE_SZ) );
}

int stateValFilter (filteredText *text, vtextSelection *sel,
		    const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, STATEVAL_SZ) );
}

int stringValFilter (filteredText *text, vtextSelection *sel,
		    const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, STRNGVAL_SZ) );
}

int tankfarmCmdStsFilter (filteredText *text, vtextSelection *sel,
			  const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, TFCMDSTS_SZ) );
}

int udbStringsFilter (filteredText *text, vtextSelection *sel,
		      const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, UDBSTRINGS_SZ) );
}



/****************************************************************
*
*    NCharFilter
*
*    Allow only N characters in a text object, where N is passed
*    in the "n" parameter. Called by various character counting
*    functions, since the filter is typedef'd in textFilters.hxx
*    we can not pass another argument
*
*
*****************************************************************/

int NCharFilter (filteredText *text, vtextSelection *sel,
	 	 const vchar **string, long *len, int n)
{
    long selectionLen, textlen;
    int ret;

    // we need to do selection length calculations because the 
    // new text is added _before_ the old selection's text is 
    // nuked.

    selectionLen = text->GetSelectionEndPosition(sel) - 
		   text->GetSelectionStartPosition(sel);

    textlen = text -> GetLength ();

    *len = vMIN(n - (textlen - selectionLen), *len);

    if (*len > 0) 
    {
	ret = vTRUE;
    } 
    else 
    {
	vwindow::Beep ();
	ret = vFALSE;
    }
    return(ret);
}

/*****************************************************************
* 
*    integerFilter
*	Allow only digits, and minus sign in first position
*
*****************************************************************/

int integerFilter (filteredText *text, vtextSelection *sel,
		   const vchar **s, long *len)
{
    int		ret = vFALSE;
    vchar	minusSign;

    minusSign = vcharFromLiteral ('-');

    if (*len > 0) 
    {
	*len = 1;
	if (vcharIsDigit (**s)) 
	{
	    ret = vTRUE;
	} 
	else if (**s == minusSign)
	{
	    long len = text->GetLength ();
	    ret = vTRUE;
	    // make sure the minus sign is the first thing
	    if (len != 0)
	    {
		vwindow::Beep();
		ret = vFALSE;
	    }
	
	}
	else 
	{
	    vwindow::Beep();
	    ret = vFALSE;
	}
    }
    else 
    {
	ret = vFALSE;
    }
    return (ret);
}


/*****************************************************************
* 
*    floatFilter
*	Allow only digits, decimal points (1), and minus sign
*	in the first position
*****************************************************************/

int floatFilter (filteredText *text, vtextSelection *sel,
	 	 const vchar **s, long *len)
{
    int		ret = vFALSE;
    vchar	decimalPoint;
    vchar	minusSign;

    decimalPoint = vcharFromLiteral ('.');
    minusSign = vcharFromLiteral ('-');

    if (*len > 0) 
    {
	*len = 1;
	if (vcharIsDigit (**s)) 
	{
	    ret = vTRUE;
	} 
	else if (**s == decimalPoint)
	{
	    long len = text->GetLength (), i;
	    ret = vTRUE;
	    // make sure there's only one decimal point
	    for (i = 0; i < len; i++)
	    {
		if (*text->GetTextAt (i) == decimalPoint) 
		{
		    ret = vFALSE;
		    vwindow::Beep();
		    break;
		}
	    }
	}
	else if (**s == minusSign)
	{
	    long len = text->GetLength ();
	    ret = vTRUE;
	    // make sure the minus sign is the first thing
	    if (len != 0)
	    {
		vwindow::Beep();
		ret = vFALSE;
	    }
	
	}
	else 
	{
	    vwindow::Beep();
	    ret = vFALSE;
	}
    }
    else 
    {
	ret = vFALSE;
    }
    return (ret);
}


/*************************************************************
*
*    numeric filter, positive number only
*
*************************************************************/

int numericFilter (filteredText *text, vtextSelection *sel,
		   const vchar **s, long *len)
{
    int ret = vFALSE;

    if (*len > 0)
    {
	*len = 1;
	if (vcharIsDigit (**s))
	{
	    ret = vTRUE;
	}
	else
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
    }
    else
    {
	ret = vFALSE;
    }
    return (ret);
}



/*****************************************************************
* 
*    protectionLevelFilter
*	Allow only 2 digits
*
*****************************************************************/

int protectionLevelFilter (filteredText *text, vtextSelection *sel,
			   const vchar **string, long *len)
{
    int ret = vFALSE;

    if (*len > 0)
    {
	*len = 1;
	if (vcharIsDigit (**string))
	{
	    ret = NCharFilter (text, sel, string, len, PRIV_LEVEL_SZ);
	}
	else
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
    }
    else
    {
	ret = vFALSE;
    }
    return (ret);
}


/*****************************************************************
* 
*    timeFilters
*
*****************************************************************/

int hhmmTimeFilter (filteredText *text, vtextSelection *sel,
		    const vchar **string, long *len)
{
    if ( (timeFilter (text, sel, string, len, HHMMTIME_SZ) ) == vTRUE)
    {
	return (hhmmFilter (text, sel, string, len) );
    }
    else
    {
	return (vFALSE);
    }
}

int hhmmssTimeFilter (filteredText *text, vtextSelection *sel,
		      const vchar **string, long *len)
{
    if ( (timeFilter (text, sel, string, len, HHMMSSTIME_SZ) ) == vTRUE)
    {
	return (hhmmssFilter (text, sel, string, len) );
    }
    else
    {
	return (vFALSE);
    }
}

int hhmmsstTimeFilter (filteredText *text, vtextSelection *sel,
		       const vchar **string, long *len)
{
    if ( (timeFilter (text, sel, string, len, HHMMSSTTIME_SZ) ) == vTRUE)
    {
	return (hhmmsstFilter (text, sel, string, len) );
    }
    else
    {
	return (vFALSE);
    }
}


// *** generic timeFilter, to check the basics
// *** NOTE: textPos -> start
// ***		starts from position 0

static int timeFilter (filteredText *text, vtextSelection *sel,
		       const vchar **string, long *len, int maskLen)
{
    int ret = vFALSE;

    if (*len > 0)
    {
	vtextitem *textitem = (vtextitem *)text -> GetUserPtr ();
	vstr *textTime = textitem -> GetTextAsString ();
	int textLen = vcharLength (textTime);
	vtextPosition *textPos = text -> GetSelectionPosition (sel);
	if (*len > 1)
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
	else if ( (vcharIsDigit (**string)) == vFALSE)
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
	// text is same length as mask, AND nothing has been selected
	// trying to pull a fast one and insert 
	else if ( ((textLen + 1) > maskLen) && (textPos -> length == 0) )
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
	// do not allow any internal insertions
	else if ( (textLen > 0) && (textPos -> start < textLen) )
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
	else
	{
	    ret = vTRUE;
	}
	vstrDestroy (textTime);
    }
    else
    {
	ret = vFALSE;
    }

    return (ret);
}


// *** hh:mm timeFilter
// *** max values are 23:59

static int hhmmFilter (filteredText *text, vtextSelection *sel,
		       const vchar **string, long *len)
{
    int ret = vTRUE;
    vtextPosition *textPos = text -> GetSelectionPosition (sel);
    vchar aNine = vcharFromLiteral ('9');
    vchar aZero = vcharFromLiteral ('0');

    if (textPos -> start == 0)
    {
	vchar aTwo  = vcharFromLiteral ('2');
	if ( ((**string) < aZero) ||
	     ((**string) > aTwo) )
	    ret = vFALSE;
    }
    else if (textPos -> start == 1)
    {
	vchar *prev = text -> GetTextAt (0);
	vchar aOne = vcharFromLiteral ('1');
	vchar aThree = vcharFromLiteral ('3');
	if ( ((*prev) == aZero) || ((*prev) == aOne) )
	{
	    if ( ((**string) < aZero) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	}
	else
	{
	    if ( ((**string) < aZero) ||
		 ((**string) > aThree) )
		ret = vFALSE;
	}
	if (ret == vTRUE)
	{
	    vchar *newString = ((vchar *)(*string));
	    *len = 2;
	    vcharCopyScribed (vcharScribeFormatLiteral ("%c:", **string),
			      newString);
	}
    }
    else if (textPos -> start == 3)
    {
	vchar aFive = vcharFromLiteral ('5');
	if ( ((**string) < aZero) ||
	     ((**string) > aFive) )
	    ret = vFALSE;
    }
    else if (textPos -> start == 4)
    {
	vchar aNine = vcharFromLiteral ('9');
	if ( ((**string) < aZero) ||
	     ((**string) > aNine) )
	    ret = vFALSE;
    }
    else
    {
	return (vFALSE);
    }

    if (ret == vFALSE)
    {
	vwindow::Beep ();
    }

    return (ret);
}


// *** hh:mm:ss timeFilter
// *** max values are 23:59:59

static int hhmmssFilter (filteredText *text, vtextSelection *sel,
			 const vchar **string, long *len)
{
    int ret = vTRUE;
    vtextPosition *textPos = text -> GetSelectionPosition (sel);
    vchar aNine = vcharFromLiteral ('9');
    vchar aZero = vcharFromLiteral ('0');

    if (textPos -> start == 0)
    {
	vchar aTwo  = vcharFromLiteral ('2');
	if ( ((**string) < aZero) ||
	     ((**string) > aTwo) )
	    ret = vFALSE;
    }
    else if (textPos -> start == 1)
    {
	vchar *prev = text -> GetTextAt (0);
	vchar aOne = vcharFromLiteral ('1');
	vchar aThree = vcharFromLiteral ('3');
	if ( ((*prev) == aZero) || ((*prev) == aOne) )
	{
	    if ( ((**string) < aZero) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	}
	else
	{
	    if ( ((**string) < aZero) ||
		 ((**string) > aThree) )
		ret = vFALSE;
	}
	if (ret == vTRUE)
	{
	    vchar *newString = ((vchar *)(*string));
	    *len = 2;
	    vcharCopyScribed (vcharScribeFormatLiteral ("%c:", **string),
			      newString);
	}
    }
    else if ( (textPos -> start == 3) || (textPos -> start == 6) )
    {
	vchar aFive = vcharFromLiteral ('5');
	if ( ((**string) < aZero) ||
	     ((**string) > aFive) )
	    ret = vFALSE;
    }
    else if ( (textPos -> start == 4) || (textPos -> start == 7) )
    {
	vchar aNine = vcharFromLiteral ('9');
	if ( ((**string) < aZero) ||
	     ((**string) > aNine) )
	    ret = vFALSE;
	else if (textPos -> start == 4)
	{
	    vchar *newString = ((vchar *)(*string));
	    *len = 2;
	    vcharCopyScribed (vcharScribeFormatLiteral ("%c:", **string),
			      newString);
	}
    }
    else
    {
	return (vFALSE);
    }

    if (ret == vFALSE)
    {
	vwindow::Beep ();
    }

    return (ret);
}


// *** hh:mm:ss.t timeFilter
// *** max values are 23:59:59.9

static int hhmmsstFilter (filteredText *text, vtextSelection *sel,
			  const vchar **string, long *len)
{
    int ret = vTRUE;
    vtextPosition *textPos = text -> GetSelectionPosition (sel);
    vchar aNine = vcharFromLiteral ('9');
    vchar aZero = vcharFromLiteral ('0');

    if (textPos -> start == 0)
    {
	vchar aTwo  = vcharFromLiteral ('2');
	if ( ((**string) < aZero) ||
	     ((**string) > aTwo) )
	    ret = vFALSE;
    }
    else if (textPos -> start == 1)
    {
	vchar *prev = text -> GetTextAt (0);
	vchar aOne = vcharFromLiteral ('1');
	vchar aThree = vcharFromLiteral ('3');
	if ( ((*prev) == aZero) || ((*prev) == aOne) )
	{
	    if ( ((**string) < aZero) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	}
	else
	{
	    if ( ((**string) < aZero) ||
		 ((**string) > aThree) )
		ret = vFALSE;
	}
	if (ret == vTRUE)
	{
	    vchar *newString = ((vchar *)(*string));
	    *len = 2;
	    vcharCopyScribed (vcharScribeFormatLiteral ("%c:", **string),
			      newString);
	}
    }
    else if  ( (textPos -> start == 3) || (textPos -> start == 6) )
    {
	vchar aFive = vcharFromLiteral ('5');
	if ( ((**string) < aZero) ||
	     ((**string) > aFive) )
	    ret = vFALSE;
    }
    else if  ( (textPos -> start == 4) || 
	       (textPos -> start == 7) ||
	       (textPos -> start == 9) )
    {
	if ( ((**string) < aZero) ||
	     ((**string) > aNine) )
	    ret = vFALSE;
	else if (textPos -> start == 4)
	{
	    vchar *newString = ((vchar *)(*string));
	    *len = 2;
	    vcharCopyScribed (vcharScribeFormatLiteral ("%c:", **string),
			      newString);
	}
	else if (textPos -> start == 7)
	{
	    vchar *newString = ((vchar *)(*string));
	    *len = 2;
	    vcharCopyScribed (vcharScribeFormatLiteral ("%c.", **string),
			      newString);
	}
    }
    else
    {
	return (vFALSE);
    }

    if (ret == vFALSE)
    {
	vwindow::Beep ();
    }

    return (ret);
}


// *** date filter
// *** MM/DD/CCYY	This function is Y2K complient
// *** 0123456789
// *** textPos 0 is the 1st character

int dateFilter (filteredText *text, vtextSelection *sel,
		    const vchar **string, long *len)
{
    int ret = vTRUE;
    vchar *prev;
    long textPos = text -> GetSelectionStart (sel);
    vchar aZero  = vcharFromLiteral ('0');
    vchar aOne   = vcharFromLiteral ('1');
    vchar aTwo   = vcharFromLiteral ('2');
    vchar aThree = vcharFromLiteral ('3');
    vchar aNine  = vcharFromLiteral ('9');

    switch (textPos)
    {
	case 0:
	    if ( ((**string) < aZero) ||
		 ((**string) > aOne) )
		ret = vFALSE;
	    break;
	case 1:
	    prev = text -> GetTextAt (0);
	    if ((*prev) == aZero)
	    {
		if ( ((**string) < aOne) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aTwo) )
		    ret = vFALSE;
	    }
		if (ret == vTRUE)
		{
			vchar *newString = ((vchar *)(*string));
			*len = 2;
			vcharCopyScribed (vcharScribeFormatLiteral ("%c/", **string),newString);
		}
	    break;
	case 2:
	case 5:
	{
	    vchar aSlash = vcharFromLiteral ('/');
	    if ((**string) != aSlash)
		ret = vFALSE;
	    break;
	}
	case 3:
	    if ( ((**string) < aZero) ||
		 ((**string) > aThree) )
		ret = vFALSE;
	    break;
	case 4:
	    prev = text -> GetTextAt (3);
	    if ((*prev) == aZero)
	    {
		if ( ((**string) < aOne) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else if ( ((*prev) == aOne) || ((*prev) == aTwo) )
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else if ((*prev) == aThree)
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aOne) )
		    ret = vFALSE;
	    }
	    else
		{
			ret = vFALSE;
		}
		if (ret == vTRUE)
		{
			vchar *newString = ((vchar *)(*string));
			*len = 2;
			vcharCopyScribed (vcharScribeFormatLiteral ("%c/", **string),newString);
		}
	    break;
	case 6:
	    if ( ((**string) < aOne) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	    break;
	case 7:
	case 8:
	case 9:
	    if ( ((**string) < aZero) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	    break;
	default:
	    ret = vFALSE;
	    break;

    } // end switch textPos

    if (ret == vFALSE)
    {
	vwindow::Beep ();
    }

    return (ret);
}


// *** date/time filter
// *** MM/DD/YY HH:MM:SS
// *** textPos 0 is the 1st character

int datetimeFilter (filteredText *text, vtextSelection *sel,
		    const vchar **string, long *len)
{
    int ret = vTRUE;
    vchar *prev;
    long textPos = text -> GetSelectionStart (sel);
    vchar aZero  = vcharFromLiteral ('0');
    vchar aOne   = vcharFromLiteral ('1');
    vchar aTwo   = vcharFromLiteral ('2');
    vchar aThree = vcharFromLiteral ('3');
    vchar aNine  = vcharFromLiteral ('9');

    switch (textPos)
    {
	case 0:
	    if ( ((**string) < aZero) ||
		 ((**string) > aOne) )
		ret = vFALSE;
	    break;

	case 1:
	    prev = text -> GetTextAt (0);
	    if ((*prev) == aZero)
	    {
		if ( ((**string) < aOne) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aTwo) )
		    ret = vFALSE;
	    }
	    break;

	case 2:
	case 5:
	{
	    vchar aSlash = vcharFromLiteral ('/');
	    if ((**string) != aSlash)
		ret = vFALSE;
	    break;
	}

	case 3:
	    if ( ((**string) < aZero) ||
		 ((**string) > aThree) )
		ret = vFALSE;
	    break;
	case 4:
	    prev = text -> GetTextAt (3);
	    if ((*prev) == aZero)
	    {
		if ( ((**string) < aOne) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else if ( ((*prev) == aOne) || ((*prev) == aTwo) )
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else if ((*prev) == aThree)
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aOne) )
		    ret = vFALSE;
	    }
	    else
		ret = vFALSE;
	    break;
	
	case 6:
	case 7:
	    if ( ((**string) < aZero) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	    break;

	case 8:
	{
	    vchar aSpace = vcharFromLiteral (' ');
	    if ((**string) != aSpace)
		ret = vFALSE;
	    break;
	}

	case 9:
	    if ( ((**string) < aZero) ||
		 ((**string) > aTwo) )
		ret = vFALSE;
	    break;

	case 10:
	    prev = text -> GetTextAt (9);
	    if ( ((*prev) == aZero) || ((*prev) == aOne) )
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aNine) )
		    ret = vFALSE;
	    }
	    else
	    {
		if ( ((**string) < aZero) ||
		     ((**string) > aThree) )
		    ret = vFALSE;
	    }
	    break;

	case 11:
	case 14:
	{
	    vchar aColon = vcharFromLiteral (':');
	    if ((**string) != aColon)
		ret = vFALSE;
	    break;
	}

	case 12:
	case 15:
	{
	    vchar aFive = vcharFromLiteral ('5');
	    if ( ((**string) < aZero) ||
		 ((**string) > aFive) )
		ret = vFALSE;
	    break;
	}

	case 13:
	case 16:
	    if ( ((**string) < aZero) ||
		 ((**string) > aNine) )
		ret = vFALSE;
	    break;

	default:
	    ret = vFALSE;
	    break;

    } // end switch textPos

    if (ret == vFALSE)
    {
	vwindow::Beep ();
    }

    return (ret);
}
